#ifndef UTILITIES_IDD_IDDENUMS_HXX
#define UTILITIES_IDD_IDDENUMS_HXX

#include <iostream>
#include <map>
#include <set>
#include <stdexcept>
#include <string>
#include <string_view>
#include <vector>

inline std::string ascii_to_lower_copy(std::string_view input) {
  std::string result{input};
  constexpr auto to_lower_diff = 'a' - 'A';
  for (auto& c : result) {
    if (c >= 'A' && c <= 'Z') {
      c += to_lower_diff;
    }
  }
  return result;
}

inline std::string ascii_to_upper_copy(std::string_view input) {
  std::string result{input};
  constexpr auto to_upper_diff = 'A' - 'a';
  for (auto& c : result) {
    if (c >= 'a' && c <= 'z') {
      c += to_upper_diff;
    }
  }
  return result;
}

template <typename T>
struct StaticInitializer
{
};

/** Base class for OPENSTUDIO_ENUMs. Comparison operators use the underlying
 * (integer) value. */
template <typename Enum>
class EnumBase : public ::StaticInitializer<Enum>
{
 public:
  /** Default constructor. EnumBase default constructor never actually called;
     * OPENSTUDIO_ENUM- generated classes have default constructors that set the
     * value to the first item in the enumerated list. */
  EnumBase() = delete;

  /** Construct from t_value. Throws std::runtime_error if t_value is not in
     * the domain. */
  EnumBase(int t_value) : m_value(lookupValue(t_value)) {}

  /** Construct from t_value. Throws std::runtime_error if t_value does not
     * match a name or a description (case insensitive). */
  EnumBase(const std::string& t_value) : m_value(lookupValue(t_value)) {}

  /** Returns the name associated with t_value, if it exists. Otherwise,
     * throws std::runtime_error. */
  static std::string valueName(int t_value) {
    const auto& m = getNames();
    auto itr = m.find(t_value);
    if (itr == m.end()) {
      throw std::runtime_error("Invalid domain for OpenStudio Enum " + Enum::enumName());
    }
    return itr->second;
  }

  /** Returns the description associated with t_value, if it exists.
     * Otherwise, throws std::runtime_error. */
  static std::string valueDescription(int t_value) {
    const std::map<int, std::string>& m = getDescriptions();
    auto itr = m.find(t_value);
    if (itr == m.end()) {
      return valueName(t_value);
    }
    return itr->second;
  }

  /** Returns the set of all values in this enum's domain. */
  static const std::set<int>& getValues() {
    static std::set<int> values(buildValues());
    return values;
  }

  /** Returns this instance's current value (as an integer). */
  int value() const {
    return m_value;
  }

  /** Returns the name associated with this instance's current value. */
  std::string valueName() const {
    return valueName(m_value);
  }

  /** Returns the description associated with this instance's current value.
     */
  std::string valueDescription() const {
    return valueDescription(m_value);
  }

  /** Set this instance's value to t_value. Throws std::runtime_error if
     *  t_value is not in the domain. */
  void setValue(int t_value) {
    m_value = lookupValue(t_value);
  }

  bool operator==(const EnumBase<Enum>& other) const {
    return (m_value == other.m_value);
  }
  bool operator!=(const EnumBase<Enum>& other) const {
    return (m_value != other.m_value);
  }
  bool operator>(const EnumBase<Enum>& other) const {
    return (m_value > other.m_value);
  }
  bool operator>=(const EnumBase<Enum>& other) const {
    return (m_value >= other.m_value);
  }
  bool operator<(const EnumBase<Enum>& other) const {
    return (m_value < other.m_value);
  }
  bool operator<=(const EnumBase<Enum>& other) const {
    return (m_value <= other.m_value);
  }

  static void initialize() {
    getValues();
    getLookupMap();
    getNames();
    getDescriptions();
  }

  /** Returns the (integer) value associated with t_name, as determined by
     * case-insensitive comparison to the enumerated names and descriptions. */
  int lookupValue(std::string t_name) {
    t_name = ascii_to_upper_copy(t_name);
    const auto& names = getLookupMap();
    auto itr = names.find(t_name);
    if (itr != names.end()) {
      return itr->second;
    }
    throw std::runtime_error("Unknown OpenStudio Enum Value '" + t_name + "' for Enum " + Enum::enumName());
  }

  /** Returns t_value if it is in the domain. Otherwise throws
     * std::runtime_error. */
  int lookupValue(int t_value) {
    const std::set<int>& values = getValues();
    if (values.count(t_value) != 0) {
      return t_value;
    } else {
      throw std::runtime_error("Unknown OpenStudio Enum Value = " + std::to_string(t_value) + " for Enum " + Enum::enumName());
    }
  }

  static const std::map<std::string, int>& getLookupMap() {
    static const std::map<std::string, int> m = buildLookupMap();
    return m;
  }

  static const std::map<int, std::string>& getNames() {
    static const std::map<int, std::string> names = buildStrings(false);
    return names;
  }

  static const std::map<int, std::string>& getDescriptions() {
    static const std::map<int, std::string> descriptions = buildStrings(true);
    return descriptions;
  }

 protected:
  int m_value;

 private:
  static std::map<int, std::string> buildStrings(bool d) {
    auto strings(Enum::buildStringVec(d));
    std::map<int, std::string> m;

    for (const auto& s : strings) {
      m[s.second] = s.first;
    }

    return m;
  }

  static std::map<std::string, int> buildLookupMap() {
    const auto& m = getNames();
    const auto& d = getDescriptions();

    std::map<std::string, int> retval;
    for (const auto& name : m) {
      retval[ascii_to_upper_copy(name.second)] = name.first;
    }

    for (const auto& description : d) {
      retval[ascii_to_upper_copy(description.second)] = description.first;
    }

    return retval;
  }

  static std::set<int> buildValues() {
    const auto& names = getNames();

    std::set<int> retvals;

    for (const auto& name : names) {
      retvals.insert(name.first);
    }

    return retvals;
  }
};

struct IddObjectType : public ::EnumBase<IddObjectType>
{
 public:
  enum domain : int;

  IddObjectType() : EnumBase<IddObjectType>(0) {}
  IddObjectType(const std::string& t_name) : EnumBase<IddObjectType>(t_name) {}
  IddObjectType(int t_value) : EnumBase<IddObjectType>(t_value) {}

  static std::string enumName() {
    return "IddObjectType";
  }

  domain value() const {
    return static_cast<domain>(EnumBase<IddObjectType>::value());
  }

 private:
  friend class EnumBase<IddObjectType>;
  using PT = std::pair<std::string, int>;
  using VecType = std::vector<PT>;
  static VecType buildStringVec(bool isd);
};

inline std::ostream& operator<<(std::ostream& os, const IddObjectType& e) {
  return os << e.valueName() << "(" << e.value() << ")";
}

/** \class IddObjectType */
enum IddObjectType::domain : int
{
  Catchall,
  UserCustom,
  LeadInput,
  EndLeadInput,
  SimulationData,
  EndSimulationData,
  Version,
  SimulationControl,
  PerformancePrecisionTradeoffs,
  Building,
  ShadowCalculation,
  SurfaceConvectionAlgorithm_Inside,
  SurfaceConvectionAlgorithm_Outside,
  HeatBalanceAlgorithm,
  HeatBalanceSettings_ConductionFiniteDifference,
  ZoneAirHeatBalanceAlgorithm,
  ZoneAirContaminantBalance,
  ZoneAirMassFlowConservation,
  ZoneCapacitanceMultiplier_ResearchSpecial,
  Timestep,
  ConvergenceLimits,
  HVACSystemRootFindingAlgorithm,
  Compliance_Building,
  Site_Location,
  Site_VariableLocation,
  SizingPeriod_DesignDay,
  SizingPeriod_WeatherFileDays,
  SizingPeriod_WeatherFileConditionType,
  RunPeriod,
  RunPeriodControl_SpecialDays,
  RunPeriodControl_DaylightSavingTime,
  WeatherProperty_SkyTemperature,
  Site_WeatherStation,
  Site_HeightVariation,
  Site_GroundTemperature_BuildingSurface,
  Site_GroundTemperature_FCfactorMethod,
  Site_GroundTemperature_Shallow,
  Site_GroundTemperature_Deep,
  Site_GroundTemperature_Undisturbed_FiniteDifference,
  Site_GroundTemperature_Undisturbed_KusudaAchenbach,
  Site_GroundTemperature_Undisturbed_Xing,
  Site_GroundDomain_Slab,
  Site_GroundDomain_Basement,
  Site_GroundReflectance,
  Site_GroundReflectance_SnowModifier,
  Site_WaterMainsTemperature,
  Site_Precipitation,
  RoofIrrigation,
  Site_SolarAndVisibleSpectrum,
  Site_SpectrumData,
  ScheduleTypeLimits,
  Schedule_Day_Hourly,
  Schedule_Day_Interval,
  Schedule_Day_List,
  Schedule_Week_Daily,
  Schedule_Week_Compact,
  Schedule_Year,
  Schedule_Compact,
  Schedule_Constant,
  Schedule_File_Shading,
  Schedule_File,
  Material,
  Material_NoMass,
  Material_InfraredTransparent,
  Material_AirGap,
  Material_RoofVegetation,
  WindowMaterial_SimpleGlazingSystem,
  WindowMaterial_Glazing,
  WindowMaterial_GlazingGroup_Thermochromic,
  WindowMaterial_Glazing_RefractionExtinctionMethod,
  WindowMaterial_Gas,
  WindowGap_SupportPillar,
  WindowGap_DeflectionState,
  WindowMaterial_GasMixture,
  WindowMaterial_Gap,
  WindowMaterial_Shade,
  WindowMaterial_ComplexShade,
  WindowMaterial_Blind,
  WindowMaterial_Screen,
  WindowMaterial_Shade_EquivalentLayer,
  WindowMaterial_Drape_EquivalentLayer,
  WindowMaterial_Blind_EquivalentLayer,
  WindowMaterial_Screen_EquivalentLayer,
  WindowMaterial_Glazing_EquivalentLayer,
  WindowMaterial_Gap_EquivalentLayer,
  MaterialProperty_MoisturePenetrationDepth_Settings,
  MaterialProperty_PhaseChange,
  MaterialProperty_PhaseChangeHysteresis,
  MaterialProperty_VariableThermalConductivity,
  MaterialProperty_VariableAbsorptance,
  MaterialProperty_HeatAndMoistureTransfer_Settings,
  MaterialProperty_HeatAndMoistureTransfer_SorptionIsotherm,
  MaterialProperty_HeatAndMoistureTransfer_Suction,
  MaterialProperty_HeatAndMoistureTransfer_Redistribution,
  MaterialProperty_HeatAndMoistureTransfer_Diffusion,
  MaterialProperty_HeatAndMoistureTransfer_ThermalConductivity,
  MaterialProperty_GlazingSpectralData,
  Construction,
  Construction_CfactorUndergroundWall,
  Construction_FfactorGroundFloor,
  ConstructionProperty_InternalHeatSource,
  Construction_AirBoundary,
  WindowThermalModel_Params,
  WindowsCalculationEngine,
  Construction_ComplexFenestrationState,
  Construction_WindowEquivalentLayer,
  Construction_WindowDataFile,
  GlobalGeometryRules,
  GeometryTransform,
  Space,
  SpaceList,
  Zone,
  ZoneList,
  ZoneGroup,
  BuildingSurface_Detailed,
  Wall_Detailed,
  RoofCeiling_Detailed,
  Floor_Detailed,
  Wall_Exterior,
  Wall_Adiabatic,
  Wall_Underground,
  Wall_Interzone,
  Roof,
  Ceiling_Adiabatic,
  Ceiling_Interzone,
  Floor_GroundContact,
  Floor_Adiabatic,
  Floor_Interzone,
  FenestrationSurface_Detailed,
  Window,
  Door,
  GlazedDoor,
  Window_Interzone,
  Door_Interzone,
  GlazedDoor_Interzone,
  WindowShadingControl,
  WindowProperty_FrameAndDivider,
  WindowProperty_AirflowControl,
  WindowProperty_StormWindow,
  InternalMass,
  Shading_Site,
  Shading_Building,
  Shading_Site_Detailed,
  Shading_Building_Detailed,
  Shading_Overhang,
  Shading_Overhang_Projection,
  Shading_Fin,
  Shading_Fin_Projection,
  Shading_Zone_Detailed,
  ShadingProperty_Reflectance,
  SurfaceProperty_HeatTransferAlgorithm,
  SurfaceProperty_HeatTransferAlgorithm_MultipleSurface,
  SurfaceProperty_HeatTransferAlgorithm_SurfaceList,
  SurfaceProperty_HeatTransferAlgorithm_Construction,
  SurfaceProperty_HeatBalanceSourceTerm,
  SurfaceControl_MovableInsulation,
  SurfaceProperty_OtherSideCoefficients,
  SurfaceProperty_OtherSideConditionsModel,
  SurfaceProperty_Underwater,
  Foundation_Kiva,
  Foundation_Kiva_Settings,
  SurfaceProperty_ExposedFoundationPerimeter,
  SurfaceConvectionAlgorithm_Inside_AdaptiveModelSelections,
  SurfaceConvectionAlgorithm_Outside_AdaptiveModelSelections,
  SurfaceConvectionAlgorithm_Inside_UserCurve,
  SurfaceConvectionAlgorithm_Outside_UserCurve,
  SurfaceProperty_ConvectionCoefficients,
  SurfaceProperty_ConvectionCoefficients_MultipleSurface,
  SurfaceProperties_VaporCoefficients,
  SurfaceProperty_ExteriorNaturalVentedCavity,
  SurfaceProperty_SolarIncidentInside,
  SurfaceProperty_IncidentSolarMultiplier,
  SurfaceProperty_LocalEnvironment,
  ZoneProperty_LocalEnvironment,
  SurfaceProperty_SurroundingSurfaces,
  SurfaceProperty_GroundSurfaces,
  ComplexFenestrationProperty_SolarAbsorbedLayers,
  ZoneProperty_UserViewFactors_BySurfaceName,
  GroundHeatTransfer_Control,
  GroundHeatTransfer_Slab_Materials,
  GroundHeatTransfer_Slab_MatlProps,
  GroundHeatTransfer_Slab_BoundConds,
  GroundHeatTransfer_Slab_BldgProps,
  GroundHeatTransfer_Slab_Insulation,
  GroundHeatTransfer_Slab_EquivalentSlab,
  GroundHeatTransfer_Slab_AutoGrid,
  GroundHeatTransfer_Slab_ManualGrid,
  GroundHeatTransfer_Slab_XFACE,
  GroundHeatTransfer_Slab_YFACE,
  GroundHeatTransfer_Slab_ZFACE,
  GroundHeatTransfer_Basement_SimParameters,
  GroundHeatTransfer_Basement_MatlProps,
  GroundHeatTransfer_Basement_Insulation,
  GroundHeatTransfer_Basement_SurfaceProps,
  GroundHeatTransfer_Basement_BldgData,
  GroundHeatTransfer_Basement_Interior,
  GroundHeatTransfer_Basement_ComBldg,
  GroundHeatTransfer_Basement_EquivSlab,
  GroundHeatTransfer_Basement_EquivAutoGrid,
  GroundHeatTransfer_Basement_AutoGrid,
  GroundHeatTransfer_Basement_ManualGrid,
  GroundHeatTransfer_Basement_XFACE,
  GroundHeatTransfer_Basement_YFACE,
  GroundHeatTransfer_Basement_ZFACE,
  RoomAirModelType,
  RoomAir_TemperaturePattern_UserDefined,
  RoomAir_TemperaturePattern_ConstantGradient,
  RoomAir_TemperaturePattern_TwoGradient,
  RoomAir_TemperaturePattern_NondimensionalHeight,
  RoomAir_TemperaturePattern_SurfaceMapping,
  RoomAir_Node,
  RoomAirSettings_OneNodeDisplacementVentilation,
  RoomAirSettings_ThreeNodeDisplacementVentilation,
  RoomAirSettings_CrossVentilation,
  RoomAirSettings_UnderFloorAirDistributionInterior,
  RoomAirSettings_UnderFloorAirDistributionExterior,
  RoomAir_Node_AirflowNetwork,
  RoomAir_Node_AirflowNetwork_AdjacentSurfaceList,
  RoomAir_Node_AirflowNetwork_InternalGains,
  RoomAir_Node_AirflowNetwork_HVACEquipment,
  RoomAirSettings_AirflowNetwork,
  People,
  ComfortViewFactorAngles,
  Lights,
  ElectricEquipment,
  GasEquipment,
  HotWaterEquipment,
  SteamEquipment,
  OtherEquipment,
  ElectricEquipment_ITE_AirCooled,
  ZoneBaseboard_OutdoorTemperatureControlled,
  SwimmingPool_Indoor,
  ZoneContaminantSourceAndSink_CarbonDioxide,
  ZoneContaminantSourceAndSink_Generic_Constant,
  SurfaceContaminantSourceAndSink_Generic_PressureDriven,
  ZoneContaminantSourceAndSink_Generic_CutoffModel,
  ZoneContaminantSourceAndSink_Generic_DecaySource,
  SurfaceContaminantSourceAndSink_Generic_BoundaryLayerDiffusion,
  SurfaceContaminantSourceAndSink_Generic_DepositionVelocitySink,
  ZoneContaminantSourceAndSink_Generic_DepositionRateSink,
  Daylighting_Controls,
  Daylighting_ReferencePoint,
  Daylighting_DELight_ComplexFenestration,
  DaylightingDevice_Tubular,
  DaylightingDevice_Shelf,
  DaylightingDevice_LightWell,
  Output_DaylightFactors,
  Output_IlluminanceMap,
  OutputControl_IlluminanceMap_Style,
  ZoneInfiltration_DesignFlowRate,
  ZoneInfiltration_EffectiveLeakageArea,
  ZoneInfiltration_FlowCoefficient,
  ZoneVentilation_DesignFlowRate,
  ZoneVentilation_WindandStackOpenArea,
  ZoneAirBalance_OutdoorAir,
  ZoneMixing,
  ZoneCrossMixing,
  ZoneRefrigerationDoorMixing,
  ZoneEarthtube,
  ZoneCoolTower_Shower,
  ZoneThermalChimney,
  AirflowNetwork_SimulationControl,
  AirflowNetwork_MultiZone_Zone,
  AirflowNetwork_MultiZone_Surface,
  AirflowNetwork_MultiZone_ReferenceCrackConditions,
  AirflowNetwork_MultiZone_Surface_Crack,
  AirflowNetwork_MultiZone_Surface_EffectiveLeakageArea,
  AirflowNetwork_MultiZone_SpecifiedFlowRate,
  AirflowNetwork_MultiZone_Component_DetailedOpening,
  AirflowNetwork_MultiZone_Component_SimpleOpening,
  AirflowNetwork_MultiZone_Component_HorizontalOpening,
  AirflowNetwork_MultiZone_Component_ZoneExhaustFan,
  AirflowNetwork_MultiZone_ExternalNode,
  AirflowNetwork_MultiZone_WindPressureCoefficientArray,
  AirflowNetwork_MultiZone_WindPressureCoefficientValues,
  AirflowNetwork_ZoneControl_PressureController,
  AirflowNetwork_Distribution_Node,
  AirflowNetwork_Distribution_Component_Leak,
  AirflowNetwork_Distribution_Component_LeakageRatio,
  AirflowNetwork_Distribution_Component_Duct,
  AirflowNetwork_Distribution_Component_Fan,
  AirflowNetwork_Distribution_Component_Coil,
  AirflowNetwork_Distribution_Component_HeatExchanger,
  AirflowNetwork_Distribution_Component_TerminalUnit,
  AirflowNetwork_Distribution_Component_ConstantPressureDrop,
  AirflowNetwork_Distribution_Component_OutdoorAirFlow,
  AirflowNetwork_Distribution_Component_ReliefAirFlow,
  AirflowNetwork_Distribution_Linkage,
  AirflowNetwork_Distribution_DuctViewFactors,
  AirflowNetwork_OccupantVentilationControl,
  AirflowNetwork_IntraZone_Node,
  AirflowNetwork_IntraZone_Linkage,
  Exterior_Lights,
  Exterior_FuelEquipment,
  Exterior_WaterEquipment,
  HVACTemplate_Thermostat,
  HVACTemplate_Zone_IdealLoadsAirSystem,
  HVACTemplate_Zone_BaseboardHeat,
  HVACTemplate_Zone_FanCoil,
  HVACTemplate_Zone_PTAC,
  HVACTemplate_Zone_PTHP,
  HVACTemplate_Zone_WaterToAirHeatPump,
  HVACTemplate_Zone_VRF,
  HVACTemplate_Zone_Unitary,
  HVACTemplate_Zone_VAV,
  HVACTemplate_Zone_VAV_FanPowered,
  HVACTemplate_Zone_VAV_HeatAndCool,
  HVACTemplate_Zone_ConstantVolume,
  HVACTemplate_Zone_DualDuct,
  HVACTemplate_System_VRF,
  HVACTemplate_System_Unitary,
  HVACTemplate_System_UnitaryHeatPump_AirToAir,
  HVACTemplate_System_UnitarySystem,
  HVACTemplate_System_VAV,
  HVACTemplate_System_PackagedVAV,
  HVACTemplate_System_ConstantVolume,
  HVACTemplate_System_DualDuct,
  HVACTemplate_System_DedicatedOutdoorAir,
  HVACTemplate_Plant_ChilledWaterLoop,
  HVACTemplate_Plant_Chiller,
  HVACTemplate_Plant_Chiller_ObjectReference,
  HVACTemplate_Plant_Tower,
  HVACTemplate_Plant_Tower_ObjectReference,
  HVACTemplate_Plant_HotWaterLoop,
  HVACTemplate_Plant_Boiler,
  HVACTemplate_Plant_Boiler_ObjectReference,
  HVACTemplate_Plant_MixedWaterLoop,
  DesignSpecification_OutdoorAir,
  DesignSpecification_OutdoorAir_SpaceList,
  DesignSpecification_ZoneAirDistribution,
  Sizing_Parameters,
  Sizing_Zone,
  DesignSpecification_ZoneHVAC_Sizing,
  DesignSpecification_AirTerminal_Sizing,
  Sizing_System,
  Sizing_Plant,
  OutputControl_Sizing_Style,
  ZoneControl_Humidistat,
  ZoneControl_Thermostat,
  ZoneControl_Thermostat_OperativeTemperature,
  ZoneControl_Thermostat_ThermalComfort,
  ZoneControl_Thermostat_TemperatureAndHumidity,
  ThermostatSetpoint_SingleHeating,
  ThermostatSetpoint_SingleCooling,
  ThermostatSetpoint_SingleHeatingOrCooling,
  ThermostatSetpoint_DualSetpoint,
  ThermostatSetpoint_ThermalComfort_Fanger_SingleHeating,
  ThermostatSetpoint_ThermalComfort_Fanger_SingleCooling,
  ThermostatSetpoint_ThermalComfort_Fanger_SingleHeatingOrCooling,
  ThermostatSetpoint_ThermalComfort_Fanger_DualSetpoint,
  ZoneControl_Thermostat_StagedDualSetpoint,
  ZoneControl_ContaminantController,
  ZoneHVAC_IdealLoadsAirSystem,
  ZoneHVAC_FourPipeFanCoil,
  ZoneHVAC_WindowAirConditioner,
  ZoneHVAC_PackagedTerminalAirConditioner,
  ZoneHVAC_PackagedTerminalHeatPump,
  ZoneHVAC_WaterToAirHeatPump,
  ZoneHVAC_Dehumidifier_DX,
  ZoneHVAC_EnergyRecoveryVentilator,
  ZoneHVAC_EnergyRecoveryVentilator_Controller,
  ZoneHVAC_UnitVentilator,
  ZoneHVAC_UnitHeater,
  ZoneHVAC_EvaporativeCoolerUnit,
  ZoneHVAC_HybridUnitaryHVAC,
  ZoneHVAC_OutdoorAirUnit,
  ZoneHVAC_OutdoorAirUnit_EquipmentList,
  ZoneHVAC_TerminalUnit_VariableRefrigerantFlow,
  ZoneHVAC_Baseboard_RadiantConvective_Water_Design,
  ZoneHVAC_Baseboard_RadiantConvective_Water,
  ZoneHVAC_Baseboard_RadiantConvective_Steam_Design,
  ZoneHVAC_Baseboard_RadiantConvective_Steam,
  ZoneHVAC_Baseboard_RadiantConvective_Electric,
  ZoneHVAC_CoolingPanel_RadiantConvective_Water,
  ZoneHVAC_Baseboard_Convective_Water,
  ZoneHVAC_Baseboard_Convective_Electric,
  ZoneHVAC_LowTemperatureRadiant_VariableFlow,
  ZoneHVAC_LowTemperatureRadiant_VariableFlow_Design,
  ZoneHVAC_LowTemperatureRadiant_ConstantFlow,
  ZoneHVAC_LowTemperatureRadiant_ConstantFlow_Design,
  ZoneHVAC_LowTemperatureRadiant_Electric,
  ZoneHVAC_LowTemperatureRadiant_SurfaceGroup,
  ZoneHVAC_HighTemperatureRadiant,
  ZoneHVAC_VentilatedSlab,
  ZoneHVAC_VentilatedSlab_SlabGroup,
  AirTerminal_SingleDuct_ConstantVolume_Reheat,
  AirTerminal_SingleDuct_ConstantVolume_NoReheat,
  AirTerminal_SingleDuct_VAV_NoReheat,
  AirTerminal_SingleDuct_VAV_Reheat,
  AirTerminal_SingleDuct_VAV_Reheat_VariableSpeedFan,
  AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat,
  AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat,
  AirTerminal_SingleDuct_SeriesPIU_Reheat,
  AirTerminal_SingleDuct_ParallelPIU_Reheat,
  AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction,
  AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam,
  AirTerminal_SingleDuct_ConstantVolume_CooledBeam,
  AirTerminal_SingleDuct_Mixer,
  AirTerminal_DualDuct_ConstantVolume,
  AirTerminal_DualDuct_VAV,
  AirTerminal_DualDuct_VAV_OutdoorAir,
  ZoneHVAC_AirDistributionUnit,
  ZoneHVAC_ExhaustControl,
  ZoneHVAC_EquipmentList,
  ZoneHVAC_EquipmentConnections,
  Fan_SystemModel,
  Fan_ConstantVolume,
  Fan_VariableVolume,
  Fan_OnOff,
  Fan_ZoneExhaust,
  FanPerformance_NightVentilation,
  Fan_ComponentModel,
  Coil_Cooling_Water,
  Coil_Cooling_Water_DetailedGeometry,
  CoilSystem_Cooling_Water,
  Coil_Cooling_DX,
  Coil_Cooling_DX_CurveFit_Performance,
  Coil_Cooling_DX_CurveFit_OperatingMode,
  Coil_Cooling_DX_CurveFit_Speed,
  Coil_Cooling_DX_SingleSpeed,
  Coil_Cooling_DX_TwoSpeed,
  Coil_Cooling_DX_MultiSpeed,
  Coil_Cooling_DX_VariableSpeed,
  Coil_Cooling_DX_TwoStageWithHumidityControlMode,
  CoilPerformance_DX_Cooling,
  Coil_Cooling_DX_VariableRefrigerantFlow,
  Coil_Heating_DX_VariableRefrigerantFlow,
  Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  Coil_Heating_Water,
  Coil_Heating_Steam,
  Coil_Heating_Electric,
  Coil_Heating_Electric_MultiStage,
  Coil_Heating_Fuel,
  Coil_Heating_Gas_MultiStage,
  Coil_Heating_Desuperheater,
  Coil_Heating_DX_SingleSpeed,
  Coil_Heating_DX_MultiSpeed,
  Coil_Heating_DX_VariableSpeed,
  Coil_Cooling_WaterToAirHeatPump_ParameterEstimation,
  Coil_Heating_WaterToAirHeatPump_ParameterEstimation,
  Coil_Cooling_WaterToAirHeatPump_EquationFit,
  Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit,
  Coil_Heating_WaterToAirHeatPump_EquationFit,
  Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit,
  Coil_WaterHeating_AirToWaterHeatPump_Pumped,
  Coil_WaterHeating_AirToWaterHeatPump_Wrapped,
  Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed,
  Coil_WaterHeating_Desuperheater,
  CoilSystem_Cooling_DX,
  CoilSystem_Heating_DX,
  CoilSystem_Cooling_Water_HeatExchangerAssisted,
  CoilSystem_Cooling_DX_HeatExchangerAssisted,
  CoilSystem_IntegratedHeatPump_AirSource,
  Coil_Cooling_DX_SingleSpeed_ThermalStorage,
  EvaporativeCooler_Direct_CelDekPad,
  EvaporativeCooler_Indirect_CelDekPad,
  EvaporativeCooler_Indirect_WetCoil,
  EvaporativeCooler_Indirect_ResearchSpecial,
  EvaporativeCooler_Direct_ResearchSpecial,
  Humidifier_Steam_Electric,
  Humidifier_Steam_Gas,
  Dehumidifier_Desiccant_NoFans,
  Dehumidifier_Desiccant_System,
  HeatExchanger_AirToAir_FlatPlate,
  HeatExchanger_AirToAir_SensibleAndLatent,
  HeatExchanger_Desiccant_BalancedFlow,
  HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1,
  AirLoopHVAC_UnitarySystem,
  UnitarySystemPerformance_Multispeed,
  AirLoopHVAC_Unitary_Furnace_HeatOnly,
  AirLoopHVAC_Unitary_Furnace_HeatCool,
  AirLoopHVAC_UnitaryHeatOnly,
  AirLoopHVAC_UnitaryHeatCool,
  AirLoopHVAC_UnitaryHeatPump_AirToAir,
  AirLoopHVAC_UnitaryHeatPump_WaterToAir,
  AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass,
  AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed,
  AirConditioner_VariableRefrigerantFlow,
  AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl,
  AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR,
  ZoneTerminalUnitList,
  Controller_WaterCoil,
  Controller_OutdoorAir,
  Controller_MechanicalVentilation,
  AirLoopHVAC_ControllerList,
  AirLoopHVAC,
  AirLoopHVAC_OutdoorAirSystem_EquipmentList,
  AirLoopHVAC_OutdoorAirSystem,
  OutdoorAir_Mixer,
  AirLoopHVAC_ZoneSplitter,
  AirLoopHVAC_SupplyPlenum,
  AirLoopHVAC_SupplyPath,
  AirLoopHVAC_ZoneMixer,
  AirLoopHVAC_ReturnPlenum,
  AirLoopHVAC_ReturnPath,
  AirLoopHVAC_ExhaustSystem,
  AirLoopHVAC_DedicatedOutdoorAirSystem,
  AirLoopHVAC_Mixer,
  AirLoopHVAC_Splitter,
  Branch,
  BranchList,
  Connector_Splitter,
  Connector_Mixer,
  ConnectorList,
  NodeList,
  OutdoorAir_Node,
  OutdoorAir_NodeList,
  Pipe_Adiabatic,
  Pipe_Adiabatic_Steam,
  Pipe_Indoor,
  Pipe_Outdoor,
  Pipe_Underground,
  PipingSystem_Underground_Domain,
  PipingSystem_Underground_PipeCircuit,
  PipingSystem_Underground_PipeSegment,
  Duct,
  Pump_VariableSpeed,
  Pump_ConstantSpeed,
  Pump_VariableSpeed_Condensate,
  HeaderedPumps_ConstantSpeed,
  HeaderedPumps_VariableSpeed,
  TemperingValve,
  LoadProfile_Plant,
  SolarCollectorPerformance_FlatPlate,
  SolarCollector_FlatPlate_Water,
  SolarCollector_FlatPlate_PhotovoltaicThermal,
  SolarCollectorPerformance_PhotovoltaicThermal_Simple,
  SolarCollectorPerformance_PhotovoltaicThermal_BIPVT,
  SolarCollector_IntegralCollectorStorage,
  SolarCollectorPerformance_IntegralCollectorStorage,
  SolarCollector_UnglazedTranspired,
  SolarCollector_UnglazedTranspired_Multisystem,
  Boiler_HotWater,
  Boiler_Steam,
  Chiller_Electric_ASHRAE205,
  Chiller_Electric_EIR,
  Chiller_Electric_ReformulatedEIR,
  Chiller_Electric,
  Chiller_Absorption_Indirect,
  Chiller_Absorption,
  Chiller_ConstantCOP,
  Chiller_EngineDriven,
  Chiller_CombustionTurbine,
  ChillerHeater_Absorption_DirectFired,
  ChillerHeater_Absorption_DoubleEffect,
  HeatPump_PlantLoop_EIR_Cooling,
  HeatPump_PlantLoop_EIR_Heating,
  HeatPump_AirToWater_FuelFired_Heating,
  HeatPump_AirToWater_FuelFired_Cooling,
  HeatPump_WaterToWater_EquationFit_Heating,
  HeatPump_WaterToWater_EquationFit_Cooling,
  HeatPump_WaterToWater_ParameterEstimation_Cooling,
  HeatPump_WaterToWater_ParameterEstimation_Heating,
  DistrictCooling,
  DistrictHeating,
  PlantComponent_TemperatureSource,
  CentralHeatPumpSystem,
  ChillerHeaterPerformance_Electric_EIR,
  CoolingTower_SingleSpeed,
  CoolingTower_TwoSpeed,
  CoolingTower_VariableSpeed_Merkel,
  CoolingTower_VariableSpeed,
  CoolingTowerPerformance_CoolTools,
  CoolingTowerPerformance_YorkCalc,
  EvaporativeFluidCooler_SingleSpeed,
  EvaporativeFluidCooler_TwoSpeed,
  FluidCooler_SingleSpeed,
  FluidCooler_TwoSpeed,
  GroundHeatExchanger_System,
  GroundHeatExchanger_Vertical_Properties,
  GroundHeatExchanger_Vertical_Array,
  GroundHeatExchanger_Vertical_Single,
  GroundHeatExchanger_ResponseFactors,
  GroundHeatExchanger_Pond,
  GroundHeatExchanger_Surface,
  GroundHeatExchanger_HorizontalTrench,
  GroundHeatExchanger_Slinky,
  HeatExchanger_FluidToFluid,
  WaterHeater_Mixed,
  WaterHeater_Stratified,
  WaterHeater_Sizing,
  WaterHeater_HeatPump_PumpedCondenser,
  WaterHeater_HeatPump_WrappedCondenser,
  ThermalStorage_Ice_Simple,
  ThermalStorage_Ice_Detailed,
  ThermalStorage_ChilledWater_Mixed,
  ThermalStorage_ChilledWater_Stratified,
  PlantLoop,
  CondenserLoop,
  PlantEquipmentList,
  CondenserEquipmentList,
  PlantEquipmentOperation_Uncontrolled,
  PlantEquipmentOperation_CoolingLoad,
  PlantEquipmentOperation_HeatingLoad,
  PlantEquipmentOperation_OutdoorDryBulb,
  PlantEquipmentOperation_OutdoorWetBulb,
  PlantEquipmentOperation_OutdoorRelativeHumidity,
  PlantEquipmentOperation_OutdoorDewpoint,
  PlantEquipmentOperation_ComponentSetpoint,
  PlantEquipmentOperation_ThermalEnergyStorage,
  PlantEquipmentOperation_OutdoorDryBulbDifference,
  PlantEquipmentOperation_OutdoorWetBulbDifference,
  PlantEquipmentOperation_OutdoorDewpointDifference,
  PlantEquipmentOperationSchemes,
  CondenserEquipmentOperationSchemes,
  EnergyManagementSystem_Sensor,
  EnergyManagementSystem_Actuator,
  EnergyManagementSystem_ProgramCallingManager,
  EnergyManagementSystem_Program,
  EnergyManagementSystem_Subroutine,
  EnergyManagementSystem_GlobalVariable,
  EnergyManagementSystem_OutputVariable,
  EnergyManagementSystem_MeteredOutputVariable,
  EnergyManagementSystem_TrendVariable,
  EnergyManagementSystem_InternalVariable,
  EnergyManagementSystem_CurveOrTableIndexVariable,
  EnergyManagementSystem_ConstructionIndexVariable,
  ExternalInterface,
  ExternalInterface_Schedule,
  ExternalInterface_Variable,
  ExternalInterface_Actuator,
  ExternalInterface_FunctionalMockupUnitImport,
  ExternalInterface_FunctionalMockupUnitImport_From_Variable,
  ExternalInterface_FunctionalMockupUnitImport_To_Schedule,
  ExternalInterface_FunctionalMockupUnitImport_To_Actuator,
  ExternalInterface_FunctionalMockupUnitImport_To_Variable,
  ExternalInterface_FunctionalMockupUnitExport_From_Variable,
  ExternalInterface_FunctionalMockupUnitExport_To_Schedule,
  ExternalInterface_FunctionalMockupUnitExport_To_Actuator,
  ExternalInterface_FunctionalMockupUnitExport_To_Variable,
  ZoneHVAC_ForcedAir_UserDefined,
  AirTerminal_SingleDuct_UserDefined,
  Coil_UserDefined,
  PlantComponent_UserDefined,
  PlantEquipmentOperation_UserDefined,
  AvailabilityManager_Scheduled,
  AvailabilityManager_ScheduledOn,
  AvailabilityManager_ScheduledOff,
  AvailabilityManager_OptimumStart,
  AvailabilityManager_NightCycle,
  AvailabilityManager_DifferentialThermostat,
  AvailabilityManager_HighTemperatureTurnOff,
  AvailabilityManager_HighTemperatureTurnOn,
  AvailabilityManager_LowTemperatureTurnOff,
  AvailabilityManager_LowTemperatureTurnOn,
  AvailabilityManager_NightVentilation,
  AvailabilityManager_HybridVentilation,
  AvailabilityManagerAssignmentList,
  SetpointManager_Scheduled,
  SetpointManager_Scheduled_DualSetpoint,
  SetpointManager_OutdoorAirReset,
  SetpointManager_SingleZone_Reheat,
  SetpointManager_SingleZone_Heating,
  SetpointManager_SingleZone_Cooling,
  SetpointManager_SingleZone_Humidity_Minimum,
  SetpointManager_SingleZone_Humidity_Maximum,
  SetpointManager_MixedAir,
  SetpointManager_OutdoorAirPretreat,
  SetpointManager_Warmest,
  SetpointManager_Coldest,
  SetpointManager_ReturnAirBypassFlow,
  SetpointManager_WarmestTemperatureFlow,
  SetpointManager_MultiZone_Heating_Average,
  SetpointManager_MultiZone_Cooling_Average,
  SetpointManager_MultiZone_MinimumHumidity_Average,
  SetpointManager_MultiZone_MaximumHumidity_Average,
  SetpointManager_MultiZone_Humidity_Minimum,
  SetpointManager_MultiZone_Humidity_Maximum,
  SetpointManager_FollowOutdoorAirTemperature,
  SetpointManager_FollowSystemNodeTemperature,
  SetpointManager_FollowGroundTemperature,
  SetpointManager_CondenserEnteringReset,
  SetpointManager_CondenserEnteringReset_Ideal,
  SetpointManager_SingleZone_OneStageCooling,
  SetpointManager_SingleZone_OneStageHeating,
  SetpointManager_ReturnTemperature_ChilledWater,
  SetpointManager_ReturnTemperature_HotWater,
  SetpointManager_SystemNodeReset_Temperature,
  SetpointManager_SystemNodeReset_Humidity,
  Refrigeration_Case,
  Refrigeration_CompressorRack,
  Refrigeration_CaseAndWalkInList,
  Refrigeration_Condenser_AirCooled,
  Refrigeration_Condenser_EvaporativeCooled,
  Refrigeration_Condenser_WaterCooled,
  Refrigeration_Condenser_Cascade,
  Refrigeration_GasCooler_AirCooled,
  Refrigeration_TransferLoadList,
  Refrigeration_Subcooler,
  Refrigeration_Compressor,
  Refrigeration_CompressorList,
  Refrigeration_System,
  Refrigeration_TranscriticalSystem,
  Refrigeration_SecondarySystem,
  Refrigeration_WalkIn,
  Refrigeration_AirChiller,
  ZoneHVAC_RefrigerationChillerSet,
  DemandManagerAssignmentList,
  DemandManager_ExteriorLights,
  DemandManager_Lights,
  DemandManager_ElectricEquipment,
  DemandManager_Thermostats,
  DemandManager_Ventilation,
  Generator_InternalCombustionEngine,
  Generator_CombustionTurbine,
  Generator_MicroTurbine,
  Generator_Photovoltaic,
  PhotovoltaicPerformance_Simple,
  PhotovoltaicPerformance_EquivalentOneDiode,
  PhotovoltaicPerformance_Sandia,
  Generator_PVWatts,
  ElectricLoadCenter_Inverter_PVWatts,
  Generator_FuelCell,
  Generator_FuelCell_PowerModule,
  Generator_FuelCell_AirSupply,
  Generator_FuelCell_WaterSupply,
  Generator_FuelCell_AuxiliaryHeater,
  Generator_FuelCell_ExhaustGasToWaterHeatExchanger,
  Generator_FuelCell_ElectricalStorage,
  Generator_FuelCell_Inverter,
  Generator_FuelCell_StackCooler,
  Generator_MicroCHP,
  Generator_MicroCHP_NonNormalizedParameters,
  Generator_FuelSupply,
  Generator_WindTurbine,
  ElectricLoadCenter_Generators,
  ElectricLoadCenter_Inverter_Simple,
  ElectricLoadCenter_Inverter_FunctionOfPower,
  ElectricLoadCenter_Inverter_LookUpTable,
  ElectricLoadCenter_Storage_Simple,
  ElectricLoadCenter_Storage_Battery,
  ElectricLoadCenter_Storage_LiIonNMCBattery,
  ElectricLoadCenter_Transformer,
  ElectricLoadCenter_Distribution,
  ElectricLoadCenter_Storage_Converter,
  WaterUse_Equipment,
  WaterUse_Connections,
  WaterUse_Storage,
  WaterUse_Well,
  WaterUse_RainCollector,
  FaultModel_TemperatureSensorOffset_OutdoorAir,
  FaultModel_HumiditySensorOffset_OutdoorAir,
  FaultModel_EnthalpySensorOffset_OutdoorAir,
  FaultModel_TemperatureSensorOffset_ReturnAir,
  FaultModel_EnthalpySensorOffset_ReturnAir,
  FaultModel_TemperatureSensorOffset_ChillerSupplyWater,
  FaultModel_TemperatureSensorOffset_CoilSupplyAir,
  FaultModel_TemperatureSensorOffset_CondenserSupplyWater,
  FaultModel_ThermostatOffset,
  FaultModel_HumidistatOffset,
  FaultModel_Fouling_AirFilter,
  FaultModel_Fouling_Boiler,
  FaultModel_Fouling_EvaporativeCooler,
  FaultModel_Fouling_Chiller,
  FaultModel_Fouling_CoolingTower,
  FaultModel_Fouling_Coil,
  Matrix_TwoDimension,
  HybridModel_Zone,
  Curve_Linear,
  Curve_QuadLinear,
  Curve_QuintLinear,
  Curve_Quadratic,
  Curve_Cubic,
  Curve_Quartic,
  Curve_Exponent,
  Curve_Bicubic,
  Curve_Biquadratic,
  Curve_QuadraticLinear,
  Curve_CubicLinear,
  Curve_Triquadratic,
  Curve_Functional_PressureDrop,
  Curve_FanPressureRise,
  Curve_ExponentialSkewNormal,
  Curve_Sigmoid,
  Curve_RectangularHyperbola1,
  Curve_RectangularHyperbola2,
  Curve_ExponentialDecay,
  Curve_DoubleExponentialDecay,
  Curve_ChillerPartLoadWithLift,
  Table_IndependentVariable,
  Table_IndependentVariableList,
  Table_Lookup,
  FluidProperties_Name,
  FluidProperties_GlycolConcentration,
  FluidProperties_Temperatures,
  FluidProperties_Saturated,
  FluidProperties_Superheated,
  FluidProperties_Concentration,
  CurrencyType,
  ComponentCost_Adjustments,
  ComponentCost_Reference,
  ComponentCost_LineItem,
  UtilityCost_Tariff,
  UtilityCost_Qualify,
  UtilityCost_Charge_Simple,
  UtilityCost_Charge_Block,
  UtilityCost_Ratchet,
  UtilityCost_Variable,
  UtilityCost_Computation,
  LifeCycleCost_Parameters,
  LifeCycleCost_RecurringCosts,
  LifeCycleCost_NonrecurringCost,
  LifeCycleCost_UsePriceEscalation,
  LifeCycleCost_UseAdjustment,
  Parametric_SetValueForRun,
  Parametric_Logic,
  Parametric_RunControl,
  Parametric_FileNameSuffix,
  Output_VariableDictionary,
  Output_Surfaces_List,
  Output_Surfaces_Drawing,
  Output_Schedules,
  Output_Constructions,
  Output_EnergyManagementSystem,
  OutputControl_SurfaceColorScheme,
  Output_Table_SummaryReports,
  Output_Table_TimeBins,
  Output_Table_Monthly,
  Output_Table_Annual,
  OutputControl_Table_Style,
  OutputControl_ReportingTolerances,
  Output_Variable,
  Output_Meter,
  Output_Meter_MeterFileOnly,
  Output_Meter_Cumulative,
  Output_Meter_Cumulative_MeterFileOnly,
  Meter_Custom,
  Meter_CustomDecrement,
  OutputControl_Files,
  OutputControl_Timestamp,
  Output_JSON,
  Output_SQLite,
  Output_EnvironmentalImpactFactors,
  EnvironmentalImpactFactors,
  FuelFactors,
  Output_Diagnostics,
  Output_DebuggingData,
  Output_PreprocessorMessage,
  PythonPlugin_SearchPaths,
  PythonPlugin_Instance,
  PythonPlugin_Variables,
  PythonPlugin_TrendVariable,
  PythonPlugin_OutputVariable,
  OS_Version,
  OS_ComponentData,
  OS_ConvergenceLimits,
  OS_ProgramControl,
  OS_HeatBalanceAlgorithm,
  OS_RunPeriod,
  OS_RunPeriodControl_DaylightSavingTime,
  OS_RunPeriodControl_SpecialDays,
  OS_ShadowCalculation,
  OS_SimulationControl,
  OS_PerformancePrecisionTradeoffs,
  OS_Sizing_Parameters,
  OS_SurfaceConvectionAlgorithm_Inside,
  OS_SurfaceConvectionAlgorithm_Outside,
  OS_Timestep,
  OS_YearDescription,
  OS_ZoneAirContaminantBalance,
  OS_ZoneAirHeatBalanceAlgorithm,
  OS_ZoneAirMassFlowConservation,
  OS_ZoneCapacitanceMultiplier_ResearchSpecial,
  OS_RadianceParameters,
  OS_ClimateZones,
  OS_Site,
  OS_Site_GroundTemperature_BuildingSurface,
  OS_Site_GroundTemperature_FCfactorMethod,
  OS_Site_GroundTemperature_Shallow,
  OS_Site_GroundTemperature_Deep,
  OS_Site_GroundTemperature_Undisturbed_KusudaAchenbach,
  OS_Site_GroundTemperature_Undisturbed_Xing,
  OS_Site_GroundReflectance,
  OS_Site_WaterMainsTemperature,
  OS_SizingPeriod_DesignDay,
  OS_SizingPeriod_WeatherFileConditionType,
  OS_SizingPeriod_WeatherFileDays,
  OS_WeatherFile,
  OS_WeatherProperty_SkyTemperature,
  OS_AdditionalProperties,
  OS_BuildingStory,
  OS_BuildingUnit,
  OS_DefaultConstructionSet,
  OS_DefaultScheduleSet,
  OS_DefaultSubSurfaceConstructions,
  OS_DefaultSurfaceConstructions,
  OS_Rendering_Color,
  OS_SpaceType,
  OS_ModelObjectList,
  OS_Material,
  OS_Material_AirGap,
  OS_Material_InfraredTransparent,
  OS_Material_NoMass,
  OS_Material_RoofVegetation,
  OS_WindowMaterial_Blind,
  OS_WindowMaterial_DaylightRedirectionDevice,
  OS_WindowMaterial_Gas,
  OS_WindowMaterial_GasMixture,
  OS_WindowMaterial_Glazing,
  OS_WindowMaterial_GlazingGroup_Thermochromic,
  OS_WindowMaterial_Glazing_RefractionExtinctionMethod,
  OS_WindowMaterial_Screen,
  OS_WindowMaterial_Shade,
  OS_WindowMaterial_SimpleGlazingSystem,
  OS_StandardsInformation_Material,
  OS_MaterialProperty_GlazingSpectralData,
  OS_MaterialProperty_MoisturePenetrationDepth_Settings,
  OS_MaterialProperty_PhaseChange,
  OS_MaterialProperty_PhaseChangeHysteresis,
  OS_Construction,
  OS_Construction_CfactorUndergroundWall,
  OS_Construction_FfactorGroundFloor,
  OS_Construction_InternalSource,
  OS_Construction_AirBoundary,
  OS_Construction_WindowDataFile,
  OS_StandardsInformation_Construction,
  OS_InternalMass_Definition,
  OS_People_Definition,
  OS_Lights_Definition,
  OS_Luminaire_Definition,
  OS_ElectricEquipment_Definition,
  OS_ElectricEquipment_ITE_AirCooled_Definition,
  OS_GasEquipment_Definition,
  OS_HotWaterEquipment_Definition,
  OS_SteamEquipment_Definition,
  OS_OtherEquipment_Definition,
  OS_Exterior_Lights_Definition,
  OS_Exterior_FuelEquipment_Definition,
  OS_Exterior_WaterEquipment_Definition,
  OS_Schedule_Compact,
  OS_Schedule_Day,
  OS_Schedule_Week,
  OS_Schedule_Year,
  OS_Schedule_Rule,
  OS_Schedule_Ruleset,
  OS_Schedule_Constant,
  OS_Schedule_FixedInterval,
  OS_Schedule_VariableInterval,
  OS_ScheduleTypeLimits,
  OS_External_File,
  OS_Schedule_File,
  OS_Building,
  OS_DaylightingDevice_Tubular,
  OS_DaylightingDevice_Shelf,
  OS_DaylightingDevice_LightWell,
  OS_Facility,
  OS_InteriorPartitionSurface,
  OS_InteriorPartitionSurfaceGroup,
  OS_ShadingSurface,
  OS_ShadingSurfaceGroup,
  OS_Space,
  OS_Surface,
  OS_SubSurface,
  OS_ShadingControl,
  OS_WindowProperty_FrameAndDivider,
  OS_SurfaceControl_MovableInsulation,
  OS_SurfaceProperty_ConvectionCoefficients,
  OS_SurfaceProperty_ConvectionCoefficients_MultipleSurface,
  OS_SurfaceProperty_ExposedFoundationPerimeter,
  OS_SurfaceProperty_OtherSideCoefficients,
  OS_SurfaceProperty_OtherSideConditionsModel,
  OS_Foundation_Kiva,
  OS_Foundation_Kiva_Settings,
  OS_ZoneProperty_UserViewFactors_BySurfaceName,
  OS_SurfaceProperty_SurroundingSurfaces,
  OS_SurfaceProperty_LocalEnvironment,
  OS_SurfaceProperty_GroundSurfaces,
  OS_SurfaceProperty_IncidentSolarMultiplier,
  OS_InternalMass,
  OS_People,
  OS_Lights,
  OS_Luminaire,
  OS_ElectricEquipment,
  OS_ElectricEquipment_ITE_AirCooled,
  OS_GasEquipment,
  OS_HotWaterEquipment,
  OS_SteamEquipment,
  OS_OtherEquipment,
  OS_SpaceInfiltration_DesignFlowRate,
  OS_SpaceInfiltration_EffectiveLeakageArea,
  OS_SpaceInfiltration_FlowCoefficient,
  OS_Exterior_Lights,
  OS_Exterior_FuelEquipment,
  OS_Exterior_WaterEquipment,
  OS_Daylighting_Control,
  OS_Glare_Sensor,
  OS_IlluminanceMap,
  OS_LightingDesignDay,
  OS_LightingSimulationControl,
  OS_LightingSimulationZone,
  OS_Refrigeration_Case,
  OS_Refrigeration_Condenser_AirCooled,
  OS_Refrigeration_Condenser_EvaporativeCooled,
  OS_Refrigeration_Condenser_WaterCooled,
  OS_Refrigeration_Condenser_Cascade,
  OS_Refrigeration_Compressor,
  OS_Refrigeration_CompressorRack,
  OS_Refrigeration_DefrostCycleParameters,
  OS_Refrigeration_GasCooler_AirCooled,
  OS_Refrigeration_Subcooler_Mechanical,
  OS_Refrigeration_Subcooler_LiquidSuction,
  OS_Refrigeration_System,
  OS_Refrigeration_TranscriticalSystem,
  OS_Refrigeration_SecondarySystem,
  OS_Refrigeration_WalkIn,
  OS_Refrigeration_WalkIn_ZoneBoundary,
  OS_Refrigeration_AirChiller,
  OS_AirConditioner_VariableRefrigerantFlow,
  OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl,
  OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_LoadingIndex,
  OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR,
  OS_AirLoopHVAC,
  OS_AirLoopHVAC_OutdoorAirSystem,
  OS_AirLoopHVAC_DedicatedOutdoorAirSystem,
  OS_AirLoopHVAC_ReturnPlenum,
  OS_AirLoopHVAC_SupplyPlenum,
  OS_AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed,
  OS_AirLoopHVAC_UnitaryCoolOnly,
  OS_AirLoopHVAC_UnitaryHeatPump_AirToAir,
  OS_AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass,
  OS_AirLoopHVAC_UnitarySystem,
  OS_UnitarySystemPerformance_Multispeed,
  OS_AirLoopHVAC_ZoneMixer,
  OS_AirLoopHVAC_ZoneSplitter,
  OS_AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat,
  OS_AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat,
  OS_AirTerminal_SingleDuct_InletSideMixer,
  OS_AirTerminal_SingleDuct_ConstantVolume_Reheat,
  OS_AirTerminal_SingleDuct_SeriesPIU_Reheat,
  OS_AirTerminal_SingleDuct_ParallelPIU_Reheat,
  OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction,
  OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam,
  OS_AirTerminal_SingleDuct_ConstantVolume_CooledBeam,
  OS_AirTerminal_SingleDuct_ConstantVolume_NoReheat,
  OS_AirTerminal_SingleDuct_VAV_NoReheat,
  OS_AirTerminal_SingleDuct_VAV_Reheat,
  OS_AirTerminal_DualDuct_VAV_OutdoorAir,
  OS_AirTerminal_DualDuct_ConstantVolume,
  OS_AirTerminal_DualDuct_VAV,
  OS_AvailabilityManagerAssignmentList,
  OS_AvailabilityManager_Scheduled,
  OS_AvailabilityManager_ScheduledOn,
  OS_AvailabilityManager_ScheduledOff,
  OS_AvailabilityManager_NightCycle,
  OS_AvailabilityManager_OptimumStart,
  OS_AvailabilityManager_DifferentialThermostat,
  OS_AvailabilityManager_HighTemperatureTurnOff,
  OS_AvailabilityManager_HighTemperatureTurnOn,
  OS_AvailabilityManager_LowTemperatureTurnOff,
  OS_AvailabilityManager_LowTemperatureTurnOn,
  OS_AvailabilityManager_NightVentilation,
  OS_AvailabilityManager_HybridVentilation,
  OS_Boiler_HotWater,
  OS_Boiler_Steam,
  OS_HeatPump_WaterToWater_EquationFit_Heating,
  OS_HeatPump_WaterToWater_EquationFit_Cooling,
  OS_HeatPump_PlantLoop_EIR_Heating,
  OS_HeatPump_PlantLoop_EIR_Cooling,
  OS_HeatPump_AirToWater_FuelFired_Heating,
  OS_HeatPump_AirToWater_FuelFired_Cooling,
  OS_Chiller_Electric_EIR,
  OS_Chiller_Electric_ReformulatedEIR,
  OS_Chiller_Electric_ASHRAE205,
  OS_CentralHeatPumpSystem,
  OS_CentralHeatPumpSystem_Module,
  OS_ChillerHeaterPerformance_Electric_EIR,
  OS_Chiller_Absorption_Indirect,
  OS_Chiller_Absorption,
  OS_CoolingTower_SingleSpeed,
  OS_CoolingTower_TwoSpeed,
  OS_CoolingTower_VariableSpeed,
  OS_CoolingTowerPerformance_CoolTools,
  OS_CoolingTowerPerformance_YorkCalc,
  OS_DistrictCooling,
  OS_DistrictHeating,
  OS_GroundHeatExchanger_Vertical,
  OS_GroundHeatExchanger_HorizontalTrench,
  OS_Coil_Cooling_CooledBeam,
  OS_Coil_Cooling_FourPipeBeam,
  OS_Coil_Heating_FourPipeBeam,
  OS_Coil_Cooling_DX,
  OS_Coil_Cooling_DX_CurveFit_Performance,
  OS_Coil_Cooling_DX_CurveFit_OperatingMode,
  OS_Coil_Cooling_DX_CurveFit_Speed,
  OS_Coil_Cooling_DX_SingleSpeed,
  OS_Coil_Cooling_DX_TwoStageWithHumidityControlMode,
  OS_CoilPerformance_DX_Cooling,
  OS_Coil_Cooling_DX_MultiSpeed,
  OS_Coil_Cooling_DX_MultiSpeed_StageData,
  OS_Coil_Cooling_DX_TwoSpeed,
  OS_CoilSystem_Cooling_Water_HeatExchangerAssisted,
  OS_CoilSystem_Cooling_DX_HeatExchangerAssisted,
  OS_CoilSystem_IntegratedHeatPump_AirSource,
  OS_Coil_Cooling_DX_VariableRefrigerantFlow,
  OS_Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  OS_Coil_Cooling_LowTemperatureRadiant_ConstantFlow,
  OS_Coil_Cooling_LowTemperatureRadiant_VariableFlow,
  OS_Coil_Cooling_Water,
  OS_Coil_Cooling_WaterToAirHeatPump_EquationFit,
  OS_Coil_Heating_DX_SingleSpeed,
  OS_Coil_Heating_DX_VariableRefrigerantFlow,
  OS_Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  OS_Coil_Heating_Electric,
  OS_Coil_Heating_Electric_MultiStage,
  OS_Coil_Heating_Electric_MultiStage_StageData,
  OS_Coil_Heating_Gas,
  OS_Coil_Heating_Gas_MultiStage,
  OS_Coil_Heating_Gas_MultiStage_StageData,
  OS_Coil_Heating_Desuperheater,
  OS_Coil_WaterHeating_Desuperheater,
  OS_Coil_Heating_LowTemperatureRadiant_ConstantFlow,
  OS_Coil_Heating_LowTemperatureRadiant_VariableFlow,
  OS_Coil_Heating_Water,
  OS_Coil_Heating_Water_Baseboard,
  OS_Coil_Heating_Water_Baseboard_Radiant,
  OS_Coil_Heating_WaterToAirHeatPump_EquationFit,
  OS_Coil_Heating_DX_MultiSpeed,
  OS_Coil_Heating_DX_MultiSpeed_StageData,
  OS_Coil_Heating_DX_VariableSpeed,
  OS_Coil_Heating_DX_VariableSpeed_SpeedData,
  OS_Coil_Cooling_DX_VariableSpeed,
  OS_Coil_Cooling_DX_VariableSpeed_SpeedData,
  OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit,
  OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData,
  OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit,
  OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData,
  OS_Coil_Cooling_Water_Panel_Radiant,
  OS_Connection,
  OS_Connector_Mixer,
  OS_Connector_Splitter,
  OS_Controller_MechanicalVentilation,
  OS_ZoneControl_ContaminantController,
  OS_Controller_OutdoorAir,
  OS_Controller_WaterCoil,
  OS_Curve_Linear,
  OS_Curve_QuadLinear,
  OS_Curve_QuintLinear,
  OS_Curve_Quadratic,
  OS_Curve_Cubic,
  OS_Curve_Quartic,
  OS_Curve_Exponent,
  OS_Curve_Bicubic,
  OS_Curve_Biquadratic,
  OS_Curve_QuadraticLinear,
  OS_Curve_Triquadratic,
  OS_Curve_Functional_PressureDrop,
  OS_Curve_FanPressureRise,
  OS_Curve_ExponentialSkewNormal,
  OS_Curve_Sigmoid,
  OS_Curve_RectangularHyperbola1,
  OS_Curve_RectangularHyperbola2,
  OS_Curve_ExponentialDecay,
  OS_Curve_DoubleExponentialDecay,
  OS_Table_MultiVariableLookup,
  OS_Table_IndependentVariable,
  OS_Table_Lookup,
  OS_DesignSpecification_OutdoorAir,
  OS_DesignSpecification_ZoneAirDistribution,
  OS_Humidifier_Steam_Electric,
  OS_Humidifier_Steam_Gas,
  OS_EvaporativeCooler_Indirect_ResearchSpecial,
  OS_EvaporativeCooler_Direct_ResearchSpecial,
  OS_EvaporativeFluidCooler_SingleSpeed,
  OS_EvaporativeFluidCooler_TwoSpeed,
  OS_Fan_SystemModel,
  OS_Fan_ConstantVolume,
  OS_Fan_VariableVolume,
  OS_Fan_OnOff,
  OS_Fan_ComponentModel,
  OS_Fan_ZoneExhaust,
  OS_HeatExchanger_AirToAir_SensibleAndLatent,
  OS_HeatExchanger_Desiccant_BalancedFlow,
  OS_HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1,
  OS_HeatExchanger_FluidToFluid,
  OS_HVACComponentList,
  OS_Node,
  OS_PortList,
  OS_Pipe_Adiabatic,
  OS_PlantLoop,
  OS_PlantEquipmentOperation_CoolingLoad,
  OS_PlantEquipmentOperation_HeatingLoad,
  OS_PlantEquipmentOperation_OutdoorDryBulb,
  OS_PlantEquipmentOperation_OutdoorWetBulb,
  OS_PlantEquipmentOperation_OutdoorRelativeHumidity,
  OS_PlantEquipmentOperation_OutdoorDewpoint,
  OS_PlantEquipmentOperation_OutdoorDryBulbDifference,
  OS_PlantEquipmentOperation_OutdoorDewpointDifference,
  OS_PlantEquipmentOperation_OutdoorWetBulbDifference,
  OS_PlantComponent_TemperatureSource,
  OS_LoadProfile_Plant,
  OS_Pump_ConstantSpeed,
  OS_Pump_VariableSpeed,
  OS_HeaderedPumps_ConstantSpeed,
  OS_HeaderedPumps_VariableSpeed,
  OS_TemperingValve,
  OS_SetpointManager_FollowOutdoorAirTemperature,
  OS_SetpointManager_MixedAir,
  OS_SetpointManager_OutdoorAirReset,
  OS_SetpointManager_OutdoorAirPretreat,
  OS_SetpointManager_SingleZone_Reheat,
  OS_SetpointManager_SingleZone_Humidity_Minimum,
  OS_SetpointManager_MultiZone_MinimumHumidity_Average,
  OS_SetpointManager_MultiZone_Humidity_Minimum,
  OS_SetpointManager_Scheduled,
  OS_SetpointManager_Scheduled_DualSetpoint,
  OS_SetpointManager_Warmest,
  OS_SetpointManager_WarmestTemperatureFlow,
  OS_SetpointManager_Coldest,
  OS_SetpointManager_FollowGroundTemperature,
  OS_SetpointManager_FollowSystemNodeTemperature,
  OS_SetpointManager_MultiZone_Cooling_Average,
  OS_SetpointManager_MultiZone_Heating_Average,
  OS_SetpointManager_MultiZone_Humidity_Maximum,
  OS_SetpointManager_MultiZone_MaximumHumidity_Average,
  OS_SetpointManager_SingleZone_OneStageCooling,
  OS_SetpointManager_SingleZone_OneStageHeating,
  OS_SetpointManager_SingleZone_Humidity_Maximum,
  OS_SetpointManager_SingleZone_Cooling,
  OS_SetpointManager_SingleZone_Heating,
  OS_SetpointManager_SystemNodeReset_Temperature,
  OS_SetpointManager_SystemNodeReset_Humidity,
  OS_Sizing_Plant,
  OS_Sizing_System,
  OS_Sizing_Zone,
  OS_Splitter,
  OS_ThermalZone,
  OS_ZoneControl_Humidistat,
  OS_ThermostatSetpoint_DualSetpoint,
  OS_ZoneControl_Thermostat_StagedDualSetpoint,
  OS_ZoneHVAC_EquipmentList,
  OS_ZoneVentilation_DesignFlowRate,
  OS_ZoneVentilation_WindandStackOpenArea,
  OS_ZoneHVAC_Baseboard_Convective_Electric,
  OS_ZoneHVAC_Baseboard_Convective_Water,
  OS_ZoneHVAC_HighTemperatureRadiant,
  OS_ZoneHVAC_LowTemperatureRadiant_Electric,
  OS_ZoneHVAC_IdealLoadsAirSystem,
  OS_ZoneHVAC_FourPipeFanCoil,
  OS_ZoneHVAC_LowTemperatureRadiant_ConstantFlow,
  OS_ZoneHVAC_LowTemperatureRadiant_VariableFlow,
  OS_ZoneHVAC_PackagedTerminalHeatPump,
  OS_ZoneHVAC_PackagedTerminalAirConditioner,
  OS_ZoneHVAC_TerminalUnit_VariableRefrigerantFlow,
  OS_ZoneHVAC_WaterToAirHeatPump,
  OS_ZoneHVAC_UnitHeater,
  OS_ZoneHVAC_Dehumidifier_DX,
  OS_ZoneHVAC_EnergyRecoveryVentilator,
  OS_ZoneHVAC_EnergyRecoveryVentilator_Controller,
  OS_ZoneHVAC_UnitVentilator,
  OS_ZoneHVAC_Baseboard_RadiantConvective_Water,
  OS_ZoneHVAC_Baseboard_RadiantConvective_Electric,
  OS_ZoneHVAC_CoolingPanel_RadiantConvective_Water,
  OS_ZoneMixing,
  OS_WaterHeater_Mixed,
  OS_WaterHeater_HeatPump,
  OS_Coil_WaterHeating_AirToWaterHeatPump,
  OS_WaterHeater_HeatPump_WrappedCondenser,
  OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed,
  OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed_SpeedData,
  OS_Coil_WaterHeating_AirToWaterHeatPump_Wrapped,
  OS_WaterHeater_Stratified,
  OS_WaterHeater_Sizing,
  OS_WaterUse_Equipment,
  OS_WaterUse_Equipment_Definition,
  OS_WaterUse_Connections,
  OS_ThermalStorage_Ice_Detailed,
  OS_ThermalStorage_ChilledWater_Stratified,
  OS_Duct,
  OS_FluidCooler_SingleSpeed,
  OS_FluidCooler_TwoSpeed,
  OS_Pipe_Indoor,
  OS_Pipe_Outdoor,
  OS_SwimmingPool_Indoor,
  OS_SolarCollectorPerformance_FlatPlate,
  OS_SolarCollector_FlatPlate_Water,
  OS_SolarCollectorPerformance_IntegralCollectorStorage,
  OS_SolarCollector_IntegralCollectorStorage,
  OS_SolarCollectorPerformance_PhotovoltaicThermal_Simple,
  OS_SolarCollectorPerformance_PhotovoltaicThermal_BIPVT,
  OS_SolarCollector_FlatPlate_PhotovoltaicThermal,
  OS_Generator_MicroTurbine,
  OS_Generator_MicroTurbine_HeatRecovery,
  OS_Generator_Photovoltaic,
  OS_PhotovoltaicPerformance_Simple,
  OS_PhotovoltaicPerformance_EquivalentOneDiode,
  OS_PhotovoltaicPerformance_Sandia,
  OS_Generator_FuelCell,
  OS_Generator_FuelCell_PowerModule,
  OS_Generator_FuelCell_AirSupply,
  OS_Generator_FuelCell_WaterSupply,
  OS_Generator_FuelCell_AuxiliaryHeater,
  OS_Generator_FuelCell_ExhaustGasToWaterHeatExchanger,
  OS_Generator_FuelCell_ElectricalStorage,
  OS_Generator_FuelCell_Inverter,
  OS_Generator_FuelCell_StackCooler,
  OS_Generator_FuelSupply,
  OS_Generator_WindTurbine,
  OS_Generator_PVWatts,
  OS_ElectricLoadCenter_Inverter_PVWatts,
  OS_ElectricLoadCenter_Inverter_Simple,
  OS_ElectricLoadCenter_Inverter_LookUpTable,
  OS_ElectricLoadCenter_Storage_Simple,
  OS_ElectricLoadCenter_Storage_LiIonNMCBattery,
  OS_ElectricLoadCenter_Transformer,
  OS_ElectricLoadCenter_Distribution,
  OS_ElectricLoadCenter_Storage_Converter,
  OS_ComponentCost_Adjustments,
  OS_LifeCycleCost,
  OS_CurrencyType,
  OS_LifeCycleCost_Parameters,
  OS_LifeCycleCost_UsePriceEscalation,
  OS_UtilityBill,
  OS_UtilityCost_Tariff,
  OS_UtilityCost_Qualify,
  OS_UtilityCost_Charge_Simple,
  OS_UtilityCost_Charge_Block,
  OS_UtilityCost_Ratchet,
  OS_UtilityCost_Variable,
  OS_UtilityCost_Computation,
  OS_Output_Schedules,
  OS_Output_Constructions,
  OS_OutputControl_Files,
  OS_OutputControl_Timestamp,
  OS_OutputControl_ReportingTolerances,
  OS_OutputControl_Table_Style,
  OS_Output_Meter,
  OS_Output_Variable,
  OS_Meter_Custom,
  OS_Meter_CustomDecrement,
  OS_Output_EnvironmentalImpactFactors,
  OS_EnvironmentalImpactFactors,
  OS_FuelFactors,
  OS_Output_Diagnostics,
  OS_Output_DebuggingData,
  OS_Output_JSON,
  OS_Output_SQLite,
  OS_Output_EnergyManagementSystem,
  OS_Output_Table_SummaryReports,
  OS_EnergyManagementSystem_Sensor,
  OS_EnergyManagementSystem_Actuator,
  OS_EnergyManagementSystem_ProgramCallingManager,
  OS_EnergyManagementSystem_Program,
  OS_EnergyManagementSystem_Subroutine,
  OS_EnergyManagementSystem_GlobalVariable,
  OS_EnergyManagementSystem_OutputVariable,
  OS_EnergyManagementSystem_MeteredOutputVariable,
  OS_EnergyManagementSystem_TrendVariable,
  OS_EnergyManagementSystem_InternalVariable,
  OS_EnergyManagementSystem_CurveOrTableIndexVariable,
  OS_EnergyManagementSystem_ConstructionIndexVariable,
  OS_PythonPlugin_Instance,
  OS_PythonPlugin_Variable,
  OS_PythonPlugin_TrendVariable,
  OS_PythonPlugin_OutputVariable,
  OS_AirflowNetworkSimulationControl,
  OS_AirflowNetworkZone,
  OS_AirflowNetworkSurface,
  OS_AirflowNetworkReferenceCrackConditions,
  OS_AirflowNetworkCrack,
  OS_AirflowNetworkEffectiveLeakageArea,
  OS_AirflowNetworkSpecifiedFlowRate,
  OS_AirflowNetworkDetailedOpening,
  OS_AirflowNetworkSimpleOpening,
  OS_AirflowNetworkHorizontalOpening,
  OS_AirflowNetworkOutdoorAirflow,
  OS_AirflowNetworkZoneExhaustFan,
  OS_AirflowNetworkExternalNode,
  OS_AirflowNetworkPressureController,
  OS_AirflowNetworkDistributionNode,
  OS_AirflowNetworkLeak,
  OS_AirflowNetworkLeakageRatio,
  OS_AirflowNetworkDuct,
  OS_AirflowNetworkFan,
  OS_AirflowNetworkEquivalentDuct,
  OS_AirflowNetworkConstantPressureDrop,
  OS_AirflowNetworkReliefAirFlow,
  OS_AirflowNetworkDistributionLinkage,
  OS_AirflowNetworkDuctViewFactors,
  OS_AirflowNetworkOccupantVentilationControl,
  OS_ExternalInterface,
  OS_ExternalInterface_Schedule,
  OS_ExternalInterface_Variable,
  OS_ExternalInterface_Actuator,
  OS_ExternalInterface_FunctionalMockupUnitImport,
  OS_ExternalInterface_FunctionalMockupUnitImport_From_Variable,
  OS_ExternalInterface_FunctionalMockupUnitImport_To_Schedule,
  OS_ExternalInterface_FunctionalMockupUnitImport_To_Actuator,
  OS_ExternalInterface_FunctionalMockupUnitImport_To_Variable,
  OS_ExternalInterface_FunctionalMockupUnitExport_From_Variable,
  OS_ExternalInterface_FunctionalMockupUnitExport_To_Schedule,
  OS_ExternalInterface_FunctionalMockupUnitExport_To_Actuator,
  OS_ExternalInterface_FunctionalMockupUnitExport_To_Variable,
  OS_Coil_UserDefined,
  OS_PlantComponent_UserDefined,
  CommentOnly,
};

IddObjectType::VecType IddObjectType::buildStringVec(bool isd) {
  struct evalue
  {
    int value;
    const char* name;
    const char* description;
  };
  const evalue a[] = {
    {IddObjectType::Catchall, "Catchall", ""},
    {IddObjectType::UserCustom, "UserCustom", ""},
    {IddObjectType::LeadInput, "LeadInput", "Lead Input"},
    {IddObjectType::EndLeadInput, "EndLeadInput", "End Lead Input"},
    {IddObjectType::SimulationData, "SimulationData", "Simulation Data"},
    {IddObjectType::EndSimulationData, "EndSimulationData", "End Simulation Data"},
    {IddObjectType::Version, "Version", "Version"},
    {IddObjectType::SimulationControl, "SimulationControl", "SimulationControl"},
    {IddObjectType::PerformancePrecisionTradeoffs, "PerformancePrecisionTradeoffs", "PerformancePrecisionTradeoffs"},
    {IddObjectType::Building, "Building", "Building"},
    {IddObjectType::ShadowCalculation, "ShadowCalculation", "ShadowCalculation"},
    {IddObjectType::SurfaceConvectionAlgorithm_Inside, "SurfaceConvectionAlgorithm_Inside", "SurfaceConvectionAlgorithm:Inside"},
    {IddObjectType::SurfaceConvectionAlgorithm_Outside, "SurfaceConvectionAlgorithm_Outside", "SurfaceConvectionAlgorithm:Outside"},
    {IddObjectType::HeatBalanceAlgorithm, "HeatBalanceAlgorithm", "HeatBalanceAlgorithm"},
    {IddObjectType::HeatBalanceSettings_ConductionFiniteDifference, "HeatBalanceSettings_ConductionFiniteDifference",
     "HeatBalanceSettings:ConductionFiniteDifference"},
    {IddObjectType::ZoneAirHeatBalanceAlgorithm, "ZoneAirHeatBalanceAlgorithm", "ZoneAirHeatBalanceAlgorithm"},
    {IddObjectType::ZoneAirContaminantBalance, "ZoneAirContaminantBalance", "ZoneAirContaminantBalance"},
    {IddObjectType::ZoneAirMassFlowConservation, "ZoneAirMassFlowConservation", "ZoneAirMassFlowConservation"},
    {IddObjectType::ZoneCapacitanceMultiplier_ResearchSpecial, "ZoneCapacitanceMultiplier_ResearchSpecial",
     "ZoneCapacitanceMultiplier:ResearchSpecial"},
    {IddObjectType::Timestep, "Timestep", "Timestep"},
    {IddObjectType::ConvergenceLimits, "ConvergenceLimits", "ConvergenceLimits"},
    {IddObjectType::HVACSystemRootFindingAlgorithm, "HVACSystemRootFindingAlgorithm", "HVACSystemRootFindingAlgorithm"},
    {IddObjectType::Compliance_Building, "Compliance_Building", "Compliance:Building"},
    {IddObjectType::Site_Location, "Site_Location", "Site:Location"},
    {IddObjectType::Site_VariableLocation, "Site_VariableLocation", "Site:VariableLocation"},
    {IddObjectType::SizingPeriod_DesignDay, "SizingPeriod_DesignDay", "SizingPeriod:DesignDay"},
    {IddObjectType::SizingPeriod_WeatherFileDays, "SizingPeriod_WeatherFileDays", "SizingPeriod:WeatherFileDays"},
    {IddObjectType::SizingPeriod_WeatherFileConditionType, "SizingPeriod_WeatherFileConditionType", "SizingPeriod:WeatherFileConditionType"},
    {IddObjectType::RunPeriod, "RunPeriod", "RunPeriod"},
    {IddObjectType::RunPeriodControl_SpecialDays, "RunPeriodControl_SpecialDays", "RunPeriodControl:SpecialDays"},
    {IddObjectType::RunPeriodControl_DaylightSavingTime, "RunPeriodControl_DaylightSavingTime", "RunPeriodControl:DaylightSavingTime"},
    {IddObjectType::WeatherProperty_SkyTemperature, "WeatherProperty_SkyTemperature", "WeatherProperty:SkyTemperature"},
    {IddObjectType::Site_WeatherStation, "Site_WeatherStation", "Site:WeatherStation"},
    {IddObjectType::Site_HeightVariation, "Site_HeightVariation", "Site:HeightVariation"},
    {IddObjectType::Site_GroundTemperature_BuildingSurface, "Site_GroundTemperature_BuildingSurface", "Site:GroundTemperature:BuildingSurface"},
    {IddObjectType::Site_GroundTemperature_FCfactorMethod, "Site_GroundTemperature_FCfactorMethod", "Site:GroundTemperature:FCfactorMethod"},
    {IddObjectType::Site_GroundTemperature_Shallow, "Site_GroundTemperature_Shallow", "Site:GroundTemperature:Shallow"},
    {IddObjectType::Site_GroundTemperature_Deep, "Site_GroundTemperature_Deep", "Site:GroundTemperature:Deep"},
    {IddObjectType::Site_GroundTemperature_Undisturbed_FiniteDifference, "Site_GroundTemperature_Undisturbed_FiniteDifference",
     "Site:GroundTemperature:Undisturbed:FiniteDifference"},
    {IddObjectType::Site_GroundTemperature_Undisturbed_KusudaAchenbach, "Site_GroundTemperature_Undisturbed_KusudaAchenbach",
     "Site:GroundTemperature:Undisturbed:KusudaAchenbach"},
    {IddObjectType::Site_GroundTemperature_Undisturbed_Xing, "Site_GroundTemperature_Undisturbed_Xing", "Site:GroundTemperature:Undisturbed:Xing"},
    {IddObjectType::Site_GroundDomain_Slab, "Site_GroundDomain_Slab", "Site:GroundDomain:Slab"},
    {IddObjectType::Site_GroundDomain_Basement, "Site_GroundDomain_Basement", "Site:GroundDomain:Basement"},
    {IddObjectType::Site_GroundReflectance, "Site_GroundReflectance", "Site:GroundReflectance"},
    {IddObjectType::Site_GroundReflectance_SnowModifier, "Site_GroundReflectance_SnowModifier", "Site:GroundReflectance:SnowModifier"},
    {IddObjectType::Site_WaterMainsTemperature, "Site_WaterMainsTemperature", "Site:WaterMainsTemperature"},
    {IddObjectType::Site_Precipitation, "Site_Precipitation", "Site:Precipitation"},
    {IddObjectType::RoofIrrigation, "RoofIrrigation", "RoofIrrigation"},
    {IddObjectType::Site_SolarAndVisibleSpectrum, "Site_SolarAndVisibleSpectrum", "Site:SolarAndVisibleSpectrum"},
    {IddObjectType::Site_SpectrumData, "Site_SpectrumData", "Site:SpectrumData"},
    {IddObjectType::ScheduleTypeLimits, "ScheduleTypeLimits", "ScheduleTypeLimits"},
    {IddObjectType::Schedule_Day_Hourly, "Schedule_Day_Hourly", "Schedule:Day:Hourly"},
    {IddObjectType::Schedule_Day_Interval, "Schedule_Day_Interval", "Schedule:Day:Interval"},
    {IddObjectType::Schedule_Day_List, "Schedule_Day_List", "Schedule:Day:List"},
    {IddObjectType::Schedule_Week_Daily, "Schedule_Week_Daily", "Schedule:Week:Daily"},
    {IddObjectType::Schedule_Week_Compact, "Schedule_Week_Compact", "Schedule:Week:Compact"},
    {IddObjectType::Schedule_Year, "Schedule_Year", "Schedule:Year"},
    {IddObjectType::Schedule_Compact, "Schedule_Compact", "Schedule:Compact"},
    {IddObjectType::Schedule_Constant, "Schedule_Constant", "Schedule:Constant"},
    {IddObjectType::Schedule_File_Shading, "Schedule_File_Shading", "Schedule:File:Shading"},
    {IddObjectType::Schedule_File, "Schedule_File", "Schedule:File"},
    {IddObjectType::Material, "Material", "Material"},
    {IddObjectType::Material_NoMass, "Material_NoMass", "Material:NoMass"},
    {IddObjectType::Material_InfraredTransparent, "Material_InfraredTransparent", "Material:InfraredTransparent"},
    {IddObjectType::Material_AirGap, "Material_AirGap", "Material:AirGap"},
    {IddObjectType::Material_RoofVegetation, "Material_RoofVegetation", "Material:RoofVegetation"},
    {IddObjectType::WindowMaterial_SimpleGlazingSystem, "WindowMaterial_SimpleGlazingSystem", "WindowMaterial:SimpleGlazingSystem"},
    {IddObjectType::WindowMaterial_Glazing, "WindowMaterial_Glazing", "WindowMaterial:Glazing"},
    {IddObjectType::WindowMaterial_GlazingGroup_Thermochromic, "WindowMaterial_GlazingGroup_Thermochromic",
     "WindowMaterial:GlazingGroup:Thermochromic"},
    {IddObjectType::WindowMaterial_Glazing_RefractionExtinctionMethod, "WindowMaterial_Glazing_RefractionExtinctionMethod",
     "WindowMaterial:Glazing:RefractionExtinctionMethod"},
    {IddObjectType::WindowMaterial_Gas, "WindowMaterial_Gas", "WindowMaterial:Gas"},
    {IddObjectType::WindowGap_SupportPillar, "WindowGap_SupportPillar", "WindowGap:SupportPillar"},
    {IddObjectType::WindowGap_DeflectionState, "WindowGap_DeflectionState", "WindowGap:DeflectionState"},
    {IddObjectType::WindowMaterial_GasMixture, "WindowMaterial_GasMixture", "WindowMaterial:GasMixture"},
    {IddObjectType::WindowMaterial_Gap, "WindowMaterial_Gap", "WindowMaterial:Gap"},
    {IddObjectType::WindowMaterial_Shade, "WindowMaterial_Shade", "WindowMaterial:Shade"},
    {IddObjectType::WindowMaterial_ComplexShade, "WindowMaterial_ComplexShade", "WindowMaterial:ComplexShade"},
    {IddObjectType::WindowMaterial_Blind, "WindowMaterial_Blind", "WindowMaterial:Blind"},
    {IddObjectType::WindowMaterial_Screen, "WindowMaterial_Screen", "WindowMaterial:Screen"},
    {IddObjectType::WindowMaterial_Shade_EquivalentLayer, "WindowMaterial_Shade_EquivalentLayer", "WindowMaterial:Shade:EquivalentLayer"},
    {IddObjectType::WindowMaterial_Drape_EquivalentLayer, "WindowMaterial_Drape_EquivalentLayer", "WindowMaterial:Drape:EquivalentLayer"},
    {IddObjectType::WindowMaterial_Blind_EquivalentLayer, "WindowMaterial_Blind_EquivalentLayer", "WindowMaterial:Blind:EquivalentLayer"},
    {IddObjectType::WindowMaterial_Screen_EquivalentLayer, "WindowMaterial_Screen_EquivalentLayer", "WindowMaterial:Screen:EquivalentLayer"},
    {IddObjectType::WindowMaterial_Glazing_EquivalentLayer, "WindowMaterial_Glazing_EquivalentLayer", "WindowMaterial:Glazing:EquivalentLayer"},
    {IddObjectType::WindowMaterial_Gap_EquivalentLayer, "WindowMaterial_Gap_EquivalentLayer", "WindowMaterial:Gap:EquivalentLayer"},
    {IddObjectType::MaterialProperty_MoisturePenetrationDepth_Settings, "MaterialProperty_MoisturePenetrationDepth_Settings",
     "MaterialProperty:MoisturePenetrationDepth:Settings"},
    {IddObjectType::MaterialProperty_PhaseChange, "MaterialProperty_PhaseChange", "MaterialProperty:PhaseChange"},
    {IddObjectType::MaterialProperty_PhaseChangeHysteresis, "MaterialProperty_PhaseChangeHysteresis", "MaterialProperty:PhaseChangeHysteresis"},
    {IddObjectType::MaterialProperty_VariableThermalConductivity, "MaterialProperty_VariableThermalConductivity",
     "MaterialProperty:VariableThermalConductivity"},
    {IddObjectType::MaterialProperty_VariableAbsorptance, "MaterialProperty_VariableAbsorptance", "MaterialProperty:VariableAbsorptance"},
    {IddObjectType::MaterialProperty_HeatAndMoistureTransfer_Settings, "MaterialProperty_HeatAndMoistureTransfer_Settings",
     "MaterialProperty:HeatAndMoistureTransfer:Settings"},
    {IddObjectType::MaterialProperty_HeatAndMoistureTransfer_SorptionIsotherm, "MaterialProperty_HeatAndMoistureTransfer_SorptionIsotherm",
     "MaterialProperty:HeatAndMoistureTransfer:SorptionIsotherm"},
    {IddObjectType::MaterialProperty_HeatAndMoistureTransfer_Suction, "MaterialProperty_HeatAndMoistureTransfer_Suction",
     "MaterialProperty:HeatAndMoistureTransfer:Suction"},
    {IddObjectType::MaterialProperty_HeatAndMoistureTransfer_Redistribution, "MaterialProperty_HeatAndMoistureTransfer_Redistribution",
     "MaterialProperty:HeatAndMoistureTransfer:Redistribution"},
    {IddObjectType::MaterialProperty_HeatAndMoistureTransfer_Diffusion, "MaterialProperty_HeatAndMoistureTransfer_Diffusion",
     "MaterialProperty:HeatAndMoistureTransfer:Diffusion"},
    {IddObjectType::MaterialProperty_HeatAndMoistureTransfer_ThermalConductivity, "MaterialProperty_HeatAndMoistureTransfer_ThermalConductivity",
     "MaterialProperty:HeatAndMoistureTransfer:ThermalConductivity"},
    {IddObjectType::MaterialProperty_GlazingSpectralData, "MaterialProperty_GlazingSpectralData", "MaterialProperty:GlazingSpectralData"},
    {IddObjectType::Construction, "Construction", "Construction"},
    {IddObjectType::Construction_CfactorUndergroundWall, "Construction_CfactorUndergroundWall", "Construction:CfactorUndergroundWall"},
    {IddObjectType::Construction_FfactorGroundFloor, "Construction_FfactorGroundFloor", "Construction:FfactorGroundFloor"},
    {IddObjectType::ConstructionProperty_InternalHeatSource, "ConstructionProperty_InternalHeatSource", "ConstructionProperty:InternalHeatSource"},
    {IddObjectType::Construction_AirBoundary, "Construction_AirBoundary", "Construction:AirBoundary"},
    {IddObjectType::WindowThermalModel_Params, "WindowThermalModel_Params", "WindowThermalModel:Params"},
    {IddObjectType::WindowsCalculationEngine, "WindowsCalculationEngine", "WindowsCalculationEngine"},
    {IddObjectType::Construction_ComplexFenestrationState, "Construction_ComplexFenestrationState", "Construction:ComplexFenestrationState"},
    {IddObjectType::Construction_WindowEquivalentLayer, "Construction_WindowEquivalentLayer", "Construction:WindowEquivalentLayer"},
    {IddObjectType::Construction_WindowDataFile, "Construction_WindowDataFile", "Construction:WindowDataFile"},
    {IddObjectType::GlobalGeometryRules, "GlobalGeometryRules", "GlobalGeometryRules"},
    {IddObjectType::GeometryTransform, "GeometryTransform", "GeometryTransform"},
    {IddObjectType::Space, "Space", "Space"},
    {IddObjectType::SpaceList, "SpaceList", "SpaceList"},
    {IddObjectType::Zone, "Zone", "Zone"},
    {IddObjectType::ZoneList, "ZoneList", "ZoneList"},
    {IddObjectType::ZoneGroup, "ZoneGroup", "ZoneGroup"},
    {IddObjectType::BuildingSurface_Detailed, "BuildingSurface_Detailed", "BuildingSurface:Detailed"},
    {IddObjectType::Wall_Detailed, "Wall_Detailed", "Wall:Detailed"},
    {IddObjectType::RoofCeiling_Detailed, "RoofCeiling_Detailed", "RoofCeiling:Detailed"},
    {IddObjectType::Floor_Detailed, "Floor_Detailed", "Floor:Detailed"},
    {IddObjectType::Wall_Exterior, "Wall_Exterior", "Wall:Exterior"},
    {IddObjectType::Wall_Adiabatic, "Wall_Adiabatic", "Wall:Adiabatic"},
    {IddObjectType::Wall_Underground, "Wall_Underground", "Wall:Underground"},
    {IddObjectType::Wall_Interzone, "Wall_Interzone", "Wall:Interzone"},
    {IddObjectType::Roof, "Roof", "Roof"},
    {IddObjectType::Ceiling_Adiabatic, "Ceiling_Adiabatic", "Ceiling:Adiabatic"},
    {IddObjectType::Ceiling_Interzone, "Ceiling_Interzone", "Ceiling:Interzone"},
    {IddObjectType::Floor_GroundContact, "Floor_GroundContact", "Floor:GroundContact"},
    {IddObjectType::Floor_Adiabatic, "Floor_Adiabatic", "Floor:Adiabatic"},
    {IddObjectType::Floor_Interzone, "Floor_Interzone", "Floor:Interzone"},
    {IddObjectType::FenestrationSurface_Detailed, "FenestrationSurface_Detailed", "FenestrationSurface:Detailed"},
    {IddObjectType::Window, "Window", "Window"},
    {IddObjectType::Door, "Door", "Door"},
    {IddObjectType::GlazedDoor, "GlazedDoor", "GlazedDoor"},
    {IddObjectType::Window_Interzone, "Window_Interzone", "Window:Interzone"},
    {IddObjectType::Door_Interzone, "Door_Interzone", "Door:Interzone"},
    {IddObjectType::GlazedDoor_Interzone, "GlazedDoor_Interzone", "GlazedDoor:Interzone"},
    {IddObjectType::WindowShadingControl, "WindowShadingControl", "WindowShadingControl"},
    {IddObjectType::WindowProperty_FrameAndDivider, "WindowProperty_FrameAndDivider", "WindowProperty:FrameAndDivider"},
    {IddObjectType::WindowProperty_AirflowControl, "WindowProperty_AirflowControl", "WindowProperty:AirflowControl"},
    {IddObjectType::WindowProperty_StormWindow, "WindowProperty_StormWindow", "WindowProperty:StormWindow"},
    {IddObjectType::InternalMass, "InternalMass", "InternalMass"},
    {IddObjectType::Shading_Site, "Shading_Site", "Shading:Site"},
    {IddObjectType::Shading_Building, "Shading_Building", "Shading:Building"},
    {IddObjectType::Shading_Site_Detailed, "Shading_Site_Detailed", "Shading:Site:Detailed"},
    {IddObjectType::Shading_Building_Detailed, "Shading_Building_Detailed", "Shading:Building:Detailed"},
    {IddObjectType::Shading_Overhang, "Shading_Overhang", "Shading:Overhang"},
    {IddObjectType::Shading_Overhang_Projection, "Shading_Overhang_Projection", "Shading:Overhang:Projection"},
    {IddObjectType::Shading_Fin, "Shading_Fin", "Shading:Fin"},
    {IddObjectType::Shading_Fin_Projection, "Shading_Fin_Projection", "Shading:Fin:Projection"},
    {IddObjectType::Shading_Zone_Detailed, "Shading_Zone_Detailed", "Shading:Zone:Detailed"},
    {IddObjectType::ShadingProperty_Reflectance, "ShadingProperty_Reflectance", "ShadingProperty:Reflectance"},
    {IddObjectType::SurfaceProperty_HeatTransferAlgorithm, "SurfaceProperty_HeatTransferAlgorithm", "SurfaceProperty:HeatTransferAlgorithm"},
    {IddObjectType::SurfaceProperty_HeatTransferAlgorithm_MultipleSurface, "SurfaceProperty_HeatTransferAlgorithm_MultipleSurface",
     "SurfaceProperty:HeatTransferAlgorithm:MultipleSurface"},
    {IddObjectType::SurfaceProperty_HeatTransferAlgorithm_SurfaceList, "SurfaceProperty_HeatTransferAlgorithm_SurfaceList",
     "SurfaceProperty:HeatTransferAlgorithm:SurfaceList"},
    {IddObjectType::SurfaceProperty_HeatTransferAlgorithm_Construction, "SurfaceProperty_HeatTransferAlgorithm_Construction",
     "SurfaceProperty:HeatTransferAlgorithm:Construction"},
    {IddObjectType::SurfaceProperty_HeatBalanceSourceTerm, "SurfaceProperty_HeatBalanceSourceTerm", "SurfaceProperty:HeatBalanceSourceTerm"},
    {IddObjectType::SurfaceControl_MovableInsulation, "SurfaceControl_MovableInsulation", "SurfaceControl:MovableInsulation"},
    {IddObjectType::SurfaceProperty_OtherSideCoefficients, "SurfaceProperty_OtherSideCoefficients", "SurfaceProperty:OtherSideCoefficients"},
    {IddObjectType::SurfaceProperty_OtherSideConditionsModel, "SurfaceProperty_OtherSideConditionsModel", "SurfaceProperty:OtherSideConditionsModel"},
    {IddObjectType::SurfaceProperty_Underwater, "SurfaceProperty_Underwater", "SurfaceProperty:Underwater"},
    {IddObjectType::Foundation_Kiva, "Foundation_Kiva", "Foundation:Kiva"},
    {IddObjectType::Foundation_Kiva_Settings, "Foundation_Kiva_Settings", "Foundation:Kiva:Settings"},
    {IddObjectType::SurfaceProperty_ExposedFoundationPerimeter, "SurfaceProperty_ExposedFoundationPerimeter",
     "SurfaceProperty:ExposedFoundationPerimeter"},
    {IddObjectType::SurfaceConvectionAlgorithm_Inside_AdaptiveModelSelections, "SurfaceConvectionAlgorithm_Inside_AdaptiveModelSelections",
     "SurfaceConvectionAlgorithm:Inside:AdaptiveModelSelections"},
    {IddObjectType::SurfaceConvectionAlgorithm_Outside_AdaptiveModelSelections, "SurfaceConvectionAlgorithm_Outside_AdaptiveModelSelections",
     "SurfaceConvectionAlgorithm:Outside:AdaptiveModelSelections"},
    {IddObjectType::SurfaceConvectionAlgorithm_Inside_UserCurve, "SurfaceConvectionAlgorithm_Inside_UserCurve",
     "SurfaceConvectionAlgorithm:Inside:UserCurve"},
    {IddObjectType::SurfaceConvectionAlgorithm_Outside_UserCurve, "SurfaceConvectionAlgorithm_Outside_UserCurve",
     "SurfaceConvectionAlgorithm:Outside:UserCurve"},
    {IddObjectType::SurfaceProperty_ConvectionCoefficients, "SurfaceProperty_ConvectionCoefficients", "SurfaceProperty:ConvectionCoefficients"},
    {IddObjectType::SurfaceProperty_ConvectionCoefficients_MultipleSurface, "SurfaceProperty_ConvectionCoefficients_MultipleSurface",
     "SurfaceProperty:ConvectionCoefficients:MultipleSurface"},
    {IddObjectType::SurfaceProperties_VaporCoefficients, "SurfaceProperties_VaporCoefficients", "SurfaceProperties:VaporCoefficients"},
    {IddObjectType::SurfaceProperty_ExteriorNaturalVentedCavity, "SurfaceProperty_ExteriorNaturalVentedCavity",
     "SurfaceProperty:ExteriorNaturalVentedCavity"},
    {IddObjectType::SurfaceProperty_SolarIncidentInside, "SurfaceProperty_SolarIncidentInside", "SurfaceProperty:SolarIncidentInside"},
    {IddObjectType::SurfaceProperty_IncidentSolarMultiplier, "SurfaceProperty_IncidentSolarMultiplier", "SurfaceProperty:IncidentSolarMultiplier"},
    {IddObjectType::SurfaceProperty_LocalEnvironment, "SurfaceProperty_LocalEnvironment", "SurfaceProperty:LocalEnvironment"},
    {IddObjectType::ZoneProperty_LocalEnvironment, "ZoneProperty_LocalEnvironment", "ZoneProperty:LocalEnvironment"},
    {IddObjectType::SurfaceProperty_SurroundingSurfaces, "SurfaceProperty_SurroundingSurfaces", "SurfaceProperty:SurroundingSurfaces"},
    {IddObjectType::SurfaceProperty_GroundSurfaces, "SurfaceProperty_GroundSurfaces", "SurfaceProperty:GroundSurfaces"},
    {IddObjectType::ComplexFenestrationProperty_SolarAbsorbedLayers, "ComplexFenestrationProperty_SolarAbsorbedLayers",
     "ComplexFenestrationProperty:SolarAbsorbedLayers"},
    {IddObjectType::ZoneProperty_UserViewFactors_BySurfaceName, "ZoneProperty_UserViewFactors_BySurfaceName",
     "ZoneProperty:UserViewFactors:BySurfaceName"},
    {IddObjectType::GroundHeatTransfer_Control, "GroundHeatTransfer_Control", "GroundHeatTransfer:Control"},
    {IddObjectType::GroundHeatTransfer_Slab_Materials, "GroundHeatTransfer_Slab_Materials", "GroundHeatTransfer:Slab:Materials"},
    {IddObjectType::GroundHeatTransfer_Slab_MatlProps, "GroundHeatTransfer_Slab_MatlProps", "GroundHeatTransfer:Slab:MatlProps"},
    {IddObjectType::GroundHeatTransfer_Slab_BoundConds, "GroundHeatTransfer_Slab_BoundConds", "GroundHeatTransfer:Slab:BoundConds"},
    {IddObjectType::GroundHeatTransfer_Slab_BldgProps, "GroundHeatTransfer_Slab_BldgProps", "GroundHeatTransfer:Slab:BldgProps"},
    {IddObjectType::GroundHeatTransfer_Slab_Insulation, "GroundHeatTransfer_Slab_Insulation", "GroundHeatTransfer:Slab:Insulation"},
    {IddObjectType::GroundHeatTransfer_Slab_EquivalentSlab, "GroundHeatTransfer_Slab_EquivalentSlab", "GroundHeatTransfer:Slab:EquivalentSlab"},
    {IddObjectType::GroundHeatTransfer_Slab_AutoGrid, "GroundHeatTransfer_Slab_AutoGrid", "GroundHeatTransfer:Slab:AutoGrid"},
    {IddObjectType::GroundHeatTransfer_Slab_ManualGrid, "GroundHeatTransfer_Slab_ManualGrid", "GroundHeatTransfer:Slab:ManualGrid"},
    {IddObjectType::GroundHeatTransfer_Slab_XFACE, "GroundHeatTransfer_Slab_XFACE", "GroundHeatTransfer:Slab:XFACE"},
    {IddObjectType::GroundHeatTransfer_Slab_YFACE, "GroundHeatTransfer_Slab_YFACE", "GroundHeatTransfer:Slab:YFACE"},
    {IddObjectType::GroundHeatTransfer_Slab_ZFACE, "GroundHeatTransfer_Slab_ZFACE", "GroundHeatTransfer:Slab:ZFACE"},
    {IddObjectType::GroundHeatTransfer_Basement_SimParameters, "GroundHeatTransfer_Basement_SimParameters",
     "GroundHeatTransfer:Basement:SimParameters"},
    {IddObjectType::GroundHeatTransfer_Basement_MatlProps, "GroundHeatTransfer_Basement_MatlProps", "GroundHeatTransfer:Basement:MatlProps"},
    {IddObjectType::GroundHeatTransfer_Basement_Insulation, "GroundHeatTransfer_Basement_Insulation", "GroundHeatTransfer:Basement:Insulation"},
    {IddObjectType::GroundHeatTransfer_Basement_SurfaceProps, "GroundHeatTransfer_Basement_SurfaceProps", "GroundHeatTransfer:Basement:SurfaceProps"},
    {IddObjectType::GroundHeatTransfer_Basement_BldgData, "GroundHeatTransfer_Basement_BldgData", "GroundHeatTransfer:Basement:BldgData"},
    {IddObjectType::GroundHeatTransfer_Basement_Interior, "GroundHeatTransfer_Basement_Interior", "GroundHeatTransfer:Basement:Interior"},
    {IddObjectType::GroundHeatTransfer_Basement_ComBldg, "GroundHeatTransfer_Basement_ComBldg", "GroundHeatTransfer:Basement:ComBldg"},
    {IddObjectType::GroundHeatTransfer_Basement_EquivSlab, "GroundHeatTransfer_Basement_EquivSlab", "GroundHeatTransfer:Basement:EquivSlab"},
    {IddObjectType::GroundHeatTransfer_Basement_EquivAutoGrid, "GroundHeatTransfer_Basement_EquivAutoGrid",
     "GroundHeatTransfer:Basement:EquivAutoGrid"},
    {IddObjectType::GroundHeatTransfer_Basement_AutoGrid, "GroundHeatTransfer_Basement_AutoGrid", "GroundHeatTransfer:Basement:AutoGrid"},
    {IddObjectType::GroundHeatTransfer_Basement_ManualGrid, "GroundHeatTransfer_Basement_ManualGrid", "GroundHeatTransfer:Basement:ManualGrid"},
    {IddObjectType::GroundHeatTransfer_Basement_XFACE, "GroundHeatTransfer_Basement_XFACE", "GroundHeatTransfer:Basement:XFACE"},
    {IddObjectType::GroundHeatTransfer_Basement_YFACE, "GroundHeatTransfer_Basement_YFACE", "GroundHeatTransfer:Basement:YFACE"},
    {IddObjectType::GroundHeatTransfer_Basement_ZFACE, "GroundHeatTransfer_Basement_ZFACE", "GroundHeatTransfer:Basement:ZFACE"},
    {IddObjectType::RoomAirModelType, "RoomAirModelType", "RoomAirModelType"},
    {IddObjectType::RoomAir_TemperaturePattern_UserDefined, "RoomAir_TemperaturePattern_UserDefined", "RoomAir:TemperaturePattern:UserDefined"},
    {IddObjectType::RoomAir_TemperaturePattern_ConstantGradient, "RoomAir_TemperaturePattern_ConstantGradient",
     "RoomAir:TemperaturePattern:ConstantGradient"},
    {IddObjectType::RoomAir_TemperaturePattern_TwoGradient, "RoomAir_TemperaturePattern_TwoGradient", "RoomAir:TemperaturePattern:TwoGradient"},
    {IddObjectType::RoomAir_TemperaturePattern_NondimensionalHeight, "RoomAir_TemperaturePattern_NondimensionalHeight",
     "RoomAir:TemperaturePattern:NondimensionalHeight"},
    {IddObjectType::RoomAir_TemperaturePattern_SurfaceMapping, "RoomAir_TemperaturePattern_SurfaceMapping",
     "RoomAir:TemperaturePattern:SurfaceMapping"},
    {IddObjectType::RoomAir_Node, "RoomAir_Node", "RoomAir:Node"},
    {IddObjectType::RoomAirSettings_OneNodeDisplacementVentilation, "RoomAirSettings_OneNodeDisplacementVentilation",
     "RoomAirSettings:OneNodeDisplacementVentilation"},
    {IddObjectType::RoomAirSettings_ThreeNodeDisplacementVentilation, "RoomAirSettings_ThreeNodeDisplacementVentilation",
     "RoomAirSettings:ThreeNodeDisplacementVentilation"},
    {IddObjectType::RoomAirSettings_CrossVentilation, "RoomAirSettings_CrossVentilation", "RoomAirSettings:CrossVentilation"},
    {IddObjectType::RoomAirSettings_UnderFloorAirDistributionInterior, "RoomAirSettings_UnderFloorAirDistributionInterior",
     "RoomAirSettings:UnderFloorAirDistributionInterior"},
    {IddObjectType::RoomAirSettings_UnderFloorAirDistributionExterior, "RoomAirSettings_UnderFloorAirDistributionExterior",
     "RoomAirSettings:UnderFloorAirDistributionExterior"},
    {IddObjectType::RoomAir_Node_AirflowNetwork, "RoomAir_Node_AirflowNetwork", "RoomAir:Node:AirflowNetwork"},
    {IddObjectType::RoomAir_Node_AirflowNetwork_AdjacentSurfaceList, "RoomAir_Node_AirflowNetwork_AdjacentSurfaceList",
     "RoomAir:Node:AirflowNetwork:AdjacentSurfaceList"},
    {IddObjectType::RoomAir_Node_AirflowNetwork_InternalGains, "RoomAir_Node_AirflowNetwork_InternalGains",
     "RoomAir:Node:AirflowNetwork:InternalGains"},
    {IddObjectType::RoomAir_Node_AirflowNetwork_HVACEquipment, "RoomAir_Node_AirflowNetwork_HVACEquipment",
     "RoomAir:Node:AirflowNetwork:HVACEquipment"},
    {IddObjectType::RoomAirSettings_AirflowNetwork, "RoomAirSettings_AirflowNetwork", "RoomAirSettings:AirflowNetwork"},
    {IddObjectType::People, "People", "People"},
    {IddObjectType::ComfortViewFactorAngles, "ComfortViewFactorAngles", "ComfortViewFactorAngles"},
    {IddObjectType::Lights, "Lights", "Lights"},
    {IddObjectType::ElectricEquipment, "ElectricEquipment", "ElectricEquipment"},
    {IddObjectType::GasEquipment, "GasEquipment", "GasEquipment"},
    {IddObjectType::HotWaterEquipment, "HotWaterEquipment", "HotWaterEquipment"},
    {IddObjectType::SteamEquipment, "SteamEquipment", "SteamEquipment"},
    {IddObjectType::OtherEquipment, "OtherEquipment", "OtherEquipment"},
    {IddObjectType::ElectricEquipment_ITE_AirCooled, "ElectricEquipment_ITE_AirCooled", "ElectricEquipment:ITE:AirCooled"},
    {IddObjectType::ZoneBaseboard_OutdoorTemperatureControlled, "ZoneBaseboard_OutdoorTemperatureControlled",
     "ZoneBaseboard:OutdoorTemperatureControlled"},
    {IddObjectType::SwimmingPool_Indoor, "SwimmingPool_Indoor", "SwimmingPool:Indoor"},
    {IddObjectType::ZoneContaminantSourceAndSink_CarbonDioxide, "ZoneContaminantSourceAndSink_CarbonDioxide",
     "ZoneContaminantSourceAndSink:CarbonDioxide"},
    {IddObjectType::ZoneContaminantSourceAndSink_Generic_Constant, "ZoneContaminantSourceAndSink_Generic_Constant",
     "ZoneContaminantSourceAndSink:Generic:Constant"},
    {IddObjectType::SurfaceContaminantSourceAndSink_Generic_PressureDriven, "SurfaceContaminantSourceAndSink_Generic_PressureDriven",
     "SurfaceContaminantSourceAndSink:Generic:PressureDriven"},
    {IddObjectType::ZoneContaminantSourceAndSink_Generic_CutoffModel, "ZoneContaminantSourceAndSink_Generic_CutoffModel",
     "ZoneContaminantSourceAndSink:Generic:CutoffModel"},
    {IddObjectType::ZoneContaminantSourceAndSink_Generic_DecaySource, "ZoneContaminantSourceAndSink_Generic_DecaySource",
     "ZoneContaminantSourceAndSink:Generic:DecaySource"},
    {IddObjectType::SurfaceContaminantSourceAndSink_Generic_BoundaryLayerDiffusion, "SurfaceContaminantSourceAndSink_Generic_BoundaryLayerDiffusion",
     "SurfaceContaminantSourceAndSink:Generic:BoundaryLayerDiffusion"},
    {IddObjectType::SurfaceContaminantSourceAndSink_Generic_DepositionVelocitySink, "SurfaceContaminantSourceAndSink_Generic_DepositionVelocitySink",
     "SurfaceContaminantSourceAndSink:Generic:DepositionVelocitySink"},
    {IddObjectType::ZoneContaminantSourceAndSink_Generic_DepositionRateSink, "ZoneContaminantSourceAndSink_Generic_DepositionRateSink",
     "ZoneContaminantSourceAndSink:Generic:DepositionRateSink"},
    {IddObjectType::Daylighting_Controls, "Daylighting_Controls", "Daylighting:Controls"},
    {IddObjectType::Daylighting_ReferencePoint, "Daylighting_ReferencePoint", "Daylighting:ReferencePoint"},
    {IddObjectType::Daylighting_DELight_ComplexFenestration, "Daylighting_DELight_ComplexFenestration", "Daylighting:DELight:ComplexFenestration"},
    {IddObjectType::DaylightingDevice_Tubular, "DaylightingDevice_Tubular", "DaylightingDevice:Tubular"},
    {IddObjectType::DaylightingDevice_Shelf, "DaylightingDevice_Shelf", "DaylightingDevice:Shelf"},
    {IddObjectType::DaylightingDevice_LightWell, "DaylightingDevice_LightWell", "DaylightingDevice:LightWell"},
    {IddObjectType::Output_DaylightFactors, "Output_DaylightFactors", "Output:DaylightFactors"},
    {IddObjectType::Output_IlluminanceMap, "Output_IlluminanceMap", "Output:IlluminanceMap"},
    {IddObjectType::OutputControl_IlluminanceMap_Style, "OutputControl_IlluminanceMap_Style", "OutputControl:IlluminanceMap:Style"},
    {IddObjectType::ZoneInfiltration_DesignFlowRate, "ZoneInfiltration_DesignFlowRate", "ZoneInfiltration:DesignFlowRate"},
    {IddObjectType::ZoneInfiltration_EffectiveLeakageArea, "ZoneInfiltration_EffectiveLeakageArea", "ZoneInfiltration:EffectiveLeakageArea"},
    {IddObjectType::ZoneInfiltration_FlowCoefficient, "ZoneInfiltration_FlowCoefficient", "ZoneInfiltration:FlowCoefficient"},
    {IddObjectType::ZoneVentilation_DesignFlowRate, "ZoneVentilation_DesignFlowRate", "ZoneVentilation:DesignFlowRate"},
    {IddObjectType::ZoneVentilation_WindandStackOpenArea, "ZoneVentilation_WindandStackOpenArea", "ZoneVentilation:WindandStackOpenArea"},
    {IddObjectType::ZoneAirBalance_OutdoorAir, "ZoneAirBalance_OutdoorAir", "ZoneAirBalance:OutdoorAir"},
    {IddObjectType::ZoneMixing, "ZoneMixing", "ZoneMixing"},
    {IddObjectType::ZoneCrossMixing, "ZoneCrossMixing", "ZoneCrossMixing"},
    {IddObjectType::ZoneRefrigerationDoorMixing, "ZoneRefrigerationDoorMixing", "ZoneRefrigerationDoorMixing"},
    {IddObjectType::ZoneEarthtube, "ZoneEarthtube", "ZoneEarthtube"},
    {IddObjectType::ZoneCoolTower_Shower, "ZoneCoolTower_Shower", "ZoneCoolTower:Shower"},
    {IddObjectType::ZoneThermalChimney, "ZoneThermalChimney", "ZoneThermalChimney"},
    {IddObjectType::AirflowNetwork_SimulationControl, "AirflowNetwork_SimulationControl", "AirflowNetwork:SimulationControl"},
    {IddObjectType::AirflowNetwork_MultiZone_Zone, "AirflowNetwork_MultiZone_Zone", "AirflowNetwork:MultiZone:Zone"},
    {IddObjectType::AirflowNetwork_MultiZone_Surface, "AirflowNetwork_MultiZone_Surface", "AirflowNetwork:MultiZone:Surface"},
    {IddObjectType::AirflowNetwork_MultiZone_ReferenceCrackConditions, "AirflowNetwork_MultiZone_ReferenceCrackConditions",
     "AirflowNetwork:MultiZone:ReferenceCrackConditions"},
    {IddObjectType::AirflowNetwork_MultiZone_Surface_Crack, "AirflowNetwork_MultiZone_Surface_Crack", "AirflowNetwork:MultiZone:Surface:Crack"},
    {IddObjectType::AirflowNetwork_MultiZone_Surface_EffectiveLeakageArea, "AirflowNetwork_MultiZone_Surface_EffectiveLeakageArea",
     "AirflowNetwork:MultiZone:Surface:EffectiveLeakageArea"},
    {IddObjectType::AirflowNetwork_MultiZone_SpecifiedFlowRate, "AirflowNetwork_MultiZone_SpecifiedFlowRate",
     "AirflowNetwork:MultiZone:SpecifiedFlowRate"},
    {IddObjectType::AirflowNetwork_MultiZone_Component_DetailedOpening, "AirflowNetwork_MultiZone_Component_DetailedOpening",
     "AirflowNetwork:MultiZone:Component:DetailedOpening"},
    {IddObjectType::AirflowNetwork_MultiZone_Component_SimpleOpening, "AirflowNetwork_MultiZone_Component_SimpleOpening",
     "AirflowNetwork:MultiZone:Component:SimpleOpening"},
    {IddObjectType::AirflowNetwork_MultiZone_Component_HorizontalOpening, "AirflowNetwork_MultiZone_Component_HorizontalOpening",
     "AirflowNetwork:MultiZone:Component:HorizontalOpening"},
    {IddObjectType::AirflowNetwork_MultiZone_Component_ZoneExhaustFan, "AirflowNetwork_MultiZone_Component_ZoneExhaustFan",
     "AirflowNetwork:MultiZone:Component:ZoneExhaustFan"},
    {IddObjectType::AirflowNetwork_MultiZone_ExternalNode, "AirflowNetwork_MultiZone_ExternalNode", "AirflowNetwork:MultiZone:ExternalNode"},
    {IddObjectType::AirflowNetwork_MultiZone_WindPressureCoefficientArray, "AirflowNetwork_MultiZone_WindPressureCoefficientArray",
     "AirflowNetwork:MultiZone:WindPressureCoefficientArray"},
    {IddObjectType::AirflowNetwork_MultiZone_WindPressureCoefficientValues, "AirflowNetwork_MultiZone_WindPressureCoefficientValues",
     "AirflowNetwork:MultiZone:WindPressureCoefficientValues"},
    {IddObjectType::AirflowNetwork_ZoneControl_PressureController, "AirflowNetwork_ZoneControl_PressureController",
     "AirflowNetwork:ZoneControl:PressureController"},
    {IddObjectType::AirflowNetwork_Distribution_Node, "AirflowNetwork_Distribution_Node", "AirflowNetwork:Distribution:Node"},
    {IddObjectType::AirflowNetwork_Distribution_Component_Leak, "AirflowNetwork_Distribution_Component_Leak",
     "AirflowNetwork:Distribution:Component:Leak"},
    {IddObjectType::AirflowNetwork_Distribution_Component_LeakageRatio, "AirflowNetwork_Distribution_Component_LeakageRatio",
     "AirflowNetwork:Distribution:Component:LeakageRatio"},
    {IddObjectType::AirflowNetwork_Distribution_Component_Duct, "AirflowNetwork_Distribution_Component_Duct",
     "AirflowNetwork:Distribution:Component:Duct"},
    {IddObjectType::AirflowNetwork_Distribution_Component_Fan, "AirflowNetwork_Distribution_Component_Fan",
     "AirflowNetwork:Distribution:Component:Fan"},
    {IddObjectType::AirflowNetwork_Distribution_Component_Coil, "AirflowNetwork_Distribution_Component_Coil",
     "AirflowNetwork:Distribution:Component:Coil"},
    {IddObjectType::AirflowNetwork_Distribution_Component_HeatExchanger, "AirflowNetwork_Distribution_Component_HeatExchanger",
     "AirflowNetwork:Distribution:Component:HeatExchanger"},
    {IddObjectType::AirflowNetwork_Distribution_Component_TerminalUnit, "AirflowNetwork_Distribution_Component_TerminalUnit",
     "AirflowNetwork:Distribution:Component:TerminalUnit"},
    {IddObjectType::AirflowNetwork_Distribution_Component_ConstantPressureDrop, "AirflowNetwork_Distribution_Component_ConstantPressureDrop",
     "AirflowNetwork:Distribution:Component:ConstantPressureDrop"},
    {IddObjectType::AirflowNetwork_Distribution_Component_OutdoorAirFlow, "AirflowNetwork_Distribution_Component_OutdoorAirFlow",
     "AirflowNetwork:Distribution:Component:OutdoorAirFlow"},
    {IddObjectType::AirflowNetwork_Distribution_Component_ReliefAirFlow, "AirflowNetwork_Distribution_Component_ReliefAirFlow",
     "AirflowNetwork:Distribution:Component:ReliefAirFlow"},
    {IddObjectType::AirflowNetwork_Distribution_Linkage, "AirflowNetwork_Distribution_Linkage", "AirflowNetwork:Distribution:Linkage"},
    {IddObjectType::AirflowNetwork_Distribution_DuctViewFactors, "AirflowNetwork_Distribution_DuctViewFactors",
     "AirflowNetwork:Distribution:DuctViewFactors"},
    {IddObjectType::AirflowNetwork_OccupantVentilationControl, "AirflowNetwork_OccupantVentilationControl",
     "AirflowNetwork:OccupantVentilationControl"},
    {IddObjectType::AirflowNetwork_IntraZone_Node, "AirflowNetwork_IntraZone_Node", "AirflowNetwork:IntraZone:Node"},
    {IddObjectType::AirflowNetwork_IntraZone_Linkage, "AirflowNetwork_IntraZone_Linkage", "AirflowNetwork:IntraZone:Linkage"},
    {IddObjectType::Exterior_Lights, "Exterior_Lights", "Exterior:Lights"},
    {IddObjectType::Exterior_FuelEquipment, "Exterior_FuelEquipment", "Exterior:FuelEquipment"},
    {IddObjectType::Exterior_WaterEquipment, "Exterior_WaterEquipment", "Exterior:WaterEquipment"},
    {IddObjectType::HVACTemplate_Thermostat, "HVACTemplate_Thermostat", "HVACTemplate:Thermostat"},
    {IddObjectType::HVACTemplate_Zone_IdealLoadsAirSystem, "HVACTemplate_Zone_IdealLoadsAirSystem", "HVACTemplate:Zone:IdealLoadsAirSystem"},
    {IddObjectType::HVACTemplate_Zone_BaseboardHeat, "HVACTemplate_Zone_BaseboardHeat", "HVACTemplate:Zone:BaseboardHeat"},
    {IddObjectType::HVACTemplate_Zone_FanCoil, "HVACTemplate_Zone_FanCoil", "HVACTemplate:Zone:FanCoil"},
    {IddObjectType::HVACTemplate_Zone_PTAC, "HVACTemplate_Zone_PTAC", "HVACTemplate:Zone:PTAC"},
    {IddObjectType::HVACTemplate_Zone_PTHP, "HVACTemplate_Zone_PTHP", "HVACTemplate:Zone:PTHP"},
    {IddObjectType::HVACTemplate_Zone_WaterToAirHeatPump, "HVACTemplate_Zone_WaterToAirHeatPump", "HVACTemplate:Zone:WaterToAirHeatPump"},
    {IddObjectType::HVACTemplate_Zone_VRF, "HVACTemplate_Zone_VRF", "HVACTemplate:Zone:VRF"},
    {IddObjectType::HVACTemplate_Zone_Unitary, "HVACTemplate_Zone_Unitary", "HVACTemplate:Zone:Unitary"},
    {IddObjectType::HVACTemplate_Zone_VAV, "HVACTemplate_Zone_VAV", "HVACTemplate:Zone:VAV"},
    {IddObjectType::HVACTemplate_Zone_VAV_FanPowered, "HVACTemplate_Zone_VAV_FanPowered", "HVACTemplate:Zone:VAV:FanPowered"},
    {IddObjectType::HVACTemplate_Zone_VAV_HeatAndCool, "HVACTemplate_Zone_VAV_HeatAndCool", "HVACTemplate:Zone:VAV:HeatAndCool"},
    {IddObjectType::HVACTemplate_Zone_ConstantVolume, "HVACTemplate_Zone_ConstantVolume", "HVACTemplate:Zone:ConstantVolume"},
    {IddObjectType::HVACTemplate_Zone_DualDuct, "HVACTemplate_Zone_DualDuct", "HVACTemplate:Zone:DualDuct"},
    {IddObjectType::HVACTemplate_System_VRF, "HVACTemplate_System_VRF", "HVACTemplate:System:VRF"},
    {IddObjectType::HVACTemplate_System_Unitary, "HVACTemplate_System_Unitary", "HVACTemplate:System:Unitary"},
    {IddObjectType::HVACTemplate_System_UnitaryHeatPump_AirToAir, "HVACTemplate_System_UnitaryHeatPump_AirToAir",
     "HVACTemplate:System:UnitaryHeatPump:AirToAir"},
    {IddObjectType::HVACTemplate_System_UnitarySystem, "HVACTemplate_System_UnitarySystem", "HVACTemplate:System:UnitarySystem"},
    {IddObjectType::HVACTemplate_System_VAV, "HVACTemplate_System_VAV", "HVACTemplate:System:VAV"},
    {IddObjectType::HVACTemplate_System_PackagedVAV, "HVACTemplate_System_PackagedVAV", "HVACTemplate:System:PackagedVAV"},
    {IddObjectType::HVACTemplate_System_ConstantVolume, "HVACTemplate_System_ConstantVolume", "HVACTemplate:System:ConstantVolume"},
    {IddObjectType::HVACTemplate_System_DualDuct, "HVACTemplate_System_DualDuct", "HVACTemplate:System:DualDuct"},
    {IddObjectType::HVACTemplate_System_DedicatedOutdoorAir, "HVACTemplate_System_DedicatedOutdoorAir", "HVACTemplate:System:DedicatedOutdoorAir"},
    {IddObjectType::HVACTemplate_Plant_ChilledWaterLoop, "HVACTemplate_Plant_ChilledWaterLoop", "HVACTemplate:Plant:ChilledWaterLoop"},
    {IddObjectType::HVACTemplate_Plant_Chiller, "HVACTemplate_Plant_Chiller", "HVACTemplate:Plant:Chiller"},
    {IddObjectType::HVACTemplate_Plant_Chiller_ObjectReference, "HVACTemplate_Plant_Chiller_ObjectReference",
     "HVACTemplate:Plant:Chiller:ObjectReference"},
    {IddObjectType::HVACTemplate_Plant_Tower, "HVACTemplate_Plant_Tower", "HVACTemplate:Plant:Tower"},
    {IddObjectType::HVACTemplate_Plant_Tower_ObjectReference, "HVACTemplate_Plant_Tower_ObjectReference", "HVACTemplate:Plant:Tower:ObjectReference"},
    {IddObjectType::HVACTemplate_Plant_HotWaterLoop, "HVACTemplate_Plant_HotWaterLoop", "HVACTemplate:Plant:HotWaterLoop"},
    {IddObjectType::HVACTemplate_Plant_Boiler, "HVACTemplate_Plant_Boiler", "HVACTemplate:Plant:Boiler"},
    {IddObjectType::HVACTemplate_Plant_Boiler_ObjectReference, "HVACTemplate_Plant_Boiler_ObjectReference",
     "HVACTemplate:Plant:Boiler:ObjectReference"},
    {IddObjectType::HVACTemplate_Plant_MixedWaterLoop, "HVACTemplate_Plant_MixedWaterLoop", "HVACTemplate:Plant:MixedWaterLoop"},
    {IddObjectType::DesignSpecification_OutdoorAir, "DesignSpecification_OutdoorAir", "DesignSpecification:OutdoorAir"},
    {IddObjectType::DesignSpecification_OutdoorAir_SpaceList, "DesignSpecification_OutdoorAir_SpaceList", "DesignSpecification:OutdoorAir:SpaceList"},
    {IddObjectType::DesignSpecification_ZoneAirDistribution, "DesignSpecification_ZoneAirDistribution", "DesignSpecification:ZoneAirDistribution"},
    {IddObjectType::Sizing_Parameters, "Sizing_Parameters", "Sizing:Parameters"},
    {IddObjectType::Sizing_Zone, "Sizing_Zone", "Sizing:Zone"},
    {IddObjectType::DesignSpecification_ZoneHVAC_Sizing, "DesignSpecification_ZoneHVAC_Sizing", "DesignSpecification:ZoneHVAC:Sizing"},
    {IddObjectType::DesignSpecification_AirTerminal_Sizing, "DesignSpecification_AirTerminal_Sizing", "DesignSpecification:AirTerminal:Sizing"},
    {IddObjectType::Sizing_System, "Sizing_System", "Sizing:System"},
    {IddObjectType::Sizing_Plant, "Sizing_Plant", "Sizing:Plant"},
    {IddObjectType::OutputControl_Sizing_Style, "OutputControl_Sizing_Style", "OutputControl:Sizing:Style"},
    {IddObjectType::ZoneControl_Humidistat, "ZoneControl_Humidistat", "ZoneControl:Humidistat"},
    {IddObjectType::ZoneControl_Thermostat, "ZoneControl_Thermostat", "ZoneControl:Thermostat"},
    {IddObjectType::ZoneControl_Thermostat_OperativeTemperature, "ZoneControl_Thermostat_OperativeTemperature",
     "ZoneControl:Thermostat:OperativeTemperature"},
    {IddObjectType::ZoneControl_Thermostat_ThermalComfort, "ZoneControl_Thermostat_ThermalComfort", "ZoneControl:Thermostat:ThermalComfort"},
    {IddObjectType::ZoneControl_Thermostat_TemperatureAndHumidity, "ZoneControl_Thermostat_TemperatureAndHumidity",
     "ZoneControl:Thermostat:TemperatureAndHumidity"},
    {IddObjectType::ThermostatSetpoint_SingleHeating, "ThermostatSetpoint_SingleHeating", "ThermostatSetpoint:SingleHeating"},
    {IddObjectType::ThermostatSetpoint_SingleCooling, "ThermostatSetpoint_SingleCooling", "ThermostatSetpoint:SingleCooling"},
    {IddObjectType::ThermostatSetpoint_SingleHeatingOrCooling, "ThermostatSetpoint_SingleHeatingOrCooling",
     "ThermostatSetpoint:SingleHeatingOrCooling"},
    {IddObjectType::ThermostatSetpoint_DualSetpoint, "ThermostatSetpoint_DualSetpoint", "ThermostatSetpoint:DualSetpoint"},
    {IddObjectType::ThermostatSetpoint_ThermalComfort_Fanger_SingleHeating, "ThermostatSetpoint_ThermalComfort_Fanger_SingleHeating",
     "ThermostatSetpoint:ThermalComfort:Fanger:SingleHeating"},
    {IddObjectType::ThermostatSetpoint_ThermalComfort_Fanger_SingleCooling, "ThermostatSetpoint_ThermalComfort_Fanger_SingleCooling",
     "ThermostatSetpoint:ThermalComfort:Fanger:SingleCooling"},
    {IddObjectType::ThermostatSetpoint_ThermalComfort_Fanger_SingleHeatingOrCooling,
     "ThermostatSetpoint_ThermalComfort_Fanger_SingleHeatingOrCooling", "ThermostatSetpoint:ThermalComfort:Fanger:SingleHeatingOrCooling"},
    {IddObjectType::ThermostatSetpoint_ThermalComfort_Fanger_DualSetpoint, "ThermostatSetpoint_ThermalComfort_Fanger_DualSetpoint",
     "ThermostatSetpoint:ThermalComfort:Fanger:DualSetpoint"},
    {IddObjectType::ZoneControl_Thermostat_StagedDualSetpoint, "ZoneControl_Thermostat_StagedDualSetpoint",
     "ZoneControl:Thermostat:StagedDualSetpoint"},
    {IddObjectType::ZoneControl_ContaminantController, "ZoneControl_ContaminantController", "ZoneControl:ContaminantController"},
    {IddObjectType::ZoneHVAC_IdealLoadsAirSystem, "ZoneHVAC_IdealLoadsAirSystem", "ZoneHVAC:IdealLoadsAirSystem"},
    {IddObjectType::ZoneHVAC_FourPipeFanCoil, "ZoneHVAC_FourPipeFanCoil", "ZoneHVAC:FourPipeFanCoil"},
    {IddObjectType::ZoneHVAC_WindowAirConditioner, "ZoneHVAC_WindowAirConditioner", "ZoneHVAC:WindowAirConditioner"},
    {IddObjectType::ZoneHVAC_PackagedTerminalAirConditioner, "ZoneHVAC_PackagedTerminalAirConditioner", "ZoneHVAC:PackagedTerminalAirConditioner"},
    {IddObjectType::ZoneHVAC_PackagedTerminalHeatPump, "ZoneHVAC_PackagedTerminalHeatPump", "ZoneHVAC:PackagedTerminalHeatPump"},
    {IddObjectType::ZoneHVAC_WaterToAirHeatPump, "ZoneHVAC_WaterToAirHeatPump", "ZoneHVAC:WaterToAirHeatPump"},
    {IddObjectType::ZoneHVAC_Dehumidifier_DX, "ZoneHVAC_Dehumidifier_DX", "ZoneHVAC:Dehumidifier:DX"},
    {IddObjectType::ZoneHVAC_EnergyRecoveryVentilator, "ZoneHVAC_EnergyRecoveryVentilator", "ZoneHVAC:EnergyRecoveryVentilator"},
    {IddObjectType::ZoneHVAC_EnergyRecoveryVentilator_Controller, "ZoneHVAC_EnergyRecoveryVentilator_Controller",
     "ZoneHVAC:EnergyRecoveryVentilator:Controller"},
    {IddObjectType::ZoneHVAC_UnitVentilator, "ZoneHVAC_UnitVentilator", "ZoneHVAC:UnitVentilator"},
    {IddObjectType::ZoneHVAC_UnitHeater, "ZoneHVAC_UnitHeater", "ZoneHVAC:UnitHeater"},
    {IddObjectType::ZoneHVAC_EvaporativeCoolerUnit, "ZoneHVAC_EvaporativeCoolerUnit", "ZoneHVAC:EvaporativeCoolerUnit"},
    {IddObjectType::ZoneHVAC_HybridUnitaryHVAC, "ZoneHVAC_HybridUnitaryHVAC", "ZoneHVAC:HybridUnitaryHVAC"},
    {IddObjectType::ZoneHVAC_OutdoorAirUnit, "ZoneHVAC_OutdoorAirUnit", "ZoneHVAC:OutdoorAirUnit"},
    {IddObjectType::ZoneHVAC_OutdoorAirUnit_EquipmentList, "ZoneHVAC_OutdoorAirUnit_EquipmentList", "ZoneHVAC:OutdoorAirUnit:EquipmentList"},
    {IddObjectType::ZoneHVAC_TerminalUnit_VariableRefrigerantFlow, "ZoneHVAC_TerminalUnit_VariableRefrigerantFlow",
     "ZoneHVAC:TerminalUnit:VariableRefrigerantFlow"},
    {IddObjectType::ZoneHVAC_Baseboard_RadiantConvective_Water_Design, "ZoneHVAC_Baseboard_RadiantConvective_Water_Design",
     "ZoneHVAC:Baseboard:RadiantConvective:Water:Design"},
    {IddObjectType::ZoneHVAC_Baseboard_RadiantConvective_Water, "ZoneHVAC_Baseboard_RadiantConvective_Water",
     "ZoneHVAC:Baseboard:RadiantConvective:Water"},
    {IddObjectType::ZoneHVAC_Baseboard_RadiantConvective_Steam_Design, "ZoneHVAC_Baseboard_RadiantConvective_Steam_Design",
     "ZoneHVAC:Baseboard:RadiantConvective:Steam:Design"},
    {IddObjectType::ZoneHVAC_Baseboard_RadiantConvective_Steam, "ZoneHVAC_Baseboard_RadiantConvective_Steam",
     "ZoneHVAC:Baseboard:RadiantConvective:Steam"},
    {IddObjectType::ZoneHVAC_Baseboard_RadiantConvective_Electric, "ZoneHVAC_Baseboard_RadiantConvective_Electric",
     "ZoneHVAC:Baseboard:RadiantConvective:Electric"},
    {IddObjectType::ZoneHVAC_CoolingPanel_RadiantConvective_Water, "ZoneHVAC_CoolingPanel_RadiantConvective_Water",
     "ZoneHVAC:CoolingPanel:RadiantConvective:Water"},
    {IddObjectType::ZoneHVAC_Baseboard_Convective_Water, "ZoneHVAC_Baseboard_Convective_Water", "ZoneHVAC:Baseboard:Convective:Water"},
    {IddObjectType::ZoneHVAC_Baseboard_Convective_Electric, "ZoneHVAC_Baseboard_Convective_Electric", "ZoneHVAC:Baseboard:Convective:Electric"},
    {IddObjectType::ZoneHVAC_LowTemperatureRadiant_VariableFlow, "ZoneHVAC_LowTemperatureRadiant_VariableFlow",
     "ZoneHVAC:LowTemperatureRadiant:VariableFlow"},
    {IddObjectType::ZoneHVAC_LowTemperatureRadiant_VariableFlow_Design, "ZoneHVAC_LowTemperatureRadiant_VariableFlow_Design",
     "ZoneHVAC:LowTemperatureRadiant:VariableFlow:Design"},
    {IddObjectType::ZoneHVAC_LowTemperatureRadiant_ConstantFlow, "ZoneHVAC_LowTemperatureRadiant_ConstantFlow",
     "ZoneHVAC:LowTemperatureRadiant:ConstantFlow"},
    {IddObjectType::ZoneHVAC_LowTemperatureRadiant_ConstantFlow_Design, "ZoneHVAC_LowTemperatureRadiant_ConstantFlow_Design",
     "ZoneHVAC:LowTemperatureRadiant:ConstantFlow:Design"},
    {IddObjectType::ZoneHVAC_LowTemperatureRadiant_Electric, "ZoneHVAC_LowTemperatureRadiant_Electric", "ZoneHVAC:LowTemperatureRadiant:Electric"},
    {IddObjectType::ZoneHVAC_LowTemperatureRadiant_SurfaceGroup, "ZoneHVAC_LowTemperatureRadiant_SurfaceGroup",
     "ZoneHVAC:LowTemperatureRadiant:SurfaceGroup"},
    {IddObjectType::ZoneHVAC_HighTemperatureRadiant, "ZoneHVAC_HighTemperatureRadiant", "ZoneHVAC:HighTemperatureRadiant"},
    {IddObjectType::ZoneHVAC_VentilatedSlab, "ZoneHVAC_VentilatedSlab", "ZoneHVAC:VentilatedSlab"},
    {IddObjectType::ZoneHVAC_VentilatedSlab_SlabGroup, "ZoneHVAC_VentilatedSlab_SlabGroup", "ZoneHVAC:VentilatedSlab:SlabGroup"},
    {IddObjectType::AirTerminal_SingleDuct_ConstantVolume_Reheat, "AirTerminal_SingleDuct_ConstantVolume_Reheat",
     "AirTerminal:SingleDuct:ConstantVolume:Reheat"},
    {IddObjectType::AirTerminal_SingleDuct_ConstantVolume_NoReheat, "AirTerminal_SingleDuct_ConstantVolume_NoReheat",
     "AirTerminal:SingleDuct:ConstantVolume:NoReheat"},
    {IddObjectType::AirTerminal_SingleDuct_VAV_NoReheat, "AirTerminal_SingleDuct_VAV_NoReheat", "AirTerminal:SingleDuct:VAV:NoReheat"},
    {IddObjectType::AirTerminal_SingleDuct_VAV_Reheat, "AirTerminal_SingleDuct_VAV_Reheat", "AirTerminal:SingleDuct:VAV:Reheat"},
    {IddObjectType::AirTerminal_SingleDuct_VAV_Reheat_VariableSpeedFan, "AirTerminal_SingleDuct_VAV_Reheat_VariableSpeedFan",
     "AirTerminal:SingleDuct:VAV:Reheat:VariableSpeedFan"},
    {IddObjectType::AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat, "AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat",
     "AirTerminal:SingleDuct:VAV:HeatAndCool:NoReheat"},
    {IddObjectType::AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat, "AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat",
     "AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat"},
    {IddObjectType::AirTerminal_SingleDuct_SeriesPIU_Reheat, "AirTerminal_SingleDuct_SeriesPIU_Reheat", "AirTerminal:SingleDuct:SeriesPIU:Reheat"},
    {IddObjectType::AirTerminal_SingleDuct_ParallelPIU_Reheat, "AirTerminal_SingleDuct_ParallelPIU_Reheat",
     "AirTerminal:SingleDuct:ParallelPIU:Reheat"},
    {IddObjectType::AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction, "AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction",
     "AirTerminal:SingleDuct:ConstantVolume:FourPipeInduction"},
    {IddObjectType::AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam, "AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam",
     "AirTerminal:SingleDuct:ConstantVolume:FourPipeBeam"},
    {IddObjectType::AirTerminal_SingleDuct_ConstantVolume_CooledBeam, "AirTerminal_SingleDuct_ConstantVolume_CooledBeam",
     "AirTerminal:SingleDuct:ConstantVolume:CooledBeam"},
    {IddObjectType::AirTerminal_SingleDuct_Mixer, "AirTerminal_SingleDuct_Mixer", "AirTerminal:SingleDuct:Mixer"},
    {IddObjectType::AirTerminal_DualDuct_ConstantVolume, "AirTerminal_DualDuct_ConstantVolume", "AirTerminal:DualDuct:ConstantVolume"},
    {IddObjectType::AirTerminal_DualDuct_VAV, "AirTerminal_DualDuct_VAV", "AirTerminal:DualDuct:VAV"},
    {IddObjectType::AirTerminal_DualDuct_VAV_OutdoorAir, "AirTerminal_DualDuct_VAV_OutdoorAir", "AirTerminal:DualDuct:VAV:OutdoorAir"},
    {IddObjectType::ZoneHVAC_AirDistributionUnit, "ZoneHVAC_AirDistributionUnit", "ZoneHVAC:AirDistributionUnit"},
    {IddObjectType::ZoneHVAC_ExhaustControl, "ZoneHVAC_ExhaustControl", "ZoneHVAC:ExhaustControl"},
    {IddObjectType::ZoneHVAC_EquipmentList, "ZoneHVAC_EquipmentList", "ZoneHVAC:EquipmentList"},
    {IddObjectType::ZoneHVAC_EquipmentConnections, "ZoneHVAC_EquipmentConnections", "ZoneHVAC:EquipmentConnections"},
    {IddObjectType::Fan_SystemModel, "Fan_SystemModel", "Fan:SystemModel"},
    {IddObjectType::Fan_ConstantVolume, "Fan_ConstantVolume", "Fan:ConstantVolume"},
    {IddObjectType::Fan_VariableVolume, "Fan_VariableVolume", "Fan:VariableVolume"},
    {IddObjectType::Fan_OnOff, "Fan_OnOff", "Fan:OnOff"},
    {IddObjectType::Fan_ZoneExhaust, "Fan_ZoneExhaust", "Fan:ZoneExhaust"},
    {IddObjectType::FanPerformance_NightVentilation, "FanPerformance_NightVentilation", "FanPerformance:NightVentilation"},
    {IddObjectType::Fan_ComponentModel, "Fan_ComponentModel", "Fan:ComponentModel"},
    {IddObjectType::Coil_Cooling_Water, "Coil_Cooling_Water", "Coil:Cooling:Water"},
    {IddObjectType::Coil_Cooling_Water_DetailedGeometry, "Coil_Cooling_Water_DetailedGeometry", "Coil:Cooling:Water:DetailedGeometry"},
    {IddObjectType::CoilSystem_Cooling_Water, "CoilSystem_Cooling_Water", "CoilSystem:Cooling:Water"},
    {IddObjectType::Coil_Cooling_DX, "Coil_Cooling_DX", "Coil:Cooling:DX"},
    {IddObjectType::Coil_Cooling_DX_CurveFit_Performance, "Coil_Cooling_DX_CurveFit_Performance", "Coil:Cooling:DX:CurveFit:Performance"},
    {IddObjectType::Coil_Cooling_DX_CurveFit_OperatingMode, "Coil_Cooling_DX_CurveFit_OperatingMode", "Coil:Cooling:DX:CurveFit:OperatingMode"},
    {IddObjectType::Coil_Cooling_DX_CurveFit_Speed, "Coil_Cooling_DX_CurveFit_Speed", "Coil:Cooling:DX:CurveFit:Speed"},
    {IddObjectType::Coil_Cooling_DX_SingleSpeed, "Coil_Cooling_DX_SingleSpeed", "Coil:Cooling:DX:SingleSpeed"},
    {IddObjectType::Coil_Cooling_DX_TwoSpeed, "Coil_Cooling_DX_TwoSpeed", "Coil:Cooling:DX:TwoSpeed"},
    {IddObjectType::Coil_Cooling_DX_MultiSpeed, "Coil_Cooling_DX_MultiSpeed", "Coil:Cooling:DX:MultiSpeed"},
    {IddObjectType::Coil_Cooling_DX_VariableSpeed, "Coil_Cooling_DX_VariableSpeed", "Coil:Cooling:DX:VariableSpeed"},
    {IddObjectType::Coil_Cooling_DX_TwoStageWithHumidityControlMode, "Coil_Cooling_DX_TwoStageWithHumidityControlMode",
     "Coil:Cooling:DX:TwoStageWithHumidityControlMode"},
    {IddObjectType::CoilPerformance_DX_Cooling, "CoilPerformance_DX_Cooling", "CoilPerformance:DX:Cooling"},
    {IddObjectType::Coil_Cooling_DX_VariableRefrigerantFlow, "Coil_Cooling_DX_VariableRefrigerantFlow", "Coil:Cooling:DX:VariableRefrigerantFlow"},
    {IddObjectType::Coil_Heating_DX_VariableRefrigerantFlow, "Coil_Heating_DX_VariableRefrigerantFlow", "Coil:Heating:DX:VariableRefrigerantFlow"},
    {IddObjectType::Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl,
     "Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl", "Coil:Cooling:DX:VariableRefrigerantFlow:FluidTemperatureControl"},
    {IddObjectType::Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl,
     "Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl", "Coil:Heating:DX:VariableRefrigerantFlow:FluidTemperatureControl"},
    {IddObjectType::Coil_Heating_Water, "Coil_Heating_Water", "Coil:Heating:Water"},
    {IddObjectType::Coil_Heating_Steam, "Coil_Heating_Steam", "Coil:Heating:Steam"},
    {IddObjectType::Coil_Heating_Electric, "Coil_Heating_Electric", "Coil:Heating:Electric"},
    {IddObjectType::Coil_Heating_Electric_MultiStage, "Coil_Heating_Electric_MultiStage", "Coil:Heating:Electric:MultiStage"},
    {IddObjectType::Coil_Heating_Fuel, "Coil_Heating_Fuel", "Coil:Heating:Fuel"},
    {IddObjectType::Coil_Heating_Gas_MultiStage, "Coil_Heating_Gas_MultiStage", "Coil:Heating:Gas:MultiStage"},
    {IddObjectType::Coil_Heating_Desuperheater, "Coil_Heating_Desuperheater", "Coil:Heating:Desuperheater"},
    {IddObjectType::Coil_Heating_DX_SingleSpeed, "Coil_Heating_DX_SingleSpeed", "Coil:Heating:DX:SingleSpeed"},
    {IddObjectType::Coil_Heating_DX_MultiSpeed, "Coil_Heating_DX_MultiSpeed", "Coil:Heating:DX:MultiSpeed"},
    {IddObjectType::Coil_Heating_DX_VariableSpeed, "Coil_Heating_DX_VariableSpeed", "Coil:Heating:DX:VariableSpeed"},
    {IddObjectType::Coil_Cooling_WaterToAirHeatPump_ParameterEstimation, "Coil_Cooling_WaterToAirHeatPump_ParameterEstimation",
     "Coil:Cooling:WaterToAirHeatPump:ParameterEstimation"},
    {IddObjectType::Coil_Heating_WaterToAirHeatPump_ParameterEstimation, "Coil_Heating_WaterToAirHeatPump_ParameterEstimation",
     "Coil:Heating:WaterToAirHeatPump:ParameterEstimation"},
    {IddObjectType::Coil_Cooling_WaterToAirHeatPump_EquationFit, "Coil_Cooling_WaterToAirHeatPump_EquationFit",
     "Coil:Cooling:WaterToAirHeatPump:EquationFit"},
    {IddObjectType::Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit, "Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit",
     "Coil:Cooling:WaterToAirHeatPump:VariableSpeedEquationFit"},
    {IddObjectType::Coil_Heating_WaterToAirHeatPump_EquationFit, "Coil_Heating_WaterToAirHeatPump_EquationFit",
     "Coil:Heating:WaterToAirHeatPump:EquationFit"},
    {IddObjectType::Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit, "Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit",
     "Coil:Heating:WaterToAirHeatPump:VariableSpeedEquationFit"},
    {IddObjectType::Coil_WaterHeating_AirToWaterHeatPump_Pumped, "Coil_WaterHeating_AirToWaterHeatPump_Pumped",
     "Coil:WaterHeating:AirToWaterHeatPump:Pumped"},
    {IddObjectType::Coil_WaterHeating_AirToWaterHeatPump_Wrapped, "Coil_WaterHeating_AirToWaterHeatPump_Wrapped",
     "Coil:WaterHeating:AirToWaterHeatPump:Wrapped"},
    {IddObjectType::Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed, "Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed",
     "Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed"},
    {IddObjectType::Coil_WaterHeating_Desuperheater, "Coil_WaterHeating_Desuperheater", "Coil:WaterHeating:Desuperheater"},
    {IddObjectType::CoilSystem_Cooling_DX, "CoilSystem_Cooling_DX", "CoilSystem:Cooling:DX"},
    {IddObjectType::CoilSystem_Heating_DX, "CoilSystem_Heating_DX", "CoilSystem:Heating:DX"},
    {IddObjectType::CoilSystem_Cooling_Water_HeatExchangerAssisted, "CoilSystem_Cooling_Water_HeatExchangerAssisted",
     "CoilSystem:Cooling:Water:HeatExchangerAssisted"},
    {IddObjectType::CoilSystem_Cooling_DX_HeatExchangerAssisted, "CoilSystem_Cooling_DX_HeatExchangerAssisted",
     "CoilSystem:Cooling:DX:HeatExchangerAssisted"},
    {IddObjectType::CoilSystem_IntegratedHeatPump_AirSource, "CoilSystem_IntegratedHeatPump_AirSource", "CoilSystem:IntegratedHeatPump:AirSource"},
    {IddObjectType::Coil_Cooling_DX_SingleSpeed_ThermalStorage, "Coil_Cooling_DX_SingleSpeed_ThermalStorage",
     "Coil:Cooling:DX:SingleSpeed:ThermalStorage"},
    {IddObjectType::EvaporativeCooler_Direct_CelDekPad, "EvaporativeCooler_Direct_CelDekPad", "EvaporativeCooler:Direct:CelDekPad"},
    {IddObjectType::EvaporativeCooler_Indirect_CelDekPad, "EvaporativeCooler_Indirect_CelDekPad", "EvaporativeCooler:Indirect:CelDekPad"},
    {IddObjectType::EvaporativeCooler_Indirect_WetCoil, "EvaporativeCooler_Indirect_WetCoil", "EvaporativeCooler:Indirect:WetCoil"},
    {IddObjectType::EvaporativeCooler_Indirect_ResearchSpecial, "EvaporativeCooler_Indirect_ResearchSpecial",
     "EvaporativeCooler:Indirect:ResearchSpecial"},
    {IddObjectType::EvaporativeCooler_Direct_ResearchSpecial, "EvaporativeCooler_Direct_ResearchSpecial", "EvaporativeCooler:Direct:ResearchSpecial"},
    {IddObjectType::Humidifier_Steam_Electric, "Humidifier_Steam_Electric", "Humidifier:Steam:Electric"},
    {IddObjectType::Humidifier_Steam_Gas, "Humidifier_Steam_Gas", "Humidifier:Steam:Gas"},
    {IddObjectType::Dehumidifier_Desiccant_NoFans, "Dehumidifier_Desiccant_NoFans", "Dehumidifier:Desiccant:NoFans"},
    {IddObjectType::Dehumidifier_Desiccant_System, "Dehumidifier_Desiccant_System", "Dehumidifier:Desiccant:System"},
    {IddObjectType::HeatExchanger_AirToAir_FlatPlate, "HeatExchanger_AirToAir_FlatPlate", "HeatExchanger:AirToAir:FlatPlate"},
    {IddObjectType::HeatExchanger_AirToAir_SensibleAndLatent, "HeatExchanger_AirToAir_SensibleAndLatent", "HeatExchanger:AirToAir:SensibleAndLatent"},
    {IddObjectType::HeatExchanger_Desiccant_BalancedFlow, "HeatExchanger_Desiccant_BalancedFlow", "HeatExchanger:Desiccant:BalancedFlow"},
    {IddObjectType::HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1, "HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1",
     "HeatExchanger:Desiccant:BalancedFlow:PerformanceDataType1"},
    {IddObjectType::AirLoopHVAC_UnitarySystem, "AirLoopHVAC_UnitarySystem", "AirLoopHVAC:UnitarySystem"},
    {IddObjectType::UnitarySystemPerformance_Multispeed, "UnitarySystemPerformance_Multispeed", "UnitarySystemPerformance:Multispeed"},
    {IddObjectType::AirLoopHVAC_Unitary_Furnace_HeatOnly, "AirLoopHVAC_Unitary_Furnace_HeatOnly", "AirLoopHVAC:Unitary:Furnace:HeatOnly"},
    {IddObjectType::AirLoopHVAC_Unitary_Furnace_HeatCool, "AirLoopHVAC_Unitary_Furnace_HeatCool", "AirLoopHVAC:Unitary:Furnace:HeatCool"},
    {IddObjectType::AirLoopHVAC_UnitaryHeatOnly, "AirLoopHVAC_UnitaryHeatOnly", "AirLoopHVAC:UnitaryHeatOnly"},
    {IddObjectType::AirLoopHVAC_UnitaryHeatCool, "AirLoopHVAC_UnitaryHeatCool", "AirLoopHVAC:UnitaryHeatCool"},
    {IddObjectType::AirLoopHVAC_UnitaryHeatPump_AirToAir, "AirLoopHVAC_UnitaryHeatPump_AirToAir", "AirLoopHVAC:UnitaryHeatPump:AirToAir"},
    {IddObjectType::AirLoopHVAC_UnitaryHeatPump_WaterToAir, "AirLoopHVAC_UnitaryHeatPump_WaterToAir", "AirLoopHVAC:UnitaryHeatPump:WaterToAir"},
    {IddObjectType::AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass, "AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass",
     "AirLoopHVAC:UnitaryHeatCool:VAVChangeoverBypass"},
    {IddObjectType::AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed, "AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed",
     "AirLoopHVAC:UnitaryHeatPump:AirToAir:MultiSpeed"},
    {IddObjectType::AirConditioner_VariableRefrigerantFlow, "AirConditioner_VariableRefrigerantFlow", "AirConditioner:VariableRefrigerantFlow"},
    {IddObjectType::AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl, "AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl",
     "AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl"},
    {IddObjectType::AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR,
     "AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR", "AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl:HR"},
    {IddObjectType::ZoneTerminalUnitList, "ZoneTerminalUnitList", "ZoneTerminalUnitList"},
    {IddObjectType::Controller_WaterCoil, "Controller_WaterCoil", "Controller:WaterCoil"},
    {IddObjectType::Controller_OutdoorAir, "Controller_OutdoorAir", "Controller:OutdoorAir"},
    {IddObjectType::Controller_MechanicalVentilation, "Controller_MechanicalVentilation", "Controller:MechanicalVentilation"},
    {IddObjectType::AirLoopHVAC_ControllerList, "AirLoopHVAC_ControllerList", "AirLoopHVAC:ControllerList"},
    {IddObjectType::AirLoopHVAC, "AirLoopHVAC", "AirLoopHVAC"},
    {IddObjectType::AirLoopHVAC_OutdoorAirSystem_EquipmentList, "AirLoopHVAC_OutdoorAirSystem_EquipmentList",
     "AirLoopHVAC:OutdoorAirSystem:EquipmentList"},
    {IddObjectType::AirLoopHVAC_OutdoorAirSystem, "AirLoopHVAC_OutdoorAirSystem", "AirLoopHVAC:OutdoorAirSystem"},
    {IddObjectType::OutdoorAir_Mixer, "OutdoorAir_Mixer", "OutdoorAir:Mixer"},
    {IddObjectType::AirLoopHVAC_ZoneSplitter, "AirLoopHVAC_ZoneSplitter", "AirLoopHVAC:ZoneSplitter"},
    {IddObjectType::AirLoopHVAC_SupplyPlenum, "AirLoopHVAC_SupplyPlenum", "AirLoopHVAC:SupplyPlenum"},
    {IddObjectType::AirLoopHVAC_SupplyPath, "AirLoopHVAC_SupplyPath", "AirLoopHVAC:SupplyPath"},
    {IddObjectType::AirLoopHVAC_ZoneMixer, "AirLoopHVAC_ZoneMixer", "AirLoopHVAC:ZoneMixer"},
    {IddObjectType::AirLoopHVAC_ReturnPlenum, "AirLoopHVAC_ReturnPlenum", "AirLoopHVAC:ReturnPlenum"},
    {IddObjectType::AirLoopHVAC_ReturnPath, "AirLoopHVAC_ReturnPath", "AirLoopHVAC:ReturnPath"},
    {IddObjectType::AirLoopHVAC_ExhaustSystem, "AirLoopHVAC_ExhaustSystem", "AirLoopHVAC:ExhaustSystem"},
    {IddObjectType::AirLoopHVAC_DedicatedOutdoorAirSystem, "AirLoopHVAC_DedicatedOutdoorAirSystem", "AirLoopHVAC:DedicatedOutdoorAirSystem"},
    {IddObjectType::AirLoopHVAC_Mixer, "AirLoopHVAC_Mixer", "AirLoopHVAC:Mixer"},
    {IddObjectType::AirLoopHVAC_Splitter, "AirLoopHVAC_Splitter", "AirLoopHVAC:Splitter"},
    {IddObjectType::Branch, "Branch", "Branch"},
    {IddObjectType::BranchList, "BranchList", "BranchList"},
    {IddObjectType::Connector_Splitter, "Connector_Splitter", "Connector:Splitter"},
    {IddObjectType::Connector_Mixer, "Connector_Mixer", "Connector:Mixer"},
    {IddObjectType::ConnectorList, "ConnectorList", "ConnectorList"},
    {IddObjectType::NodeList, "NodeList", "NodeList"},
    {IddObjectType::OutdoorAir_Node, "OutdoorAir_Node", "OutdoorAir:Node"},
    {IddObjectType::OutdoorAir_NodeList, "OutdoorAir_NodeList", "OutdoorAir:NodeList"},
    {IddObjectType::Pipe_Adiabatic, "Pipe_Adiabatic", "Pipe:Adiabatic"},
    {IddObjectType::Pipe_Adiabatic_Steam, "Pipe_Adiabatic_Steam", "Pipe:Adiabatic:Steam"},
    {IddObjectType::Pipe_Indoor, "Pipe_Indoor", "Pipe:Indoor"},
    {IddObjectType::Pipe_Outdoor, "Pipe_Outdoor", "Pipe:Outdoor"},
    {IddObjectType::Pipe_Underground, "Pipe_Underground", "Pipe:Underground"},
    {IddObjectType::PipingSystem_Underground_Domain, "PipingSystem_Underground_Domain", "PipingSystem:Underground:Domain"},
    {IddObjectType::PipingSystem_Underground_PipeCircuit, "PipingSystem_Underground_PipeCircuit", "PipingSystem:Underground:PipeCircuit"},
    {IddObjectType::PipingSystem_Underground_PipeSegment, "PipingSystem_Underground_PipeSegment", "PipingSystem:Underground:PipeSegment"},
    {IddObjectType::Duct, "Duct", "Duct"},
    {IddObjectType::Pump_VariableSpeed, "Pump_VariableSpeed", "Pump:VariableSpeed"},
    {IddObjectType::Pump_ConstantSpeed, "Pump_ConstantSpeed", "Pump:ConstantSpeed"},
    {IddObjectType::Pump_VariableSpeed_Condensate, "Pump_VariableSpeed_Condensate", "Pump:VariableSpeed:Condensate"},
    {IddObjectType::HeaderedPumps_ConstantSpeed, "HeaderedPumps_ConstantSpeed", "HeaderedPumps:ConstantSpeed"},
    {IddObjectType::HeaderedPumps_VariableSpeed, "HeaderedPumps_VariableSpeed", "HeaderedPumps:VariableSpeed"},
    {IddObjectType::TemperingValve, "TemperingValve", "TemperingValve"},
    {IddObjectType::LoadProfile_Plant, "LoadProfile_Plant", "LoadProfile:Plant"},
    {IddObjectType::SolarCollectorPerformance_FlatPlate, "SolarCollectorPerformance_FlatPlate", "SolarCollectorPerformance:FlatPlate"},
    {IddObjectType::SolarCollector_FlatPlate_Water, "SolarCollector_FlatPlate_Water", "SolarCollector:FlatPlate:Water"},
    {IddObjectType::SolarCollector_FlatPlate_PhotovoltaicThermal, "SolarCollector_FlatPlate_PhotovoltaicThermal",
     "SolarCollector:FlatPlate:PhotovoltaicThermal"},
    {IddObjectType::SolarCollectorPerformance_PhotovoltaicThermal_Simple, "SolarCollectorPerformance_PhotovoltaicThermal_Simple",
     "SolarCollectorPerformance:PhotovoltaicThermal:Simple"},
    {IddObjectType::SolarCollectorPerformance_PhotovoltaicThermal_BIPVT, "SolarCollectorPerformance_PhotovoltaicThermal_BIPVT",
     "SolarCollectorPerformance:PhotovoltaicThermal:BIPVT"},
    {IddObjectType::SolarCollector_IntegralCollectorStorage, "SolarCollector_IntegralCollectorStorage", "SolarCollector:IntegralCollectorStorage"},
    {IddObjectType::SolarCollectorPerformance_IntegralCollectorStorage, "SolarCollectorPerformance_IntegralCollectorStorage",
     "SolarCollectorPerformance:IntegralCollectorStorage"},
    {IddObjectType::SolarCollector_UnglazedTranspired, "SolarCollector_UnglazedTranspired", "SolarCollector:UnglazedTranspired"},
    {IddObjectType::SolarCollector_UnglazedTranspired_Multisystem, "SolarCollector_UnglazedTranspired_Multisystem",
     "SolarCollector:UnglazedTranspired:Multisystem"},
    {IddObjectType::Boiler_HotWater, "Boiler_HotWater", "Boiler:HotWater"},
    {IddObjectType::Boiler_Steam, "Boiler_Steam", "Boiler:Steam"},
    {IddObjectType::Chiller_Electric_ASHRAE205, "Chiller_Electric_ASHRAE205", "Chiller:Electric:ASHRAE205"},
    {IddObjectType::Chiller_Electric_EIR, "Chiller_Electric_EIR", "Chiller:Electric:EIR"},
    {IddObjectType::Chiller_Electric_ReformulatedEIR, "Chiller_Electric_ReformulatedEIR", "Chiller:Electric:ReformulatedEIR"},
    {IddObjectType::Chiller_Electric, "Chiller_Electric", "Chiller:Electric"},
    {IddObjectType::Chiller_Absorption_Indirect, "Chiller_Absorption_Indirect", "Chiller:Absorption:Indirect"},
    {IddObjectType::Chiller_Absorption, "Chiller_Absorption", "Chiller:Absorption"},
    {IddObjectType::Chiller_ConstantCOP, "Chiller_ConstantCOP", "Chiller:ConstantCOP"},
    {IddObjectType::Chiller_EngineDriven, "Chiller_EngineDriven", "Chiller:EngineDriven"},
    {IddObjectType::Chiller_CombustionTurbine, "Chiller_CombustionTurbine", "Chiller:CombustionTurbine"},
    {IddObjectType::ChillerHeater_Absorption_DirectFired, "ChillerHeater_Absorption_DirectFired", "ChillerHeater:Absorption:DirectFired"},
    {IddObjectType::ChillerHeater_Absorption_DoubleEffect, "ChillerHeater_Absorption_DoubleEffect", "ChillerHeater:Absorption:DoubleEffect"},
    {IddObjectType::HeatPump_PlantLoop_EIR_Cooling, "HeatPump_PlantLoop_EIR_Cooling", "HeatPump:PlantLoop:EIR:Cooling"},
    {IddObjectType::HeatPump_PlantLoop_EIR_Heating, "HeatPump_PlantLoop_EIR_Heating", "HeatPump:PlantLoop:EIR:Heating"},
    {IddObjectType::HeatPump_AirToWater_FuelFired_Heating, "HeatPump_AirToWater_FuelFired_Heating", "HeatPump:AirToWater:FuelFired:Heating"},
    {IddObjectType::HeatPump_AirToWater_FuelFired_Cooling, "HeatPump_AirToWater_FuelFired_Cooling", "HeatPump:AirToWater:FuelFired:Cooling"},
    {IddObjectType::HeatPump_WaterToWater_EquationFit_Heating, "HeatPump_WaterToWater_EquationFit_Heating",
     "HeatPump:WaterToWater:EquationFit:Heating"},
    {IddObjectType::HeatPump_WaterToWater_EquationFit_Cooling, "HeatPump_WaterToWater_EquationFit_Cooling",
     "HeatPump:WaterToWater:EquationFit:Cooling"},
    {IddObjectType::HeatPump_WaterToWater_ParameterEstimation_Cooling, "HeatPump_WaterToWater_ParameterEstimation_Cooling",
     "HeatPump:WaterToWater:ParameterEstimation:Cooling"},
    {IddObjectType::HeatPump_WaterToWater_ParameterEstimation_Heating, "HeatPump_WaterToWater_ParameterEstimation_Heating",
     "HeatPump:WaterToWater:ParameterEstimation:Heating"},
    {IddObjectType::DistrictCooling, "DistrictCooling", "DistrictCooling"},
    {IddObjectType::DistrictHeating, "DistrictHeating", "DistrictHeating"},
    {IddObjectType::PlantComponent_TemperatureSource, "PlantComponent_TemperatureSource", "PlantComponent:TemperatureSource"},
    {IddObjectType::CentralHeatPumpSystem, "CentralHeatPumpSystem", "CentralHeatPumpSystem"},
    {IddObjectType::ChillerHeaterPerformance_Electric_EIR, "ChillerHeaterPerformance_Electric_EIR", "ChillerHeaterPerformance:Electric:EIR"},
    {IddObjectType::CoolingTower_SingleSpeed, "CoolingTower_SingleSpeed", "CoolingTower:SingleSpeed"},
    {IddObjectType::CoolingTower_TwoSpeed, "CoolingTower_TwoSpeed", "CoolingTower:TwoSpeed"},
    {IddObjectType::CoolingTower_VariableSpeed_Merkel, "CoolingTower_VariableSpeed_Merkel", "CoolingTower:VariableSpeed:Merkel"},
    {IddObjectType::CoolingTower_VariableSpeed, "CoolingTower_VariableSpeed", "CoolingTower:VariableSpeed"},
    {IddObjectType::CoolingTowerPerformance_CoolTools, "CoolingTowerPerformance_CoolTools", "CoolingTowerPerformance:CoolTools"},
    {IddObjectType::CoolingTowerPerformance_YorkCalc, "CoolingTowerPerformance_YorkCalc", "CoolingTowerPerformance:YorkCalc"},
    {IddObjectType::EvaporativeFluidCooler_SingleSpeed, "EvaporativeFluidCooler_SingleSpeed", "EvaporativeFluidCooler:SingleSpeed"},
    {IddObjectType::EvaporativeFluidCooler_TwoSpeed, "EvaporativeFluidCooler_TwoSpeed", "EvaporativeFluidCooler:TwoSpeed"},
    {IddObjectType::FluidCooler_SingleSpeed, "FluidCooler_SingleSpeed", "FluidCooler:SingleSpeed"},
    {IddObjectType::FluidCooler_TwoSpeed, "FluidCooler_TwoSpeed", "FluidCooler:TwoSpeed"},
    {IddObjectType::GroundHeatExchanger_System, "GroundHeatExchanger_System", "GroundHeatExchanger:System"},
    {IddObjectType::GroundHeatExchanger_Vertical_Properties, "GroundHeatExchanger_Vertical_Properties", "GroundHeatExchanger:Vertical:Properties"},
    {IddObjectType::GroundHeatExchanger_Vertical_Array, "GroundHeatExchanger_Vertical_Array", "GroundHeatExchanger:Vertical:Array"},
    {IddObjectType::GroundHeatExchanger_Vertical_Single, "GroundHeatExchanger_Vertical_Single", "GroundHeatExchanger:Vertical:Single"},
    {IddObjectType::GroundHeatExchanger_ResponseFactors, "GroundHeatExchanger_ResponseFactors", "GroundHeatExchanger:ResponseFactors"},
    {IddObjectType::GroundHeatExchanger_Pond, "GroundHeatExchanger_Pond", "GroundHeatExchanger:Pond"},
    {IddObjectType::GroundHeatExchanger_Surface, "GroundHeatExchanger_Surface", "GroundHeatExchanger:Surface"},
    {IddObjectType::GroundHeatExchanger_HorizontalTrench, "GroundHeatExchanger_HorizontalTrench", "GroundHeatExchanger:HorizontalTrench"},
    {IddObjectType::GroundHeatExchanger_Slinky, "GroundHeatExchanger_Slinky", "GroundHeatExchanger:Slinky"},
    {IddObjectType::HeatExchanger_FluidToFluid, "HeatExchanger_FluidToFluid", "HeatExchanger:FluidToFluid"},
    {IddObjectType::WaterHeater_Mixed, "WaterHeater_Mixed", "WaterHeater:Mixed"},
    {IddObjectType::WaterHeater_Stratified, "WaterHeater_Stratified", "WaterHeater:Stratified"},
    {IddObjectType::WaterHeater_Sizing, "WaterHeater_Sizing", "WaterHeater:Sizing"},
    {IddObjectType::WaterHeater_HeatPump_PumpedCondenser, "WaterHeater_HeatPump_PumpedCondenser", "WaterHeater:HeatPump:PumpedCondenser"},
    {IddObjectType::WaterHeater_HeatPump_WrappedCondenser, "WaterHeater_HeatPump_WrappedCondenser", "WaterHeater:HeatPump:WrappedCondenser"},
    {IddObjectType::ThermalStorage_Ice_Simple, "ThermalStorage_Ice_Simple", "ThermalStorage:Ice:Simple"},
    {IddObjectType::ThermalStorage_Ice_Detailed, "ThermalStorage_Ice_Detailed", "ThermalStorage:Ice:Detailed"},
    {IddObjectType::ThermalStorage_ChilledWater_Mixed, "ThermalStorage_ChilledWater_Mixed", "ThermalStorage:ChilledWater:Mixed"},
    {IddObjectType::ThermalStorage_ChilledWater_Stratified, "ThermalStorage_ChilledWater_Stratified", "ThermalStorage:ChilledWater:Stratified"},
    {IddObjectType::PlantLoop, "PlantLoop", "PlantLoop"},
    {IddObjectType::CondenserLoop, "CondenserLoop", "CondenserLoop"},
    {IddObjectType::PlantEquipmentList, "PlantEquipmentList", "PlantEquipmentList"},
    {IddObjectType::CondenserEquipmentList, "CondenserEquipmentList", "CondenserEquipmentList"},
    {IddObjectType::PlantEquipmentOperation_Uncontrolled, "PlantEquipmentOperation_Uncontrolled", "PlantEquipmentOperation:Uncontrolled"},
    {IddObjectType::PlantEquipmentOperation_CoolingLoad, "PlantEquipmentOperation_CoolingLoad", "PlantEquipmentOperation:CoolingLoad"},
    {IddObjectType::PlantEquipmentOperation_HeatingLoad, "PlantEquipmentOperation_HeatingLoad", "PlantEquipmentOperation:HeatingLoad"},
    {IddObjectType::PlantEquipmentOperation_OutdoorDryBulb, "PlantEquipmentOperation_OutdoorDryBulb", "PlantEquipmentOperation:OutdoorDryBulb"},
    {IddObjectType::PlantEquipmentOperation_OutdoorWetBulb, "PlantEquipmentOperation_OutdoorWetBulb", "PlantEquipmentOperation:OutdoorWetBulb"},
    {IddObjectType::PlantEquipmentOperation_OutdoorRelativeHumidity, "PlantEquipmentOperation_OutdoorRelativeHumidity",
     "PlantEquipmentOperation:OutdoorRelativeHumidity"},
    {IddObjectType::PlantEquipmentOperation_OutdoorDewpoint, "PlantEquipmentOperation_OutdoorDewpoint", "PlantEquipmentOperation:OutdoorDewpoint"},
    {IddObjectType::PlantEquipmentOperation_ComponentSetpoint, "PlantEquipmentOperation_ComponentSetpoint",
     "PlantEquipmentOperation:ComponentSetpoint"},
    {IddObjectType::PlantEquipmentOperation_ThermalEnergyStorage, "PlantEquipmentOperation_ThermalEnergyStorage",
     "PlantEquipmentOperation:ThermalEnergyStorage"},
    {IddObjectType::PlantEquipmentOperation_OutdoorDryBulbDifference, "PlantEquipmentOperation_OutdoorDryBulbDifference",
     "PlantEquipmentOperation:OutdoorDryBulbDifference"},
    {IddObjectType::PlantEquipmentOperation_OutdoorWetBulbDifference, "PlantEquipmentOperation_OutdoorWetBulbDifference",
     "PlantEquipmentOperation:OutdoorWetBulbDifference"},
    {IddObjectType::PlantEquipmentOperation_OutdoorDewpointDifference, "PlantEquipmentOperation_OutdoorDewpointDifference",
     "PlantEquipmentOperation:OutdoorDewpointDifference"},
    {IddObjectType::PlantEquipmentOperationSchemes, "PlantEquipmentOperationSchemes", "PlantEquipmentOperationSchemes"},
    {IddObjectType::CondenserEquipmentOperationSchemes, "CondenserEquipmentOperationSchemes", "CondenserEquipmentOperationSchemes"},
    {IddObjectType::EnergyManagementSystem_Sensor, "EnergyManagementSystem_Sensor", "EnergyManagementSystem:Sensor"},
    {IddObjectType::EnergyManagementSystem_Actuator, "EnergyManagementSystem_Actuator", "EnergyManagementSystem:Actuator"},
    {IddObjectType::EnergyManagementSystem_ProgramCallingManager, "EnergyManagementSystem_ProgramCallingManager",
     "EnergyManagementSystem:ProgramCallingManager"},
    {IddObjectType::EnergyManagementSystem_Program, "EnergyManagementSystem_Program", "EnergyManagementSystem:Program"},
    {IddObjectType::EnergyManagementSystem_Subroutine, "EnergyManagementSystem_Subroutine", "EnergyManagementSystem:Subroutine"},
    {IddObjectType::EnergyManagementSystem_GlobalVariable, "EnergyManagementSystem_GlobalVariable", "EnergyManagementSystem:GlobalVariable"},
    {IddObjectType::EnergyManagementSystem_OutputVariable, "EnergyManagementSystem_OutputVariable", "EnergyManagementSystem:OutputVariable"},
    {IddObjectType::EnergyManagementSystem_MeteredOutputVariable, "EnergyManagementSystem_MeteredOutputVariable",
     "EnergyManagementSystem:MeteredOutputVariable"},
    {IddObjectType::EnergyManagementSystem_TrendVariable, "EnergyManagementSystem_TrendVariable", "EnergyManagementSystem:TrendVariable"},
    {IddObjectType::EnergyManagementSystem_InternalVariable, "EnergyManagementSystem_InternalVariable", "EnergyManagementSystem:InternalVariable"},
    {IddObjectType::EnergyManagementSystem_CurveOrTableIndexVariable, "EnergyManagementSystem_CurveOrTableIndexVariable",
     "EnergyManagementSystem:CurveOrTableIndexVariable"},
    {IddObjectType::EnergyManagementSystem_ConstructionIndexVariable, "EnergyManagementSystem_ConstructionIndexVariable",
     "EnergyManagementSystem:ConstructionIndexVariable"},
    {IddObjectType::ExternalInterface, "ExternalInterface", "ExternalInterface"},
    {IddObjectType::ExternalInterface_Schedule, "ExternalInterface_Schedule", "ExternalInterface:Schedule"},
    {IddObjectType::ExternalInterface_Variable, "ExternalInterface_Variable", "ExternalInterface:Variable"},
    {IddObjectType::ExternalInterface_Actuator, "ExternalInterface_Actuator", "ExternalInterface:Actuator"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitImport, "ExternalInterface_FunctionalMockupUnitImport",
     "ExternalInterface:FunctionalMockupUnitImport"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitImport_From_Variable, "ExternalInterface_FunctionalMockupUnitImport_From_Variable",
     "ExternalInterface:FunctionalMockupUnitImport:From:Variable"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitImport_To_Schedule, "ExternalInterface_FunctionalMockupUnitImport_To_Schedule",
     "ExternalInterface:FunctionalMockupUnitImport:To:Schedule"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitImport_To_Actuator, "ExternalInterface_FunctionalMockupUnitImport_To_Actuator",
     "ExternalInterface:FunctionalMockupUnitImport:To:Actuator"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitImport_To_Variable, "ExternalInterface_FunctionalMockupUnitImport_To_Variable",
     "ExternalInterface:FunctionalMockupUnitImport:To:Variable"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitExport_From_Variable, "ExternalInterface_FunctionalMockupUnitExport_From_Variable",
     "ExternalInterface:FunctionalMockupUnitExport:From:Variable"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitExport_To_Schedule, "ExternalInterface_FunctionalMockupUnitExport_To_Schedule",
     "ExternalInterface:FunctionalMockupUnitExport:To:Schedule"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitExport_To_Actuator, "ExternalInterface_FunctionalMockupUnitExport_To_Actuator",
     "ExternalInterface:FunctionalMockupUnitExport:To:Actuator"},
    {IddObjectType::ExternalInterface_FunctionalMockupUnitExport_To_Variable, "ExternalInterface_FunctionalMockupUnitExport_To_Variable",
     "ExternalInterface:FunctionalMockupUnitExport:To:Variable"},
    {IddObjectType::ZoneHVAC_ForcedAir_UserDefined, "ZoneHVAC_ForcedAir_UserDefined", "ZoneHVAC:ForcedAir:UserDefined"},
    {IddObjectType::AirTerminal_SingleDuct_UserDefined, "AirTerminal_SingleDuct_UserDefined", "AirTerminal:SingleDuct:UserDefined"},
    {IddObjectType::Coil_UserDefined, "Coil_UserDefined", "Coil:UserDefined"},
    {IddObjectType::PlantComponent_UserDefined, "PlantComponent_UserDefined", "PlantComponent:UserDefined"},
    {IddObjectType::PlantEquipmentOperation_UserDefined, "PlantEquipmentOperation_UserDefined", "PlantEquipmentOperation:UserDefined"},
    {IddObjectType::AvailabilityManager_Scheduled, "AvailabilityManager_Scheduled", "AvailabilityManager:Scheduled"},
    {IddObjectType::AvailabilityManager_ScheduledOn, "AvailabilityManager_ScheduledOn", "AvailabilityManager:ScheduledOn"},
    {IddObjectType::AvailabilityManager_ScheduledOff, "AvailabilityManager_ScheduledOff", "AvailabilityManager:ScheduledOff"},
    {IddObjectType::AvailabilityManager_OptimumStart, "AvailabilityManager_OptimumStart", "AvailabilityManager:OptimumStart"},
    {IddObjectType::AvailabilityManager_NightCycle, "AvailabilityManager_NightCycle", "AvailabilityManager:NightCycle"},
    {IddObjectType::AvailabilityManager_DifferentialThermostat, "AvailabilityManager_DifferentialThermostat",
     "AvailabilityManager:DifferentialThermostat"},
    {IddObjectType::AvailabilityManager_HighTemperatureTurnOff, "AvailabilityManager_HighTemperatureTurnOff",
     "AvailabilityManager:HighTemperatureTurnOff"},
    {IddObjectType::AvailabilityManager_HighTemperatureTurnOn, "AvailabilityManager_HighTemperatureTurnOn",
     "AvailabilityManager:HighTemperatureTurnOn"},
    {IddObjectType::AvailabilityManager_LowTemperatureTurnOff, "AvailabilityManager_LowTemperatureTurnOff",
     "AvailabilityManager:LowTemperatureTurnOff"},
    {IddObjectType::AvailabilityManager_LowTemperatureTurnOn, "AvailabilityManager_LowTemperatureTurnOn", "AvailabilityManager:LowTemperatureTurnOn"},
    {IddObjectType::AvailabilityManager_NightVentilation, "AvailabilityManager_NightVentilation", "AvailabilityManager:NightVentilation"},
    {IddObjectType::AvailabilityManager_HybridVentilation, "AvailabilityManager_HybridVentilation", "AvailabilityManager:HybridVentilation"},
    {IddObjectType::AvailabilityManagerAssignmentList, "AvailabilityManagerAssignmentList", "AvailabilityManagerAssignmentList"},
    {IddObjectType::SetpointManager_Scheduled, "SetpointManager_Scheduled", "SetpointManager:Scheduled"},
    {IddObjectType::SetpointManager_Scheduled_DualSetpoint, "SetpointManager_Scheduled_DualSetpoint", "SetpointManager:Scheduled:DualSetpoint"},
    {IddObjectType::SetpointManager_OutdoorAirReset, "SetpointManager_OutdoorAirReset", "SetpointManager:OutdoorAirReset"},
    {IddObjectType::SetpointManager_SingleZone_Reheat, "SetpointManager_SingleZone_Reheat", "SetpointManager:SingleZone:Reheat"},
    {IddObjectType::SetpointManager_SingleZone_Heating, "SetpointManager_SingleZone_Heating", "SetpointManager:SingleZone:Heating"},
    {IddObjectType::SetpointManager_SingleZone_Cooling, "SetpointManager_SingleZone_Cooling", "SetpointManager:SingleZone:Cooling"},
    {IddObjectType::SetpointManager_SingleZone_Humidity_Minimum, "SetpointManager_SingleZone_Humidity_Minimum",
     "SetpointManager:SingleZone:Humidity:Minimum"},
    {IddObjectType::SetpointManager_SingleZone_Humidity_Maximum, "SetpointManager_SingleZone_Humidity_Maximum",
     "SetpointManager:SingleZone:Humidity:Maximum"},
    {IddObjectType::SetpointManager_MixedAir, "SetpointManager_MixedAir", "SetpointManager:MixedAir"},
    {IddObjectType::SetpointManager_OutdoorAirPretreat, "SetpointManager_OutdoorAirPretreat", "SetpointManager:OutdoorAirPretreat"},
    {IddObjectType::SetpointManager_Warmest, "SetpointManager_Warmest", "SetpointManager:Warmest"},
    {IddObjectType::SetpointManager_Coldest, "SetpointManager_Coldest", "SetpointManager:Coldest"},
    {IddObjectType::SetpointManager_ReturnAirBypassFlow, "SetpointManager_ReturnAirBypassFlow", "SetpointManager:ReturnAirBypassFlow"},
    {IddObjectType::SetpointManager_WarmestTemperatureFlow, "SetpointManager_WarmestTemperatureFlow", "SetpointManager:WarmestTemperatureFlow"},
    {IddObjectType::SetpointManager_MultiZone_Heating_Average, "SetpointManager_MultiZone_Heating_Average",
     "SetpointManager:MultiZone:Heating:Average"},
    {IddObjectType::SetpointManager_MultiZone_Cooling_Average, "SetpointManager_MultiZone_Cooling_Average",
     "SetpointManager:MultiZone:Cooling:Average"},
    {IddObjectType::SetpointManager_MultiZone_MinimumHumidity_Average, "SetpointManager_MultiZone_MinimumHumidity_Average",
     "SetpointManager:MultiZone:MinimumHumidity:Average"},
    {IddObjectType::SetpointManager_MultiZone_MaximumHumidity_Average, "SetpointManager_MultiZone_MaximumHumidity_Average",
     "SetpointManager:MultiZone:MaximumHumidity:Average"},
    {IddObjectType::SetpointManager_MultiZone_Humidity_Minimum, "SetpointManager_MultiZone_Humidity_Minimum",
     "SetpointManager:MultiZone:Humidity:Minimum"},
    {IddObjectType::SetpointManager_MultiZone_Humidity_Maximum, "SetpointManager_MultiZone_Humidity_Maximum",
     "SetpointManager:MultiZone:Humidity:Maximum"},
    {IddObjectType::SetpointManager_FollowOutdoorAirTemperature, "SetpointManager_FollowOutdoorAirTemperature",
     "SetpointManager:FollowOutdoorAirTemperature"},
    {IddObjectType::SetpointManager_FollowSystemNodeTemperature, "SetpointManager_FollowSystemNodeTemperature",
     "SetpointManager:FollowSystemNodeTemperature"},
    {IddObjectType::SetpointManager_FollowGroundTemperature, "SetpointManager_FollowGroundTemperature", "SetpointManager:FollowGroundTemperature"},
    {IddObjectType::SetpointManager_CondenserEnteringReset, "SetpointManager_CondenserEnteringReset", "SetpointManager:CondenserEnteringReset"},
    {IddObjectType::SetpointManager_CondenserEnteringReset_Ideal, "SetpointManager_CondenserEnteringReset_Ideal",
     "SetpointManager:CondenserEnteringReset:Ideal"},
    {IddObjectType::SetpointManager_SingleZone_OneStageCooling, "SetpointManager_SingleZone_OneStageCooling",
     "SetpointManager:SingleZone:OneStageCooling"},
    {IddObjectType::SetpointManager_SingleZone_OneStageHeating, "SetpointManager_SingleZone_OneStageHeating",
     "SetpointManager:SingleZone:OneStageHeating"},
    {IddObjectType::SetpointManager_ReturnTemperature_ChilledWater, "SetpointManager_ReturnTemperature_ChilledWater",
     "SetpointManager:ReturnTemperature:ChilledWater"},
    {IddObjectType::SetpointManager_ReturnTemperature_HotWater, "SetpointManager_ReturnTemperature_HotWater",
     "SetpointManager:ReturnTemperature:HotWater"},
    {IddObjectType::SetpointManager_SystemNodeReset_Temperature, "SetpointManager_SystemNodeReset_Temperature",
     "SetpointManager:SystemNodeReset:Temperature"},
    {IddObjectType::SetpointManager_SystemNodeReset_Humidity, "SetpointManager_SystemNodeReset_Humidity", "SetpointManager:SystemNodeReset:Humidity"},
    {IddObjectType::Refrigeration_Case, "Refrigeration_Case", "Refrigeration:Case"},
    {IddObjectType::Refrigeration_CompressorRack, "Refrigeration_CompressorRack", "Refrigeration:CompressorRack"},
    {IddObjectType::Refrigeration_CaseAndWalkInList, "Refrigeration_CaseAndWalkInList", "Refrigeration:CaseAndWalkInList"},
    {IddObjectType::Refrigeration_Condenser_AirCooled, "Refrigeration_Condenser_AirCooled", "Refrigeration:Condenser:AirCooled"},
    {IddObjectType::Refrigeration_Condenser_EvaporativeCooled, "Refrigeration_Condenser_EvaporativeCooled",
     "Refrigeration:Condenser:EvaporativeCooled"},
    {IddObjectType::Refrigeration_Condenser_WaterCooled, "Refrigeration_Condenser_WaterCooled", "Refrigeration:Condenser:WaterCooled"},
    {IddObjectType::Refrigeration_Condenser_Cascade, "Refrigeration_Condenser_Cascade", "Refrigeration:Condenser:Cascade"},
    {IddObjectType::Refrigeration_GasCooler_AirCooled, "Refrigeration_GasCooler_AirCooled", "Refrigeration:GasCooler:AirCooled"},
    {IddObjectType::Refrigeration_TransferLoadList, "Refrigeration_TransferLoadList", "Refrigeration:TransferLoadList"},
    {IddObjectType::Refrigeration_Subcooler, "Refrigeration_Subcooler", "Refrigeration:Subcooler"},
    {IddObjectType::Refrigeration_Compressor, "Refrigeration_Compressor", "Refrigeration:Compressor"},
    {IddObjectType::Refrigeration_CompressorList, "Refrigeration_CompressorList", "Refrigeration:CompressorList"},
    {IddObjectType::Refrigeration_System, "Refrigeration_System", "Refrigeration:System"},
    {IddObjectType::Refrigeration_TranscriticalSystem, "Refrigeration_TranscriticalSystem", "Refrigeration:TranscriticalSystem"},
    {IddObjectType::Refrigeration_SecondarySystem, "Refrigeration_SecondarySystem", "Refrigeration:SecondarySystem"},
    {IddObjectType::Refrigeration_WalkIn, "Refrigeration_WalkIn", "Refrigeration:WalkIn"},
    {IddObjectType::Refrigeration_AirChiller, "Refrigeration_AirChiller", "Refrigeration:AirChiller"},
    {IddObjectType::ZoneHVAC_RefrigerationChillerSet, "ZoneHVAC_RefrigerationChillerSet", "ZoneHVAC:RefrigerationChillerSet"},
    {IddObjectType::DemandManagerAssignmentList, "DemandManagerAssignmentList", "DemandManagerAssignmentList"},
    {IddObjectType::DemandManager_ExteriorLights, "DemandManager_ExteriorLights", "DemandManager:ExteriorLights"},
    {IddObjectType::DemandManager_Lights, "DemandManager_Lights", "DemandManager:Lights"},
    {IddObjectType::DemandManager_ElectricEquipment, "DemandManager_ElectricEquipment", "DemandManager:ElectricEquipment"},
    {IddObjectType::DemandManager_Thermostats, "DemandManager_Thermostats", "DemandManager:Thermostats"},
    {IddObjectType::DemandManager_Ventilation, "DemandManager_Ventilation", "DemandManager:Ventilation"},
    {IddObjectType::Generator_InternalCombustionEngine, "Generator_InternalCombustionEngine", "Generator:InternalCombustionEngine"},
    {IddObjectType::Generator_CombustionTurbine, "Generator_CombustionTurbine", "Generator:CombustionTurbine"},
    {IddObjectType::Generator_MicroTurbine, "Generator_MicroTurbine", "Generator:MicroTurbine"},
    {IddObjectType::Generator_Photovoltaic, "Generator_Photovoltaic", "Generator:Photovoltaic"},
    {IddObjectType::PhotovoltaicPerformance_Simple, "PhotovoltaicPerformance_Simple", "PhotovoltaicPerformance:Simple"},
    {IddObjectType::PhotovoltaicPerformance_EquivalentOneDiode, "PhotovoltaicPerformance_EquivalentOneDiode",
     "PhotovoltaicPerformance:EquivalentOne-Diode"},
    {IddObjectType::PhotovoltaicPerformance_Sandia, "PhotovoltaicPerformance_Sandia", "PhotovoltaicPerformance:Sandia"},
    {IddObjectType::Generator_PVWatts, "Generator_PVWatts", "Generator:PVWatts"},
    {IddObjectType::ElectricLoadCenter_Inverter_PVWatts, "ElectricLoadCenter_Inverter_PVWatts", "ElectricLoadCenter:Inverter:PVWatts"},
    {IddObjectType::Generator_FuelCell, "Generator_FuelCell", "Generator:FuelCell"},
    {IddObjectType::Generator_FuelCell_PowerModule, "Generator_FuelCell_PowerModule", "Generator:FuelCell:PowerModule"},
    {IddObjectType::Generator_FuelCell_AirSupply, "Generator_FuelCell_AirSupply", "Generator:FuelCell:AirSupply"},
    {IddObjectType::Generator_FuelCell_WaterSupply, "Generator_FuelCell_WaterSupply", "Generator:FuelCell:WaterSupply"},
    {IddObjectType::Generator_FuelCell_AuxiliaryHeater, "Generator_FuelCell_AuxiliaryHeater", "Generator:FuelCell:AuxiliaryHeater"},
    {IddObjectType::Generator_FuelCell_ExhaustGasToWaterHeatExchanger, "Generator_FuelCell_ExhaustGasToWaterHeatExchanger",
     "Generator:FuelCell:ExhaustGasToWaterHeatExchanger"},
    {IddObjectType::Generator_FuelCell_ElectricalStorage, "Generator_FuelCell_ElectricalStorage", "Generator:FuelCell:ElectricalStorage"},
    {IddObjectType::Generator_FuelCell_Inverter, "Generator_FuelCell_Inverter", "Generator:FuelCell:Inverter"},
    {IddObjectType::Generator_FuelCell_StackCooler, "Generator_FuelCell_StackCooler", "Generator:FuelCell:StackCooler"},
    {IddObjectType::Generator_MicroCHP, "Generator_MicroCHP", "Generator:MicroCHP"},
    {IddObjectType::Generator_MicroCHP_NonNormalizedParameters, "Generator_MicroCHP_NonNormalizedParameters",
     "Generator:MicroCHP:NonNormalizedParameters"},
    {IddObjectType::Generator_FuelSupply, "Generator_FuelSupply", "Generator:FuelSupply"},
    {IddObjectType::Generator_WindTurbine, "Generator_WindTurbine", "Generator:WindTurbine"},
    {IddObjectType::ElectricLoadCenter_Generators, "ElectricLoadCenter_Generators", "ElectricLoadCenter:Generators"},
    {IddObjectType::ElectricLoadCenter_Inverter_Simple, "ElectricLoadCenter_Inverter_Simple", "ElectricLoadCenter:Inverter:Simple"},
    {IddObjectType::ElectricLoadCenter_Inverter_FunctionOfPower, "ElectricLoadCenter_Inverter_FunctionOfPower",
     "ElectricLoadCenter:Inverter:FunctionOfPower"},
    {IddObjectType::ElectricLoadCenter_Inverter_LookUpTable, "ElectricLoadCenter_Inverter_LookUpTable", "ElectricLoadCenter:Inverter:LookUpTable"},
    {IddObjectType::ElectricLoadCenter_Storage_Simple, "ElectricLoadCenter_Storage_Simple", "ElectricLoadCenter:Storage:Simple"},
    {IddObjectType::ElectricLoadCenter_Storage_Battery, "ElectricLoadCenter_Storage_Battery", "ElectricLoadCenter:Storage:Battery"},
    {IddObjectType::ElectricLoadCenter_Storage_LiIonNMCBattery, "ElectricLoadCenter_Storage_LiIonNMCBattery",
     "ElectricLoadCenter:Storage:LiIonNMCBattery"},
    {IddObjectType::ElectricLoadCenter_Transformer, "ElectricLoadCenter_Transformer", "ElectricLoadCenter:Transformer"},
    {IddObjectType::ElectricLoadCenter_Distribution, "ElectricLoadCenter_Distribution", "ElectricLoadCenter:Distribution"},
    {IddObjectType::ElectricLoadCenter_Storage_Converter, "ElectricLoadCenter_Storage_Converter", "ElectricLoadCenter:Storage:Converter"},
    {IddObjectType::WaterUse_Equipment, "WaterUse_Equipment", "WaterUse:Equipment"},
    {IddObjectType::WaterUse_Connections, "WaterUse_Connections", "WaterUse:Connections"},
    {IddObjectType::WaterUse_Storage, "WaterUse_Storage", "WaterUse:Storage"},
    {IddObjectType::WaterUse_Well, "WaterUse_Well", "WaterUse:Well"},
    {IddObjectType::WaterUse_RainCollector, "WaterUse_RainCollector", "WaterUse:RainCollector"},
    {IddObjectType::FaultModel_TemperatureSensorOffset_OutdoorAir, "FaultModel_TemperatureSensorOffset_OutdoorAir",
     "FaultModel:TemperatureSensorOffset:OutdoorAir"},
    {IddObjectType::FaultModel_HumiditySensorOffset_OutdoorAir, "FaultModel_HumiditySensorOffset_OutdoorAir",
     "FaultModel:HumiditySensorOffset:OutdoorAir"},
    {IddObjectType::FaultModel_EnthalpySensorOffset_OutdoorAir, "FaultModel_EnthalpySensorOffset_OutdoorAir",
     "FaultModel:EnthalpySensorOffset:OutdoorAir"},
    {IddObjectType::FaultModel_TemperatureSensorOffset_ReturnAir, "FaultModel_TemperatureSensorOffset_ReturnAir",
     "FaultModel:TemperatureSensorOffset:ReturnAir"},
    {IddObjectType::FaultModel_EnthalpySensorOffset_ReturnAir, "FaultModel_EnthalpySensorOffset_ReturnAir",
     "FaultModel:EnthalpySensorOffset:ReturnAir"},
    {IddObjectType::FaultModel_TemperatureSensorOffset_ChillerSupplyWater, "FaultModel_TemperatureSensorOffset_ChillerSupplyWater",
     "FaultModel:TemperatureSensorOffset:ChillerSupplyWater"},
    {IddObjectType::FaultModel_TemperatureSensorOffset_CoilSupplyAir, "FaultModel_TemperatureSensorOffset_CoilSupplyAir",
     "FaultModel:TemperatureSensorOffset:CoilSupplyAir"},
    {IddObjectType::FaultModel_TemperatureSensorOffset_CondenserSupplyWater, "FaultModel_TemperatureSensorOffset_CondenserSupplyWater",
     "FaultModel:TemperatureSensorOffset:CondenserSupplyWater"},
    {IddObjectType::FaultModel_ThermostatOffset, "FaultModel_ThermostatOffset", "FaultModel:ThermostatOffset"},
    {IddObjectType::FaultModel_HumidistatOffset, "FaultModel_HumidistatOffset", "FaultModel:HumidistatOffset"},
    {IddObjectType::FaultModel_Fouling_AirFilter, "FaultModel_Fouling_AirFilter", "FaultModel:Fouling:AirFilter"},
    {IddObjectType::FaultModel_Fouling_Boiler, "FaultModel_Fouling_Boiler", "FaultModel:Fouling:Boiler"},
    {IddObjectType::FaultModel_Fouling_EvaporativeCooler, "FaultModel_Fouling_EvaporativeCooler", "FaultModel:Fouling:EvaporativeCooler"},
    {IddObjectType::FaultModel_Fouling_Chiller, "FaultModel_Fouling_Chiller", "FaultModel:Fouling:Chiller"},
    {IddObjectType::FaultModel_Fouling_CoolingTower, "FaultModel_Fouling_CoolingTower", "FaultModel:Fouling:CoolingTower"},
    {IddObjectType::FaultModel_Fouling_Coil, "FaultModel_Fouling_Coil", "FaultModel:Fouling:Coil"},
    {IddObjectType::Matrix_TwoDimension, "Matrix_TwoDimension", "Matrix:TwoDimension"},
    {IddObjectType::HybridModel_Zone, "HybridModel_Zone", "HybridModel:Zone"},
    {IddObjectType::Curve_Linear, "Curve_Linear", "Curve:Linear"},
    {IddObjectType::Curve_QuadLinear, "Curve_QuadLinear", "Curve:QuadLinear"},
    {IddObjectType::Curve_QuintLinear, "Curve_QuintLinear", "Curve:QuintLinear"},
    {IddObjectType::Curve_Quadratic, "Curve_Quadratic", "Curve:Quadratic"},
    {IddObjectType::Curve_Cubic, "Curve_Cubic", "Curve:Cubic"},
    {IddObjectType::Curve_Quartic, "Curve_Quartic", "Curve:Quartic"},
    {IddObjectType::Curve_Exponent, "Curve_Exponent", "Curve:Exponent"},
    {IddObjectType::Curve_Bicubic, "Curve_Bicubic", "Curve:Bicubic"},
    {IddObjectType::Curve_Biquadratic, "Curve_Biquadratic", "Curve:Biquadratic"},
    {IddObjectType::Curve_QuadraticLinear, "Curve_QuadraticLinear", "Curve:QuadraticLinear"},
    {IddObjectType::Curve_CubicLinear, "Curve_CubicLinear", "Curve:CubicLinear"},
    {IddObjectType::Curve_Triquadratic, "Curve_Triquadratic", "Curve:Triquadratic"},
    {IddObjectType::Curve_Functional_PressureDrop, "Curve_Functional_PressureDrop", "Curve:Functional:PressureDrop"},
    {IddObjectType::Curve_FanPressureRise, "Curve_FanPressureRise", "Curve:FanPressureRise"},
    {IddObjectType::Curve_ExponentialSkewNormal, "Curve_ExponentialSkewNormal", "Curve:ExponentialSkewNormal"},
    {IddObjectType::Curve_Sigmoid, "Curve_Sigmoid", "Curve:Sigmoid"},
    {IddObjectType::Curve_RectangularHyperbola1, "Curve_RectangularHyperbola1", "Curve:RectangularHyperbola1"},
    {IddObjectType::Curve_RectangularHyperbola2, "Curve_RectangularHyperbola2", "Curve:RectangularHyperbola2"},
    {IddObjectType::Curve_ExponentialDecay, "Curve_ExponentialDecay", "Curve:ExponentialDecay"},
    {IddObjectType::Curve_DoubleExponentialDecay, "Curve_DoubleExponentialDecay", "Curve:DoubleExponentialDecay"},
    {IddObjectType::Curve_ChillerPartLoadWithLift, "Curve_ChillerPartLoadWithLift", "Curve:ChillerPartLoadWithLift"},
    {IddObjectType::Table_IndependentVariable, "Table_IndependentVariable", "Table:IndependentVariable"},
    {IddObjectType::Table_IndependentVariableList, "Table_IndependentVariableList", "Table:IndependentVariableList"},
    {IddObjectType::Table_Lookup, "Table_Lookup", "Table:Lookup"},
    {IddObjectType::FluidProperties_Name, "FluidProperties_Name", "FluidProperties:Name"},
    {IddObjectType::FluidProperties_GlycolConcentration, "FluidProperties_GlycolConcentration", "FluidProperties:GlycolConcentration"},
    {IddObjectType::FluidProperties_Temperatures, "FluidProperties_Temperatures", "FluidProperties:Temperatures"},
    {IddObjectType::FluidProperties_Saturated, "FluidProperties_Saturated", "FluidProperties:Saturated"},
    {IddObjectType::FluidProperties_Superheated, "FluidProperties_Superheated", "FluidProperties:Superheated"},
    {IddObjectType::FluidProperties_Concentration, "FluidProperties_Concentration", "FluidProperties:Concentration"},
    {IddObjectType::CurrencyType, "CurrencyType", "CurrencyType"},
    {IddObjectType::ComponentCost_Adjustments, "ComponentCost_Adjustments", "ComponentCost:Adjustments"},
    {IddObjectType::ComponentCost_Reference, "ComponentCost_Reference", "ComponentCost:Reference"},
    {IddObjectType::ComponentCost_LineItem, "ComponentCost_LineItem", "ComponentCost:LineItem"},
    {IddObjectType::UtilityCost_Tariff, "UtilityCost_Tariff", "UtilityCost:Tariff"},
    {IddObjectType::UtilityCost_Qualify, "UtilityCost_Qualify", "UtilityCost:Qualify"},
    {IddObjectType::UtilityCost_Charge_Simple, "UtilityCost_Charge_Simple", "UtilityCost:Charge:Simple"},
    {IddObjectType::UtilityCost_Charge_Block, "UtilityCost_Charge_Block", "UtilityCost:Charge:Block"},
    {IddObjectType::UtilityCost_Ratchet, "UtilityCost_Ratchet", "UtilityCost:Ratchet"},
    {IddObjectType::UtilityCost_Variable, "UtilityCost_Variable", "UtilityCost:Variable"},
    {IddObjectType::UtilityCost_Computation, "UtilityCost_Computation", "UtilityCost:Computation"},
    {IddObjectType::LifeCycleCost_Parameters, "LifeCycleCost_Parameters", "LifeCycleCost:Parameters"},
    {IddObjectType::LifeCycleCost_RecurringCosts, "LifeCycleCost_RecurringCosts", "LifeCycleCost:RecurringCosts"},
    {IddObjectType::LifeCycleCost_NonrecurringCost, "LifeCycleCost_NonrecurringCost", "LifeCycleCost:NonrecurringCost"},
    {IddObjectType::LifeCycleCost_UsePriceEscalation, "LifeCycleCost_UsePriceEscalation", "LifeCycleCost:UsePriceEscalation"},
    {IddObjectType::LifeCycleCost_UseAdjustment, "LifeCycleCost_UseAdjustment", "LifeCycleCost:UseAdjustment"},
    {IddObjectType::Parametric_SetValueForRun, "Parametric_SetValueForRun", "Parametric:SetValueForRun"},
    {IddObjectType::Parametric_Logic, "Parametric_Logic", "Parametric:Logic"},
    {IddObjectType::Parametric_RunControl, "Parametric_RunControl", "Parametric:RunControl"},
    {IddObjectType::Parametric_FileNameSuffix, "Parametric_FileNameSuffix", "Parametric:FileNameSuffix"},
    {IddObjectType::Output_VariableDictionary, "Output_VariableDictionary", "Output:VariableDictionary"},
    {IddObjectType::Output_Surfaces_List, "Output_Surfaces_List", "Output:Surfaces:List"},
    {IddObjectType::Output_Surfaces_Drawing, "Output_Surfaces_Drawing", "Output:Surfaces:Drawing"},
    {IddObjectType::Output_Schedules, "Output_Schedules", "Output:Schedules"},
    {IddObjectType::Output_Constructions, "Output_Constructions", "Output:Constructions"},
    {IddObjectType::Output_EnergyManagementSystem, "Output_EnergyManagementSystem", "Output:EnergyManagementSystem"},
    {IddObjectType::OutputControl_SurfaceColorScheme, "OutputControl_SurfaceColorScheme", "OutputControl:SurfaceColorScheme"},
    {IddObjectType::Output_Table_SummaryReports, "Output_Table_SummaryReports", "Output:Table:SummaryReports"},
    {IddObjectType::Output_Table_TimeBins, "Output_Table_TimeBins", "Output:Table:TimeBins"},
    {IddObjectType::Output_Table_Monthly, "Output_Table_Monthly", "Output:Table:Monthly"},
    {IddObjectType::Output_Table_Annual, "Output_Table_Annual", "Output:Table:Annual"},
    {IddObjectType::OutputControl_Table_Style, "OutputControl_Table_Style", "OutputControl:Table:Style"},
    {IddObjectType::OutputControl_ReportingTolerances, "OutputControl_ReportingTolerances", "OutputControl:ReportingTolerances"},
    {IddObjectType::Output_Variable, "Output_Variable", "Output:Variable"},
    {IddObjectType::Output_Meter, "Output_Meter", "Output:Meter"},
    {IddObjectType::Output_Meter_MeterFileOnly, "Output_Meter_MeterFileOnly", "Output:Meter:MeterFileOnly"},
    {IddObjectType::Output_Meter_Cumulative, "Output_Meter_Cumulative", "Output:Meter:Cumulative"},
    {IddObjectType::Output_Meter_Cumulative_MeterFileOnly, "Output_Meter_Cumulative_MeterFileOnly", "Output:Meter:Cumulative:MeterFileOnly"},
    {IddObjectType::Meter_Custom, "Meter_Custom", "Meter:Custom"},
    {IddObjectType::Meter_CustomDecrement, "Meter_CustomDecrement", "Meter:CustomDecrement"},
    {IddObjectType::OutputControl_Files, "OutputControl_Files", "OutputControl:Files"},
    {IddObjectType::OutputControl_Timestamp, "OutputControl_Timestamp", "OutputControl:Timestamp"},
    {IddObjectType::Output_JSON, "Output_JSON", "Output:JSON"},
    {IddObjectType::Output_SQLite, "Output_SQLite", "Output:SQLite"},
    {IddObjectType::Output_EnvironmentalImpactFactors, "Output_EnvironmentalImpactFactors", "Output:EnvironmentalImpactFactors"},
    {IddObjectType::EnvironmentalImpactFactors, "EnvironmentalImpactFactors", "EnvironmentalImpactFactors"},
    {IddObjectType::FuelFactors, "FuelFactors", "FuelFactors"},
    {IddObjectType::Output_Diagnostics, "Output_Diagnostics", "Output:Diagnostics"},
    {IddObjectType::Output_DebuggingData, "Output_DebuggingData", "Output:DebuggingData"},
    {IddObjectType::Output_PreprocessorMessage, "Output_PreprocessorMessage", "Output:PreprocessorMessage"},
    {IddObjectType::PythonPlugin_SearchPaths, "PythonPlugin_SearchPaths", "PythonPlugin:SearchPaths"},
    {IddObjectType::PythonPlugin_Instance, "PythonPlugin_Instance", "PythonPlugin:Instance"},
    {IddObjectType::PythonPlugin_Variables, "PythonPlugin_Variables", "PythonPlugin:Variables"},
    {IddObjectType::PythonPlugin_TrendVariable, "PythonPlugin_TrendVariable", "PythonPlugin:TrendVariable"},
    {IddObjectType::PythonPlugin_OutputVariable, "PythonPlugin_OutputVariable", "PythonPlugin:OutputVariable"},
    {IddObjectType::OS_Version, "OS_Version", "OS:Version"},
    {IddObjectType::OS_ComponentData, "OS_ComponentData", "OS:ComponentData"},
    {IddObjectType::OS_ConvergenceLimits, "OS_ConvergenceLimits", "OS:ConvergenceLimits"},
    {IddObjectType::OS_ProgramControl, "OS_ProgramControl", "OS:ProgramControl"},
    {IddObjectType::OS_HeatBalanceAlgorithm, "OS_HeatBalanceAlgorithm", "OS:HeatBalanceAlgorithm"},
    {IddObjectType::OS_RunPeriod, "OS_RunPeriod", "OS:RunPeriod"},
    {IddObjectType::OS_RunPeriodControl_DaylightSavingTime, "OS_RunPeriodControl_DaylightSavingTime", "OS:RunPeriodControl:DaylightSavingTime"},
    {IddObjectType::OS_RunPeriodControl_SpecialDays, "OS_RunPeriodControl_SpecialDays", "OS:RunPeriodControl:SpecialDays"},
    {IddObjectType::OS_ShadowCalculation, "OS_ShadowCalculation", "OS:ShadowCalculation"},
    {IddObjectType::OS_SimulationControl, "OS_SimulationControl", "OS:SimulationControl"},
    {IddObjectType::OS_PerformancePrecisionTradeoffs, "OS_PerformancePrecisionTradeoffs", "OS:PerformancePrecisionTradeoffs"},
    {IddObjectType::OS_Sizing_Parameters, "OS_Sizing_Parameters", "OS:Sizing:Parameters"},
    {IddObjectType::OS_SurfaceConvectionAlgorithm_Inside, "OS_SurfaceConvectionAlgorithm_Inside", "OS:SurfaceConvectionAlgorithm:Inside"},
    {IddObjectType::OS_SurfaceConvectionAlgorithm_Outside, "OS_SurfaceConvectionAlgorithm_Outside", "OS:SurfaceConvectionAlgorithm:Outside"},
    {IddObjectType::OS_Timestep, "OS_Timestep", "OS:Timestep"},
    {IddObjectType::OS_YearDescription, "OS_YearDescription", "OS:YearDescription"},
    {IddObjectType::OS_ZoneAirContaminantBalance, "OS_ZoneAirContaminantBalance", "OS:ZoneAirContaminantBalance"},
    {IddObjectType::OS_ZoneAirHeatBalanceAlgorithm, "OS_ZoneAirHeatBalanceAlgorithm", "OS:ZoneAirHeatBalanceAlgorithm"},
    {IddObjectType::OS_ZoneAirMassFlowConservation, "OS_ZoneAirMassFlowConservation", "OS:ZoneAirMassFlowConservation"},
    {IddObjectType::OS_ZoneCapacitanceMultiplier_ResearchSpecial, "OS_ZoneCapacitanceMultiplier_ResearchSpecial",
     "OS:ZoneCapacitanceMultiplier:ResearchSpecial"},
    {IddObjectType::OS_RadianceParameters, "OS_RadianceParameters", "OS:RadianceParameters"},
    {IddObjectType::OS_ClimateZones, "OS_ClimateZones", "OS:ClimateZones"},
    {IddObjectType::OS_Site, "OS_Site", "OS:Site"},
    {IddObjectType::OS_Site_GroundTemperature_BuildingSurface, "OS_Site_GroundTemperature_BuildingSurface",
     "OS:Site:GroundTemperature:BuildingSurface"},
    {IddObjectType::OS_Site_GroundTemperature_FCfactorMethod, "OS_Site_GroundTemperature_FCfactorMethod", "OS:Site:GroundTemperature:FCfactorMethod"},
    {IddObjectType::OS_Site_GroundTemperature_Shallow, "OS_Site_GroundTemperature_Shallow", "OS:Site:GroundTemperature:Shallow"},
    {IddObjectType::OS_Site_GroundTemperature_Deep, "OS_Site_GroundTemperature_Deep", "OS:Site:GroundTemperature:Deep"},
    {IddObjectType::OS_Site_GroundTemperature_Undisturbed_KusudaAchenbach, "OS_Site_GroundTemperature_Undisturbed_KusudaAchenbach",
     "OS:Site:GroundTemperature:Undisturbed:KusudaAchenbach"},
    {IddObjectType::OS_Site_GroundTemperature_Undisturbed_Xing, "OS_Site_GroundTemperature_Undisturbed_Xing",
     "OS:Site:GroundTemperature:Undisturbed:Xing"},
    {IddObjectType::OS_Site_GroundReflectance, "OS_Site_GroundReflectance", "OS:Site:GroundReflectance"},
    {IddObjectType::OS_Site_WaterMainsTemperature, "OS_Site_WaterMainsTemperature", "OS:Site:WaterMainsTemperature"},
    {IddObjectType::OS_SizingPeriod_DesignDay, "OS_SizingPeriod_DesignDay", "OS:SizingPeriod:DesignDay"},
    {IddObjectType::OS_SizingPeriod_WeatherFileConditionType, "OS_SizingPeriod_WeatherFileConditionType", "OS:SizingPeriod:WeatherFileConditionType"},
    {IddObjectType::OS_SizingPeriod_WeatherFileDays, "OS_SizingPeriod_WeatherFileDays", "OS:SizingPeriod:WeatherFileDays"},
    {IddObjectType::OS_WeatherFile, "OS_WeatherFile", "OS:WeatherFile"},
    {IddObjectType::OS_WeatherProperty_SkyTemperature, "OS_WeatherProperty_SkyTemperature", "OS:WeatherProperty:SkyTemperature"},
    {IddObjectType::OS_AdditionalProperties, "OS_AdditionalProperties", "OS:AdditionalProperties"},
    {IddObjectType::OS_BuildingStory, "OS_BuildingStory", "OS:BuildingStory"},
    {IddObjectType::OS_BuildingUnit, "OS_BuildingUnit", "OS:BuildingUnit"},
    {IddObjectType::OS_DefaultConstructionSet, "OS_DefaultConstructionSet", "OS:DefaultConstructionSet"},
    {IddObjectType::OS_DefaultScheduleSet, "OS_DefaultScheduleSet", "OS:DefaultScheduleSet"},
    {IddObjectType::OS_DefaultSubSurfaceConstructions, "OS_DefaultSubSurfaceConstructions", "OS:DefaultSubSurfaceConstructions"},
    {IddObjectType::OS_DefaultSurfaceConstructions, "OS_DefaultSurfaceConstructions", "OS:DefaultSurfaceConstructions"},
    {IddObjectType::OS_Rendering_Color, "OS_Rendering_Color", "OS:Rendering:Color"},
    {IddObjectType::OS_SpaceType, "OS_SpaceType", "OS:SpaceType"},
    {IddObjectType::OS_ModelObjectList, "OS_ModelObjectList", "OS:ModelObjectList"},
    {IddObjectType::OS_Material, "OS_Material", "OS:Material"},
    {IddObjectType::OS_Material_AirGap, "OS_Material_AirGap", "OS:Material:AirGap"},
    {IddObjectType::OS_Material_InfraredTransparent, "OS_Material_InfraredTransparent", "OS:Material:InfraredTransparent"},
    {IddObjectType::OS_Material_NoMass, "OS_Material_NoMass", "OS:Material:NoMass"},
    {IddObjectType::OS_Material_RoofVegetation, "OS_Material_RoofVegetation", "OS:Material:RoofVegetation"},
    {IddObjectType::OS_WindowMaterial_Blind, "OS_WindowMaterial_Blind", "OS:WindowMaterial:Blind"},
    {IddObjectType::OS_WindowMaterial_DaylightRedirectionDevice, "OS_WindowMaterial_DaylightRedirectionDevice",
     "OS:WindowMaterial:DaylightRedirectionDevice"},
    {IddObjectType::OS_WindowMaterial_Gas, "OS_WindowMaterial_Gas", "OS:WindowMaterial:Gas"},
    {IddObjectType::OS_WindowMaterial_GasMixture, "OS_WindowMaterial_GasMixture", "OS:WindowMaterial:GasMixture"},
    {IddObjectType::OS_WindowMaterial_Glazing, "OS_WindowMaterial_Glazing", "OS:WindowMaterial:Glazing"},
    {IddObjectType::OS_WindowMaterial_GlazingGroup_Thermochromic, "OS_WindowMaterial_GlazingGroup_Thermochromic",
     "OS:WindowMaterial:GlazingGroup:Thermochromic"},
    {IddObjectType::OS_WindowMaterial_Glazing_RefractionExtinctionMethod, "OS_WindowMaterial_Glazing_RefractionExtinctionMethod",
     "OS:WindowMaterial:Glazing:RefractionExtinctionMethod"},
    {IddObjectType::OS_WindowMaterial_Screen, "OS_WindowMaterial_Screen", "OS:WindowMaterial:Screen"},
    {IddObjectType::OS_WindowMaterial_Shade, "OS_WindowMaterial_Shade", "OS:WindowMaterial:Shade"},
    {IddObjectType::OS_WindowMaterial_SimpleGlazingSystem, "OS_WindowMaterial_SimpleGlazingSystem", "OS:WindowMaterial:SimpleGlazingSystem"},
    {IddObjectType::OS_StandardsInformation_Material, "OS_StandardsInformation_Material", "OS:StandardsInformation:Material"},
    {IddObjectType::OS_MaterialProperty_GlazingSpectralData, "OS_MaterialProperty_GlazingSpectralData", "OS:MaterialProperty:GlazingSpectralData"},
    {IddObjectType::OS_MaterialProperty_MoisturePenetrationDepth_Settings, "OS_MaterialProperty_MoisturePenetrationDepth_Settings",
     "OS:MaterialProperty:MoisturePenetrationDepth:Settings"},
    {IddObjectType::OS_MaterialProperty_PhaseChange, "OS_MaterialProperty_PhaseChange", "OS:MaterialProperty:PhaseChange"},
    {IddObjectType::OS_MaterialProperty_PhaseChangeHysteresis, "OS_MaterialProperty_PhaseChangeHysteresis",
     "OS:MaterialProperty:PhaseChangeHysteresis"},
    {IddObjectType::OS_Construction, "OS_Construction", "OS:Construction"},
    {IddObjectType::OS_Construction_CfactorUndergroundWall, "OS_Construction_CfactorUndergroundWall", "OS:Construction:CfactorUndergroundWall"},
    {IddObjectType::OS_Construction_FfactorGroundFloor, "OS_Construction_FfactorGroundFloor", "OS:Construction:FfactorGroundFloor"},
    {IddObjectType::OS_Construction_InternalSource, "OS_Construction_InternalSource", "OS:Construction:InternalSource"},
    {IddObjectType::OS_Construction_AirBoundary, "OS_Construction_AirBoundary", "OS:Construction:AirBoundary"},
    {IddObjectType::OS_Construction_WindowDataFile, "OS_Construction_WindowDataFile", "OS:Construction:WindowDataFile"},
    {IddObjectType::OS_StandardsInformation_Construction, "OS_StandardsInformation_Construction", "OS:StandardsInformation:Construction"},
    {IddObjectType::OS_InternalMass_Definition, "OS_InternalMass_Definition", "OS:InternalMass:Definition"},
    {IddObjectType::OS_People_Definition, "OS_People_Definition", "OS:People:Definition"},
    {IddObjectType::OS_Lights_Definition, "OS_Lights_Definition", "OS:Lights:Definition"},
    {IddObjectType::OS_Luminaire_Definition, "OS_Luminaire_Definition", "OS:Luminaire:Definition"},
    {IddObjectType::OS_ElectricEquipment_Definition, "OS_ElectricEquipment_Definition", "OS:ElectricEquipment:Definition"},
    {IddObjectType::OS_ElectricEquipment_ITE_AirCooled_Definition, "OS_ElectricEquipment_ITE_AirCooled_Definition",
     "OS:ElectricEquipment:ITE:AirCooled:Definition"},
    {IddObjectType::OS_GasEquipment_Definition, "OS_GasEquipment_Definition", "OS:GasEquipment:Definition"},
    {IddObjectType::OS_HotWaterEquipment_Definition, "OS_HotWaterEquipment_Definition", "OS:HotWaterEquipment:Definition"},
    {IddObjectType::OS_SteamEquipment_Definition, "OS_SteamEquipment_Definition", "OS:SteamEquipment:Definition"},
    {IddObjectType::OS_OtherEquipment_Definition, "OS_OtherEquipment_Definition", "OS:OtherEquipment:Definition"},
    {IddObjectType::OS_Exterior_Lights_Definition, "OS_Exterior_Lights_Definition", "OS:Exterior:Lights:Definition"},
    {IddObjectType::OS_Exterior_FuelEquipment_Definition, "OS_Exterior_FuelEquipment_Definition", "OS:Exterior:FuelEquipment:Definition"},
    {IddObjectType::OS_Exterior_WaterEquipment_Definition, "OS_Exterior_WaterEquipment_Definition", "OS:Exterior:WaterEquipment:Definition"},
    {IddObjectType::OS_Schedule_Compact, "OS_Schedule_Compact", "OS:Schedule:Compact"},
    {IddObjectType::OS_Schedule_Day, "OS_Schedule_Day", "OS:Schedule:Day"},
    {IddObjectType::OS_Schedule_Week, "OS_Schedule_Week", "OS:Schedule:Week"},
    {IddObjectType::OS_Schedule_Year, "OS_Schedule_Year", "OS:Schedule:Year"},
    {IddObjectType::OS_Schedule_Rule, "OS_Schedule_Rule", "OS:Schedule:Rule"},
    {IddObjectType::OS_Schedule_Ruleset, "OS_Schedule_Ruleset", "OS:Schedule:Ruleset"},
    {IddObjectType::OS_Schedule_Constant, "OS_Schedule_Constant", "OS:Schedule:Constant"},
    {IddObjectType::OS_Schedule_FixedInterval, "OS_Schedule_FixedInterval", "OS:Schedule:FixedInterval"},
    {IddObjectType::OS_Schedule_VariableInterval, "OS_Schedule_VariableInterval", "OS:Schedule:VariableInterval"},
    {IddObjectType::OS_ScheduleTypeLimits, "OS_ScheduleTypeLimits", "OS:ScheduleTypeLimits"},
    {IddObjectType::OS_External_File, "OS_External_File", "OS:External:File"},
    {IddObjectType::OS_Schedule_File, "OS_Schedule_File", "OS:Schedule:File"},
    {IddObjectType::OS_Building, "OS_Building", "OS:Building"},
    {IddObjectType::OS_DaylightingDevice_Tubular, "OS_DaylightingDevice_Tubular", "OS:DaylightingDevice:Tubular"},
    {IddObjectType::OS_DaylightingDevice_Shelf, "OS_DaylightingDevice_Shelf", "OS:DaylightingDevice:Shelf"},
    {IddObjectType::OS_DaylightingDevice_LightWell, "OS_DaylightingDevice_LightWell", "OS:DaylightingDevice:LightWell"},
    {IddObjectType::OS_Facility, "OS_Facility", "OS:Facility"},
    {IddObjectType::OS_InteriorPartitionSurface, "OS_InteriorPartitionSurface", "OS:InteriorPartitionSurface"},
    {IddObjectType::OS_InteriorPartitionSurfaceGroup, "OS_InteriorPartitionSurfaceGroup", "OS:InteriorPartitionSurfaceGroup"},
    {IddObjectType::OS_ShadingSurface, "OS_ShadingSurface", "OS:ShadingSurface"},
    {IddObjectType::OS_ShadingSurfaceGroup, "OS_ShadingSurfaceGroup", "OS:ShadingSurfaceGroup"},
    {IddObjectType::OS_Space, "OS_Space", "OS:Space"},
    {IddObjectType::OS_Surface, "OS_Surface", "OS:Surface"},
    {IddObjectType::OS_SubSurface, "OS_SubSurface", "OS:SubSurface"},
    {IddObjectType::OS_ShadingControl, "OS_ShadingControl", "OS:ShadingControl"},
    {IddObjectType::OS_WindowProperty_FrameAndDivider, "OS_WindowProperty_FrameAndDivider", "OS:WindowProperty:FrameAndDivider"},
    {IddObjectType::OS_SurfaceControl_MovableInsulation, "OS_SurfaceControl_MovableInsulation", "OS:SurfaceControl:MovableInsulation"},
    {IddObjectType::OS_SurfaceProperty_ConvectionCoefficients, "OS_SurfaceProperty_ConvectionCoefficients",
     "OS:SurfaceProperty:ConvectionCoefficients"},
    {IddObjectType::OS_SurfaceProperty_ConvectionCoefficients_MultipleSurface, "OS_SurfaceProperty_ConvectionCoefficients_MultipleSurface",
     "OS:SurfaceProperty:ConvectionCoefficients:MultipleSurface"},
    {IddObjectType::OS_SurfaceProperty_ExposedFoundationPerimeter, "OS_SurfaceProperty_ExposedFoundationPerimeter",
     "OS:SurfaceProperty:ExposedFoundationPerimeter"},
    {IddObjectType::OS_SurfaceProperty_OtherSideCoefficients, "OS_SurfaceProperty_OtherSideCoefficients", "OS:SurfaceProperty:OtherSideCoefficients"},
    {IddObjectType::OS_SurfaceProperty_OtherSideConditionsModel, "OS_SurfaceProperty_OtherSideConditionsModel",
     "OS:SurfaceProperty:OtherSideConditionsModel"},
    {IddObjectType::OS_Foundation_Kiva, "OS_Foundation_Kiva", "OS:Foundation:Kiva"},
    {IddObjectType::OS_Foundation_Kiva_Settings, "OS_Foundation_Kiva_Settings", "OS:Foundation:Kiva:Settings"},
    {IddObjectType::OS_ZoneProperty_UserViewFactors_BySurfaceName, "OS_ZoneProperty_UserViewFactors_BySurfaceName",
     "OS:ZoneProperty:UserViewFactors:BySurfaceName"},
    {IddObjectType::OS_SurfaceProperty_SurroundingSurfaces, "OS_SurfaceProperty_SurroundingSurfaces", "OS:SurfaceProperty:SurroundingSurfaces"},
    {IddObjectType::OS_SurfaceProperty_LocalEnvironment, "OS_SurfaceProperty_LocalEnvironment", "OS:SurfaceProperty:LocalEnvironment"},
    {IddObjectType::OS_SurfaceProperty_GroundSurfaces, "OS_SurfaceProperty_GroundSurfaces", "OS:SurfaceProperty:GroundSurfaces"},
    {IddObjectType::OS_SurfaceProperty_IncidentSolarMultiplier, "OS_SurfaceProperty_IncidentSolarMultiplier",
     "OS:SurfaceProperty:IncidentSolarMultiplier"},
    {IddObjectType::OS_InternalMass, "OS_InternalMass", "OS:InternalMass"},
    {IddObjectType::OS_People, "OS_People", "OS:People"},
    {IddObjectType::OS_Lights, "OS_Lights", "OS:Lights"},
    {IddObjectType::OS_Luminaire, "OS_Luminaire", "OS:Luminaire"},
    {IddObjectType::OS_ElectricEquipment, "OS_ElectricEquipment", "OS:ElectricEquipment"},
    {IddObjectType::OS_ElectricEquipment_ITE_AirCooled, "OS_ElectricEquipment_ITE_AirCooled", "OS:ElectricEquipment:ITE:AirCooled"},
    {IddObjectType::OS_GasEquipment, "OS_GasEquipment", "OS:GasEquipment"},
    {IddObjectType::OS_HotWaterEquipment, "OS_HotWaterEquipment", "OS:HotWaterEquipment"},
    {IddObjectType::OS_SteamEquipment, "OS_SteamEquipment", "OS:SteamEquipment"},
    {IddObjectType::OS_OtherEquipment, "OS_OtherEquipment", "OS:OtherEquipment"},
    {IddObjectType::OS_SpaceInfiltration_DesignFlowRate, "OS_SpaceInfiltration_DesignFlowRate", "OS:SpaceInfiltration:DesignFlowRate"},
    {IddObjectType::OS_SpaceInfiltration_EffectiveLeakageArea, "OS_SpaceInfiltration_EffectiveLeakageArea",
     "OS:SpaceInfiltration:EffectiveLeakageArea"},
    {IddObjectType::OS_SpaceInfiltration_FlowCoefficient, "OS_SpaceInfiltration_FlowCoefficient", "OS:SpaceInfiltration:FlowCoefficient"},
    {IddObjectType::OS_Exterior_Lights, "OS_Exterior_Lights", "OS:Exterior:Lights"},
    {IddObjectType::OS_Exterior_FuelEquipment, "OS_Exterior_FuelEquipment", "OS:Exterior:FuelEquipment"},
    {IddObjectType::OS_Exterior_WaterEquipment, "OS_Exterior_WaterEquipment", "OS:Exterior:WaterEquipment"},
    {IddObjectType::OS_Daylighting_Control, "OS_Daylighting_Control", "OS:Daylighting:Control"},
    {IddObjectType::OS_Glare_Sensor, "OS_Glare_Sensor", "OS:Glare:Sensor"},
    {IddObjectType::OS_IlluminanceMap, "OS_IlluminanceMap", "OS:IlluminanceMap"},
    {IddObjectType::OS_LightingDesignDay, "OS_LightingDesignDay", "OS:LightingDesignDay"},
    {IddObjectType::OS_LightingSimulationControl, "OS_LightingSimulationControl", "OS:LightingSimulationControl"},
    {IddObjectType::OS_LightingSimulationZone, "OS_LightingSimulationZone", "OS:LightingSimulationZone"},
    {IddObjectType::OS_Refrigeration_Case, "OS_Refrigeration_Case", "OS:Refrigeration:Case"},
    {IddObjectType::OS_Refrigeration_Condenser_AirCooled, "OS_Refrigeration_Condenser_AirCooled", "OS:Refrigeration:Condenser:AirCooled"},
    {IddObjectType::OS_Refrigeration_Condenser_EvaporativeCooled, "OS_Refrigeration_Condenser_EvaporativeCooled",
     "OS:Refrigeration:Condenser:EvaporativeCooled"},
    {IddObjectType::OS_Refrigeration_Condenser_WaterCooled, "OS_Refrigeration_Condenser_WaterCooled", "OS:Refrigeration:Condenser:WaterCooled"},
    {IddObjectType::OS_Refrigeration_Condenser_Cascade, "OS_Refrigeration_Condenser_Cascade", "OS:Refrigeration:Condenser:Cascade"},
    {IddObjectType::OS_Refrigeration_Compressor, "OS_Refrigeration_Compressor", "OS:Refrigeration:Compressor"},
    {IddObjectType::OS_Refrigeration_CompressorRack, "OS_Refrigeration_CompressorRack", "OS:Refrigeration:CompressorRack"},
    {IddObjectType::OS_Refrigeration_DefrostCycleParameters, "OS_Refrigeration_DefrostCycleParameters", "OS:Refrigeration:DefrostCycleParameters"},
    {IddObjectType::OS_Refrigeration_GasCooler_AirCooled, "OS_Refrigeration_GasCooler_AirCooled", "OS:Refrigeration:GasCooler:AirCooled"},
    {IddObjectType::OS_Refrigeration_Subcooler_Mechanical, "OS_Refrigeration_Subcooler_Mechanical", "OS:Refrigeration:Subcooler:Mechanical"},
    {IddObjectType::OS_Refrigeration_Subcooler_LiquidSuction, "OS_Refrigeration_Subcooler_LiquidSuction", "OS:Refrigeration:Subcooler:LiquidSuction"},
    {IddObjectType::OS_Refrigeration_System, "OS_Refrigeration_System", "OS:Refrigeration:System"},
    {IddObjectType::OS_Refrigeration_TranscriticalSystem, "OS_Refrigeration_TranscriticalSystem", "OS:Refrigeration:TranscriticalSystem"},
    {IddObjectType::OS_Refrigeration_SecondarySystem, "OS_Refrigeration_SecondarySystem", "OS:Refrigeration:SecondarySystem"},
    {IddObjectType::OS_Refrigeration_WalkIn, "OS_Refrigeration_WalkIn", "OS:Refrigeration:WalkIn"},
    {IddObjectType::OS_Refrigeration_WalkIn_ZoneBoundary, "OS_Refrigeration_WalkIn_ZoneBoundary", "OS:Refrigeration:WalkIn:ZoneBoundary"},
    {IddObjectType::OS_Refrigeration_AirChiller, "OS_Refrigeration_AirChiller", "OS:Refrigeration:AirChiller"},
    {IddObjectType::OS_AirConditioner_VariableRefrigerantFlow, "OS_AirConditioner_VariableRefrigerantFlow",
     "OS:AirConditioner:VariableRefrigerantFlow"},
    {IddObjectType::OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl,
     "OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl", "OS:AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl"},
    {IddObjectType::OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_LoadingIndex,
     "OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_LoadingIndex",
     "OS:AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl:LoadingIndex"},
    {IddObjectType::OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR,
     "OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR", "OS:AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl:HR"},
    {IddObjectType::OS_AirLoopHVAC, "OS_AirLoopHVAC", "OS:AirLoopHVAC"},
    {IddObjectType::OS_AirLoopHVAC_OutdoorAirSystem, "OS_AirLoopHVAC_OutdoorAirSystem", "OS:AirLoopHVAC:OutdoorAirSystem"},
    {IddObjectType::OS_AirLoopHVAC_DedicatedOutdoorAirSystem, "OS_AirLoopHVAC_DedicatedOutdoorAirSystem", "OS:AirLoopHVAC:DedicatedOutdoorAirSystem"},
    {IddObjectType::OS_AirLoopHVAC_ReturnPlenum, "OS_AirLoopHVAC_ReturnPlenum", "OS:AirLoopHVAC:ReturnPlenum"},
    {IddObjectType::OS_AirLoopHVAC_SupplyPlenum, "OS_AirLoopHVAC_SupplyPlenum", "OS:AirLoopHVAC:SupplyPlenum"},
    {IddObjectType::OS_AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed, "OS_AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed",
     "OS:AirLoopHVAC:UnitaryHeatPump:AirToAir:MultiSpeed"},
    {IddObjectType::OS_AirLoopHVAC_UnitaryCoolOnly, "OS_AirLoopHVAC_UnitaryCoolOnly", "OS:AirLoopHVAC:UnitaryCoolOnly"},
    {IddObjectType::OS_AirLoopHVAC_UnitaryHeatPump_AirToAir, "OS_AirLoopHVAC_UnitaryHeatPump_AirToAir", "OS:AirLoopHVAC:UnitaryHeatPump:AirToAir"},
    {IddObjectType::OS_AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass, "OS_AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass",
     "OS:AirLoopHVAC:UnitaryHeatCool:VAVChangeoverBypass"},
    {IddObjectType::OS_AirLoopHVAC_UnitarySystem, "OS_AirLoopHVAC_UnitarySystem", "OS:AirLoopHVAC:UnitarySystem"},
    {IddObjectType::OS_UnitarySystemPerformance_Multispeed, "OS_UnitarySystemPerformance_Multispeed", "OS:UnitarySystemPerformance:Multispeed"},
    {IddObjectType::OS_AirLoopHVAC_ZoneMixer, "OS_AirLoopHVAC_ZoneMixer", "OS:AirLoopHVAC:ZoneMixer"},
    {IddObjectType::OS_AirLoopHVAC_ZoneSplitter, "OS_AirLoopHVAC_ZoneSplitter", "OS:AirLoopHVAC:ZoneSplitter"},
    {IddObjectType::OS_AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat, "OS_AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat",
     "OS:AirTerminal:SingleDuct:VAV:HeatAndCool:NoReheat"},
    {IddObjectType::OS_AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat, "OS_AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat",
     "OS:AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat"},
    {IddObjectType::OS_AirTerminal_SingleDuct_InletSideMixer, "OS_AirTerminal_SingleDuct_InletSideMixer", "OS:AirTerminal:SingleDuct:InletSideMixer"},
    {IddObjectType::OS_AirTerminal_SingleDuct_ConstantVolume_Reheat, "OS_AirTerminal_SingleDuct_ConstantVolume_Reheat",
     "OS:AirTerminal:SingleDuct:ConstantVolume:Reheat"},
    {IddObjectType::OS_AirTerminal_SingleDuct_SeriesPIU_Reheat, "OS_AirTerminal_SingleDuct_SeriesPIU_Reheat",
     "OS:AirTerminal:SingleDuct:SeriesPIU:Reheat"},
    {IddObjectType::OS_AirTerminal_SingleDuct_ParallelPIU_Reheat, "OS_AirTerminal_SingleDuct_ParallelPIU_Reheat",
     "OS:AirTerminal:SingleDuct:ParallelPIU:Reheat"},
    {IddObjectType::OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction, "OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction",
     "OS:AirTerminal:SingleDuct:ConstantVolume:FourPipeInduction"},
    {IddObjectType::OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam, "OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam",
     "OS:AirTerminal:SingleDuct:ConstantVolume:FourPipeBeam"},
    {IddObjectType::OS_AirTerminal_SingleDuct_ConstantVolume_CooledBeam, "OS_AirTerminal_SingleDuct_ConstantVolume_CooledBeam",
     "OS:AirTerminal:SingleDuct:ConstantVolume:CooledBeam"},
    {IddObjectType::OS_AirTerminal_SingleDuct_ConstantVolume_NoReheat, "OS_AirTerminal_SingleDuct_ConstantVolume_NoReheat",
     "OS:AirTerminal:SingleDuct:ConstantVolume:NoReheat"},
    {IddObjectType::OS_AirTerminal_SingleDuct_VAV_NoReheat, "OS_AirTerminal_SingleDuct_VAV_NoReheat", "OS:AirTerminal:SingleDuct:VAV:NoReheat"},
    {IddObjectType::OS_AirTerminal_SingleDuct_VAV_Reheat, "OS_AirTerminal_SingleDuct_VAV_Reheat", "OS:AirTerminal:SingleDuct:VAV:Reheat"},
    {IddObjectType::OS_AirTerminal_DualDuct_VAV_OutdoorAir, "OS_AirTerminal_DualDuct_VAV_OutdoorAir", "OS:AirTerminal:DualDuct:VAV:OutdoorAir"},
    {IddObjectType::OS_AirTerminal_DualDuct_ConstantVolume, "OS_AirTerminal_DualDuct_ConstantVolume", "OS:AirTerminal:DualDuct:ConstantVolume"},
    {IddObjectType::OS_AirTerminal_DualDuct_VAV, "OS_AirTerminal_DualDuct_VAV", "OS:AirTerminal:DualDuct:VAV"},
    {IddObjectType::OS_AvailabilityManagerAssignmentList, "OS_AvailabilityManagerAssignmentList", "OS:AvailabilityManagerAssignmentList"},
    {IddObjectType::OS_AvailabilityManager_Scheduled, "OS_AvailabilityManager_Scheduled", "OS:AvailabilityManager:Scheduled"},
    {IddObjectType::OS_AvailabilityManager_ScheduledOn, "OS_AvailabilityManager_ScheduledOn", "OS:AvailabilityManager:ScheduledOn"},
    {IddObjectType::OS_AvailabilityManager_ScheduledOff, "OS_AvailabilityManager_ScheduledOff", "OS:AvailabilityManager:ScheduledOff"},
    {IddObjectType::OS_AvailabilityManager_NightCycle, "OS_AvailabilityManager_NightCycle", "OS:AvailabilityManager:NightCycle"},
    {IddObjectType::OS_AvailabilityManager_OptimumStart, "OS_AvailabilityManager_OptimumStart", "OS:AvailabilityManager:OptimumStart"},
    {IddObjectType::OS_AvailabilityManager_DifferentialThermostat, "OS_AvailabilityManager_DifferentialThermostat",
     "OS:AvailabilityManager:DifferentialThermostat"},
    {IddObjectType::OS_AvailabilityManager_HighTemperatureTurnOff, "OS_AvailabilityManager_HighTemperatureTurnOff",
     "OS:AvailabilityManager:HighTemperatureTurnOff"},
    {IddObjectType::OS_AvailabilityManager_HighTemperatureTurnOn, "OS_AvailabilityManager_HighTemperatureTurnOn",
     "OS:AvailabilityManager:HighTemperatureTurnOn"},
    {IddObjectType::OS_AvailabilityManager_LowTemperatureTurnOff, "OS_AvailabilityManager_LowTemperatureTurnOff",
     "OS:AvailabilityManager:LowTemperatureTurnOff"},
    {IddObjectType::OS_AvailabilityManager_LowTemperatureTurnOn, "OS_AvailabilityManager_LowTemperatureTurnOn",
     "OS:AvailabilityManager:LowTemperatureTurnOn"},
    {IddObjectType::OS_AvailabilityManager_NightVentilation, "OS_AvailabilityManager_NightVentilation", "OS:AvailabilityManager:NightVentilation"},
    {IddObjectType::OS_AvailabilityManager_HybridVentilation, "OS_AvailabilityManager_HybridVentilation", "OS:AvailabilityManager:HybridVentilation"},
    {IddObjectType::OS_Boiler_HotWater, "OS_Boiler_HotWater", "OS:Boiler:HotWater"},
    {IddObjectType::OS_Boiler_Steam, "OS_Boiler_Steam", "OS:Boiler:Steam"},
    {IddObjectType::OS_HeatPump_WaterToWater_EquationFit_Heating, "OS_HeatPump_WaterToWater_EquationFit_Heating",
     "OS:HeatPump:WaterToWater:EquationFit:Heating"},
    {IddObjectType::OS_HeatPump_WaterToWater_EquationFit_Cooling, "OS_HeatPump_WaterToWater_EquationFit_Cooling",
     "OS:HeatPump:WaterToWater:EquationFit:Cooling"},
    {IddObjectType::OS_HeatPump_PlantLoop_EIR_Heating, "OS_HeatPump_PlantLoop_EIR_Heating", "OS:HeatPump:PlantLoop:EIR:Heating"},
    {IddObjectType::OS_HeatPump_PlantLoop_EIR_Cooling, "OS_HeatPump_PlantLoop_EIR_Cooling", "OS:HeatPump:PlantLoop:EIR:Cooling"},
    {IddObjectType::OS_HeatPump_AirToWater_FuelFired_Heating, "OS_HeatPump_AirToWater_FuelFired_Heating", "OS:HeatPump:AirToWater:FuelFired:Heating"},
    {IddObjectType::OS_HeatPump_AirToWater_FuelFired_Cooling, "OS_HeatPump_AirToWater_FuelFired_Cooling", "OS:HeatPump:AirToWater:FuelFired:Cooling"},
    {IddObjectType::OS_Chiller_Electric_EIR, "OS_Chiller_Electric_EIR", "OS:Chiller:Electric:EIR"},
    {IddObjectType::OS_Chiller_Electric_ReformulatedEIR, "OS_Chiller_Electric_ReformulatedEIR", "OS:Chiller:Electric:ReformulatedEIR"},
    {IddObjectType::OS_Chiller_Electric_ASHRAE205, "OS_Chiller_Electric_ASHRAE205", "OS:Chiller:Electric:ASHRAE205"},
    {IddObjectType::OS_CentralHeatPumpSystem, "OS_CentralHeatPumpSystem", "OS:CentralHeatPumpSystem"},
    {IddObjectType::OS_CentralHeatPumpSystem_Module, "OS_CentralHeatPumpSystem_Module", "OS:CentralHeatPumpSystem:Module"},
    {IddObjectType::OS_ChillerHeaterPerformance_Electric_EIR, "OS_ChillerHeaterPerformance_Electric_EIR", "OS:ChillerHeaterPerformance:Electric:EIR"},
    {IddObjectType::OS_Chiller_Absorption_Indirect, "OS_Chiller_Absorption_Indirect", "OS:Chiller:Absorption:Indirect"},
    {IddObjectType::OS_Chiller_Absorption, "OS_Chiller_Absorption", "OS:Chiller:Absorption"},
    {IddObjectType::OS_CoolingTower_SingleSpeed, "OS_CoolingTower_SingleSpeed", "OS:CoolingTower:SingleSpeed"},
    {IddObjectType::OS_CoolingTower_TwoSpeed, "OS_CoolingTower_TwoSpeed", "OS:CoolingTower:TwoSpeed"},
    {IddObjectType::OS_CoolingTower_VariableSpeed, "OS_CoolingTower_VariableSpeed", "OS:CoolingTower:VariableSpeed"},
    {IddObjectType::OS_CoolingTowerPerformance_CoolTools, "OS_CoolingTowerPerformance_CoolTools", "OS:CoolingTowerPerformance:CoolTools"},
    {IddObjectType::OS_CoolingTowerPerformance_YorkCalc, "OS_CoolingTowerPerformance_YorkCalc", "OS:CoolingTowerPerformance:YorkCalc"},
    {IddObjectType::OS_DistrictCooling, "OS_DistrictCooling", "OS:DistrictCooling"},
    {IddObjectType::OS_DistrictHeating, "OS_DistrictHeating", "OS:DistrictHeating"},
    {IddObjectType::OS_GroundHeatExchanger_Vertical, "OS_GroundHeatExchanger_Vertical", "OS:GroundHeatExchanger:Vertical"},
    {IddObjectType::OS_GroundHeatExchanger_HorizontalTrench, "OS_GroundHeatExchanger_HorizontalTrench", "OS:GroundHeatExchanger:HorizontalTrench"},
    {IddObjectType::OS_Coil_Cooling_CooledBeam, "OS_Coil_Cooling_CooledBeam", "OS:Coil:Cooling:CooledBeam"},
    {IddObjectType::OS_Coil_Cooling_FourPipeBeam, "OS_Coil_Cooling_FourPipeBeam", "OS:Coil:Cooling:FourPipeBeam"},
    {IddObjectType::OS_Coil_Heating_FourPipeBeam, "OS_Coil_Heating_FourPipeBeam", "OS:Coil:Heating:FourPipeBeam"},
    {IddObjectType::OS_Coil_Cooling_DX, "OS_Coil_Cooling_DX", "OS:Coil:Cooling:DX"},
    {IddObjectType::OS_Coil_Cooling_DX_CurveFit_Performance, "OS_Coil_Cooling_DX_CurveFit_Performance", "OS:Coil:Cooling:DX:CurveFit:Performance"},
    {IddObjectType::OS_Coil_Cooling_DX_CurveFit_OperatingMode, "OS_Coil_Cooling_DX_CurveFit_OperatingMode",
     "OS:Coil:Cooling:DX:CurveFit:OperatingMode"},
    {IddObjectType::OS_Coil_Cooling_DX_CurveFit_Speed, "OS_Coil_Cooling_DX_CurveFit_Speed", "OS:Coil:Cooling:DX:CurveFit:Speed"},
    {IddObjectType::OS_Coil_Cooling_DX_SingleSpeed, "OS_Coil_Cooling_DX_SingleSpeed", "OS:Coil:Cooling:DX:SingleSpeed"},
    {IddObjectType::OS_Coil_Cooling_DX_TwoStageWithHumidityControlMode, "OS_Coil_Cooling_DX_TwoStageWithHumidityControlMode",
     "OS:Coil:Cooling:DX:TwoStageWithHumidityControlMode"},
    {IddObjectType::OS_CoilPerformance_DX_Cooling, "OS_CoilPerformance_DX_Cooling", "OS:CoilPerformance:DX:Cooling"},
    {IddObjectType::OS_Coil_Cooling_DX_MultiSpeed, "OS_Coil_Cooling_DX_MultiSpeed", "OS:Coil:Cooling:DX:MultiSpeed"},
    {IddObjectType::OS_Coil_Cooling_DX_MultiSpeed_StageData, "OS_Coil_Cooling_DX_MultiSpeed_StageData", "OS:Coil:Cooling:DX:MultiSpeed:StageData"},
    {IddObjectType::OS_Coil_Cooling_DX_TwoSpeed, "OS_Coil_Cooling_DX_TwoSpeed", "OS:Coil:Cooling:DX:TwoSpeed"},
    {IddObjectType::OS_CoilSystem_Cooling_Water_HeatExchangerAssisted, "OS_CoilSystem_Cooling_Water_HeatExchangerAssisted",
     "OS:CoilSystem:Cooling:Water:HeatExchangerAssisted"},
    {IddObjectType::OS_CoilSystem_Cooling_DX_HeatExchangerAssisted, "OS_CoilSystem_Cooling_DX_HeatExchangerAssisted",
     "OS:CoilSystem:Cooling:DX:HeatExchangerAssisted"},
    {IddObjectType::OS_CoilSystem_IntegratedHeatPump_AirSource, "OS_CoilSystem_IntegratedHeatPump_AirSource",
     "OS:CoilSystem:IntegratedHeatPump:AirSource"},
    {IddObjectType::OS_Coil_Cooling_DX_VariableRefrigerantFlow, "OS_Coil_Cooling_DX_VariableRefrigerantFlow",
     "OS:Coil:Cooling:DX:VariableRefrigerantFlow"},
    {IddObjectType::OS_Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl,
     "OS_Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl", "OS:Coil:Cooling:DX:VariableRefrigerantFlow:FluidTemperatureControl"},
    {IddObjectType::OS_Coil_Cooling_LowTemperatureRadiant_ConstantFlow, "OS_Coil_Cooling_LowTemperatureRadiant_ConstantFlow",
     "OS:Coil:Cooling:LowTemperatureRadiant:ConstantFlow"},
    {IddObjectType::OS_Coil_Cooling_LowTemperatureRadiant_VariableFlow, "OS_Coil_Cooling_LowTemperatureRadiant_VariableFlow",
     "OS:Coil:Cooling:LowTemperatureRadiant:VariableFlow"},
    {IddObjectType::OS_Coil_Cooling_Water, "OS_Coil_Cooling_Water", "OS:Coil:Cooling:Water"},
    {IddObjectType::OS_Coil_Cooling_WaterToAirHeatPump_EquationFit, "OS_Coil_Cooling_WaterToAirHeatPump_EquationFit",
     "OS:Coil:Cooling:WaterToAirHeatPump:EquationFit"},
    {IddObjectType::OS_Coil_Heating_DX_SingleSpeed, "OS_Coil_Heating_DX_SingleSpeed", "OS:Coil:Heating:DX:SingleSpeed"},
    {IddObjectType::OS_Coil_Heating_DX_VariableRefrigerantFlow, "OS_Coil_Heating_DX_VariableRefrigerantFlow",
     "OS:Coil:Heating:DX:VariableRefrigerantFlow"},
    {IddObjectType::OS_Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl,
     "OS_Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl", "OS:Coil:Heating:DX:VariableRefrigerantFlow:FluidTemperatureControl"},
    {IddObjectType::OS_Coil_Heating_Electric, "OS_Coil_Heating_Electric", "OS:Coil:Heating:Electric"},
    {IddObjectType::OS_Coil_Heating_Electric_MultiStage, "OS_Coil_Heating_Electric_MultiStage", "OS:Coil:Heating:Electric:MultiStage"},
    {IddObjectType::OS_Coil_Heating_Electric_MultiStage_StageData, "OS_Coil_Heating_Electric_MultiStage_StageData",
     "OS:Coil:Heating:Electric:MultiStage:StageData"},
    {IddObjectType::OS_Coil_Heating_Gas, "OS_Coil_Heating_Gas", "OS:Coil:Heating:Gas"},
    {IddObjectType::OS_Coil_Heating_Gas_MultiStage, "OS_Coil_Heating_Gas_MultiStage", "OS:Coil:Heating:Gas:MultiStage"},
    {IddObjectType::OS_Coil_Heating_Gas_MultiStage_StageData, "OS_Coil_Heating_Gas_MultiStage_StageData", "OS:Coil:Heating:Gas:MultiStage:StageData"},
    {IddObjectType::OS_Coil_Heating_Desuperheater, "OS_Coil_Heating_Desuperheater", "OS:Coil:Heating:Desuperheater"},
    {IddObjectType::OS_Coil_WaterHeating_Desuperheater, "OS_Coil_WaterHeating_Desuperheater", "OS:Coil:WaterHeating:Desuperheater"},
    {IddObjectType::OS_Coil_Heating_LowTemperatureRadiant_ConstantFlow, "OS_Coil_Heating_LowTemperatureRadiant_ConstantFlow",
     "OS:Coil:Heating:LowTemperatureRadiant:ConstantFlow"},
    {IddObjectType::OS_Coil_Heating_LowTemperatureRadiant_VariableFlow, "OS_Coil_Heating_LowTemperatureRadiant_VariableFlow",
     "OS:Coil:Heating:LowTemperatureRadiant:VariableFlow"},
    {IddObjectType::OS_Coil_Heating_Water, "OS_Coil_Heating_Water", "OS:Coil:Heating:Water"},
    {IddObjectType::OS_Coil_Heating_Water_Baseboard, "OS_Coil_Heating_Water_Baseboard", "OS:Coil:Heating:Water:Baseboard"},
    {IddObjectType::OS_Coil_Heating_Water_Baseboard_Radiant, "OS_Coil_Heating_Water_Baseboard_Radiant", "OS:Coil:Heating:Water:Baseboard:Radiant"},
    {IddObjectType::OS_Coil_Heating_WaterToAirHeatPump_EquationFit, "OS_Coil_Heating_WaterToAirHeatPump_EquationFit",
     "OS:Coil:Heating:WaterToAirHeatPump:EquationFit"},
    {IddObjectType::OS_Coil_Heating_DX_MultiSpeed, "OS_Coil_Heating_DX_MultiSpeed", "OS:Coil:Heating:DX:MultiSpeed"},
    {IddObjectType::OS_Coil_Heating_DX_MultiSpeed_StageData, "OS_Coil_Heating_DX_MultiSpeed_StageData", "OS:Coil:Heating:DX:MultiSpeed:StageData"},
    {IddObjectType::OS_Coil_Heating_DX_VariableSpeed, "OS_Coil_Heating_DX_VariableSpeed", "OS:Coil:Heating:DX:VariableSpeed"},
    {IddObjectType::OS_Coil_Heating_DX_VariableSpeed_SpeedData, "OS_Coil_Heating_DX_VariableSpeed_SpeedData",
     "OS:Coil:Heating:DX:VariableSpeed:SpeedData"},
    {IddObjectType::OS_Coil_Cooling_DX_VariableSpeed, "OS_Coil_Cooling_DX_VariableSpeed", "OS:Coil:Cooling:DX:VariableSpeed"},
    {IddObjectType::OS_Coil_Cooling_DX_VariableSpeed_SpeedData, "OS_Coil_Cooling_DX_VariableSpeed_SpeedData",
     "OS:Coil:Cooling:DX:VariableSpeed:SpeedData"},
    {IddObjectType::OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit, "OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit",
     "OS:Coil:Heating:WaterToAirHeatPump:VariableSpeedEquationFit"},
    {IddObjectType::OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData,
     "OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData",
     "OS:Coil:Heating:WaterToAirHeatPump:VariableSpeedEquationFit:SpeedData"},
    {IddObjectType::OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit, "OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit",
     "OS:Coil:Cooling:WaterToAirHeatPump:VariableSpeedEquationFit"},
    {IddObjectType::OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData,
     "OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData",
     "OS:Coil:Cooling:WaterToAirHeatPump:VariableSpeedEquationFit:SpeedData"},
    {IddObjectType::OS_Coil_Cooling_Water_Panel_Radiant, "OS_Coil_Cooling_Water_Panel_Radiant", "OS:Coil:Cooling:Water:Panel:Radiant"},
    {IddObjectType::OS_Connection, "OS_Connection", "OS:Connection"},
    {IddObjectType::OS_Connector_Mixer, "OS_Connector_Mixer", "OS:Connector:Mixer"},
    {IddObjectType::OS_Connector_Splitter, "OS_Connector_Splitter", "OS:Connector:Splitter"},
    {IddObjectType::OS_Controller_MechanicalVentilation, "OS_Controller_MechanicalVentilation", "OS:Controller:MechanicalVentilation"},
    {IddObjectType::OS_ZoneControl_ContaminantController, "OS_ZoneControl_ContaminantController", "OS:ZoneControl:ContaminantController"},
    {IddObjectType::OS_Controller_OutdoorAir, "OS_Controller_OutdoorAir", "OS:Controller:OutdoorAir"},
    {IddObjectType::OS_Controller_WaterCoil, "OS_Controller_WaterCoil", "OS:Controller:WaterCoil"},
    {IddObjectType::OS_Curve_Linear, "OS_Curve_Linear", "OS:Curve:Linear"},
    {IddObjectType::OS_Curve_QuadLinear, "OS_Curve_QuadLinear", "OS:Curve:QuadLinear"},
    {IddObjectType::OS_Curve_QuintLinear, "OS_Curve_QuintLinear", "OS:Curve:QuintLinear"},
    {IddObjectType::OS_Curve_Quadratic, "OS_Curve_Quadratic", "OS:Curve:Quadratic"},
    {IddObjectType::OS_Curve_Cubic, "OS_Curve_Cubic", "OS:Curve:Cubic"},
    {IddObjectType::OS_Curve_Quartic, "OS_Curve_Quartic", "OS:Curve:Quartic"},
    {IddObjectType::OS_Curve_Exponent, "OS_Curve_Exponent", "OS:Curve:Exponent"},
    {IddObjectType::OS_Curve_Bicubic, "OS_Curve_Bicubic", "OS:Curve:Bicubic"},
    {IddObjectType::OS_Curve_Biquadratic, "OS_Curve_Biquadratic", "OS:Curve:Biquadratic"},
    {IddObjectType::OS_Curve_QuadraticLinear, "OS_Curve_QuadraticLinear", "OS:Curve:QuadraticLinear"},
    {IddObjectType::OS_Curve_Triquadratic, "OS_Curve_Triquadratic", "OS:Curve:Triquadratic"},
    {IddObjectType::OS_Curve_Functional_PressureDrop, "OS_Curve_Functional_PressureDrop", "OS:Curve:Functional:PressureDrop"},
    {IddObjectType::OS_Curve_FanPressureRise, "OS_Curve_FanPressureRise", "OS:Curve:FanPressureRise"},
    {IddObjectType::OS_Curve_ExponentialSkewNormal, "OS_Curve_ExponentialSkewNormal", "OS:Curve:ExponentialSkewNormal"},
    {IddObjectType::OS_Curve_Sigmoid, "OS_Curve_Sigmoid", "OS:Curve:Sigmoid"},
    {IddObjectType::OS_Curve_RectangularHyperbola1, "OS_Curve_RectangularHyperbola1", "OS:Curve:RectangularHyperbola1"},
    {IddObjectType::OS_Curve_RectangularHyperbola2, "OS_Curve_RectangularHyperbola2", "OS:Curve:RectangularHyperbola2"},
    {IddObjectType::OS_Curve_ExponentialDecay, "OS_Curve_ExponentialDecay", "OS:Curve:ExponentialDecay"},
    {IddObjectType::OS_Curve_DoubleExponentialDecay, "OS_Curve_DoubleExponentialDecay", "OS:Curve:DoubleExponentialDecay"},
    {IddObjectType::OS_Table_MultiVariableLookup, "OS_Table_MultiVariableLookup", "OS:Table:MultiVariableLookup"},
    {IddObjectType::OS_Table_IndependentVariable, "OS_Table_IndependentVariable", "OS:Table:IndependentVariable"},
    {IddObjectType::OS_Table_Lookup, "OS_Table_Lookup", "OS:Table:Lookup"},
    {IddObjectType::OS_DesignSpecification_OutdoorAir, "OS_DesignSpecification_OutdoorAir", "OS:DesignSpecification:OutdoorAir"},
    {IddObjectType::OS_DesignSpecification_ZoneAirDistribution, "OS_DesignSpecification_ZoneAirDistribution",
     "OS:DesignSpecification:ZoneAirDistribution"},
    {IddObjectType::OS_Humidifier_Steam_Electric, "OS_Humidifier_Steam_Electric", "OS:Humidifier:Steam:Electric"},
    {IddObjectType::OS_Humidifier_Steam_Gas, "OS_Humidifier_Steam_Gas", "OS:Humidifier:Steam:Gas"},
    {IddObjectType::OS_EvaporativeCooler_Indirect_ResearchSpecial, "OS_EvaporativeCooler_Indirect_ResearchSpecial",
     "OS:EvaporativeCooler:Indirect:ResearchSpecial"},
    {IddObjectType::OS_EvaporativeCooler_Direct_ResearchSpecial, "OS_EvaporativeCooler_Direct_ResearchSpecial",
     "OS:EvaporativeCooler:Direct:ResearchSpecial"},
    {IddObjectType::OS_EvaporativeFluidCooler_SingleSpeed, "OS_EvaporativeFluidCooler_SingleSpeed", "OS:EvaporativeFluidCooler:SingleSpeed"},
    {IddObjectType::OS_EvaporativeFluidCooler_TwoSpeed, "OS_EvaporativeFluidCooler_TwoSpeed", "OS:EvaporativeFluidCooler:TwoSpeed"},
    {IddObjectType::OS_Fan_SystemModel, "OS_Fan_SystemModel", "OS:Fan:SystemModel"},
    {IddObjectType::OS_Fan_ConstantVolume, "OS_Fan_ConstantVolume", "OS:Fan:ConstantVolume"},
    {IddObjectType::OS_Fan_VariableVolume, "OS_Fan_VariableVolume", "OS:Fan:VariableVolume"},
    {IddObjectType::OS_Fan_OnOff, "OS_Fan_OnOff", "OS:Fan:OnOff"},
    {IddObjectType::OS_Fan_ComponentModel, "OS_Fan_ComponentModel", "OS:Fan:ComponentModel"},
    {IddObjectType::OS_Fan_ZoneExhaust, "OS_Fan_ZoneExhaust", "OS:Fan:ZoneExhaust"},
    {IddObjectType::OS_HeatExchanger_AirToAir_SensibleAndLatent, "OS_HeatExchanger_AirToAir_SensibleAndLatent",
     "OS:HeatExchanger:AirToAir:SensibleAndLatent"},
    {IddObjectType::OS_HeatExchanger_Desiccant_BalancedFlow, "OS_HeatExchanger_Desiccant_BalancedFlow", "OS:HeatExchanger:Desiccant:BalancedFlow"},
    {IddObjectType::OS_HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1, "OS_HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1",
     "OS:HeatExchanger:Desiccant:BalancedFlow:PerformanceDataType1"},
    {IddObjectType::OS_HeatExchanger_FluidToFluid, "OS_HeatExchanger_FluidToFluid", "OS:HeatExchanger:FluidToFluid"},
    {IddObjectType::OS_HVACComponentList, "OS_HVACComponentList", "OS:HVACComponentList"},
    {IddObjectType::OS_Node, "OS_Node", "OS:Node"},
    {IddObjectType::OS_PortList, "OS_PortList", "OS:PortList"},
    {IddObjectType::OS_Pipe_Adiabatic, "OS_Pipe_Adiabatic", "OS:Pipe:Adiabatic"},
    {IddObjectType::OS_PlantLoop, "OS_PlantLoop", "OS:PlantLoop"},
    {IddObjectType::OS_PlantEquipmentOperation_CoolingLoad, "OS_PlantEquipmentOperation_CoolingLoad", "OS:PlantEquipmentOperation:CoolingLoad"},
    {IddObjectType::OS_PlantEquipmentOperation_HeatingLoad, "OS_PlantEquipmentOperation_HeatingLoad", "OS:PlantEquipmentOperation:HeatingLoad"},
    {IddObjectType::OS_PlantEquipmentOperation_OutdoorDryBulb, "OS_PlantEquipmentOperation_OutdoorDryBulb",
     "OS:PlantEquipmentOperation:OutdoorDryBulb"},
    {IddObjectType::OS_PlantEquipmentOperation_OutdoorWetBulb, "OS_PlantEquipmentOperation_OutdoorWetBulb",
     "OS:PlantEquipmentOperation:OutdoorWetBulb"},
    {IddObjectType::OS_PlantEquipmentOperation_OutdoorRelativeHumidity, "OS_PlantEquipmentOperation_OutdoorRelativeHumidity",
     "OS:PlantEquipmentOperation:OutdoorRelativeHumidity"},
    {IddObjectType::OS_PlantEquipmentOperation_OutdoorDewpoint, "OS_PlantEquipmentOperation_OutdoorDewpoint",
     "OS:PlantEquipmentOperation:OutdoorDewpoint"},
    {IddObjectType::OS_PlantEquipmentOperation_OutdoorDryBulbDifference, "OS_PlantEquipmentOperation_OutdoorDryBulbDifference",
     "OS:PlantEquipmentOperation:OutdoorDryBulbDifference"},
    {IddObjectType::OS_PlantEquipmentOperation_OutdoorDewpointDifference, "OS_PlantEquipmentOperation_OutdoorDewpointDifference",
     "OS:PlantEquipmentOperation:OutdoorDewpointDifference"},
    {IddObjectType::OS_PlantEquipmentOperation_OutdoorWetBulbDifference, "OS_PlantEquipmentOperation_OutdoorWetBulbDifference",
     "OS:PlantEquipmentOperation:OutdoorWetBulbDifference"},
    {IddObjectType::OS_PlantComponent_TemperatureSource, "OS_PlantComponent_TemperatureSource", "OS:PlantComponent:TemperatureSource"},
    {IddObjectType::OS_LoadProfile_Plant, "OS_LoadProfile_Plant", "OS:LoadProfile:Plant"},
    {IddObjectType::OS_Pump_ConstantSpeed, "OS_Pump_ConstantSpeed", "OS:Pump:ConstantSpeed"},
    {IddObjectType::OS_Pump_VariableSpeed, "OS_Pump_VariableSpeed", "OS:Pump:VariableSpeed"},
    {IddObjectType::OS_HeaderedPumps_ConstantSpeed, "OS_HeaderedPumps_ConstantSpeed", "OS:HeaderedPumps:ConstantSpeed"},
    {IddObjectType::OS_HeaderedPumps_VariableSpeed, "OS_HeaderedPumps_VariableSpeed", "OS:HeaderedPumps:VariableSpeed"},
    {IddObjectType::OS_TemperingValve, "OS_TemperingValve", "OS:TemperingValve"},
    {IddObjectType::OS_SetpointManager_FollowOutdoorAirTemperature, "OS_SetpointManager_FollowOutdoorAirTemperature",
     "OS:SetpointManager:FollowOutdoorAirTemperature"},
    {IddObjectType::OS_SetpointManager_MixedAir, "OS_SetpointManager_MixedAir", "OS:SetpointManager:MixedAir"},
    {IddObjectType::OS_SetpointManager_OutdoorAirReset, "OS_SetpointManager_OutdoorAirReset", "OS:SetpointManager:OutdoorAirReset"},
    {IddObjectType::OS_SetpointManager_OutdoorAirPretreat, "OS_SetpointManager_OutdoorAirPretreat", "OS:SetpointManager:OutdoorAirPretreat"},
    {IddObjectType::OS_SetpointManager_SingleZone_Reheat, "OS_SetpointManager_SingleZone_Reheat", "OS:SetpointManager:SingleZone:Reheat"},
    {IddObjectType::OS_SetpointManager_SingleZone_Humidity_Minimum, "OS_SetpointManager_SingleZone_Humidity_Minimum",
     "OS:SetpointManager:SingleZone:Humidity:Minimum"},
    {IddObjectType::OS_SetpointManager_MultiZone_MinimumHumidity_Average, "OS_SetpointManager_MultiZone_MinimumHumidity_Average",
     "OS:SetpointManager:MultiZone:MinimumHumidity:Average"},
    {IddObjectType::OS_SetpointManager_MultiZone_Humidity_Minimum, "OS_SetpointManager_MultiZone_Humidity_Minimum",
     "OS:SetpointManager:MultiZone:Humidity:Minimum"},
    {IddObjectType::OS_SetpointManager_Scheduled, "OS_SetpointManager_Scheduled", "OS:SetpointManager:Scheduled"},
    {IddObjectType::OS_SetpointManager_Scheduled_DualSetpoint, "OS_SetpointManager_Scheduled_DualSetpoint",
     "OS:SetpointManager:Scheduled:DualSetpoint"},
    {IddObjectType::OS_SetpointManager_Warmest, "OS_SetpointManager_Warmest", "OS:SetpointManager:Warmest"},
    {IddObjectType::OS_SetpointManager_WarmestTemperatureFlow, "OS_SetpointManager_WarmestTemperatureFlow",
     "OS:SetpointManager:WarmestTemperatureFlow"},
    {IddObjectType::OS_SetpointManager_Coldest, "OS_SetpointManager_Coldest", "OS:SetpointManager:Coldest"},
    {IddObjectType::OS_SetpointManager_FollowGroundTemperature, "OS_SetpointManager_FollowGroundTemperature",
     "OS:SetpointManager:FollowGroundTemperature"},
    {IddObjectType::OS_SetpointManager_FollowSystemNodeTemperature, "OS_SetpointManager_FollowSystemNodeTemperature",
     "OS:SetpointManager:FollowSystemNodeTemperature"},
    {IddObjectType::OS_SetpointManager_MultiZone_Cooling_Average, "OS_SetpointManager_MultiZone_Cooling_Average",
     "OS:SetpointManager:MultiZone:Cooling:Average"},
    {IddObjectType::OS_SetpointManager_MultiZone_Heating_Average, "OS_SetpointManager_MultiZone_Heating_Average",
     "OS:SetpointManager:MultiZone:Heating:Average"},
    {IddObjectType::OS_SetpointManager_MultiZone_Humidity_Maximum, "OS_SetpointManager_MultiZone_Humidity_Maximum",
     "OS:SetpointManager:MultiZone:Humidity:Maximum"},
    {IddObjectType::OS_SetpointManager_MultiZone_MaximumHumidity_Average, "OS_SetpointManager_MultiZone_MaximumHumidity_Average",
     "OS:SetpointManager:MultiZone:MaximumHumidity:Average"},
    {IddObjectType::OS_SetpointManager_SingleZone_OneStageCooling, "OS_SetpointManager_SingleZone_OneStageCooling",
     "OS:SetpointManager:SingleZone:OneStageCooling"},
    {IddObjectType::OS_SetpointManager_SingleZone_OneStageHeating, "OS_SetpointManager_SingleZone_OneStageHeating",
     "OS:SetpointManager:SingleZone:OneStageHeating"},
    {IddObjectType::OS_SetpointManager_SingleZone_Humidity_Maximum, "OS_SetpointManager_SingleZone_Humidity_Maximum",
     "OS:SetpointManager:SingleZone:Humidity:Maximum"},
    {IddObjectType::OS_SetpointManager_SingleZone_Cooling, "OS_SetpointManager_SingleZone_Cooling", "OS:SetpointManager:SingleZone:Cooling"},
    {IddObjectType::OS_SetpointManager_SingleZone_Heating, "OS_SetpointManager_SingleZone_Heating", "OS:SetpointManager:SingleZone:Heating"},
    {IddObjectType::OS_SetpointManager_SystemNodeReset_Temperature, "OS_SetpointManager_SystemNodeReset_Temperature",
     "OS:SetpointManager:SystemNodeReset:Temperature"},
    {IddObjectType::OS_SetpointManager_SystemNodeReset_Humidity, "OS_SetpointManager_SystemNodeReset_Humidity",
     "OS:SetpointManager:SystemNodeReset:Humidity"},
    {IddObjectType::OS_Sizing_Plant, "OS_Sizing_Plant", "OS:Sizing:Plant"},
    {IddObjectType::OS_Sizing_System, "OS_Sizing_System", "OS:Sizing:System"},
    {IddObjectType::OS_Sizing_Zone, "OS_Sizing_Zone", "OS:Sizing:Zone"},
    {IddObjectType::OS_Splitter, "OS_Splitter", "OS:Splitter"},
    {IddObjectType::OS_ThermalZone, "OS_ThermalZone", "OS:ThermalZone"},
    {IddObjectType::OS_ZoneControl_Humidistat, "OS_ZoneControl_Humidistat", "OS:ZoneControl:Humidistat"},
    {IddObjectType::OS_ThermostatSetpoint_DualSetpoint, "OS_ThermostatSetpoint_DualSetpoint", "OS:ThermostatSetpoint:DualSetpoint"},
    {IddObjectType::OS_ZoneControl_Thermostat_StagedDualSetpoint, "OS_ZoneControl_Thermostat_StagedDualSetpoint",
     "OS:ZoneControl:Thermostat:StagedDualSetpoint"},
    {IddObjectType::OS_ZoneHVAC_EquipmentList, "OS_ZoneHVAC_EquipmentList", "OS:ZoneHVAC:EquipmentList"},
    {IddObjectType::OS_ZoneVentilation_DesignFlowRate, "OS_ZoneVentilation_DesignFlowRate", "OS:ZoneVentilation:DesignFlowRate"},
    {IddObjectType::OS_ZoneVentilation_WindandStackOpenArea, "OS_ZoneVentilation_WindandStackOpenArea", "OS:ZoneVentilation:WindandStackOpenArea"},
    {IddObjectType::OS_ZoneHVAC_Baseboard_Convective_Electric, "OS_ZoneHVAC_Baseboard_Convective_Electric",
     "OS:ZoneHVAC:Baseboard:Convective:Electric"},
    {IddObjectType::OS_ZoneHVAC_Baseboard_Convective_Water, "OS_ZoneHVAC_Baseboard_Convective_Water", "OS:ZoneHVAC:Baseboard:Convective:Water"},
    {IddObjectType::OS_ZoneHVAC_HighTemperatureRadiant, "OS_ZoneHVAC_HighTemperatureRadiant", "OS:ZoneHVAC:HighTemperatureRadiant"},
    {IddObjectType::OS_ZoneHVAC_LowTemperatureRadiant_Electric, "OS_ZoneHVAC_LowTemperatureRadiant_Electric",
     "OS:ZoneHVAC:LowTemperatureRadiant:Electric"},
    {IddObjectType::OS_ZoneHVAC_IdealLoadsAirSystem, "OS_ZoneHVAC_IdealLoadsAirSystem", "OS:ZoneHVAC:IdealLoadsAirSystem"},
    {IddObjectType::OS_ZoneHVAC_FourPipeFanCoil, "OS_ZoneHVAC_FourPipeFanCoil", "OS:ZoneHVAC:FourPipeFanCoil"},
    {IddObjectType::OS_ZoneHVAC_LowTemperatureRadiant_ConstantFlow, "OS_ZoneHVAC_LowTemperatureRadiant_ConstantFlow",
     "OS:ZoneHVAC:LowTemperatureRadiant:ConstantFlow"},
    {IddObjectType::OS_ZoneHVAC_LowTemperatureRadiant_VariableFlow, "OS_ZoneHVAC_LowTemperatureRadiant_VariableFlow",
     "OS:ZoneHVAC:LowTemperatureRadiant:VariableFlow"},
    {IddObjectType::OS_ZoneHVAC_PackagedTerminalHeatPump, "OS_ZoneHVAC_PackagedTerminalHeatPump", "OS:ZoneHVAC:PackagedTerminalHeatPump"},
    {IddObjectType::OS_ZoneHVAC_PackagedTerminalAirConditioner, "OS_ZoneHVAC_PackagedTerminalAirConditioner",
     "OS:ZoneHVAC:PackagedTerminalAirConditioner"},
    {IddObjectType::OS_ZoneHVAC_TerminalUnit_VariableRefrigerantFlow, "OS_ZoneHVAC_TerminalUnit_VariableRefrigerantFlow",
     "OS:ZoneHVAC:TerminalUnit:VariableRefrigerantFlow"},
    {IddObjectType::OS_ZoneHVAC_WaterToAirHeatPump, "OS_ZoneHVAC_WaterToAirHeatPump", "OS:ZoneHVAC:WaterToAirHeatPump"},
    {IddObjectType::OS_ZoneHVAC_UnitHeater, "OS_ZoneHVAC_UnitHeater", "OS:ZoneHVAC:UnitHeater"},
    {IddObjectType::OS_ZoneHVAC_Dehumidifier_DX, "OS_ZoneHVAC_Dehumidifier_DX", "OS:ZoneHVAC:Dehumidifier:DX"},
    {IddObjectType::OS_ZoneHVAC_EnergyRecoveryVentilator, "OS_ZoneHVAC_EnergyRecoveryVentilator", "OS:ZoneHVAC:EnergyRecoveryVentilator"},
    {IddObjectType::OS_ZoneHVAC_EnergyRecoveryVentilator_Controller, "OS_ZoneHVAC_EnergyRecoveryVentilator_Controller",
     "OS:ZoneHVAC:EnergyRecoveryVentilator:Controller"},
    {IddObjectType::OS_ZoneHVAC_UnitVentilator, "OS_ZoneHVAC_UnitVentilator", "OS:ZoneHVAC:UnitVentilator"},
    {IddObjectType::OS_ZoneHVAC_Baseboard_RadiantConvective_Water, "OS_ZoneHVAC_Baseboard_RadiantConvective_Water",
     "OS:ZoneHVAC:Baseboard:RadiantConvective:Water"},
    {IddObjectType::OS_ZoneHVAC_Baseboard_RadiantConvective_Electric, "OS_ZoneHVAC_Baseboard_RadiantConvective_Electric",
     "OS:ZoneHVAC:Baseboard:RadiantConvective:Electric"},
    {IddObjectType::OS_ZoneHVAC_CoolingPanel_RadiantConvective_Water, "OS_ZoneHVAC_CoolingPanel_RadiantConvective_Water",
     "OS:ZoneHVAC:CoolingPanel:RadiantConvective:Water"},
    {IddObjectType::OS_ZoneMixing, "OS_ZoneMixing", "OS:ZoneMixing"},
    {IddObjectType::OS_WaterHeater_Mixed, "OS_WaterHeater_Mixed", "OS:WaterHeater:Mixed"},
    {IddObjectType::OS_WaterHeater_HeatPump, "OS_WaterHeater_HeatPump", "OS:WaterHeater:HeatPump"},
    {IddObjectType::OS_Coil_WaterHeating_AirToWaterHeatPump, "OS_Coil_WaterHeating_AirToWaterHeatPump", "OS:Coil:WaterHeating:AirToWaterHeatPump"},
    {IddObjectType::OS_WaterHeater_HeatPump_WrappedCondenser, "OS_WaterHeater_HeatPump_WrappedCondenser", "OS:WaterHeater:HeatPump:WrappedCondenser"},
    {IddObjectType::OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed, "OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed",
     "OS:Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed"},
    {IddObjectType::OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed_SpeedData,
     "OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed_SpeedData", "OS:Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed:SpeedData"},
    {IddObjectType::OS_Coil_WaterHeating_AirToWaterHeatPump_Wrapped, "OS_Coil_WaterHeating_AirToWaterHeatPump_Wrapped",
     "OS:Coil:WaterHeating:AirToWaterHeatPump:Wrapped"},
    {IddObjectType::OS_WaterHeater_Stratified, "OS_WaterHeater_Stratified", "OS:WaterHeater:Stratified"},
    {IddObjectType::OS_WaterHeater_Sizing, "OS_WaterHeater_Sizing", "OS:WaterHeater:Sizing"},
    {IddObjectType::OS_WaterUse_Equipment, "OS_WaterUse_Equipment", "OS:WaterUse:Equipment"},
    {IddObjectType::OS_WaterUse_Equipment_Definition, "OS_WaterUse_Equipment_Definition", "OS:WaterUse:Equipment:Definition"},
    {IddObjectType::OS_WaterUse_Connections, "OS_WaterUse_Connections", "OS:WaterUse:Connections"},
    {IddObjectType::OS_ThermalStorage_Ice_Detailed, "OS_ThermalStorage_Ice_Detailed", "OS:ThermalStorage:Ice:Detailed"},
    {IddObjectType::OS_ThermalStorage_ChilledWater_Stratified, "OS_ThermalStorage_ChilledWater_Stratified",
     "OS:ThermalStorage:ChilledWater:Stratified"},
    {IddObjectType::OS_Duct, "OS_Duct", "OS:Duct"},
    {IddObjectType::OS_FluidCooler_SingleSpeed, "OS_FluidCooler_SingleSpeed", "OS:FluidCooler:SingleSpeed"},
    {IddObjectType::OS_FluidCooler_TwoSpeed, "OS_FluidCooler_TwoSpeed", "OS:FluidCooler:TwoSpeed"},
    {IddObjectType::OS_Pipe_Indoor, "OS_Pipe_Indoor", "OS:Pipe:Indoor"},
    {IddObjectType::OS_Pipe_Outdoor, "OS_Pipe_Outdoor", "OS:Pipe:Outdoor"},
    {IddObjectType::OS_SwimmingPool_Indoor, "OS_SwimmingPool_Indoor", "OS:SwimmingPool:Indoor"},
    {IddObjectType::OS_SolarCollectorPerformance_FlatPlate, "OS_SolarCollectorPerformance_FlatPlate", "OS:SolarCollectorPerformance:FlatPlate"},
    {IddObjectType::OS_SolarCollector_FlatPlate_Water, "OS_SolarCollector_FlatPlate_Water", "OS:SolarCollector:FlatPlate:Water"},
    {IddObjectType::OS_SolarCollectorPerformance_IntegralCollectorStorage, "OS_SolarCollectorPerformance_IntegralCollectorStorage",
     "OS:SolarCollectorPerformance:IntegralCollectorStorage"},
    {IddObjectType::OS_SolarCollector_IntegralCollectorStorage, "OS_SolarCollector_IntegralCollectorStorage",
     "OS:SolarCollector:IntegralCollectorStorage"},
    {IddObjectType::OS_SolarCollectorPerformance_PhotovoltaicThermal_Simple, "OS_SolarCollectorPerformance_PhotovoltaicThermal_Simple",
     "OS:SolarCollectorPerformance:PhotovoltaicThermal:Simple"},
    {IddObjectType::OS_SolarCollectorPerformance_PhotovoltaicThermal_BIPVT, "OS_SolarCollectorPerformance_PhotovoltaicThermal_BIPVT",
     "OS:SolarCollectorPerformance:PhotovoltaicThermal:BIPVT"},
    {IddObjectType::OS_SolarCollector_FlatPlate_PhotovoltaicThermal, "OS_SolarCollector_FlatPlate_PhotovoltaicThermal",
     "OS:SolarCollector:FlatPlate:PhotovoltaicThermal"},
    {IddObjectType::OS_Generator_MicroTurbine, "OS_Generator_MicroTurbine", "OS:Generator:MicroTurbine"},
    {IddObjectType::OS_Generator_MicroTurbine_HeatRecovery, "OS_Generator_MicroTurbine_HeatRecovery", "OS:Generator:MicroTurbine:HeatRecovery"},
    {IddObjectType::OS_Generator_Photovoltaic, "OS_Generator_Photovoltaic", "OS:Generator:Photovoltaic"},
    {IddObjectType::OS_PhotovoltaicPerformance_Simple, "OS_PhotovoltaicPerformance_Simple", "OS:PhotovoltaicPerformance:Simple"},
    {IddObjectType::OS_PhotovoltaicPerformance_EquivalentOneDiode, "OS_PhotovoltaicPerformance_EquivalentOneDiode",
     "OS:PhotovoltaicPerformance:EquivalentOne-Diode"},
    {IddObjectType::OS_PhotovoltaicPerformance_Sandia, "OS_PhotovoltaicPerformance_Sandia", "OS:PhotovoltaicPerformance:Sandia"},
    {IddObjectType::OS_Generator_FuelCell, "OS_Generator_FuelCell", "OS:Generator:FuelCell"},
    {IddObjectType::OS_Generator_FuelCell_PowerModule, "OS_Generator_FuelCell_PowerModule", "OS:Generator:FuelCell:PowerModule"},
    {IddObjectType::OS_Generator_FuelCell_AirSupply, "OS_Generator_FuelCell_AirSupply", "OS:Generator:FuelCell:AirSupply"},
    {IddObjectType::OS_Generator_FuelCell_WaterSupply, "OS_Generator_FuelCell_WaterSupply", "OS:Generator:FuelCell:WaterSupply"},
    {IddObjectType::OS_Generator_FuelCell_AuxiliaryHeater, "OS_Generator_FuelCell_AuxiliaryHeater", "OS:Generator:FuelCell:AuxiliaryHeater"},
    {IddObjectType::OS_Generator_FuelCell_ExhaustGasToWaterHeatExchanger, "OS_Generator_FuelCell_ExhaustGasToWaterHeatExchanger",
     "OS:Generator:FuelCell:ExhaustGasToWaterHeatExchanger"},
    {IddObjectType::OS_Generator_FuelCell_ElectricalStorage, "OS_Generator_FuelCell_ElectricalStorage", "OS:Generator:FuelCell:ElectricalStorage"},
    {IddObjectType::OS_Generator_FuelCell_Inverter, "OS_Generator_FuelCell_Inverter", "OS:Generator:FuelCell:Inverter"},
    {IddObjectType::OS_Generator_FuelCell_StackCooler, "OS_Generator_FuelCell_StackCooler", "OS:Generator:FuelCell:StackCooler"},
    {IddObjectType::OS_Generator_FuelSupply, "OS_Generator_FuelSupply", "OS:Generator:FuelSupply"},
    {IddObjectType::OS_Generator_WindTurbine, "OS_Generator_WindTurbine", "OS:Generator:WindTurbine"},
    {IddObjectType::OS_Generator_PVWatts, "OS_Generator_PVWatts", "OS:Generator:PVWatts"},
    {IddObjectType::OS_ElectricLoadCenter_Inverter_PVWatts, "OS_ElectricLoadCenter_Inverter_PVWatts", "OS:ElectricLoadCenter:Inverter:PVWatts"},
    {IddObjectType::OS_ElectricLoadCenter_Inverter_Simple, "OS_ElectricLoadCenter_Inverter_Simple", "OS:ElectricLoadCenter:Inverter:Simple"},
    {IddObjectType::OS_ElectricLoadCenter_Inverter_LookUpTable, "OS_ElectricLoadCenter_Inverter_LookUpTable",
     "OS:ElectricLoadCenter:Inverter:LookUpTable"},
    {IddObjectType::OS_ElectricLoadCenter_Storage_Simple, "OS_ElectricLoadCenter_Storage_Simple", "OS:ElectricLoadCenter:Storage:Simple"},
    {IddObjectType::OS_ElectricLoadCenter_Storage_LiIonNMCBattery, "OS_ElectricLoadCenter_Storage_LiIonNMCBattery",
     "OS:ElectricLoadCenter:Storage:LiIonNMCBattery"},
    {IddObjectType::OS_ElectricLoadCenter_Transformer, "OS_ElectricLoadCenter_Transformer", "OS:ElectricLoadCenter:Transformer"},
    {IddObjectType::OS_ElectricLoadCenter_Distribution, "OS_ElectricLoadCenter_Distribution", "OS:ElectricLoadCenter:Distribution"},
    {IddObjectType::OS_ElectricLoadCenter_Storage_Converter, "OS_ElectricLoadCenter_Storage_Converter", "OS:ElectricLoadCenter:Storage:Converter"},
    {IddObjectType::OS_ComponentCost_Adjustments, "OS_ComponentCost_Adjustments", "OS:ComponentCost:Adjustments"},
    {IddObjectType::OS_LifeCycleCost, "OS_LifeCycleCost", "OS:LifeCycleCost"},
    {IddObjectType::OS_CurrencyType, "OS_CurrencyType", "OS:CurrencyType"},
    {IddObjectType::OS_LifeCycleCost_Parameters, "OS_LifeCycleCost_Parameters", "OS:LifeCycleCost:Parameters"},
    {IddObjectType::OS_LifeCycleCost_UsePriceEscalation, "OS_LifeCycleCost_UsePriceEscalation", "OS:LifeCycleCost:UsePriceEscalation"},
    {IddObjectType::OS_UtilityBill, "OS_UtilityBill", "OS:UtilityBill"},
    {IddObjectType::OS_UtilityCost_Tariff, "OS_UtilityCost_Tariff", "OS:UtilityCost:Tariff"},
    {IddObjectType::OS_UtilityCost_Qualify, "OS_UtilityCost_Qualify", "OS:UtilityCost:Qualify"},
    {IddObjectType::OS_UtilityCost_Charge_Simple, "OS_UtilityCost_Charge_Simple", "OS:UtilityCost:Charge:Simple"},
    {IddObjectType::OS_UtilityCost_Charge_Block, "OS_UtilityCost_Charge_Block", "OS:UtilityCost:Charge:Block"},
    {IddObjectType::OS_UtilityCost_Ratchet, "OS_UtilityCost_Ratchet", "OS:UtilityCost:Ratchet"},
    {IddObjectType::OS_UtilityCost_Variable, "OS_UtilityCost_Variable", "OS:UtilityCost:Variable"},
    {IddObjectType::OS_UtilityCost_Computation, "OS_UtilityCost_Computation", "OS:UtilityCost:Computation"},
    {IddObjectType::OS_Output_Schedules, "OS_Output_Schedules", "OS:Output:Schedules"},
    {IddObjectType::OS_Output_Constructions, "OS_Output_Constructions", "OS:Output:Constructions"},
    {IddObjectType::OS_OutputControl_Files, "OS_OutputControl_Files", "OS:OutputControl:Files"},
    {IddObjectType::OS_OutputControl_Timestamp, "OS_OutputControl_Timestamp", "OS:OutputControl:Timestamp"},
    {IddObjectType::OS_OutputControl_ReportingTolerances, "OS_OutputControl_ReportingTolerances", "OS:OutputControl:ReportingTolerances"},
    {IddObjectType::OS_OutputControl_Table_Style, "OS_OutputControl_Table_Style", "OS:OutputControl:Table:Style"},
    {IddObjectType::OS_Output_Meter, "OS_Output_Meter", "OS:Output:Meter"},
    {IddObjectType::OS_Output_Variable, "OS_Output_Variable", "OS:Output:Variable"},
    {IddObjectType::OS_Meter_Custom, "OS_Meter_Custom", "OS:Meter:Custom"},
    {IddObjectType::OS_Meter_CustomDecrement, "OS_Meter_CustomDecrement", "OS:Meter:CustomDecrement"},
    {IddObjectType::OS_Output_EnvironmentalImpactFactors, "OS_Output_EnvironmentalImpactFactors", "OS:Output:EnvironmentalImpactFactors"},
    {IddObjectType::OS_EnvironmentalImpactFactors, "OS_EnvironmentalImpactFactors", "OS:EnvironmentalImpactFactors"},
    {IddObjectType::OS_FuelFactors, "OS_FuelFactors", "OS:FuelFactors"},
    {IddObjectType::OS_Output_Diagnostics, "OS_Output_Diagnostics", "OS:Output:Diagnostics"},
    {IddObjectType::OS_Output_DebuggingData, "OS_Output_DebuggingData", "OS:Output:DebuggingData"},
    {IddObjectType::OS_Output_JSON, "OS_Output_JSON", "OS:Output:JSON"},
    {IddObjectType::OS_Output_SQLite, "OS_Output_SQLite", "OS:Output:SQLite"},
    {IddObjectType::OS_Output_EnergyManagementSystem, "OS_Output_EnergyManagementSystem", "OS:Output:EnergyManagementSystem"},
    {IddObjectType::OS_Output_Table_SummaryReports, "OS_Output_Table_SummaryReports", "OS:Output:Table:SummaryReports"},
    {IddObjectType::OS_EnergyManagementSystem_Sensor, "OS_EnergyManagementSystem_Sensor", "OS:EnergyManagementSystem:Sensor"},
    {IddObjectType::OS_EnergyManagementSystem_Actuator, "OS_EnergyManagementSystem_Actuator", "OS:EnergyManagementSystem:Actuator"},
    {IddObjectType::OS_EnergyManagementSystem_ProgramCallingManager, "OS_EnergyManagementSystem_ProgramCallingManager",
     "OS:EnergyManagementSystem:ProgramCallingManager"},
    {IddObjectType::OS_EnergyManagementSystem_Program, "OS_EnergyManagementSystem_Program", "OS:EnergyManagementSystem:Program"},
    {IddObjectType::OS_EnergyManagementSystem_Subroutine, "OS_EnergyManagementSystem_Subroutine", "OS:EnergyManagementSystem:Subroutine"},
    {IddObjectType::OS_EnergyManagementSystem_GlobalVariable, "OS_EnergyManagementSystem_GlobalVariable", "OS:EnergyManagementSystem:GlobalVariable"},
    {IddObjectType::OS_EnergyManagementSystem_OutputVariable, "OS_EnergyManagementSystem_OutputVariable", "OS:EnergyManagementSystem:OutputVariable"},
    {IddObjectType::OS_EnergyManagementSystem_MeteredOutputVariable, "OS_EnergyManagementSystem_MeteredOutputVariable",
     "OS:EnergyManagementSystem:MeteredOutputVariable"},
    {IddObjectType::OS_EnergyManagementSystem_TrendVariable, "OS_EnergyManagementSystem_TrendVariable", "OS:EnergyManagementSystem:TrendVariable"},
    {IddObjectType::OS_EnergyManagementSystem_InternalVariable, "OS_EnergyManagementSystem_InternalVariable",
     "OS:EnergyManagementSystem:InternalVariable"},
    {IddObjectType::OS_EnergyManagementSystem_CurveOrTableIndexVariable, "OS_EnergyManagementSystem_CurveOrTableIndexVariable",
     "OS:EnergyManagementSystem:CurveOrTableIndexVariable"},
    {IddObjectType::OS_EnergyManagementSystem_ConstructionIndexVariable, "OS_EnergyManagementSystem_ConstructionIndexVariable",
     "OS:EnergyManagementSystem:ConstructionIndexVariable"},
    {IddObjectType::OS_PythonPlugin_Instance, "OS_PythonPlugin_Instance", "OS:PythonPlugin:Instance"},
    {IddObjectType::OS_PythonPlugin_Variable, "OS_PythonPlugin_Variable", "OS:PythonPlugin:Variable"},
    {IddObjectType::OS_PythonPlugin_TrendVariable, "OS_PythonPlugin_TrendVariable", "OS:PythonPlugin:TrendVariable"},
    {IddObjectType::OS_PythonPlugin_OutputVariable, "OS_PythonPlugin_OutputVariable", "OS:PythonPlugin:OutputVariable"},
    {IddObjectType::OS_AirflowNetworkSimulationControl, "OS_AirflowNetworkSimulationControl", "OS:AirflowNetworkSimulationControl"},
    {IddObjectType::OS_AirflowNetworkZone, "OS_AirflowNetworkZone", "OS:AirflowNetworkZone"},
    {IddObjectType::OS_AirflowNetworkSurface, "OS_AirflowNetworkSurface", "OS:AirflowNetworkSurface"},
    {IddObjectType::OS_AirflowNetworkReferenceCrackConditions, "OS_AirflowNetworkReferenceCrackConditions",
     "OS:AirflowNetworkReferenceCrackConditions"},
    {IddObjectType::OS_AirflowNetworkCrack, "OS_AirflowNetworkCrack", "OS:AirflowNetworkCrack"},
    {IddObjectType::OS_AirflowNetworkEffectiveLeakageArea, "OS_AirflowNetworkEffectiveLeakageArea", "OS:AirflowNetworkEffectiveLeakageArea"},
    {IddObjectType::OS_AirflowNetworkSpecifiedFlowRate, "OS_AirflowNetworkSpecifiedFlowRate", "OS:AirflowNetworkSpecifiedFlowRate"},
    {IddObjectType::OS_AirflowNetworkDetailedOpening, "OS_AirflowNetworkDetailedOpening", "OS:AirflowNetworkDetailedOpening"},
    {IddObjectType::OS_AirflowNetworkSimpleOpening, "OS_AirflowNetworkSimpleOpening", "OS:AirflowNetworkSimpleOpening"},
    {IddObjectType::OS_AirflowNetworkHorizontalOpening, "OS_AirflowNetworkHorizontalOpening", "OS:AirflowNetworkHorizontalOpening"},
    {IddObjectType::OS_AirflowNetworkOutdoorAirflow, "OS_AirflowNetworkOutdoorAirflow", "OS:AirflowNetworkOutdoorAirflow"},
    {IddObjectType::OS_AirflowNetworkZoneExhaustFan, "OS_AirflowNetworkZoneExhaustFan", "OS:AirflowNetworkZoneExhaustFan"},
    {IddObjectType::OS_AirflowNetworkExternalNode, "OS_AirflowNetworkExternalNode", "OS:AirflowNetworkExternalNode"},
    {IddObjectType::OS_AirflowNetworkPressureController, "OS_AirflowNetworkPressureController", "OS:AirflowNetworkPressureController"},
    {IddObjectType::OS_AirflowNetworkDistributionNode, "OS_AirflowNetworkDistributionNode", "OS:AirflowNetworkDistributionNode"},
    {IddObjectType::OS_AirflowNetworkLeak, "OS_AirflowNetworkLeak", "OS:AirflowNetworkLeak"},
    {IddObjectType::OS_AirflowNetworkLeakageRatio, "OS_AirflowNetworkLeakageRatio", "OS:AirflowNetworkLeakageRatio"},
    {IddObjectType::OS_AirflowNetworkDuct, "OS_AirflowNetworkDuct", "OS:AirflowNetworkDuct"},
    {IddObjectType::OS_AirflowNetworkFan, "OS_AirflowNetworkFan", "OS:AirflowNetworkFan"},
    {IddObjectType::OS_AirflowNetworkEquivalentDuct, "OS_AirflowNetworkEquivalentDuct", "OS:AirflowNetworkEquivalentDuct"},
    {IddObjectType::OS_AirflowNetworkConstantPressureDrop, "OS_AirflowNetworkConstantPressureDrop", "OS:AirflowNetworkConstantPressureDrop"},
    {IddObjectType::OS_AirflowNetworkReliefAirFlow, "OS_AirflowNetworkReliefAirFlow", "OS:AirflowNetworkReliefAirFlow"},
    {IddObjectType::OS_AirflowNetworkDistributionLinkage, "OS_AirflowNetworkDistributionLinkage", "OS:AirflowNetworkDistributionLinkage"},
    {IddObjectType::OS_AirflowNetworkDuctViewFactors, "OS_AirflowNetworkDuctViewFactors", "OS:AirflowNetworkDuctViewFactors"},
    {IddObjectType::OS_AirflowNetworkOccupantVentilationControl, "OS_AirflowNetworkOccupantVentilationControl",
     "OS:AirflowNetworkOccupantVentilationControl"},
    {IddObjectType::OS_ExternalInterface, "OS_ExternalInterface", "OS:ExternalInterface"},
    {IddObjectType::OS_ExternalInterface_Schedule, "OS_ExternalInterface_Schedule", "OS:ExternalInterface:Schedule"},
    {IddObjectType::OS_ExternalInterface_Variable, "OS_ExternalInterface_Variable", "OS:ExternalInterface:Variable"},
    {IddObjectType::OS_ExternalInterface_Actuator, "OS_ExternalInterface_Actuator", "OS:ExternalInterface:Actuator"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitImport, "OS_ExternalInterface_FunctionalMockupUnitImport",
     "OS:ExternalInterface:FunctionalMockupUnitImport"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitImport_From_Variable, "OS_ExternalInterface_FunctionalMockupUnitImport_From_Variable",
     "OS:ExternalInterface:FunctionalMockupUnitImport:From:Variable"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitImport_To_Schedule, "OS_ExternalInterface_FunctionalMockupUnitImport_To_Schedule",
     "OS:ExternalInterface:FunctionalMockupUnitImport:To:Schedule"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitImport_To_Actuator, "OS_ExternalInterface_FunctionalMockupUnitImport_To_Actuator",
     "OS:ExternalInterface:FunctionalMockupUnitImport:To:Actuator"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitImport_To_Variable, "OS_ExternalInterface_FunctionalMockupUnitImport_To_Variable",
     "OS:ExternalInterface:FunctionalMockupUnitImport:To:Variable"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitExport_From_Variable, "OS_ExternalInterface_FunctionalMockupUnitExport_From_Variable",
     "OS:ExternalInterface:FunctionalMockupUnitExport:From:Variable"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitExport_To_Schedule, "OS_ExternalInterface_FunctionalMockupUnitExport_To_Schedule",
     "OS:ExternalInterface:FunctionalMockupUnitExport:To:Schedule"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitExport_To_Actuator, "OS_ExternalInterface_FunctionalMockupUnitExport_To_Actuator",
     "OS:ExternalInterface:FunctionalMockupUnitExport:To:Actuator"},
    {IddObjectType::OS_ExternalInterface_FunctionalMockupUnitExport_To_Variable, "OS_ExternalInterface_FunctionalMockupUnitExport_To_Variable",
     "OS:ExternalInterface:FunctionalMockupUnitExport:To:Variable"},
    {IddObjectType::OS_Coil_UserDefined, "OS_Coil_UserDefined", "OS:Coil:UserDefined"},
    {IddObjectType::OS_PlantComponent_UserDefined, "OS_PlantComponent_UserDefined", "OS:PlantComponent:UserDefined"},
    {IddObjectType::CommentOnly, "CommentOnly", ""},
    {0, 0, 0}};
  VecType v;
  int i = 0;
  while (!(a[i].value == 0 && a[i].name == 0 && a[i].description == 0)) {
    if (isd) {
      std::string description = a[i].description;
      if (!description.empty()) {
        v.push_back(PT(description, a[i].value));
      }
    } else {
      v.push_back(PT(a[i].name, a[i].value));
    }
    ++i;
  }
  return v;
}

#endif  // UTILITIES_IDD_IDDENUMS_HXX
