#ifndef UTILITIES_IDD_IDDENUMS_CONSTEXPR_HXX
#define UTILITIES_IDD_IDDENUMS_CONSTEXPR_HXX

#include <algorithm>
#include <cctype>
#include <compare>
#include <iostream>
#include <numeric>
#include <map>
#include <set>
#include <stdexcept>
#include <string>
#include <string_view>
#include <vector>

#include <fmt/format.h>

namespace constexpr_enum {

inline std::string ascii_to_lower_copy(std::string_view input) {
  std::string result{input};
  constexpr auto to_lower_diff = 'a' - 'A';
  for (auto& c : result) {
    if (c >= 'A' && c <= 'Z') {
      c += to_lower_diff;
    }
  }
  return result;
}

constexpr char charToUpper(const char c) {
  constexpr auto to_upper_diff = 'A' - 'a';
  if (c >= 'a' && c <= 'z') {
    return c + to_upper_diff;
  }
  return c;
}

inline std::string ascii_to_upper_copy(std::string_view input) {
  std::string result{input};
  constexpr auto to_upper_diff = 'A' - 'a';
  for (auto& c : result) {
    if (c >= 'a' && c <= 'z') {
      c += to_upper_diff;
    }
  }
  return result;
}

/** Base class for OPENSTUDIO_ENUMs. Comparison operators use the underlying
 * (integer) value. */
template <typename Enum>
class EnumBase
{
 public:
  /** Default constructor. EnumBase default constructor never actually called;
     * OPENSTUDIO_ENUM- generated classes have default constructors that set the
     * value to the first item in the enumerated list. */
  EnumBase() = delete;

  /** Construct from t_value. Throws std::runtime_error if t_value is not in
     * the domain. */
  constexpr EnumBase(int t_value) : m_value(lookupValue(t_value)) {}

  /** Construct from t_value. Throws std::runtime_error if t_value does not
     * match a name or a description (case insensitive). */
  constexpr EnumBase(std::string_view t_value) : m_value(lookupValue(t_value)) {}

  /** Returns the name associated with t_value, if it exists. Otherwise,
     * throws std::runtime_error. */
  static constexpr std::string_view valueName(int t_value) {
    t_value = lookupValue(t_value);
    return Enum::names[static_cast<size_t>(t_value)];
  }

  /** Returns the description associated with t_value, if it exists.
     * Otherwise, throws std::runtime_error. */
  static constexpr std::string_view valueDescription(int t_value) {
    t_value = lookupValue(t_value);
    auto& desc = Enum::descriptions[static_cast<size_t>(t_value)];
    return desc.empty() ? valueName(t_value) : desc;
  }

  /** Returns the set of all values in this enum's domain. */
  static constexpr auto getValues() {
    std::array<int, Enum::Size> values;
    std::iota(values.begin(), values.end(), typename Enum::SizeType(0));
    return values;
  }

  /** Returns this instance's current value (as an integer). */
  constexpr int value() const {
    return m_value;
  }

  /** Returns the name associated with this instance's current value. */
  constexpr std::string_view valueName() const {
    return valueName(m_value);
  }

  /** Returns the description associated with this instance's current value.
     */
  constexpr std::string_view valueDescription() const {
    return valueDescription(m_value);
  }

  /** Set this instance's value to t_value. Throws std::runtime_error if
     *  t_value is not in the domain. */
  constexpr void setValue(int t_value) {
    m_value = lookupValue(t_value);
  }

#if 1
  constexpr auto operator<=>(const EnumBase<Enum>&) const = default;
#else
  constexpr bool operator==(const EnumBase<Enum>& other) const {
    return (m_value == other.m_value);
  }
  constexpr bool operator!=(const EnumBase<Enum>& other) const {
    return (m_value != other.m_value);
  }
  constexpr bool operator>(const EnumBase<Enum>& other) const {
    return (m_value > other.m_value);
  }
  constexpr bool operator>=(const EnumBase<Enum>& other) const {
    return (m_value >= other.m_value);
  }
  constexpr bool operator<(const EnumBase<Enum>& other) const {
    return (m_value < other.m_value);
  }
  constexpr bool operator<=(const EnumBase<Enum>& other) const {
    return (m_value <= other.m_value);
  }
#endif

  /** Returns the (integer) value associated with t_name, as determined by
     * case-insensitive comparison to the enumerated names and descriptions. */
  constexpr static int lookupValue(std::string_view t_name) {
    // auto nameUC = ascii_to_upper_copy(t_name);  // TODO: make this constexpr

    std::array<char, Enum::max_element_size> arr{};  // vector<char> works on recent clang, not on GCC 10
    std::transform(t_name.begin(), t_name.end(), arr.begin(), charToUpper);
    std::string_view nameUC{arr.data()};

#if 0
    auto predicate_caseinsensitive =
      [&t_name](std::string_view nameUC) {
        if (nameUC.size() != t_name.size()) {
          return false;
        }
        constexpr auto to_upper_diff = 'A' - 'a';
        for (size_t i = 0; i < t_name.size(); ++i) {
          auto c = t_name[i];
          if (c >= 'a' && c <= 'z') {
            c += to_upper_diff;
          }
          if (c != nameUC[i]) {
            return false;
          }
        }
      }
#endif

    auto itNames = std::find(Enum::namesUC.cbegin(), Enum::namesUC.cend(), nameUC);
    if (itNames != Enum::namesUC.cend()) {
      return std::distance(Enum::namesUC.cbegin(), itNames);
    }
    auto itDescs = std::find(Enum::descriptionsUC.cbegin(), Enum::descriptionsUC.cend(), nameUC);
    if (itDescs != Enum::descriptionsUC.cend()) {
      return std::distance(Enum::descriptionsUC.cbegin(), itDescs);
    }

    // if consteval only in C++23
    // if consteval {
    if (!std::is_constant_evaluated()) {
      throw std::runtime_error(fmt::format("Unknown OpenStudio Enum Value '{}' for Enum {}", t_name, Enum::enumName()));
    } else {
      // Or return -1 ?
      throw std::invalid_argument("Unknown OpenStudio Enum Value");
    }
  }

  /** Returns t_value if it is in the domain. Otherwise throws
     * std::runtime_error. */
  static constexpr int lookupValue(int t_value) {
    if (t_value < Enum::Size) {
      return t_value;
    } else {
      throw std::runtime_error(fmt::format("Unknown OpenStudio Enum Value = {} for Enum {}", t_value, Enum::enumName()));
    }
  }

 protected:
  int m_value;
};

struct IddObjectType : public EnumBase<IddObjectType>
{
 public:
  using SizeType = int;  // Or uint16_t? This provides 65535 possible keys which is more than enough
  static constexpr SizeType Size = 1396;

  enum domain : SizeType;

  constexpr IddObjectType() : EnumBase<IddObjectType>(0) {}
  constexpr IddObjectType(std::string_view t_name) : EnumBase<IddObjectType>(t_name) {}
  constexpr IddObjectType(int t_value) : EnumBase<IddObjectType>(t_value) {}

  static std::string_view enumName() {
    return "IddObjectType";
  }

  constexpr domain value() const {
    return static_cast<domain>(EnumBase<IddObjectType>::value());
  }

 protected:
  friend class EnumBase<IddObjectType>;

 public:
  static constexpr std::array<std::string_view, Size> names{
    "Catchall",
    "UserCustom",
    "LeadInput",
    "EndLeadInput",
    "SimulationData",
    "EndSimulationData",
    "Version",
    "SimulationControl",
    "PerformancePrecisionTradeoffs",
    "Building",
    "ShadowCalculation",
    "SurfaceConvectionAlgorithm_Inside",
    "SurfaceConvectionAlgorithm_Outside",
    "HeatBalanceAlgorithm",
    "HeatBalanceSettings_ConductionFiniteDifference",
    "ZoneAirHeatBalanceAlgorithm",
    "ZoneAirContaminantBalance",
    "ZoneAirMassFlowConservation",
    "ZoneCapacitanceMultiplier_ResearchSpecial",
    "Timestep",
    "ConvergenceLimits",
    "HVACSystemRootFindingAlgorithm",
    "Compliance_Building",
    "Site_Location",
    "Site_VariableLocation",
    "SizingPeriod_DesignDay",
    "SizingPeriod_WeatherFileDays",
    "SizingPeriod_WeatherFileConditionType",
    "RunPeriod",
    "RunPeriodControl_SpecialDays",
    "RunPeriodControl_DaylightSavingTime",
    "WeatherProperty_SkyTemperature",
    "Site_WeatherStation",
    "Site_HeightVariation",
    "Site_GroundTemperature_BuildingSurface",
    "Site_GroundTemperature_FCfactorMethod",
    "Site_GroundTemperature_Shallow",
    "Site_GroundTemperature_Deep",
    "Site_GroundTemperature_Undisturbed_FiniteDifference",
    "Site_GroundTemperature_Undisturbed_KusudaAchenbach",
    "Site_GroundTemperature_Undisturbed_Xing",
    "Site_GroundDomain_Slab",
    "Site_GroundDomain_Basement",
    "Site_GroundReflectance",
    "Site_GroundReflectance_SnowModifier",
    "Site_WaterMainsTemperature",
    "Site_Precipitation",
    "RoofIrrigation",
    "Site_SolarAndVisibleSpectrum",
    "Site_SpectrumData",
    "ScheduleTypeLimits",
    "Schedule_Day_Hourly",
    "Schedule_Day_Interval",
    "Schedule_Day_List",
    "Schedule_Week_Daily",
    "Schedule_Week_Compact",
    "Schedule_Year",
    "Schedule_Compact",
    "Schedule_Constant",
    "Schedule_File_Shading",
    "Schedule_File",
    "Material",
    "Material_NoMass",
    "Material_InfraredTransparent",
    "Material_AirGap",
    "Material_RoofVegetation",
    "WindowMaterial_SimpleGlazingSystem",
    "WindowMaterial_Glazing",
    "WindowMaterial_GlazingGroup_Thermochromic",
    "WindowMaterial_Glazing_RefractionExtinctionMethod",
    "WindowMaterial_Gas",
    "WindowGap_SupportPillar",
    "WindowGap_DeflectionState",
    "WindowMaterial_GasMixture",
    "WindowMaterial_Gap",
    "WindowMaterial_Shade",
    "WindowMaterial_ComplexShade",
    "WindowMaterial_Blind",
    "WindowMaterial_Screen",
    "WindowMaterial_Shade_EquivalentLayer",
    "WindowMaterial_Drape_EquivalentLayer",
    "WindowMaterial_Blind_EquivalentLayer",
    "WindowMaterial_Screen_EquivalentLayer",
    "WindowMaterial_Glazing_EquivalentLayer",
    "WindowMaterial_Gap_EquivalentLayer",
    "MaterialProperty_MoisturePenetrationDepth_Settings",
    "MaterialProperty_PhaseChange",
    "MaterialProperty_PhaseChangeHysteresis",
    "MaterialProperty_VariableThermalConductivity",
    "MaterialProperty_VariableAbsorptance",
    "MaterialProperty_HeatAndMoistureTransfer_Settings",
    "MaterialProperty_HeatAndMoistureTransfer_SorptionIsotherm",
    "MaterialProperty_HeatAndMoistureTransfer_Suction",
    "MaterialProperty_HeatAndMoistureTransfer_Redistribution",
    "MaterialProperty_HeatAndMoistureTransfer_Diffusion",
    "MaterialProperty_HeatAndMoistureTransfer_ThermalConductivity",
    "MaterialProperty_GlazingSpectralData",
    "Construction",
    "Construction_CfactorUndergroundWall",
    "Construction_FfactorGroundFloor",
    "ConstructionProperty_InternalHeatSource",
    "Construction_AirBoundary",
    "WindowThermalModel_Params",
    "WindowsCalculationEngine",
    "Construction_ComplexFenestrationState",
    "Construction_WindowEquivalentLayer",
    "Construction_WindowDataFile",
    "GlobalGeometryRules",
    "GeometryTransform",
    "Space",
    "SpaceList",
    "Zone",
    "ZoneList",
    "ZoneGroup",
    "BuildingSurface_Detailed",
    "Wall_Detailed",
    "RoofCeiling_Detailed",
    "Floor_Detailed",
    "Wall_Exterior",
    "Wall_Adiabatic",
    "Wall_Underground",
    "Wall_Interzone",
    "Roof",
    "Ceiling_Adiabatic",
    "Ceiling_Interzone",
    "Floor_GroundContact",
    "Floor_Adiabatic",
    "Floor_Interzone",
    "FenestrationSurface_Detailed",
    "Window",
    "Door",
    "GlazedDoor",
    "Window_Interzone",
    "Door_Interzone",
    "GlazedDoor_Interzone",
    "WindowShadingControl",
    "WindowProperty_FrameAndDivider",
    "WindowProperty_AirflowControl",
    "WindowProperty_StormWindow",
    "InternalMass",
    "Shading_Site",
    "Shading_Building",
    "Shading_Site_Detailed",
    "Shading_Building_Detailed",
    "Shading_Overhang",
    "Shading_Overhang_Projection",
    "Shading_Fin",
    "Shading_Fin_Projection",
    "Shading_Zone_Detailed",
    "ShadingProperty_Reflectance",
    "SurfaceProperty_HeatTransferAlgorithm",
    "SurfaceProperty_HeatTransferAlgorithm_MultipleSurface",
    "SurfaceProperty_HeatTransferAlgorithm_SurfaceList",
    "SurfaceProperty_HeatTransferAlgorithm_Construction",
    "SurfaceProperty_HeatBalanceSourceTerm",
    "SurfaceControl_MovableInsulation",
    "SurfaceProperty_OtherSideCoefficients",
    "SurfaceProperty_OtherSideConditionsModel",
    "SurfaceProperty_Underwater",
    "Foundation_Kiva",
    "Foundation_Kiva_Settings",
    "SurfaceProperty_ExposedFoundationPerimeter",
    "SurfaceConvectionAlgorithm_Inside_AdaptiveModelSelections",
    "SurfaceConvectionAlgorithm_Outside_AdaptiveModelSelections",
    "SurfaceConvectionAlgorithm_Inside_UserCurve",
    "SurfaceConvectionAlgorithm_Outside_UserCurve",
    "SurfaceProperty_ConvectionCoefficients",
    "SurfaceProperty_ConvectionCoefficients_MultipleSurface",
    "SurfaceProperties_VaporCoefficients",
    "SurfaceProperty_ExteriorNaturalVentedCavity",
    "SurfaceProperty_SolarIncidentInside",
    "SurfaceProperty_IncidentSolarMultiplier",
    "SurfaceProperty_LocalEnvironment",
    "ZoneProperty_LocalEnvironment",
    "SurfaceProperty_SurroundingSurfaces",
    "SurfaceProperty_GroundSurfaces",
    "ComplexFenestrationProperty_SolarAbsorbedLayers",
    "ZoneProperty_UserViewFactors_BySurfaceName",
    "GroundHeatTransfer_Control",
    "GroundHeatTransfer_Slab_Materials",
    "GroundHeatTransfer_Slab_MatlProps",
    "GroundHeatTransfer_Slab_BoundConds",
    "GroundHeatTransfer_Slab_BldgProps",
    "GroundHeatTransfer_Slab_Insulation",
    "GroundHeatTransfer_Slab_EquivalentSlab",
    "GroundHeatTransfer_Slab_AutoGrid",
    "GroundHeatTransfer_Slab_ManualGrid",
    "GroundHeatTransfer_Slab_XFACE",
    "GroundHeatTransfer_Slab_YFACE",
    "GroundHeatTransfer_Slab_ZFACE",
    "GroundHeatTransfer_Basement_SimParameters",
    "GroundHeatTransfer_Basement_MatlProps",
    "GroundHeatTransfer_Basement_Insulation",
    "GroundHeatTransfer_Basement_SurfaceProps",
    "GroundHeatTransfer_Basement_BldgData",
    "GroundHeatTransfer_Basement_Interior",
    "GroundHeatTransfer_Basement_ComBldg",
    "GroundHeatTransfer_Basement_EquivSlab",
    "GroundHeatTransfer_Basement_EquivAutoGrid",
    "GroundHeatTransfer_Basement_AutoGrid",
    "GroundHeatTransfer_Basement_ManualGrid",
    "GroundHeatTransfer_Basement_XFACE",
    "GroundHeatTransfer_Basement_YFACE",
    "GroundHeatTransfer_Basement_ZFACE",
    "RoomAirModelType",
    "RoomAir_TemperaturePattern_UserDefined",
    "RoomAir_TemperaturePattern_ConstantGradient",
    "RoomAir_TemperaturePattern_TwoGradient",
    "RoomAir_TemperaturePattern_NondimensionalHeight",
    "RoomAir_TemperaturePattern_SurfaceMapping",
    "RoomAir_Node",
    "RoomAirSettings_OneNodeDisplacementVentilation",
    "RoomAirSettings_ThreeNodeDisplacementVentilation",
    "RoomAirSettings_CrossVentilation",
    "RoomAirSettings_UnderFloorAirDistributionInterior",
    "RoomAirSettings_UnderFloorAirDistributionExterior",
    "RoomAir_Node_AirflowNetwork",
    "RoomAir_Node_AirflowNetwork_AdjacentSurfaceList",
    "RoomAir_Node_AirflowNetwork_InternalGains",
    "RoomAir_Node_AirflowNetwork_HVACEquipment",
    "RoomAirSettings_AirflowNetwork",
    "People",
    "ComfortViewFactorAngles",
    "Lights",
    "ElectricEquipment",
    "GasEquipment",
    "HotWaterEquipment",
    "SteamEquipment",
    "OtherEquipment",
    "ElectricEquipment_ITE_AirCooled",
    "ZoneBaseboard_OutdoorTemperatureControlled",
    "SwimmingPool_Indoor",
    "ZoneContaminantSourceAndSink_CarbonDioxide",
    "ZoneContaminantSourceAndSink_Generic_Constant",
    "SurfaceContaminantSourceAndSink_Generic_PressureDriven",
    "ZoneContaminantSourceAndSink_Generic_CutoffModel",
    "ZoneContaminantSourceAndSink_Generic_DecaySource",
    "SurfaceContaminantSourceAndSink_Generic_BoundaryLayerDiffusion",
    "SurfaceContaminantSourceAndSink_Generic_DepositionVelocitySink",
    "ZoneContaminantSourceAndSink_Generic_DepositionRateSink",
    "Daylighting_Controls",
    "Daylighting_ReferencePoint",
    "Daylighting_DELight_ComplexFenestration",
    "DaylightingDevice_Tubular",
    "DaylightingDevice_Shelf",
    "DaylightingDevice_LightWell",
    "Output_DaylightFactors",
    "Output_IlluminanceMap",
    "OutputControl_IlluminanceMap_Style",
    "ZoneInfiltration_DesignFlowRate",
    "ZoneInfiltration_EffectiveLeakageArea",
    "ZoneInfiltration_FlowCoefficient",
    "ZoneVentilation_DesignFlowRate",
    "ZoneVentilation_WindandStackOpenArea",
    "ZoneAirBalance_OutdoorAir",
    "ZoneMixing",
    "ZoneCrossMixing",
    "ZoneRefrigerationDoorMixing",
    "ZoneEarthtube",
    "ZoneCoolTower_Shower",
    "ZoneThermalChimney",
    "AirflowNetwork_SimulationControl",
    "AirflowNetwork_MultiZone_Zone",
    "AirflowNetwork_MultiZone_Surface",
    "AirflowNetwork_MultiZone_ReferenceCrackConditions",
    "AirflowNetwork_MultiZone_Surface_Crack",
    "AirflowNetwork_MultiZone_Surface_EffectiveLeakageArea",
    "AirflowNetwork_MultiZone_SpecifiedFlowRate",
    "AirflowNetwork_MultiZone_Component_DetailedOpening",
    "AirflowNetwork_MultiZone_Component_SimpleOpening",
    "AirflowNetwork_MultiZone_Component_HorizontalOpening",
    "AirflowNetwork_MultiZone_Component_ZoneExhaustFan",
    "AirflowNetwork_MultiZone_ExternalNode",
    "AirflowNetwork_MultiZone_WindPressureCoefficientArray",
    "AirflowNetwork_MultiZone_WindPressureCoefficientValues",
    "AirflowNetwork_ZoneControl_PressureController",
    "AirflowNetwork_Distribution_Node",
    "AirflowNetwork_Distribution_Component_Leak",
    "AirflowNetwork_Distribution_Component_LeakageRatio",
    "AirflowNetwork_Distribution_Component_Duct",
    "AirflowNetwork_Distribution_Component_Fan",
    "AirflowNetwork_Distribution_Component_Coil",
    "AirflowNetwork_Distribution_Component_HeatExchanger",
    "AirflowNetwork_Distribution_Component_TerminalUnit",
    "AirflowNetwork_Distribution_Component_ConstantPressureDrop",
    "AirflowNetwork_Distribution_Component_OutdoorAirFlow",
    "AirflowNetwork_Distribution_Component_ReliefAirFlow",
    "AirflowNetwork_Distribution_Linkage",
    "AirflowNetwork_Distribution_DuctViewFactors",
    "AirflowNetwork_OccupantVentilationControl",
    "AirflowNetwork_IntraZone_Node",
    "AirflowNetwork_IntraZone_Linkage",
    "Exterior_Lights",
    "Exterior_FuelEquipment",
    "Exterior_WaterEquipment",
    "HVACTemplate_Thermostat",
    "HVACTemplate_Zone_IdealLoadsAirSystem",
    "HVACTemplate_Zone_BaseboardHeat",
    "HVACTemplate_Zone_FanCoil",
    "HVACTemplate_Zone_PTAC",
    "HVACTemplate_Zone_PTHP",
    "HVACTemplate_Zone_WaterToAirHeatPump",
    "HVACTemplate_Zone_VRF",
    "HVACTemplate_Zone_Unitary",
    "HVACTemplate_Zone_VAV",
    "HVACTemplate_Zone_VAV_FanPowered",
    "HVACTemplate_Zone_VAV_HeatAndCool",
    "HVACTemplate_Zone_ConstantVolume",
    "HVACTemplate_Zone_DualDuct",
    "HVACTemplate_System_VRF",
    "HVACTemplate_System_Unitary",
    "HVACTemplate_System_UnitaryHeatPump_AirToAir",
    "HVACTemplate_System_UnitarySystem",
    "HVACTemplate_System_VAV",
    "HVACTemplate_System_PackagedVAV",
    "HVACTemplate_System_ConstantVolume",
    "HVACTemplate_System_DualDuct",
    "HVACTemplate_System_DedicatedOutdoorAir",
    "HVACTemplate_Plant_ChilledWaterLoop",
    "HVACTemplate_Plant_Chiller",
    "HVACTemplate_Plant_Chiller_ObjectReference",
    "HVACTemplate_Plant_Tower",
    "HVACTemplate_Plant_Tower_ObjectReference",
    "HVACTemplate_Plant_HotWaterLoop",
    "HVACTemplate_Plant_Boiler",
    "HVACTemplate_Plant_Boiler_ObjectReference",
    "HVACTemplate_Plant_MixedWaterLoop",
    "DesignSpecification_OutdoorAir",
    "DesignSpecification_OutdoorAir_SpaceList",
    "DesignSpecification_ZoneAirDistribution",
    "Sizing_Parameters",
    "Sizing_Zone",
    "DesignSpecification_ZoneHVAC_Sizing",
    "DesignSpecification_AirTerminal_Sizing",
    "Sizing_System",
    "Sizing_Plant",
    "OutputControl_Sizing_Style",
    "ZoneControl_Humidistat",
    "ZoneControl_Thermostat",
    "ZoneControl_Thermostat_OperativeTemperature",
    "ZoneControl_Thermostat_ThermalComfort",
    "ZoneControl_Thermostat_TemperatureAndHumidity",
    "ThermostatSetpoint_SingleHeating",
    "ThermostatSetpoint_SingleCooling",
    "ThermostatSetpoint_SingleHeatingOrCooling",
    "ThermostatSetpoint_DualSetpoint",
    "ThermostatSetpoint_ThermalComfort_Fanger_SingleHeating",
    "ThermostatSetpoint_ThermalComfort_Fanger_SingleCooling",
    "ThermostatSetpoint_ThermalComfort_Fanger_SingleHeatingOrCooling",
    "ThermostatSetpoint_ThermalComfort_Fanger_DualSetpoint",
    "ZoneControl_Thermostat_StagedDualSetpoint",
    "ZoneControl_ContaminantController",
    "ZoneHVAC_IdealLoadsAirSystem",
    "ZoneHVAC_FourPipeFanCoil",
    "ZoneHVAC_WindowAirConditioner",
    "ZoneHVAC_PackagedTerminalAirConditioner",
    "ZoneHVAC_PackagedTerminalHeatPump",
    "ZoneHVAC_WaterToAirHeatPump",
    "ZoneHVAC_Dehumidifier_DX",
    "ZoneHVAC_EnergyRecoveryVentilator",
    "ZoneHVAC_EnergyRecoveryVentilator_Controller",
    "ZoneHVAC_UnitVentilator",
    "ZoneHVAC_UnitHeater",
    "ZoneHVAC_EvaporativeCoolerUnit",
    "ZoneHVAC_HybridUnitaryHVAC",
    "ZoneHVAC_OutdoorAirUnit",
    "ZoneHVAC_OutdoorAirUnit_EquipmentList",
    "ZoneHVAC_TerminalUnit_VariableRefrigerantFlow",
    "ZoneHVAC_Baseboard_RadiantConvective_Water_Design",
    "ZoneHVAC_Baseboard_RadiantConvective_Water",
    "ZoneHVAC_Baseboard_RadiantConvective_Steam_Design",
    "ZoneHVAC_Baseboard_RadiantConvective_Steam",
    "ZoneHVAC_Baseboard_RadiantConvective_Electric",
    "ZoneHVAC_CoolingPanel_RadiantConvective_Water",
    "ZoneHVAC_Baseboard_Convective_Water",
    "ZoneHVAC_Baseboard_Convective_Electric",
    "ZoneHVAC_LowTemperatureRadiant_VariableFlow",
    "ZoneHVAC_LowTemperatureRadiant_VariableFlow_Design",
    "ZoneHVAC_LowTemperatureRadiant_ConstantFlow",
    "ZoneHVAC_LowTemperatureRadiant_ConstantFlow_Design",
    "ZoneHVAC_LowTemperatureRadiant_Electric",
    "ZoneHVAC_LowTemperatureRadiant_SurfaceGroup",
    "ZoneHVAC_HighTemperatureRadiant",
    "ZoneHVAC_VentilatedSlab",
    "ZoneHVAC_VentilatedSlab_SlabGroup",
    "AirTerminal_SingleDuct_ConstantVolume_Reheat",
    "AirTerminal_SingleDuct_ConstantVolume_NoReheat",
    "AirTerminal_SingleDuct_VAV_NoReheat",
    "AirTerminal_SingleDuct_VAV_Reheat",
    "AirTerminal_SingleDuct_VAV_Reheat_VariableSpeedFan",
    "AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat",
    "AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat",
    "AirTerminal_SingleDuct_SeriesPIU_Reheat",
    "AirTerminal_SingleDuct_ParallelPIU_Reheat",
    "AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction",
    "AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam",
    "AirTerminal_SingleDuct_ConstantVolume_CooledBeam",
    "AirTerminal_SingleDuct_Mixer",
    "AirTerminal_DualDuct_ConstantVolume",
    "AirTerminal_DualDuct_VAV",
    "AirTerminal_DualDuct_VAV_OutdoorAir",
    "ZoneHVAC_AirDistributionUnit",
    "ZoneHVAC_ExhaustControl",
    "ZoneHVAC_EquipmentList",
    "ZoneHVAC_EquipmentConnections",
    "Fan_SystemModel",
    "Fan_ConstantVolume",
    "Fan_VariableVolume",
    "Fan_OnOff",
    "Fan_ZoneExhaust",
    "FanPerformance_NightVentilation",
    "Fan_ComponentModel",
    "Coil_Cooling_Water",
    "Coil_Cooling_Water_DetailedGeometry",
    "CoilSystem_Cooling_Water",
    "Coil_Cooling_DX",
    "Coil_Cooling_DX_CurveFit_Performance",
    "Coil_Cooling_DX_CurveFit_OperatingMode",
    "Coil_Cooling_DX_CurveFit_Speed",
    "Coil_Cooling_DX_SingleSpeed",
    "Coil_Cooling_DX_TwoSpeed",
    "Coil_Cooling_DX_MultiSpeed",
    "Coil_Cooling_DX_VariableSpeed",
    "Coil_Cooling_DX_TwoStageWithHumidityControlMode",
    "CoilPerformance_DX_Cooling",
    "Coil_Cooling_DX_VariableRefrigerantFlow",
    "Coil_Heating_DX_VariableRefrigerantFlow",
    "Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl",
    "Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl",
    "Coil_Heating_Water",
    "Coil_Heating_Steam",
    "Coil_Heating_Electric",
    "Coil_Heating_Electric_MultiStage",
    "Coil_Heating_Fuel",
    "Coil_Heating_Gas_MultiStage",
    "Coil_Heating_Desuperheater",
    "Coil_Heating_DX_SingleSpeed",
    "Coil_Heating_DX_MultiSpeed",
    "Coil_Heating_DX_VariableSpeed",
    "Coil_Cooling_WaterToAirHeatPump_ParameterEstimation",
    "Coil_Heating_WaterToAirHeatPump_ParameterEstimation",
    "Coil_Cooling_WaterToAirHeatPump_EquationFit",
    "Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit",
    "Coil_Heating_WaterToAirHeatPump_EquationFit",
    "Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit",
    "Coil_WaterHeating_AirToWaterHeatPump_Pumped",
    "Coil_WaterHeating_AirToWaterHeatPump_Wrapped",
    "Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed",
    "Coil_WaterHeating_Desuperheater",
    "CoilSystem_Cooling_DX",
    "CoilSystem_Heating_DX",
    "CoilSystem_Cooling_Water_HeatExchangerAssisted",
    "CoilSystem_Cooling_DX_HeatExchangerAssisted",
    "CoilSystem_IntegratedHeatPump_AirSource",
    "Coil_Cooling_DX_SingleSpeed_ThermalStorage",
    "EvaporativeCooler_Direct_CelDekPad",
    "EvaporativeCooler_Indirect_CelDekPad",
    "EvaporativeCooler_Indirect_WetCoil",
    "EvaporativeCooler_Indirect_ResearchSpecial",
    "EvaporativeCooler_Direct_ResearchSpecial",
    "Humidifier_Steam_Electric",
    "Humidifier_Steam_Gas",
    "Dehumidifier_Desiccant_NoFans",
    "Dehumidifier_Desiccant_System",
    "HeatExchanger_AirToAir_FlatPlate",
    "HeatExchanger_AirToAir_SensibleAndLatent",
    "HeatExchanger_Desiccant_BalancedFlow",
    "HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1",
    "AirLoopHVAC_UnitarySystem",
    "UnitarySystemPerformance_Multispeed",
    "AirLoopHVAC_Unitary_Furnace_HeatOnly",
    "AirLoopHVAC_Unitary_Furnace_HeatCool",
    "AirLoopHVAC_UnitaryHeatOnly",
    "AirLoopHVAC_UnitaryHeatCool",
    "AirLoopHVAC_UnitaryHeatPump_AirToAir",
    "AirLoopHVAC_UnitaryHeatPump_WaterToAir",
    "AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass",
    "AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed",
    "AirConditioner_VariableRefrigerantFlow",
    "AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl",
    "AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR",
    "ZoneTerminalUnitList",
    "Controller_WaterCoil",
    "Controller_OutdoorAir",
    "Controller_MechanicalVentilation",
    "AirLoopHVAC_ControllerList",
    "AirLoopHVAC",
    "AirLoopHVAC_OutdoorAirSystem_EquipmentList",
    "AirLoopHVAC_OutdoorAirSystem",
    "OutdoorAir_Mixer",
    "AirLoopHVAC_ZoneSplitter",
    "AirLoopHVAC_SupplyPlenum",
    "AirLoopHVAC_SupplyPath",
    "AirLoopHVAC_ZoneMixer",
    "AirLoopHVAC_ReturnPlenum",
    "AirLoopHVAC_ReturnPath",
    "AirLoopHVAC_ExhaustSystem",
    "AirLoopHVAC_DedicatedOutdoorAirSystem",
    "AirLoopHVAC_Mixer",
    "AirLoopHVAC_Splitter",
    "Branch",
    "BranchList",
    "Connector_Splitter",
    "Connector_Mixer",
    "ConnectorList",
    "NodeList",
    "OutdoorAir_Node",
    "OutdoorAir_NodeList",
    "Pipe_Adiabatic",
    "Pipe_Adiabatic_Steam",
    "Pipe_Indoor",
    "Pipe_Outdoor",
    "Pipe_Underground",
    "PipingSystem_Underground_Domain",
    "PipingSystem_Underground_PipeCircuit",
    "PipingSystem_Underground_PipeSegment",
    "Duct",
    "Pump_VariableSpeed",
    "Pump_ConstantSpeed",
    "Pump_VariableSpeed_Condensate",
    "HeaderedPumps_ConstantSpeed",
    "HeaderedPumps_VariableSpeed",
    "TemperingValve",
    "LoadProfile_Plant",
    "SolarCollectorPerformance_FlatPlate",
    "SolarCollector_FlatPlate_Water",
    "SolarCollector_FlatPlate_PhotovoltaicThermal",
    "SolarCollectorPerformance_PhotovoltaicThermal_Simple",
    "SolarCollectorPerformance_PhotovoltaicThermal_BIPVT",
    "SolarCollector_IntegralCollectorStorage",
    "SolarCollectorPerformance_IntegralCollectorStorage",
    "SolarCollector_UnglazedTranspired",
    "SolarCollector_UnglazedTranspired_Multisystem",
    "Boiler_HotWater",
    "Boiler_Steam",
    "Chiller_Electric_ASHRAE205",
    "Chiller_Electric_EIR",
    "Chiller_Electric_ReformulatedEIR",
    "Chiller_Electric",
    "Chiller_Absorption_Indirect",
    "Chiller_Absorption",
    "Chiller_ConstantCOP",
    "Chiller_EngineDriven",
    "Chiller_CombustionTurbine",
    "ChillerHeater_Absorption_DirectFired",
    "ChillerHeater_Absorption_DoubleEffect",
    "HeatPump_PlantLoop_EIR_Cooling",
    "HeatPump_PlantLoop_EIR_Heating",
    "HeatPump_AirToWater_FuelFired_Heating",
    "HeatPump_AirToWater_FuelFired_Cooling",
    "HeatPump_WaterToWater_EquationFit_Heating",
    "HeatPump_WaterToWater_EquationFit_Cooling",
    "HeatPump_WaterToWater_ParameterEstimation_Cooling",
    "HeatPump_WaterToWater_ParameterEstimation_Heating",
    "DistrictCooling",
    "DistrictHeating",
    "PlantComponent_TemperatureSource",
    "CentralHeatPumpSystem",
    "ChillerHeaterPerformance_Electric_EIR",
    "CoolingTower_SingleSpeed",
    "CoolingTower_TwoSpeed",
    "CoolingTower_VariableSpeed_Merkel",
    "CoolingTower_VariableSpeed",
    "CoolingTowerPerformance_CoolTools",
    "CoolingTowerPerformance_YorkCalc",
    "EvaporativeFluidCooler_SingleSpeed",
    "EvaporativeFluidCooler_TwoSpeed",
    "FluidCooler_SingleSpeed",
    "FluidCooler_TwoSpeed",
    "GroundHeatExchanger_System",
    "GroundHeatExchanger_Vertical_Properties",
    "GroundHeatExchanger_Vertical_Array",
    "GroundHeatExchanger_Vertical_Single",
    "GroundHeatExchanger_ResponseFactors",
    "GroundHeatExchanger_Pond",
    "GroundHeatExchanger_Surface",
    "GroundHeatExchanger_HorizontalTrench",
    "GroundHeatExchanger_Slinky",
    "HeatExchanger_FluidToFluid",
    "WaterHeater_Mixed",
    "WaterHeater_Stratified",
    "WaterHeater_Sizing",
    "WaterHeater_HeatPump_PumpedCondenser",
    "WaterHeater_HeatPump_WrappedCondenser",
    "ThermalStorage_Ice_Simple",
    "ThermalStorage_Ice_Detailed",
    "ThermalStorage_ChilledWater_Mixed",
    "ThermalStorage_ChilledWater_Stratified",
    "PlantLoop",
    "CondenserLoop",
    "PlantEquipmentList",
    "CondenserEquipmentList",
    "PlantEquipmentOperation_Uncontrolled",
    "PlantEquipmentOperation_CoolingLoad",
    "PlantEquipmentOperation_HeatingLoad",
    "PlantEquipmentOperation_OutdoorDryBulb",
    "PlantEquipmentOperation_OutdoorWetBulb",
    "PlantEquipmentOperation_OutdoorRelativeHumidity",
    "PlantEquipmentOperation_OutdoorDewpoint",
    "PlantEquipmentOperation_ComponentSetpoint",
    "PlantEquipmentOperation_ThermalEnergyStorage",
    "PlantEquipmentOperation_OutdoorDryBulbDifference",
    "PlantEquipmentOperation_OutdoorWetBulbDifference",
    "PlantEquipmentOperation_OutdoorDewpointDifference",
    "PlantEquipmentOperationSchemes",
    "CondenserEquipmentOperationSchemes",
    "EnergyManagementSystem_Sensor",
    "EnergyManagementSystem_Actuator",
    "EnergyManagementSystem_ProgramCallingManager",
    "EnergyManagementSystem_Program",
    "EnergyManagementSystem_Subroutine",
    "EnergyManagementSystem_GlobalVariable",
    "EnergyManagementSystem_OutputVariable",
    "EnergyManagementSystem_MeteredOutputVariable",
    "EnergyManagementSystem_TrendVariable",
    "EnergyManagementSystem_InternalVariable",
    "EnergyManagementSystem_CurveOrTableIndexVariable",
    "EnergyManagementSystem_ConstructionIndexVariable",
    "ExternalInterface",
    "ExternalInterface_Schedule",
    "ExternalInterface_Variable",
    "ExternalInterface_Actuator",
    "ExternalInterface_FunctionalMockupUnitImport",
    "ExternalInterface_FunctionalMockupUnitImport_From_Variable",
    "ExternalInterface_FunctionalMockupUnitImport_To_Schedule",
    "ExternalInterface_FunctionalMockupUnitImport_To_Actuator",
    "ExternalInterface_FunctionalMockupUnitImport_To_Variable",
    "ExternalInterface_FunctionalMockupUnitExport_From_Variable",
    "ExternalInterface_FunctionalMockupUnitExport_To_Schedule",
    "ExternalInterface_FunctionalMockupUnitExport_To_Actuator",
    "ExternalInterface_FunctionalMockupUnitExport_To_Variable",
    "ZoneHVAC_ForcedAir_UserDefined",
    "AirTerminal_SingleDuct_UserDefined",
    "Coil_UserDefined",
    "PlantComponent_UserDefined",
    "PlantEquipmentOperation_UserDefined",
    "AvailabilityManager_Scheduled",
    "AvailabilityManager_ScheduledOn",
    "AvailabilityManager_ScheduledOff",
    "AvailabilityManager_OptimumStart",
    "AvailabilityManager_NightCycle",
    "AvailabilityManager_DifferentialThermostat",
    "AvailabilityManager_HighTemperatureTurnOff",
    "AvailabilityManager_HighTemperatureTurnOn",
    "AvailabilityManager_LowTemperatureTurnOff",
    "AvailabilityManager_LowTemperatureTurnOn",
    "AvailabilityManager_NightVentilation",
    "AvailabilityManager_HybridVentilation",
    "AvailabilityManagerAssignmentList",
    "SetpointManager_Scheduled",
    "SetpointManager_Scheduled_DualSetpoint",
    "SetpointManager_OutdoorAirReset",
    "SetpointManager_SingleZone_Reheat",
    "SetpointManager_SingleZone_Heating",
    "SetpointManager_SingleZone_Cooling",
    "SetpointManager_SingleZone_Humidity_Minimum",
    "SetpointManager_SingleZone_Humidity_Maximum",
    "SetpointManager_MixedAir",
    "SetpointManager_OutdoorAirPretreat",
    "SetpointManager_Warmest",
    "SetpointManager_Coldest",
    "SetpointManager_ReturnAirBypassFlow",
    "SetpointManager_WarmestTemperatureFlow",
    "SetpointManager_MultiZone_Heating_Average",
    "SetpointManager_MultiZone_Cooling_Average",
    "SetpointManager_MultiZone_MinimumHumidity_Average",
    "SetpointManager_MultiZone_MaximumHumidity_Average",
    "SetpointManager_MultiZone_Humidity_Minimum",
    "SetpointManager_MultiZone_Humidity_Maximum",
    "SetpointManager_FollowOutdoorAirTemperature",
    "SetpointManager_FollowSystemNodeTemperature",
    "SetpointManager_FollowGroundTemperature",
    "SetpointManager_CondenserEnteringReset",
    "SetpointManager_CondenserEnteringReset_Ideal",
    "SetpointManager_SingleZone_OneStageCooling",
    "SetpointManager_SingleZone_OneStageHeating",
    "SetpointManager_ReturnTemperature_ChilledWater",
    "SetpointManager_ReturnTemperature_HotWater",
    "SetpointManager_SystemNodeReset_Temperature",
    "SetpointManager_SystemNodeReset_Humidity",
    "Refrigeration_Case",
    "Refrigeration_CompressorRack",
    "Refrigeration_CaseAndWalkInList",
    "Refrigeration_Condenser_AirCooled",
    "Refrigeration_Condenser_EvaporativeCooled",
    "Refrigeration_Condenser_WaterCooled",
    "Refrigeration_Condenser_Cascade",
    "Refrigeration_GasCooler_AirCooled",
    "Refrigeration_TransferLoadList",
    "Refrigeration_Subcooler",
    "Refrigeration_Compressor",
    "Refrigeration_CompressorList",
    "Refrigeration_System",
    "Refrigeration_TranscriticalSystem",
    "Refrigeration_SecondarySystem",
    "Refrigeration_WalkIn",
    "Refrigeration_AirChiller",
    "ZoneHVAC_RefrigerationChillerSet",
    "DemandManagerAssignmentList",
    "DemandManager_ExteriorLights",
    "DemandManager_Lights",
    "DemandManager_ElectricEquipment",
    "DemandManager_Thermostats",
    "DemandManager_Ventilation",
    "Generator_InternalCombustionEngine",
    "Generator_CombustionTurbine",
    "Generator_MicroTurbine",
    "Generator_Photovoltaic",
    "PhotovoltaicPerformance_Simple",
    "PhotovoltaicPerformance_EquivalentOneDiode",
    "PhotovoltaicPerformance_Sandia",
    "Generator_PVWatts",
    "ElectricLoadCenter_Inverter_PVWatts",
    "Generator_FuelCell",
    "Generator_FuelCell_PowerModule",
    "Generator_FuelCell_AirSupply",
    "Generator_FuelCell_WaterSupply",
    "Generator_FuelCell_AuxiliaryHeater",
    "Generator_FuelCell_ExhaustGasToWaterHeatExchanger",
    "Generator_FuelCell_ElectricalStorage",
    "Generator_FuelCell_Inverter",
    "Generator_FuelCell_StackCooler",
    "Generator_MicroCHP",
    "Generator_MicroCHP_NonNormalizedParameters",
    "Generator_FuelSupply",
    "Generator_WindTurbine",
    "ElectricLoadCenter_Generators",
    "ElectricLoadCenter_Inverter_Simple",
    "ElectricLoadCenter_Inverter_FunctionOfPower",
    "ElectricLoadCenter_Inverter_LookUpTable",
    "ElectricLoadCenter_Storage_Simple",
    "ElectricLoadCenter_Storage_Battery",
    "ElectricLoadCenter_Storage_LiIonNMCBattery",
    "ElectricLoadCenter_Transformer",
    "ElectricLoadCenter_Distribution",
    "ElectricLoadCenter_Storage_Converter",
    "WaterUse_Equipment",
    "WaterUse_Connections",
    "WaterUse_Storage",
    "WaterUse_Well",
    "WaterUse_RainCollector",
    "FaultModel_TemperatureSensorOffset_OutdoorAir",
    "FaultModel_HumiditySensorOffset_OutdoorAir",
    "FaultModel_EnthalpySensorOffset_OutdoorAir",
    "FaultModel_TemperatureSensorOffset_ReturnAir",
    "FaultModel_EnthalpySensorOffset_ReturnAir",
    "FaultModel_TemperatureSensorOffset_ChillerSupplyWater",
    "FaultModel_TemperatureSensorOffset_CoilSupplyAir",
    "FaultModel_TemperatureSensorOffset_CondenserSupplyWater",
    "FaultModel_ThermostatOffset",
    "FaultModel_HumidistatOffset",
    "FaultModel_Fouling_AirFilter",
    "FaultModel_Fouling_Boiler",
    "FaultModel_Fouling_EvaporativeCooler",
    "FaultModel_Fouling_Chiller",
    "FaultModel_Fouling_CoolingTower",
    "FaultModel_Fouling_Coil",
    "Matrix_TwoDimension",
    "HybridModel_Zone",
    "Curve_Linear",
    "Curve_QuadLinear",
    "Curve_QuintLinear",
    "Curve_Quadratic",
    "Curve_Cubic",
    "Curve_Quartic",
    "Curve_Exponent",
    "Curve_Bicubic",
    "Curve_Biquadratic",
    "Curve_QuadraticLinear",
    "Curve_CubicLinear",
    "Curve_Triquadratic",
    "Curve_Functional_PressureDrop",
    "Curve_FanPressureRise",
    "Curve_ExponentialSkewNormal",
    "Curve_Sigmoid",
    "Curve_RectangularHyperbola1",
    "Curve_RectangularHyperbola2",
    "Curve_ExponentialDecay",
    "Curve_DoubleExponentialDecay",
    "Curve_ChillerPartLoadWithLift",
    "Table_IndependentVariable",
    "Table_IndependentVariableList",
    "Table_Lookup",
    "FluidProperties_Name",
    "FluidProperties_GlycolConcentration",
    "FluidProperties_Temperatures",
    "FluidProperties_Saturated",
    "FluidProperties_Superheated",
    "FluidProperties_Concentration",
    "CurrencyType",
    "ComponentCost_Adjustments",
    "ComponentCost_Reference",
    "ComponentCost_LineItem",
    "UtilityCost_Tariff",
    "UtilityCost_Qualify",
    "UtilityCost_Charge_Simple",
    "UtilityCost_Charge_Block",
    "UtilityCost_Ratchet",
    "UtilityCost_Variable",
    "UtilityCost_Computation",
    "LifeCycleCost_Parameters",
    "LifeCycleCost_RecurringCosts",
    "LifeCycleCost_NonrecurringCost",
    "LifeCycleCost_UsePriceEscalation",
    "LifeCycleCost_UseAdjustment",
    "Parametric_SetValueForRun",
    "Parametric_Logic",
    "Parametric_RunControl",
    "Parametric_FileNameSuffix",
    "Output_VariableDictionary",
    "Output_Surfaces_List",
    "Output_Surfaces_Drawing",
    "Output_Schedules",
    "Output_Constructions",
    "Output_EnergyManagementSystem",
    "OutputControl_SurfaceColorScheme",
    "Output_Table_SummaryReports",
    "Output_Table_TimeBins",
    "Output_Table_Monthly",
    "Output_Table_Annual",
    "OutputControl_Table_Style",
    "OutputControl_ReportingTolerances",
    "Output_Variable",
    "Output_Meter",
    "Output_Meter_MeterFileOnly",
    "Output_Meter_Cumulative",
    "Output_Meter_Cumulative_MeterFileOnly",
    "Meter_Custom",
    "Meter_CustomDecrement",
    "OutputControl_Files",
    "OutputControl_Timestamp",
    "Output_JSON",
    "Output_SQLite",
    "Output_EnvironmentalImpactFactors",
    "EnvironmentalImpactFactors",
    "FuelFactors",
    "Output_Diagnostics",
    "Output_DebuggingData",
    "Output_PreprocessorMessage",
    "PythonPlugin_SearchPaths",
    "PythonPlugin_Instance",
    "PythonPlugin_Variables",
    "PythonPlugin_TrendVariable",
    "PythonPlugin_OutputVariable",
    "OS_Version",
    "OS_ComponentData",
    "OS_ConvergenceLimits",
    "OS_ProgramControl",
    "OS_HeatBalanceAlgorithm",
    "OS_RunPeriod",
    "OS_RunPeriodControl_DaylightSavingTime",
    "OS_RunPeriodControl_SpecialDays",
    "OS_ShadowCalculation",
    "OS_SimulationControl",
    "OS_PerformancePrecisionTradeoffs",
    "OS_Sizing_Parameters",
    "OS_SurfaceConvectionAlgorithm_Inside",
    "OS_SurfaceConvectionAlgorithm_Outside",
    "OS_Timestep",
    "OS_YearDescription",
    "OS_ZoneAirContaminantBalance",
    "OS_ZoneAirHeatBalanceAlgorithm",
    "OS_ZoneAirMassFlowConservation",
    "OS_ZoneCapacitanceMultiplier_ResearchSpecial",
    "OS_RadianceParameters",
    "OS_ClimateZones",
    "OS_Site",
    "OS_Site_GroundTemperature_BuildingSurface",
    "OS_Site_GroundTemperature_FCfactorMethod",
    "OS_Site_GroundTemperature_Shallow",
    "OS_Site_GroundTemperature_Deep",
    "OS_Site_GroundTemperature_Undisturbed_KusudaAchenbach",
    "OS_Site_GroundTemperature_Undisturbed_Xing",
    "OS_Site_GroundReflectance",
    "OS_Site_WaterMainsTemperature",
    "OS_SizingPeriod_DesignDay",
    "OS_SizingPeriod_WeatherFileConditionType",
    "OS_SizingPeriod_WeatherFileDays",
    "OS_WeatherFile",
    "OS_WeatherProperty_SkyTemperature",
    "OS_AdditionalProperties",
    "OS_BuildingStory",
    "OS_BuildingUnit",
    "OS_DefaultConstructionSet",
    "OS_DefaultScheduleSet",
    "OS_DefaultSubSurfaceConstructions",
    "OS_DefaultSurfaceConstructions",
    "OS_Rendering_Color",
    "OS_SpaceType",
    "OS_ModelObjectList",
    "OS_Material",
    "OS_Material_AirGap",
    "OS_Material_InfraredTransparent",
    "OS_Material_NoMass",
    "OS_Material_RoofVegetation",
    "OS_WindowMaterial_Blind",
    "OS_WindowMaterial_DaylightRedirectionDevice",
    "OS_WindowMaterial_Gas",
    "OS_WindowMaterial_GasMixture",
    "OS_WindowMaterial_Glazing",
    "OS_WindowMaterial_GlazingGroup_Thermochromic",
    "OS_WindowMaterial_Glazing_RefractionExtinctionMethod",
    "OS_WindowMaterial_Screen",
    "OS_WindowMaterial_Shade",
    "OS_WindowMaterial_SimpleGlazingSystem",
    "OS_StandardsInformation_Material",
    "OS_MaterialProperty_GlazingSpectralData",
    "OS_MaterialProperty_MoisturePenetrationDepth_Settings",
    "OS_MaterialProperty_PhaseChange",
    "OS_MaterialProperty_PhaseChangeHysteresis",
    "OS_Construction",
    "OS_Construction_CfactorUndergroundWall",
    "OS_Construction_FfactorGroundFloor",
    "OS_Construction_InternalSource",
    "OS_Construction_AirBoundary",
    "OS_Construction_WindowDataFile",
    "OS_StandardsInformation_Construction",
    "OS_InternalMass_Definition",
    "OS_People_Definition",
    "OS_Lights_Definition",
    "OS_Luminaire_Definition",
    "OS_ElectricEquipment_Definition",
    "OS_ElectricEquipment_ITE_AirCooled_Definition",
    "OS_GasEquipment_Definition",
    "OS_HotWaterEquipment_Definition",
    "OS_SteamEquipment_Definition",
    "OS_OtherEquipment_Definition",
    "OS_Exterior_Lights_Definition",
    "OS_Exterior_FuelEquipment_Definition",
    "OS_Exterior_WaterEquipment_Definition",
    "OS_Schedule_Compact",
    "OS_Schedule_Day",
    "OS_Schedule_Week",
    "OS_Schedule_Year",
    "OS_Schedule_Rule",
    "OS_Schedule_Ruleset",
    "OS_Schedule_Constant",
    "OS_Schedule_FixedInterval",
    "OS_Schedule_VariableInterval",
    "OS_ScheduleTypeLimits",
    "OS_External_File",
    "OS_Schedule_File",
    "OS_Building",
    "OS_DaylightingDevice_Tubular",
    "OS_DaylightingDevice_Shelf",
    "OS_DaylightingDevice_LightWell",
    "OS_Facility",
    "OS_InteriorPartitionSurface",
    "OS_InteriorPartitionSurfaceGroup",
    "OS_ShadingSurface",
    "OS_ShadingSurfaceGroup",
    "OS_Space",
    "OS_Surface",
    "OS_SubSurface",
    "OS_ShadingControl",
    "OS_WindowProperty_FrameAndDivider",
    "OS_SurfaceControl_MovableInsulation",
    "OS_SurfaceProperty_ConvectionCoefficients",
    "OS_SurfaceProperty_ConvectionCoefficients_MultipleSurface",
    "OS_SurfaceProperty_ExposedFoundationPerimeter",
    "OS_SurfaceProperty_OtherSideCoefficients",
    "OS_SurfaceProperty_OtherSideConditionsModel",
    "OS_Foundation_Kiva",
    "OS_Foundation_Kiva_Settings",
    "OS_ZoneProperty_UserViewFactors_BySurfaceName",
    "OS_SurfaceProperty_SurroundingSurfaces",
    "OS_SurfaceProperty_LocalEnvironment",
    "OS_SurfaceProperty_GroundSurfaces",
    "OS_SurfaceProperty_IncidentSolarMultiplier",
    "OS_InternalMass",
    "OS_People",
    "OS_Lights",
    "OS_Luminaire",
    "OS_ElectricEquipment",
    "OS_ElectricEquipment_ITE_AirCooled",
    "OS_GasEquipment",
    "OS_HotWaterEquipment",
    "OS_SteamEquipment",
    "OS_OtherEquipment",
    "OS_SpaceInfiltration_DesignFlowRate",
    "OS_SpaceInfiltration_EffectiveLeakageArea",
    "OS_SpaceInfiltration_FlowCoefficient",
    "OS_Exterior_Lights",
    "OS_Exterior_FuelEquipment",
    "OS_Exterior_WaterEquipment",
    "OS_Daylighting_Control",
    "OS_Glare_Sensor",
    "OS_IlluminanceMap",
    "OS_LightingDesignDay",
    "OS_LightingSimulationControl",
    "OS_LightingSimulationZone",
    "OS_Refrigeration_Case",
    "OS_Refrigeration_Condenser_AirCooled",
    "OS_Refrigeration_Condenser_EvaporativeCooled",
    "OS_Refrigeration_Condenser_WaterCooled",
    "OS_Refrigeration_Condenser_Cascade",
    "OS_Refrigeration_Compressor",
    "OS_Refrigeration_CompressorRack",
    "OS_Refrigeration_DefrostCycleParameters",
    "OS_Refrigeration_GasCooler_AirCooled",
    "OS_Refrigeration_Subcooler_Mechanical",
    "OS_Refrigeration_Subcooler_LiquidSuction",
    "OS_Refrigeration_System",
    "OS_Refrigeration_TranscriticalSystem",
    "OS_Refrigeration_SecondarySystem",
    "OS_Refrigeration_WalkIn",
    "OS_Refrigeration_WalkIn_ZoneBoundary",
    "OS_Refrigeration_AirChiller",
    "OS_AirConditioner_VariableRefrigerantFlow",
    "OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl",
    "OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_LoadingIndex",
    "OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR",
    "OS_AirLoopHVAC",
    "OS_AirLoopHVAC_OutdoorAirSystem",
    "OS_AirLoopHVAC_DedicatedOutdoorAirSystem",
    "OS_AirLoopHVAC_ReturnPlenum",
    "OS_AirLoopHVAC_SupplyPlenum",
    "OS_AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed",
    "OS_AirLoopHVAC_UnitaryCoolOnly",
    "OS_AirLoopHVAC_UnitaryHeatPump_AirToAir",
    "OS_AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass",
    "OS_AirLoopHVAC_UnitarySystem",
    "OS_UnitarySystemPerformance_Multispeed",
    "OS_AirLoopHVAC_ZoneMixer",
    "OS_AirLoopHVAC_ZoneSplitter",
    "OS_AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat",
    "OS_AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat",
    "OS_AirTerminal_SingleDuct_InletSideMixer",
    "OS_AirTerminal_SingleDuct_ConstantVolume_Reheat",
    "OS_AirTerminal_SingleDuct_SeriesPIU_Reheat",
    "OS_AirTerminal_SingleDuct_ParallelPIU_Reheat",
    "OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction",
    "OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam",
    "OS_AirTerminal_SingleDuct_ConstantVolume_CooledBeam",
    "OS_AirTerminal_SingleDuct_ConstantVolume_NoReheat",
    "OS_AirTerminal_SingleDuct_VAV_NoReheat",
    "OS_AirTerminal_SingleDuct_VAV_Reheat",
    "OS_AirTerminal_DualDuct_VAV_OutdoorAir",
    "OS_AirTerminal_DualDuct_ConstantVolume",
    "OS_AirTerminal_DualDuct_VAV",
    "OS_AvailabilityManagerAssignmentList",
    "OS_AvailabilityManager_Scheduled",
    "OS_AvailabilityManager_ScheduledOn",
    "OS_AvailabilityManager_ScheduledOff",
    "OS_AvailabilityManager_NightCycle",
    "OS_AvailabilityManager_OptimumStart",
    "OS_AvailabilityManager_DifferentialThermostat",
    "OS_AvailabilityManager_HighTemperatureTurnOff",
    "OS_AvailabilityManager_HighTemperatureTurnOn",
    "OS_AvailabilityManager_LowTemperatureTurnOff",
    "OS_AvailabilityManager_LowTemperatureTurnOn",
    "OS_AvailabilityManager_NightVentilation",
    "OS_AvailabilityManager_HybridVentilation",
    "OS_Boiler_HotWater",
    "OS_Boiler_Steam",
    "OS_HeatPump_WaterToWater_EquationFit_Heating",
    "OS_HeatPump_WaterToWater_EquationFit_Cooling",
    "OS_HeatPump_PlantLoop_EIR_Heating",
    "OS_HeatPump_PlantLoop_EIR_Cooling",
    "OS_HeatPump_AirToWater_FuelFired_Heating",
    "OS_HeatPump_AirToWater_FuelFired_Cooling",
    "OS_Chiller_Electric_EIR",
    "OS_Chiller_Electric_ReformulatedEIR",
    "OS_Chiller_Electric_ASHRAE205",
    "OS_CentralHeatPumpSystem",
    "OS_CentralHeatPumpSystem_Module",
    "OS_ChillerHeaterPerformance_Electric_EIR",
    "OS_Chiller_Absorption_Indirect",
    "OS_Chiller_Absorption",
    "OS_CoolingTower_SingleSpeed",
    "OS_CoolingTower_TwoSpeed",
    "OS_CoolingTower_VariableSpeed",
    "OS_CoolingTowerPerformance_CoolTools",
    "OS_CoolingTowerPerformance_YorkCalc",
    "OS_DistrictCooling",
    "OS_DistrictHeating",
    "OS_GroundHeatExchanger_Vertical",
    "OS_GroundHeatExchanger_HorizontalTrench",
    "OS_Coil_Cooling_CooledBeam",
    "OS_Coil_Cooling_FourPipeBeam",
    "OS_Coil_Heating_FourPipeBeam",
    "OS_Coil_Cooling_DX",
    "OS_Coil_Cooling_DX_CurveFit_Performance",
    "OS_Coil_Cooling_DX_CurveFit_OperatingMode",
    "OS_Coil_Cooling_DX_CurveFit_Speed",
    "OS_Coil_Cooling_DX_SingleSpeed",
    "OS_Coil_Cooling_DX_TwoStageWithHumidityControlMode",
    "OS_CoilPerformance_DX_Cooling",
    "OS_Coil_Cooling_DX_MultiSpeed",
    "OS_Coil_Cooling_DX_MultiSpeed_StageData",
    "OS_Coil_Cooling_DX_TwoSpeed",
    "OS_CoilSystem_Cooling_Water_HeatExchangerAssisted",
    "OS_CoilSystem_Cooling_DX_HeatExchangerAssisted",
    "OS_CoilSystem_IntegratedHeatPump_AirSource",
    "OS_Coil_Cooling_DX_VariableRefrigerantFlow",
    "OS_Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl",
    "OS_Coil_Cooling_LowTemperatureRadiant_ConstantFlow",
    "OS_Coil_Cooling_LowTemperatureRadiant_VariableFlow",
    "OS_Coil_Cooling_Water",
    "OS_Coil_Cooling_WaterToAirHeatPump_EquationFit",
    "OS_Coil_Heating_DX_SingleSpeed",
    "OS_Coil_Heating_DX_VariableRefrigerantFlow",
    "OS_Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl",
    "OS_Coil_Heating_Electric",
    "OS_Coil_Heating_Electric_MultiStage",
    "OS_Coil_Heating_Electric_MultiStage_StageData",
    "OS_Coil_Heating_Gas",
    "OS_Coil_Heating_Gas_MultiStage",
    "OS_Coil_Heating_Gas_MultiStage_StageData",
    "OS_Coil_Heating_Desuperheater",
    "OS_Coil_WaterHeating_Desuperheater",
    "OS_Coil_Heating_LowTemperatureRadiant_ConstantFlow",
    "OS_Coil_Heating_LowTemperatureRadiant_VariableFlow",
    "OS_Coil_Heating_Water",
    "OS_Coil_Heating_Water_Baseboard",
    "OS_Coil_Heating_Water_Baseboard_Radiant",
    "OS_Coil_Heating_WaterToAirHeatPump_EquationFit",
    "OS_Coil_Heating_DX_MultiSpeed",
    "OS_Coil_Heating_DX_MultiSpeed_StageData",
    "OS_Coil_Heating_DX_VariableSpeed",
    "OS_Coil_Heating_DX_VariableSpeed_SpeedData",
    "OS_Coil_Cooling_DX_VariableSpeed",
    "OS_Coil_Cooling_DX_VariableSpeed_SpeedData",
    "OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit",
    "OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData",
    "OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit",
    "OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData",
    "OS_Coil_Cooling_Water_Panel_Radiant",
    "OS_Connection",
    "OS_Connector_Mixer",
    "OS_Connector_Splitter",
    "OS_Controller_MechanicalVentilation",
    "OS_ZoneControl_ContaminantController",
    "OS_Controller_OutdoorAir",
    "OS_Controller_WaterCoil",
    "OS_Curve_Linear",
    "OS_Curve_QuadLinear",
    "OS_Curve_QuintLinear",
    "OS_Curve_Quadratic",
    "OS_Curve_Cubic",
    "OS_Curve_Quartic",
    "OS_Curve_Exponent",
    "OS_Curve_Bicubic",
    "OS_Curve_Biquadratic",
    "OS_Curve_QuadraticLinear",
    "OS_Curve_Triquadratic",
    "OS_Curve_Functional_PressureDrop",
    "OS_Curve_FanPressureRise",
    "OS_Curve_ExponentialSkewNormal",
    "OS_Curve_Sigmoid",
    "OS_Curve_RectangularHyperbola1",
    "OS_Curve_RectangularHyperbola2",
    "OS_Curve_ExponentialDecay",
    "OS_Curve_DoubleExponentialDecay",
    "OS_Table_MultiVariableLookup",
    "OS_Table_IndependentVariable",
    "OS_Table_Lookup",
    "OS_DesignSpecification_OutdoorAir",
    "OS_DesignSpecification_ZoneAirDistribution",
    "OS_Humidifier_Steam_Electric",
    "OS_Humidifier_Steam_Gas",
    "OS_EvaporativeCooler_Indirect_ResearchSpecial",
    "OS_EvaporativeCooler_Direct_ResearchSpecial",
    "OS_EvaporativeFluidCooler_SingleSpeed",
    "OS_EvaporativeFluidCooler_TwoSpeed",
    "OS_Fan_SystemModel",
    "OS_Fan_ConstantVolume",
    "OS_Fan_VariableVolume",
    "OS_Fan_OnOff",
    "OS_Fan_ComponentModel",
    "OS_Fan_ZoneExhaust",
    "OS_HeatExchanger_AirToAir_SensibleAndLatent",
    "OS_HeatExchanger_Desiccant_BalancedFlow",
    "OS_HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1",
    "OS_HeatExchanger_FluidToFluid",
    "OS_HVACComponentList",
    "OS_Node",
    "OS_PortList",
    "OS_Pipe_Adiabatic",
    "OS_PlantLoop",
    "OS_PlantEquipmentOperation_CoolingLoad",
    "OS_PlantEquipmentOperation_HeatingLoad",
    "OS_PlantEquipmentOperation_OutdoorDryBulb",
    "OS_PlantEquipmentOperation_OutdoorWetBulb",
    "OS_PlantEquipmentOperation_OutdoorRelativeHumidity",
    "OS_PlantEquipmentOperation_OutdoorDewpoint",
    "OS_PlantEquipmentOperation_OutdoorDryBulbDifference",
    "OS_PlantEquipmentOperation_OutdoorDewpointDifference",
    "OS_PlantEquipmentOperation_OutdoorWetBulbDifference",
    "OS_PlantComponent_TemperatureSource",
    "OS_LoadProfile_Plant",
    "OS_Pump_ConstantSpeed",
    "OS_Pump_VariableSpeed",
    "OS_HeaderedPumps_ConstantSpeed",
    "OS_HeaderedPumps_VariableSpeed",
    "OS_TemperingValve",
    "OS_SetpointManager_FollowOutdoorAirTemperature",
    "OS_SetpointManager_MixedAir",
    "OS_SetpointManager_OutdoorAirReset",
    "OS_SetpointManager_OutdoorAirPretreat",
    "OS_SetpointManager_SingleZone_Reheat",
    "OS_SetpointManager_SingleZone_Humidity_Minimum",
    "OS_SetpointManager_MultiZone_MinimumHumidity_Average",
    "OS_SetpointManager_MultiZone_Humidity_Minimum",
    "OS_SetpointManager_Scheduled",
    "OS_SetpointManager_Scheduled_DualSetpoint",
    "OS_SetpointManager_Warmest",
    "OS_SetpointManager_WarmestTemperatureFlow",
    "OS_SetpointManager_Coldest",
    "OS_SetpointManager_FollowGroundTemperature",
    "OS_SetpointManager_FollowSystemNodeTemperature",
    "OS_SetpointManager_MultiZone_Cooling_Average",
    "OS_SetpointManager_MultiZone_Heating_Average",
    "OS_SetpointManager_MultiZone_Humidity_Maximum",
    "OS_SetpointManager_MultiZone_MaximumHumidity_Average",
    "OS_SetpointManager_SingleZone_OneStageCooling",
    "OS_SetpointManager_SingleZone_OneStageHeating",
    "OS_SetpointManager_SingleZone_Humidity_Maximum",
    "OS_SetpointManager_SingleZone_Cooling",
    "OS_SetpointManager_SingleZone_Heating",
    "OS_SetpointManager_SystemNodeReset_Temperature",
    "OS_SetpointManager_SystemNodeReset_Humidity",
    "OS_Sizing_Plant",
    "OS_Sizing_System",
    "OS_Sizing_Zone",
    "OS_Splitter",
    "OS_ThermalZone",
    "OS_ZoneControl_Humidistat",
    "OS_ThermostatSetpoint_DualSetpoint",
    "OS_ZoneControl_Thermostat_StagedDualSetpoint",
    "OS_ZoneHVAC_EquipmentList",
    "OS_ZoneVentilation_DesignFlowRate",
    "OS_ZoneVentilation_WindandStackOpenArea",
    "OS_ZoneHVAC_Baseboard_Convective_Electric",
    "OS_ZoneHVAC_Baseboard_Convective_Water",
    "OS_ZoneHVAC_HighTemperatureRadiant",
    "OS_ZoneHVAC_LowTemperatureRadiant_Electric",
    "OS_ZoneHVAC_IdealLoadsAirSystem",
    "OS_ZoneHVAC_FourPipeFanCoil",
    "OS_ZoneHVAC_LowTemperatureRadiant_ConstantFlow",
    "OS_ZoneHVAC_LowTemperatureRadiant_VariableFlow",
    "OS_ZoneHVAC_PackagedTerminalHeatPump",
    "OS_ZoneHVAC_PackagedTerminalAirConditioner",
    "OS_ZoneHVAC_TerminalUnit_VariableRefrigerantFlow",
    "OS_ZoneHVAC_WaterToAirHeatPump",
    "OS_ZoneHVAC_UnitHeater",
    "OS_ZoneHVAC_Dehumidifier_DX",
    "OS_ZoneHVAC_EnergyRecoveryVentilator",
    "OS_ZoneHVAC_EnergyRecoveryVentilator_Controller",
    "OS_ZoneHVAC_UnitVentilator",
    "OS_ZoneHVAC_Baseboard_RadiantConvective_Water",
    "OS_ZoneHVAC_Baseboard_RadiantConvective_Electric",
    "OS_ZoneHVAC_CoolingPanel_RadiantConvective_Water",
    "OS_ZoneMixing",
    "OS_WaterHeater_Mixed",
    "OS_WaterHeater_HeatPump",
    "OS_Coil_WaterHeating_AirToWaterHeatPump",
    "OS_WaterHeater_HeatPump_WrappedCondenser",
    "OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed",
    "OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed_SpeedData",
    "OS_Coil_WaterHeating_AirToWaterHeatPump_Wrapped",
    "OS_WaterHeater_Stratified",
    "OS_WaterHeater_Sizing",
    "OS_WaterUse_Equipment",
    "OS_WaterUse_Equipment_Definition",
    "OS_WaterUse_Connections",
    "OS_ThermalStorage_Ice_Detailed",
    "OS_ThermalStorage_ChilledWater_Stratified",
    "OS_Duct",
    "OS_FluidCooler_SingleSpeed",
    "OS_FluidCooler_TwoSpeed",
    "OS_Pipe_Indoor",
    "OS_Pipe_Outdoor",
    "OS_SwimmingPool_Indoor",
    "OS_SolarCollectorPerformance_FlatPlate",
    "OS_SolarCollector_FlatPlate_Water",
    "OS_SolarCollectorPerformance_IntegralCollectorStorage",
    "OS_SolarCollector_IntegralCollectorStorage",
    "OS_SolarCollectorPerformance_PhotovoltaicThermal_Simple",
    "OS_SolarCollectorPerformance_PhotovoltaicThermal_BIPVT",
    "OS_SolarCollector_FlatPlate_PhotovoltaicThermal",
    "OS_Generator_MicroTurbine",
    "OS_Generator_MicroTurbine_HeatRecovery",
    "OS_Generator_Photovoltaic",
    "OS_PhotovoltaicPerformance_Simple",
    "OS_PhotovoltaicPerformance_EquivalentOneDiode",
    "OS_PhotovoltaicPerformance_Sandia",
    "OS_Generator_FuelCell",
    "OS_Generator_FuelCell_PowerModule",
    "OS_Generator_FuelCell_AirSupply",
    "OS_Generator_FuelCell_WaterSupply",
    "OS_Generator_FuelCell_AuxiliaryHeater",
    "OS_Generator_FuelCell_ExhaustGasToWaterHeatExchanger",
    "OS_Generator_FuelCell_ElectricalStorage",
    "OS_Generator_FuelCell_Inverter",
    "OS_Generator_FuelCell_StackCooler",
    "OS_Generator_FuelSupply",
    "OS_Generator_WindTurbine",
    "OS_Generator_PVWatts",
    "OS_ElectricLoadCenter_Inverter_PVWatts",
    "OS_ElectricLoadCenter_Inverter_Simple",
    "OS_ElectricLoadCenter_Inverter_LookUpTable",
    "OS_ElectricLoadCenter_Storage_Simple",
    "OS_ElectricLoadCenter_Storage_LiIonNMCBattery",
    "OS_ElectricLoadCenter_Transformer",
    "OS_ElectricLoadCenter_Distribution",
    "OS_ElectricLoadCenter_Storage_Converter",
    "OS_ComponentCost_Adjustments",
    "OS_LifeCycleCost",
    "OS_CurrencyType",
    "OS_LifeCycleCost_Parameters",
    "OS_LifeCycleCost_UsePriceEscalation",
    "OS_UtilityBill",
    "OS_UtilityCost_Tariff",
    "OS_UtilityCost_Qualify",
    "OS_UtilityCost_Charge_Simple",
    "OS_UtilityCost_Charge_Block",
    "OS_UtilityCost_Ratchet",
    "OS_UtilityCost_Variable",
    "OS_UtilityCost_Computation",
    "OS_Output_Schedules",
    "OS_Output_Constructions",
    "OS_OutputControl_Files",
    "OS_OutputControl_Timestamp",
    "OS_OutputControl_ReportingTolerances",
    "OS_OutputControl_Table_Style",
    "OS_Output_Meter",
    "OS_Output_Variable",
    "OS_Meter_Custom",
    "OS_Meter_CustomDecrement",
    "OS_Output_EnvironmentalImpactFactors",
    "OS_EnvironmentalImpactFactors",
    "OS_FuelFactors",
    "OS_Output_Diagnostics",
    "OS_Output_DebuggingData",
    "OS_Output_JSON",
    "OS_Output_SQLite",
    "OS_Output_EnergyManagementSystem",
    "OS_Output_Table_SummaryReports",
    "OS_EnergyManagementSystem_Sensor",
    "OS_EnergyManagementSystem_Actuator",
    "OS_EnergyManagementSystem_ProgramCallingManager",
    "OS_EnergyManagementSystem_Program",
    "OS_EnergyManagementSystem_Subroutine",
    "OS_EnergyManagementSystem_GlobalVariable",
    "OS_EnergyManagementSystem_OutputVariable",
    "OS_EnergyManagementSystem_MeteredOutputVariable",
    "OS_EnergyManagementSystem_TrendVariable",
    "OS_EnergyManagementSystem_InternalVariable",
    "OS_EnergyManagementSystem_CurveOrTableIndexVariable",
    "OS_EnergyManagementSystem_ConstructionIndexVariable",
    "OS_PythonPlugin_Instance",
    "OS_PythonPlugin_Variable",
    "OS_PythonPlugin_TrendVariable",
    "OS_PythonPlugin_OutputVariable",
    "OS_AirflowNetworkSimulationControl",
    "OS_AirflowNetworkZone",
    "OS_AirflowNetworkSurface",
    "OS_AirflowNetworkReferenceCrackConditions",
    "OS_AirflowNetworkCrack",
    "OS_AirflowNetworkEffectiveLeakageArea",
    "OS_AirflowNetworkSpecifiedFlowRate",
    "OS_AirflowNetworkDetailedOpening",
    "OS_AirflowNetworkSimpleOpening",
    "OS_AirflowNetworkHorizontalOpening",
    "OS_AirflowNetworkOutdoorAirflow",
    "OS_AirflowNetworkZoneExhaustFan",
    "OS_AirflowNetworkExternalNode",
    "OS_AirflowNetworkPressureController",
    "OS_AirflowNetworkDistributionNode",
    "OS_AirflowNetworkLeak",
    "OS_AirflowNetworkLeakageRatio",
    "OS_AirflowNetworkDuct",
    "OS_AirflowNetworkFan",
    "OS_AirflowNetworkEquivalentDuct",
    "OS_AirflowNetworkConstantPressureDrop",
    "OS_AirflowNetworkReliefAirFlow",
    "OS_AirflowNetworkDistributionLinkage",
    "OS_AirflowNetworkDuctViewFactors",
    "OS_AirflowNetworkOccupantVentilationControl",
    "OS_ExternalInterface",
    "OS_ExternalInterface_Schedule",
    "OS_ExternalInterface_Variable",
    "OS_ExternalInterface_Actuator",
    "OS_ExternalInterface_FunctionalMockupUnitImport",
    "OS_ExternalInterface_FunctionalMockupUnitImport_From_Variable",
    "OS_ExternalInterface_FunctionalMockupUnitImport_To_Schedule",
    "OS_ExternalInterface_FunctionalMockupUnitImport_To_Actuator",
    "OS_ExternalInterface_FunctionalMockupUnitImport_To_Variable",
    "OS_ExternalInterface_FunctionalMockupUnitExport_From_Variable",
    "OS_ExternalInterface_FunctionalMockupUnitExport_To_Schedule",
    "OS_ExternalInterface_FunctionalMockupUnitExport_To_Actuator",
    "OS_ExternalInterface_FunctionalMockupUnitExport_To_Variable",
    "OS_Coil_UserDefined",
    "OS_PlantComponent_UserDefined",
    "CommentOnly",
  };

  static constexpr std::array<std::string_view, Size> namesUC{
    "CATCHALL",
    "USERCUSTOM",
    "LEADINPUT",
    "ENDLEADINPUT",
    "SIMULATIONDATA",
    "ENDSIMULATIONDATA",
    "VERSION",
    "SIMULATIONCONTROL",
    "PERFORMANCEPRECISIONTRADEOFFS",
    "BUILDING",
    "SHADOWCALCULATION",
    "SURFACECONVECTIONALGORITHM_INSIDE",
    "SURFACECONVECTIONALGORITHM_OUTSIDE",
    "HEATBALANCEALGORITHM",
    "HEATBALANCESETTINGS_CONDUCTIONFINITEDIFFERENCE",
    "ZONEAIRHEATBALANCEALGORITHM",
    "ZONEAIRCONTAMINANTBALANCE",
    "ZONEAIRMASSFLOWCONSERVATION",
    "ZONECAPACITANCEMULTIPLIER_RESEARCHSPECIAL",
    "TIMESTEP",
    "CONVERGENCELIMITS",
    "HVACSYSTEMROOTFINDINGALGORITHM",
    "COMPLIANCE_BUILDING",
    "SITE_LOCATION",
    "SITE_VARIABLELOCATION",
    "SIZINGPERIOD_DESIGNDAY",
    "SIZINGPERIOD_WEATHERFILEDAYS",
    "SIZINGPERIOD_WEATHERFILECONDITIONTYPE",
    "RUNPERIOD",
    "RUNPERIODCONTROL_SPECIALDAYS",
    "RUNPERIODCONTROL_DAYLIGHTSAVINGTIME",
    "WEATHERPROPERTY_SKYTEMPERATURE",
    "SITE_WEATHERSTATION",
    "SITE_HEIGHTVARIATION",
    "SITE_GROUNDTEMPERATURE_BUILDINGSURFACE",
    "SITE_GROUNDTEMPERATURE_FCFACTORMETHOD",
    "SITE_GROUNDTEMPERATURE_SHALLOW",
    "SITE_GROUNDTEMPERATURE_DEEP",
    "SITE_GROUNDTEMPERATURE_UNDISTURBED_FINITEDIFFERENCE",
    "SITE_GROUNDTEMPERATURE_UNDISTURBED_KUSUDAACHENBACH",
    "SITE_GROUNDTEMPERATURE_UNDISTURBED_XING",
    "SITE_GROUNDDOMAIN_SLAB",
    "SITE_GROUNDDOMAIN_BASEMENT",
    "SITE_GROUNDREFLECTANCE",
    "SITE_GROUNDREFLECTANCE_SNOWMODIFIER",
    "SITE_WATERMAINSTEMPERATURE",
    "SITE_PRECIPITATION",
    "ROOFIRRIGATION",
    "SITE_SOLARANDVISIBLESPECTRUM",
    "SITE_SPECTRUMDATA",
    "SCHEDULETYPELIMITS",
    "SCHEDULE_DAY_HOURLY",
    "SCHEDULE_DAY_INTERVAL",
    "SCHEDULE_DAY_LIST",
    "SCHEDULE_WEEK_DAILY",
    "SCHEDULE_WEEK_COMPACT",
    "SCHEDULE_YEAR",
    "SCHEDULE_COMPACT",
    "SCHEDULE_CONSTANT",
    "SCHEDULE_FILE_SHADING",
    "SCHEDULE_FILE",
    "MATERIAL",
    "MATERIAL_NOMASS",
    "MATERIAL_INFRAREDTRANSPARENT",
    "MATERIAL_AIRGAP",
    "MATERIAL_ROOFVEGETATION",
    "WINDOWMATERIAL_SIMPLEGLAZINGSYSTEM",
    "WINDOWMATERIAL_GLAZING",
    "WINDOWMATERIAL_GLAZINGGROUP_THERMOCHROMIC",
    "WINDOWMATERIAL_GLAZING_REFRACTIONEXTINCTIONMETHOD",
    "WINDOWMATERIAL_GAS",
    "WINDOWGAP_SUPPORTPILLAR",
    "WINDOWGAP_DEFLECTIONSTATE",
    "WINDOWMATERIAL_GASMIXTURE",
    "WINDOWMATERIAL_GAP",
    "WINDOWMATERIAL_SHADE",
    "WINDOWMATERIAL_COMPLEXSHADE",
    "WINDOWMATERIAL_BLIND",
    "WINDOWMATERIAL_SCREEN",
    "WINDOWMATERIAL_SHADE_EQUIVALENTLAYER",
    "WINDOWMATERIAL_DRAPE_EQUIVALENTLAYER",
    "WINDOWMATERIAL_BLIND_EQUIVALENTLAYER",
    "WINDOWMATERIAL_SCREEN_EQUIVALENTLAYER",
    "WINDOWMATERIAL_GLAZING_EQUIVALENTLAYER",
    "WINDOWMATERIAL_GAP_EQUIVALENTLAYER",
    "MATERIALPROPERTY_MOISTUREPENETRATIONDEPTH_SETTINGS",
    "MATERIALPROPERTY_PHASECHANGE",
    "MATERIALPROPERTY_PHASECHANGEHYSTERESIS",
    "MATERIALPROPERTY_VARIABLETHERMALCONDUCTIVITY",
    "MATERIALPROPERTY_VARIABLEABSORPTANCE",
    "MATERIALPROPERTY_HEATANDMOISTURETRANSFER_SETTINGS",
    "MATERIALPROPERTY_HEATANDMOISTURETRANSFER_SORPTIONISOTHERM",
    "MATERIALPROPERTY_HEATANDMOISTURETRANSFER_SUCTION",
    "MATERIALPROPERTY_HEATANDMOISTURETRANSFER_REDISTRIBUTION",
    "MATERIALPROPERTY_HEATANDMOISTURETRANSFER_DIFFUSION",
    "MATERIALPROPERTY_HEATANDMOISTURETRANSFER_THERMALCONDUCTIVITY",
    "MATERIALPROPERTY_GLAZINGSPECTRALDATA",
    "CONSTRUCTION",
    "CONSTRUCTION_CFACTORUNDERGROUNDWALL",
    "CONSTRUCTION_FFACTORGROUNDFLOOR",
    "CONSTRUCTIONPROPERTY_INTERNALHEATSOURCE",
    "CONSTRUCTION_AIRBOUNDARY",
    "WINDOWTHERMALMODEL_PARAMS",
    "WINDOWSCALCULATIONENGINE",
    "CONSTRUCTION_COMPLEXFENESTRATIONSTATE",
    "CONSTRUCTION_WINDOWEQUIVALENTLAYER",
    "CONSTRUCTION_WINDOWDATAFILE",
    "GLOBALGEOMETRYRULES",
    "GEOMETRYTRANSFORM",
    "SPACE",
    "SPACELIST",
    "ZONE",
    "ZONELIST",
    "ZONEGROUP",
    "BUILDINGSURFACE_DETAILED",
    "WALL_DETAILED",
    "ROOFCEILING_DETAILED",
    "FLOOR_DETAILED",
    "WALL_EXTERIOR",
    "WALL_ADIABATIC",
    "WALL_UNDERGROUND",
    "WALL_INTERZONE",
    "ROOF",
    "CEILING_ADIABATIC",
    "CEILING_INTERZONE",
    "FLOOR_GROUNDCONTACT",
    "FLOOR_ADIABATIC",
    "FLOOR_INTERZONE",
    "FENESTRATIONSURFACE_DETAILED",
    "WINDOW",
    "DOOR",
    "GLAZEDDOOR",
    "WINDOW_INTERZONE",
    "DOOR_INTERZONE",
    "GLAZEDDOOR_INTERZONE",
    "WINDOWSHADINGCONTROL",
    "WINDOWPROPERTY_FRAMEANDDIVIDER",
    "WINDOWPROPERTY_AIRFLOWCONTROL",
    "WINDOWPROPERTY_STORMWINDOW",
    "INTERNALMASS",
    "SHADING_SITE",
    "SHADING_BUILDING",
    "SHADING_SITE_DETAILED",
    "SHADING_BUILDING_DETAILED",
    "SHADING_OVERHANG",
    "SHADING_OVERHANG_PROJECTION",
    "SHADING_FIN",
    "SHADING_FIN_PROJECTION",
    "SHADING_ZONE_DETAILED",
    "SHADINGPROPERTY_REFLECTANCE",
    "SURFACEPROPERTY_HEATTRANSFERALGORITHM",
    "SURFACEPROPERTY_HEATTRANSFERALGORITHM_MULTIPLESURFACE",
    "SURFACEPROPERTY_HEATTRANSFERALGORITHM_SURFACELIST",
    "SURFACEPROPERTY_HEATTRANSFERALGORITHM_CONSTRUCTION",
    "SURFACEPROPERTY_HEATBALANCESOURCETERM",
    "SURFACECONTROL_MOVABLEINSULATION",
    "SURFACEPROPERTY_OTHERSIDECOEFFICIENTS",
    "SURFACEPROPERTY_OTHERSIDECONDITIONSMODEL",
    "SURFACEPROPERTY_UNDERWATER",
    "FOUNDATION_KIVA",
    "FOUNDATION_KIVA_SETTINGS",
    "SURFACEPROPERTY_EXPOSEDFOUNDATIONPERIMETER",
    "SURFACECONVECTIONALGORITHM_INSIDE_ADAPTIVEMODELSELECTIONS",
    "SURFACECONVECTIONALGORITHM_OUTSIDE_ADAPTIVEMODELSELECTIONS",
    "SURFACECONVECTIONALGORITHM_INSIDE_USERCURVE",
    "SURFACECONVECTIONALGORITHM_OUTSIDE_USERCURVE",
    "SURFACEPROPERTY_CONVECTIONCOEFFICIENTS",
    "SURFACEPROPERTY_CONVECTIONCOEFFICIENTS_MULTIPLESURFACE",
    "SURFACEPROPERTIES_VAPORCOEFFICIENTS",
    "SURFACEPROPERTY_EXTERIORNATURALVENTEDCAVITY",
    "SURFACEPROPERTY_SOLARINCIDENTINSIDE",
    "SURFACEPROPERTY_INCIDENTSOLARMULTIPLIER",
    "SURFACEPROPERTY_LOCALENVIRONMENT",
    "ZONEPROPERTY_LOCALENVIRONMENT",
    "SURFACEPROPERTY_SURROUNDINGSURFACES",
    "SURFACEPROPERTY_GROUNDSURFACES",
    "COMPLEXFENESTRATIONPROPERTY_SOLARABSORBEDLAYERS",
    "ZONEPROPERTY_USERVIEWFACTORS_BYSURFACENAME",
    "GROUNDHEATTRANSFER_CONTROL",
    "GROUNDHEATTRANSFER_SLAB_MATERIALS",
    "GROUNDHEATTRANSFER_SLAB_MATLPROPS",
    "GROUNDHEATTRANSFER_SLAB_BOUNDCONDS",
    "GROUNDHEATTRANSFER_SLAB_BLDGPROPS",
    "GROUNDHEATTRANSFER_SLAB_INSULATION",
    "GROUNDHEATTRANSFER_SLAB_EQUIVALENTSLAB",
    "GROUNDHEATTRANSFER_SLAB_AUTOGRID",
    "GROUNDHEATTRANSFER_SLAB_MANUALGRID",
    "GROUNDHEATTRANSFER_SLAB_XFACE",
    "GROUNDHEATTRANSFER_SLAB_YFACE",
    "GROUNDHEATTRANSFER_SLAB_ZFACE",
    "GROUNDHEATTRANSFER_BASEMENT_SIMPARAMETERS",
    "GROUNDHEATTRANSFER_BASEMENT_MATLPROPS",
    "GROUNDHEATTRANSFER_BASEMENT_INSULATION",
    "GROUNDHEATTRANSFER_BASEMENT_SURFACEPROPS",
    "GROUNDHEATTRANSFER_BASEMENT_BLDGDATA",
    "GROUNDHEATTRANSFER_BASEMENT_INTERIOR",
    "GROUNDHEATTRANSFER_BASEMENT_COMBLDG",
    "GROUNDHEATTRANSFER_BASEMENT_EQUIVSLAB",
    "GROUNDHEATTRANSFER_BASEMENT_EQUIVAUTOGRID",
    "GROUNDHEATTRANSFER_BASEMENT_AUTOGRID",
    "GROUNDHEATTRANSFER_BASEMENT_MANUALGRID",
    "GROUNDHEATTRANSFER_BASEMENT_XFACE",
    "GROUNDHEATTRANSFER_BASEMENT_YFACE",
    "GROUNDHEATTRANSFER_BASEMENT_ZFACE",
    "ROOMAIRMODELTYPE",
    "ROOMAIR_TEMPERATUREPATTERN_USERDEFINED",
    "ROOMAIR_TEMPERATUREPATTERN_CONSTANTGRADIENT",
    "ROOMAIR_TEMPERATUREPATTERN_TWOGRADIENT",
    "ROOMAIR_TEMPERATUREPATTERN_NONDIMENSIONALHEIGHT",
    "ROOMAIR_TEMPERATUREPATTERN_SURFACEMAPPING",
    "ROOMAIR_NODE",
    "ROOMAIRSETTINGS_ONENODEDISPLACEMENTVENTILATION",
    "ROOMAIRSETTINGS_THREENODEDISPLACEMENTVENTILATION",
    "ROOMAIRSETTINGS_CROSSVENTILATION",
    "ROOMAIRSETTINGS_UNDERFLOORAIRDISTRIBUTIONINTERIOR",
    "ROOMAIRSETTINGS_UNDERFLOORAIRDISTRIBUTIONEXTERIOR",
    "ROOMAIR_NODE_AIRFLOWNETWORK",
    "ROOMAIR_NODE_AIRFLOWNETWORK_ADJACENTSURFACELIST",
    "ROOMAIR_NODE_AIRFLOWNETWORK_INTERNALGAINS",
    "ROOMAIR_NODE_AIRFLOWNETWORK_HVACEQUIPMENT",
    "ROOMAIRSETTINGS_AIRFLOWNETWORK",
    "PEOPLE",
    "COMFORTVIEWFACTORANGLES",
    "LIGHTS",
    "ELECTRICEQUIPMENT",
    "GASEQUIPMENT",
    "HOTWATEREQUIPMENT",
    "STEAMEQUIPMENT",
    "OTHEREQUIPMENT",
    "ELECTRICEQUIPMENT_ITE_AIRCOOLED",
    "ZONEBASEBOARD_OUTDOORTEMPERATURECONTROLLED",
    "SWIMMINGPOOL_INDOOR",
    "ZONECONTAMINANTSOURCEANDSINK_CARBONDIOXIDE",
    "ZONECONTAMINANTSOURCEANDSINK_GENERIC_CONSTANT",
    "SURFACECONTAMINANTSOURCEANDSINK_GENERIC_PRESSUREDRIVEN",
    "ZONECONTAMINANTSOURCEANDSINK_GENERIC_CUTOFFMODEL",
    "ZONECONTAMINANTSOURCEANDSINK_GENERIC_DECAYSOURCE",
    "SURFACECONTAMINANTSOURCEANDSINK_GENERIC_BOUNDARYLAYERDIFFUSION",
    "SURFACECONTAMINANTSOURCEANDSINK_GENERIC_DEPOSITIONVELOCITYSINK",
    "ZONECONTAMINANTSOURCEANDSINK_GENERIC_DEPOSITIONRATESINK",
    "DAYLIGHTING_CONTROLS",
    "DAYLIGHTING_REFERENCEPOINT",
    "DAYLIGHTING_DELIGHT_COMPLEXFENESTRATION",
    "DAYLIGHTINGDEVICE_TUBULAR",
    "DAYLIGHTINGDEVICE_SHELF",
    "DAYLIGHTINGDEVICE_LIGHTWELL",
    "OUTPUT_DAYLIGHTFACTORS",
    "OUTPUT_ILLUMINANCEMAP",
    "OUTPUTCONTROL_ILLUMINANCEMAP_STYLE",
    "ZONEINFILTRATION_DESIGNFLOWRATE",
    "ZONEINFILTRATION_EFFECTIVELEAKAGEAREA",
    "ZONEINFILTRATION_FLOWCOEFFICIENT",
    "ZONEVENTILATION_DESIGNFLOWRATE",
    "ZONEVENTILATION_WINDANDSTACKOPENAREA",
    "ZONEAIRBALANCE_OUTDOORAIR",
    "ZONEMIXING",
    "ZONECROSSMIXING",
    "ZONEREFRIGERATIONDOORMIXING",
    "ZONEEARTHTUBE",
    "ZONECOOLTOWER_SHOWER",
    "ZONETHERMALCHIMNEY",
    "AIRFLOWNETWORK_SIMULATIONCONTROL",
    "AIRFLOWNETWORK_MULTIZONE_ZONE",
    "AIRFLOWNETWORK_MULTIZONE_SURFACE",
    "AIRFLOWNETWORK_MULTIZONE_REFERENCECRACKCONDITIONS",
    "AIRFLOWNETWORK_MULTIZONE_SURFACE_CRACK",
    "AIRFLOWNETWORK_MULTIZONE_SURFACE_EFFECTIVELEAKAGEAREA",
    "AIRFLOWNETWORK_MULTIZONE_SPECIFIEDFLOWRATE",
    "AIRFLOWNETWORK_MULTIZONE_COMPONENT_DETAILEDOPENING",
    "AIRFLOWNETWORK_MULTIZONE_COMPONENT_SIMPLEOPENING",
    "AIRFLOWNETWORK_MULTIZONE_COMPONENT_HORIZONTALOPENING",
    "AIRFLOWNETWORK_MULTIZONE_COMPONENT_ZONEEXHAUSTFAN",
    "AIRFLOWNETWORK_MULTIZONE_EXTERNALNODE",
    "AIRFLOWNETWORK_MULTIZONE_WINDPRESSURECOEFFICIENTARRAY",
    "AIRFLOWNETWORK_MULTIZONE_WINDPRESSURECOEFFICIENTVALUES",
    "AIRFLOWNETWORK_ZONECONTROL_PRESSURECONTROLLER",
    "AIRFLOWNETWORK_DISTRIBUTION_NODE",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_LEAK",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_LEAKAGERATIO",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_DUCT",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_FAN",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_COIL",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_HEATEXCHANGER",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_TERMINALUNIT",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_CONSTANTPRESSUREDROP",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_OUTDOORAIRFLOW",
    "AIRFLOWNETWORK_DISTRIBUTION_COMPONENT_RELIEFAIRFLOW",
    "AIRFLOWNETWORK_DISTRIBUTION_LINKAGE",
    "AIRFLOWNETWORK_DISTRIBUTION_DUCTVIEWFACTORS",
    "AIRFLOWNETWORK_OCCUPANTVENTILATIONCONTROL",
    "AIRFLOWNETWORK_INTRAZONE_NODE",
    "AIRFLOWNETWORK_INTRAZONE_LINKAGE",
    "EXTERIOR_LIGHTS",
    "EXTERIOR_FUELEQUIPMENT",
    "EXTERIOR_WATEREQUIPMENT",
    "HVACTEMPLATE_THERMOSTAT",
    "HVACTEMPLATE_ZONE_IDEALLOADSAIRSYSTEM",
    "HVACTEMPLATE_ZONE_BASEBOARDHEAT",
    "HVACTEMPLATE_ZONE_FANCOIL",
    "HVACTEMPLATE_ZONE_PTAC",
    "HVACTEMPLATE_ZONE_PTHP",
    "HVACTEMPLATE_ZONE_WATERTOAIRHEATPUMP",
    "HVACTEMPLATE_ZONE_VRF",
    "HVACTEMPLATE_ZONE_UNITARY",
    "HVACTEMPLATE_ZONE_VAV",
    "HVACTEMPLATE_ZONE_VAV_FANPOWERED",
    "HVACTEMPLATE_ZONE_VAV_HEATANDCOOL",
    "HVACTEMPLATE_ZONE_CONSTANTVOLUME",
    "HVACTEMPLATE_ZONE_DUALDUCT",
    "HVACTEMPLATE_SYSTEM_VRF",
    "HVACTEMPLATE_SYSTEM_UNITARY",
    "HVACTEMPLATE_SYSTEM_UNITARYHEATPUMP_AIRTOAIR",
    "HVACTEMPLATE_SYSTEM_UNITARYSYSTEM",
    "HVACTEMPLATE_SYSTEM_VAV",
    "HVACTEMPLATE_SYSTEM_PACKAGEDVAV",
    "HVACTEMPLATE_SYSTEM_CONSTANTVOLUME",
    "HVACTEMPLATE_SYSTEM_DUALDUCT",
    "HVACTEMPLATE_SYSTEM_DEDICATEDOUTDOORAIR",
    "HVACTEMPLATE_PLANT_CHILLEDWATERLOOP",
    "HVACTEMPLATE_PLANT_CHILLER",
    "HVACTEMPLATE_PLANT_CHILLER_OBJECTREFERENCE",
    "HVACTEMPLATE_PLANT_TOWER",
    "HVACTEMPLATE_PLANT_TOWER_OBJECTREFERENCE",
    "HVACTEMPLATE_PLANT_HOTWATERLOOP",
    "HVACTEMPLATE_PLANT_BOILER",
    "HVACTEMPLATE_PLANT_BOILER_OBJECTREFERENCE",
    "HVACTEMPLATE_PLANT_MIXEDWATERLOOP",
    "DESIGNSPECIFICATION_OUTDOORAIR",
    "DESIGNSPECIFICATION_OUTDOORAIR_SPACELIST",
    "DESIGNSPECIFICATION_ZONEAIRDISTRIBUTION",
    "SIZING_PARAMETERS",
    "SIZING_ZONE",
    "DESIGNSPECIFICATION_ZONEHVAC_SIZING",
    "DESIGNSPECIFICATION_AIRTERMINAL_SIZING",
    "SIZING_SYSTEM",
    "SIZING_PLANT",
    "OUTPUTCONTROL_SIZING_STYLE",
    "ZONECONTROL_HUMIDISTAT",
    "ZONECONTROL_THERMOSTAT",
    "ZONECONTROL_THERMOSTAT_OPERATIVETEMPERATURE",
    "ZONECONTROL_THERMOSTAT_THERMALCOMFORT",
    "ZONECONTROL_THERMOSTAT_TEMPERATUREANDHUMIDITY",
    "THERMOSTATSETPOINT_SINGLEHEATING",
    "THERMOSTATSETPOINT_SINGLECOOLING",
    "THERMOSTATSETPOINT_SINGLEHEATINGORCOOLING",
    "THERMOSTATSETPOINT_DUALSETPOINT",
    "THERMOSTATSETPOINT_THERMALCOMFORT_FANGER_SINGLEHEATING",
    "THERMOSTATSETPOINT_THERMALCOMFORT_FANGER_SINGLECOOLING",
    "THERMOSTATSETPOINT_THERMALCOMFORT_FANGER_SINGLEHEATINGORCOOLING",
    "THERMOSTATSETPOINT_THERMALCOMFORT_FANGER_DUALSETPOINT",
    "ZONECONTROL_THERMOSTAT_STAGEDDUALSETPOINT",
    "ZONECONTROL_CONTAMINANTCONTROLLER",
    "ZONEHVAC_IDEALLOADSAIRSYSTEM",
    "ZONEHVAC_FOURPIPEFANCOIL",
    "ZONEHVAC_WINDOWAIRCONDITIONER",
    "ZONEHVAC_PACKAGEDTERMINALAIRCONDITIONER",
    "ZONEHVAC_PACKAGEDTERMINALHEATPUMP",
    "ZONEHVAC_WATERTOAIRHEATPUMP",
    "ZONEHVAC_DEHUMIDIFIER_DX",
    "ZONEHVAC_ENERGYRECOVERYVENTILATOR",
    "ZONEHVAC_ENERGYRECOVERYVENTILATOR_CONTROLLER",
    "ZONEHVAC_UNITVENTILATOR",
    "ZONEHVAC_UNITHEATER",
    "ZONEHVAC_EVAPORATIVECOOLERUNIT",
    "ZONEHVAC_HYBRIDUNITARYHVAC",
    "ZONEHVAC_OUTDOORAIRUNIT",
    "ZONEHVAC_OUTDOORAIRUNIT_EQUIPMENTLIST",
    "ZONEHVAC_TERMINALUNIT_VARIABLEREFRIGERANTFLOW",
    "ZONEHVAC_BASEBOARD_RADIANTCONVECTIVE_WATER_DESIGN",
    "ZONEHVAC_BASEBOARD_RADIANTCONVECTIVE_WATER",
    "ZONEHVAC_BASEBOARD_RADIANTCONVECTIVE_STEAM_DESIGN",
    "ZONEHVAC_BASEBOARD_RADIANTCONVECTIVE_STEAM",
    "ZONEHVAC_BASEBOARD_RADIANTCONVECTIVE_ELECTRIC",
    "ZONEHVAC_COOLINGPANEL_RADIANTCONVECTIVE_WATER",
    "ZONEHVAC_BASEBOARD_CONVECTIVE_WATER",
    "ZONEHVAC_BASEBOARD_CONVECTIVE_ELECTRIC",
    "ZONEHVAC_LOWTEMPERATURERADIANT_VARIABLEFLOW",
    "ZONEHVAC_LOWTEMPERATURERADIANT_VARIABLEFLOW_DESIGN",
    "ZONEHVAC_LOWTEMPERATURERADIANT_CONSTANTFLOW",
    "ZONEHVAC_LOWTEMPERATURERADIANT_CONSTANTFLOW_DESIGN",
    "ZONEHVAC_LOWTEMPERATURERADIANT_ELECTRIC",
    "ZONEHVAC_LOWTEMPERATURERADIANT_SURFACEGROUP",
    "ZONEHVAC_HIGHTEMPERATURERADIANT",
    "ZONEHVAC_VENTILATEDSLAB",
    "ZONEHVAC_VENTILATEDSLAB_SLABGROUP",
    "AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_REHEAT",
    "AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_NOREHEAT",
    "AIRTERMINAL_SINGLEDUCT_VAV_NOREHEAT",
    "AIRTERMINAL_SINGLEDUCT_VAV_REHEAT",
    "AIRTERMINAL_SINGLEDUCT_VAV_REHEAT_VARIABLESPEEDFAN",
    "AIRTERMINAL_SINGLEDUCT_VAV_HEATANDCOOL_NOREHEAT",
    "AIRTERMINAL_SINGLEDUCT_VAV_HEATANDCOOL_REHEAT",
    "AIRTERMINAL_SINGLEDUCT_SERIESPIU_REHEAT",
    "AIRTERMINAL_SINGLEDUCT_PARALLELPIU_REHEAT",
    "AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_FOURPIPEINDUCTION",
    "AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_FOURPIPEBEAM",
    "AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_COOLEDBEAM",
    "AIRTERMINAL_SINGLEDUCT_MIXER",
    "AIRTERMINAL_DUALDUCT_CONSTANTVOLUME",
    "AIRTERMINAL_DUALDUCT_VAV",
    "AIRTERMINAL_DUALDUCT_VAV_OUTDOORAIR",
    "ZONEHVAC_AIRDISTRIBUTIONUNIT",
    "ZONEHVAC_EXHAUSTCONTROL",
    "ZONEHVAC_EQUIPMENTLIST",
    "ZONEHVAC_EQUIPMENTCONNECTIONS",
    "FAN_SYSTEMMODEL",
    "FAN_CONSTANTVOLUME",
    "FAN_VARIABLEVOLUME",
    "FAN_ONOFF",
    "FAN_ZONEEXHAUST",
    "FANPERFORMANCE_NIGHTVENTILATION",
    "FAN_COMPONENTMODEL",
    "COIL_COOLING_WATER",
    "COIL_COOLING_WATER_DETAILEDGEOMETRY",
    "COILSYSTEM_COOLING_WATER",
    "COIL_COOLING_DX",
    "COIL_COOLING_DX_CURVEFIT_PERFORMANCE",
    "COIL_COOLING_DX_CURVEFIT_OPERATINGMODE",
    "COIL_COOLING_DX_CURVEFIT_SPEED",
    "COIL_COOLING_DX_SINGLESPEED",
    "COIL_COOLING_DX_TWOSPEED",
    "COIL_COOLING_DX_MULTISPEED",
    "COIL_COOLING_DX_VARIABLESPEED",
    "COIL_COOLING_DX_TWOSTAGEWITHHUMIDITYCONTROLMODE",
    "COILPERFORMANCE_DX_COOLING",
    "COIL_COOLING_DX_VARIABLEREFRIGERANTFLOW",
    "COIL_HEATING_DX_VARIABLEREFRIGERANTFLOW",
    "COIL_COOLING_DX_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL",
    "COIL_HEATING_DX_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL",
    "COIL_HEATING_WATER",
    "COIL_HEATING_STEAM",
    "COIL_HEATING_ELECTRIC",
    "COIL_HEATING_ELECTRIC_MULTISTAGE",
    "COIL_HEATING_FUEL",
    "COIL_HEATING_GAS_MULTISTAGE",
    "COIL_HEATING_DESUPERHEATER",
    "COIL_HEATING_DX_SINGLESPEED",
    "COIL_HEATING_DX_MULTISPEED",
    "COIL_HEATING_DX_VARIABLESPEED",
    "COIL_COOLING_WATERTOAIRHEATPUMP_PARAMETERESTIMATION",
    "COIL_HEATING_WATERTOAIRHEATPUMP_PARAMETERESTIMATION",
    "COIL_COOLING_WATERTOAIRHEATPUMP_EQUATIONFIT",
    "COIL_COOLING_WATERTOAIRHEATPUMP_VARIABLESPEEDEQUATIONFIT",
    "COIL_HEATING_WATERTOAIRHEATPUMP_EQUATIONFIT",
    "COIL_HEATING_WATERTOAIRHEATPUMP_VARIABLESPEEDEQUATIONFIT",
    "COIL_WATERHEATING_AIRTOWATERHEATPUMP_PUMPED",
    "COIL_WATERHEATING_AIRTOWATERHEATPUMP_WRAPPED",
    "COIL_WATERHEATING_AIRTOWATERHEATPUMP_VARIABLESPEED",
    "COIL_WATERHEATING_DESUPERHEATER",
    "COILSYSTEM_COOLING_DX",
    "COILSYSTEM_HEATING_DX",
    "COILSYSTEM_COOLING_WATER_HEATEXCHANGERASSISTED",
    "COILSYSTEM_COOLING_DX_HEATEXCHANGERASSISTED",
    "COILSYSTEM_INTEGRATEDHEATPUMP_AIRSOURCE",
    "COIL_COOLING_DX_SINGLESPEED_THERMALSTORAGE",
    "EVAPORATIVECOOLER_DIRECT_CELDEKPAD",
    "EVAPORATIVECOOLER_INDIRECT_CELDEKPAD",
    "EVAPORATIVECOOLER_INDIRECT_WETCOIL",
    "EVAPORATIVECOOLER_INDIRECT_RESEARCHSPECIAL",
    "EVAPORATIVECOOLER_DIRECT_RESEARCHSPECIAL",
    "HUMIDIFIER_STEAM_ELECTRIC",
    "HUMIDIFIER_STEAM_GAS",
    "DEHUMIDIFIER_DESICCANT_NOFANS",
    "DEHUMIDIFIER_DESICCANT_SYSTEM",
    "HEATEXCHANGER_AIRTOAIR_FLATPLATE",
    "HEATEXCHANGER_AIRTOAIR_SENSIBLEANDLATENT",
    "HEATEXCHANGER_DESICCANT_BALANCEDFLOW",
    "HEATEXCHANGER_DESICCANT_BALANCEDFLOW_PERFORMANCEDATATYPE1",
    "AIRLOOPHVAC_UNITARYSYSTEM",
    "UNITARYSYSTEMPERFORMANCE_MULTISPEED",
    "AIRLOOPHVAC_UNITARY_FURNACE_HEATONLY",
    "AIRLOOPHVAC_UNITARY_FURNACE_HEATCOOL",
    "AIRLOOPHVAC_UNITARYHEATONLY",
    "AIRLOOPHVAC_UNITARYHEATCOOL",
    "AIRLOOPHVAC_UNITARYHEATPUMP_AIRTOAIR",
    "AIRLOOPHVAC_UNITARYHEATPUMP_WATERTOAIR",
    "AIRLOOPHVAC_UNITARYHEATCOOL_VAVCHANGEOVERBYPASS",
    "AIRLOOPHVAC_UNITARYHEATPUMP_AIRTOAIR_MULTISPEED",
    "AIRCONDITIONER_VARIABLEREFRIGERANTFLOW",
    "AIRCONDITIONER_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL",
    "AIRCONDITIONER_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL_HR",
    "ZONETERMINALUNITLIST",
    "CONTROLLER_WATERCOIL",
    "CONTROLLER_OUTDOORAIR",
    "CONTROLLER_MECHANICALVENTILATION",
    "AIRLOOPHVAC_CONTROLLERLIST",
    "AIRLOOPHVAC",
    "AIRLOOPHVAC_OUTDOORAIRSYSTEM_EQUIPMENTLIST",
    "AIRLOOPHVAC_OUTDOORAIRSYSTEM",
    "OUTDOORAIR_MIXER",
    "AIRLOOPHVAC_ZONESPLITTER",
    "AIRLOOPHVAC_SUPPLYPLENUM",
    "AIRLOOPHVAC_SUPPLYPATH",
    "AIRLOOPHVAC_ZONEMIXER",
    "AIRLOOPHVAC_RETURNPLENUM",
    "AIRLOOPHVAC_RETURNPATH",
    "AIRLOOPHVAC_EXHAUSTSYSTEM",
    "AIRLOOPHVAC_DEDICATEDOUTDOORAIRSYSTEM",
    "AIRLOOPHVAC_MIXER",
    "AIRLOOPHVAC_SPLITTER",
    "BRANCH",
    "BRANCHLIST",
    "CONNECTOR_SPLITTER",
    "CONNECTOR_MIXER",
    "CONNECTORLIST",
    "NODELIST",
    "OUTDOORAIR_NODE",
    "OUTDOORAIR_NODELIST",
    "PIPE_ADIABATIC",
    "PIPE_ADIABATIC_STEAM",
    "PIPE_INDOOR",
    "PIPE_OUTDOOR",
    "PIPE_UNDERGROUND",
    "PIPINGSYSTEM_UNDERGROUND_DOMAIN",
    "PIPINGSYSTEM_UNDERGROUND_PIPECIRCUIT",
    "PIPINGSYSTEM_UNDERGROUND_PIPESEGMENT",
    "DUCT",
    "PUMP_VARIABLESPEED",
    "PUMP_CONSTANTSPEED",
    "PUMP_VARIABLESPEED_CONDENSATE",
    "HEADEREDPUMPS_CONSTANTSPEED",
    "HEADEREDPUMPS_VARIABLESPEED",
    "TEMPERINGVALVE",
    "LOADPROFILE_PLANT",
    "SOLARCOLLECTORPERFORMANCE_FLATPLATE",
    "SOLARCOLLECTOR_FLATPLATE_WATER",
    "SOLARCOLLECTOR_FLATPLATE_PHOTOVOLTAICTHERMAL",
    "SOLARCOLLECTORPERFORMANCE_PHOTOVOLTAICTHERMAL_SIMPLE",
    "SOLARCOLLECTORPERFORMANCE_PHOTOVOLTAICTHERMAL_BIPVT",
    "SOLARCOLLECTOR_INTEGRALCOLLECTORSTORAGE",
    "SOLARCOLLECTORPERFORMANCE_INTEGRALCOLLECTORSTORAGE",
    "SOLARCOLLECTOR_UNGLAZEDTRANSPIRED",
    "SOLARCOLLECTOR_UNGLAZEDTRANSPIRED_MULTISYSTEM",
    "BOILER_HOTWATER",
    "BOILER_STEAM",
    "CHILLER_ELECTRIC_ASHRAE205",
    "CHILLER_ELECTRIC_EIR",
    "CHILLER_ELECTRIC_REFORMULATEDEIR",
    "CHILLER_ELECTRIC",
    "CHILLER_ABSORPTION_INDIRECT",
    "CHILLER_ABSORPTION",
    "CHILLER_CONSTANTCOP",
    "CHILLER_ENGINEDRIVEN",
    "CHILLER_COMBUSTIONTURBINE",
    "CHILLERHEATER_ABSORPTION_DIRECTFIRED",
    "CHILLERHEATER_ABSORPTION_DOUBLEEFFECT",
    "HEATPUMP_PLANTLOOP_EIR_COOLING",
    "HEATPUMP_PLANTLOOP_EIR_HEATING",
    "HEATPUMP_AIRTOWATER_FUELFIRED_HEATING",
    "HEATPUMP_AIRTOWATER_FUELFIRED_COOLING",
    "HEATPUMP_WATERTOWATER_EQUATIONFIT_HEATING",
    "HEATPUMP_WATERTOWATER_EQUATIONFIT_COOLING",
    "HEATPUMP_WATERTOWATER_PARAMETERESTIMATION_COOLING",
    "HEATPUMP_WATERTOWATER_PARAMETERESTIMATION_HEATING",
    "DISTRICTCOOLING",
    "DISTRICTHEATING",
    "PLANTCOMPONENT_TEMPERATURESOURCE",
    "CENTRALHEATPUMPSYSTEM",
    "CHILLERHEATERPERFORMANCE_ELECTRIC_EIR",
    "COOLINGTOWER_SINGLESPEED",
    "COOLINGTOWER_TWOSPEED",
    "COOLINGTOWER_VARIABLESPEED_MERKEL",
    "COOLINGTOWER_VARIABLESPEED",
    "COOLINGTOWERPERFORMANCE_COOLTOOLS",
    "COOLINGTOWERPERFORMANCE_YORKCALC",
    "EVAPORATIVEFLUIDCOOLER_SINGLESPEED",
    "EVAPORATIVEFLUIDCOOLER_TWOSPEED",
    "FLUIDCOOLER_SINGLESPEED",
    "FLUIDCOOLER_TWOSPEED",
    "GROUNDHEATEXCHANGER_SYSTEM",
    "GROUNDHEATEXCHANGER_VERTICAL_PROPERTIES",
    "GROUNDHEATEXCHANGER_VERTICAL_ARRAY",
    "GROUNDHEATEXCHANGER_VERTICAL_SINGLE",
    "GROUNDHEATEXCHANGER_RESPONSEFACTORS",
    "GROUNDHEATEXCHANGER_POND",
    "GROUNDHEATEXCHANGER_SURFACE",
    "GROUNDHEATEXCHANGER_HORIZONTALTRENCH",
    "GROUNDHEATEXCHANGER_SLINKY",
    "HEATEXCHANGER_FLUIDTOFLUID",
    "WATERHEATER_MIXED",
    "WATERHEATER_STRATIFIED",
    "WATERHEATER_SIZING",
    "WATERHEATER_HEATPUMP_PUMPEDCONDENSER",
    "WATERHEATER_HEATPUMP_WRAPPEDCONDENSER",
    "THERMALSTORAGE_ICE_SIMPLE",
    "THERMALSTORAGE_ICE_DETAILED",
    "THERMALSTORAGE_CHILLEDWATER_MIXED",
    "THERMALSTORAGE_CHILLEDWATER_STRATIFIED",
    "PLANTLOOP",
    "CONDENSERLOOP",
    "PLANTEQUIPMENTLIST",
    "CONDENSEREQUIPMENTLIST",
    "PLANTEQUIPMENTOPERATION_UNCONTROLLED",
    "PLANTEQUIPMENTOPERATION_COOLINGLOAD",
    "PLANTEQUIPMENTOPERATION_HEATINGLOAD",
    "PLANTEQUIPMENTOPERATION_OUTDOORDRYBULB",
    "PLANTEQUIPMENTOPERATION_OUTDOORWETBULB",
    "PLANTEQUIPMENTOPERATION_OUTDOORRELATIVEHUMIDITY",
    "PLANTEQUIPMENTOPERATION_OUTDOORDEWPOINT",
    "PLANTEQUIPMENTOPERATION_COMPONENTSETPOINT",
    "PLANTEQUIPMENTOPERATION_THERMALENERGYSTORAGE",
    "PLANTEQUIPMENTOPERATION_OUTDOORDRYBULBDIFFERENCE",
    "PLANTEQUIPMENTOPERATION_OUTDOORWETBULBDIFFERENCE",
    "PLANTEQUIPMENTOPERATION_OUTDOORDEWPOINTDIFFERENCE",
    "PLANTEQUIPMENTOPERATIONSCHEMES",
    "CONDENSEREQUIPMENTOPERATIONSCHEMES",
    "ENERGYMANAGEMENTSYSTEM_SENSOR",
    "ENERGYMANAGEMENTSYSTEM_ACTUATOR",
    "ENERGYMANAGEMENTSYSTEM_PROGRAMCALLINGMANAGER",
    "ENERGYMANAGEMENTSYSTEM_PROGRAM",
    "ENERGYMANAGEMENTSYSTEM_SUBROUTINE",
    "ENERGYMANAGEMENTSYSTEM_GLOBALVARIABLE",
    "ENERGYMANAGEMENTSYSTEM_OUTPUTVARIABLE",
    "ENERGYMANAGEMENTSYSTEM_METEREDOUTPUTVARIABLE",
    "ENERGYMANAGEMENTSYSTEM_TRENDVARIABLE",
    "ENERGYMANAGEMENTSYSTEM_INTERNALVARIABLE",
    "ENERGYMANAGEMENTSYSTEM_CURVEORTABLEINDEXVARIABLE",
    "ENERGYMANAGEMENTSYSTEM_CONSTRUCTIONINDEXVARIABLE",
    "EXTERNALINTERFACE",
    "EXTERNALINTERFACE_SCHEDULE",
    "EXTERNALINTERFACE_VARIABLE",
    "EXTERNALINTERFACE_ACTUATOR",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_FROM_VARIABLE",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_TO_SCHEDULE",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_TO_ACTUATOR",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_TO_VARIABLE",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_FROM_VARIABLE",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_TO_SCHEDULE",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_TO_ACTUATOR",
    "EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_TO_VARIABLE",
    "ZONEHVAC_FORCEDAIR_USERDEFINED",
    "AIRTERMINAL_SINGLEDUCT_USERDEFINED",
    "COIL_USERDEFINED",
    "PLANTCOMPONENT_USERDEFINED",
    "PLANTEQUIPMENTOPERATION_USERDEFINED",
    "AVAILABILITYMANAGER_SCHEDULED",
    "AVAILABILITYMANAGER_SCHEDULEDON",
    "AVAILABILITYMANAGER_SCHEDULEDOFF",
    "AVAILABILITYMANAGER_OPTIMUMSTART",
    "AVAILABILITYMANAGER_NIGHTCYCLE",
    "AVAILABILITYMANAGER_DIFFERENTIALTHERMOSTAT",
    "AVAILABILITYMANAGER_HIGHTEMPERATURETURNOFF",
    "AVAILABILITYMANAGER_HIGHTEMPERATURETURNON",
    "AVAILABILITYMANAGER_LOWTEMPERATURETURNOFF",
    "AVAILABILITYMANAGER_LOWTEMPERATURETURNON",
    "AVAILABILITYMANAGER_NIGHTVENTILATION",
    "AVAILABILITYMANAGER_HYBRIDVENTILATION",
    "AVAILABILITYMANAGERASSIGNMENTLIST",
    "SETPOINTMANAGER_SCHEDULED",
    "SETPOINTMANAGER_SCHEDULED_DUALSETPOINT",
    "SETPOINTMANAGER_OUTDOORAIRRESET",
    "SETPOINTMANAGER_SINGLEZONE_REHEAT",
    "SETPOINTMANAGER_SINGLEZONE_HEATING",
    "SETPOINTMANAGER_SINGLEZONE_COOLING",
    "SETPOINTMANAGER_SINGLEZONE_HUMIDITY_MINIMUM",
    "SETPOINTMANAGER_SINGLEZONE_HUMIDITY_MAXIMUM",
    "SETPOINTMANAGER_MIXEDAIR",
    "SETPOINTMANAGER_OUTDOORAIRPRETREAT",
    "SETPOINTMANAGER_WARMEST",
    "SETPOINTMANAGER_COLDEST",
    "SETPOINTMANAGER_RETURNAIRBYPASSFLOW",
    "SETPOINTMANAGER_WARMESTTEMPERATUREFLOW",
    "SETPOINTMANAGER_MULTIZONE_HEATING_AVERAGE",
    "SETPOINTMANAGER_MULTIZONE_COOLING_AVERAGE",
    "SETPOINTMANAGER_MULTIZONE_MINIMUMHUMIDITY_AVERAGE",
    "SETPOINTMANAGER_MULTIZONE_MAXIMUMHUMIDITY_AVERAGE",
    "SETPOINTMANAGER_MULTIZONE_HUMIDITY_MINIMUM",
    "SETPOINTMANAGER_MULTIZONE_HUMIDITY_MAXIMUM",
    "SETPOINTMANAGER_FOLLOWOUTDOORAIRTEMPERATURE",
    "SETPOINTMANAGER_FOLLOWSYSTEMNODETEMPERATURE",
    "SETPOINTMANAGER_FOLLOWGROUNDTEMPERATURE",
    "SETPOINTMANAGER_CONDENSERENTERINGRESET",
    "SETPOINTMANAGER_CONDENSERENTERINGRESET_IDEAL",
    "SETPOINTMANAGER_SINGLEZONE_ONESTAGECOOLING",
    "SETPOINTMANAGER_SINGLEZONE_ONESTAGEHEATING",
    "SETPOINTMANAGER_RETURNTEMPERATURE_CHILLEDWATER",
    "SETPOINTMANAGER_RETURNTEMPERATURE_HOTWATER",
    "SETPOINTMANAGER_SYSTEMNODERESET_TEMPERATURE",
    "SETPOINTMANAGER_SYSTEMNODERESET_HUMIDITY",
    "REFRIGERATION_CASE",
    "REFRIGERATION_COMPRESSORRACK",
    "REFRIGERATION_CASEANDWALKINLIST",
    "REFRIGERATION_CONDENSER_AIRCOOLED",
    "REFRIGERATION_CONDENSER_EVAPORATIVECOOLED",
    "REFRIGERATION_CONDENSER_WATERCOOLED",
    "REFRIGERATION_CONDENSER_CASCADE",
    "REFRIGERATION_GASCOOLER_AIRCOOLED",
    "REFRIGERATION_TRANSFERLOADLIST",
    "REFRIGERATION_SUBCOOLER",
    "REFRIGERATION_COMPRESSOR",
    "REFRIGERATION_COMPRESSORLIST",
    "REFRIGERATION_SYSTEM",
    "REFRIGERATION_TRANSCRITICALSYSTEM",
    "REFRIGERATION_SECONDARYSYSTEM",
    "REFRIGERATION_WALKIN",
    "REFRIGERATION_AIRCHILLER",
    "ZONEHVAC_REFRIGERATIONCHILLERSET",
    "DEMANDMANAGERASSIGNMENTLIST",
    "DEMANDMANAGER_EXTERIORLIGHTS",
    "DEMANDMANAGER_LIGHTS",
    "DEMANDMANAGER_ELECTRICEQUIPMENT",
    "DEMANDMANAGER_THERMOSTATS",
    "DEMANDMANAGER_VENTILATION",
    "GENERATOR_INTERNALCOMBUSTIONENGINE",
    "GENERATOR_COMBUSTIONTURBINE",
    "GENERATOR_MICROTURBINE",
    "GENERATOR_PHOTOVOLTAIC",
    "PHOTOVOLTAICPERFORMANCE_SIMPLE",
    "PHOTOVOLTAICPERFORMANCE_EQUIVALENTONEDIODE",
    "PHOTOVOLTAICPERFORMANCE_SANDIA",
    "GENERATOR_PVWATTS",
    "ELECTRICLOADCENTER_INVERTER_PVWATTS",
    "GENERATOR_FUELCELL",
    "GENERATOR_FUELCELL_POWERMODULE",
    "GENERATOR_FUELCELL_AIRSUPPLY",
    "GENERATOR_FUELCELL_WATERSUPPLY",
    "GENERATOR_FUELCELL_AUXILIARYHEATER",
    "GENERATOR_FUELCELL_EXHAUSTGASTOWATERHEATEXCHANGER",
    "GENERATOR_FUELCELL_ELECTRICALSTORAGE",
    "GENERATOR_FUELCELL_INVERTER",
    "GENERATOR_FUELCELL_STACKCOOLER",
    "GENERATOR_MICROCHP",
    "GENERATOR_MICROCHP_NONNORMALIZEDPARAMETERS",
    "GENERATOR_FUELSUPPLY",
    "GENERATOR_WINDTURBINE",
    "ELECTRICLOADCENTER_GENERATORS",
    "ELECTRICLOADCENTER_INVERTER_SIMPLE",
    "ELECTRICLOADCENTER_INVERTER_FUNCTIONOFPOWER",
    "ELECTRICLOADCENTER_INVERTER_LOOKUPTABLE",
    "ELECTRICLOADCENTER_STORAGE_SIMPLE",
    "ELECTRICLOADCENTER_STORAGE_BATTERY",
    "ELECTRICLOADCENTER_STORAGE_LIIONNMCBATTERY",
    "ELECTRICLOADCENTER_TRANSFORMER",
    "ELECTRICLOADCENTER_DISTRIBUTION",
    "ELECTRICLOADCENTER_STORAGE_CONVERTER",
    "WATERUSE_EQUIPMENT",
    "WATERUSE_CONNECTIONS",
    "WATERUSE_STORAGE",
    "WATERUSE_WELL",
    "WATERUSE_RAINCOLLECTOR",
    "FAULTMODEL_TEMPERATURESENSOROFFSET_OUTDOORAIR",
    "FAULTMODEL_HUMIDITYSENSOROFFSET_OUTDOORAIR",
    "FAULTMODEL_ENTHALPYSENSOROFFSET_OUTDOORAIR",
    "FAULTMODEL_TEMPERATURESENSOROFFSET_RETURNAIR",
    "FAULTMODEL_ENTHALPYSENSOROFFSET_RETURNAIR",
    "FAULTMODEL_TEMPERATURESENSOROFFSET_CHILLERSUPPLYWATER",
    "FAULTMODEL_TEMPERATURESENSOROFFSET_COILSUPPLYAIR",
    "FAULTMODEL_TEMPERATURESENSOROFFSET_CONDENSERSUPPLYWATER",
    "FAULTMODEL_THERMOSTATOFFSET",
    "FAULTMODEL_HUMIDISTATOFFSET",
    "FAULTMODEL_FOULING_AIRFILTER",
    "FAULTMODEL_FOULING_BOILER",
    "FAULTMODEL_FOULING_EVAPORATIVECOOLER",
    "FAULTMODEL_FOULING_CHILLER",
    "FAULTMODEL_FOULING_COOLINGTOWER",
    "FAULTMODEL_FOULING_COIL",
    "MATRIX_TWODIMENSION",
    "HYBRIDMODEL_ZONE",
    "CURVE_LINEAR",
    "CURVE_QUADLINEAR",
    "CURVE_QUINTLINEAR",
    "CURVE_QUADRATIC",
    "CURVE_CUBIC",
    "CURVE_QUARTIC",
    "CURVE_EXPONENT",
    "CURVE_BICUBIC",
    "CURVE_BIQUADRATIC",
    "CURVE_QUADRATICLINEAR",
    "CURVE_CUBICLINEAR",
    "CURVE_TRIQUADRATIC",
    "CURVE_FUNCTIONAL_PRESSUREDROP",
    "CURVE_FANPRESSURERISE",
    "CURVE_EXPONENTIALSKEWNORMAL",
    "CURVE_SIGMOID",
    "CURVE_RECTANGULARHYPERBOLA1",
    "CURVE_RECTANGULARHYPERBOLA2",
    "CURVE_EXPONENTIALDECAY",
    "CURVE_DOUBLEEXPONENTIALDECAY",
    "CURVE_CHILLERPARTLOADWITHLIFT",
    "TABLE_INDEPENDENTVARIABLE",
    "TABLE_INDEPENDENTVARIABLELIST",
    "TABLE_LOOKUP",
    "FLUIDPROPERTIES_NAME",
    "FLUIDPROPERTIES_GLYCOLCONCENTRATION",
    "FLUIDPROPERTIES_TEMPERATURES",
    "FLUIDPROPERTIES_SATURATED",
    "FLUIDPROPERTIES_SUPERHEATED",
    "FLUIDPROPERTIES_CONCENTRATION",
    "CURRENCYTYPE",
    "COMPONENTCOST_ADJUSTMENTS",
    "COMPONENTCOST_REFERENCE",
    "COMPONENTCOST_LINEITEM",
    "UTILITYCOST_TARIFF",
    "UTILITYCOST_QUALIFY",
    "UTILITYCOST_CHARGE_SIMPLE",
    "UTILITYCOST_CHARGE_BLOCK",
    "UTILITYCOST_RATCHET",
    "UTILITYCOST_VARIABLE",
    "UTILITYCOST_COMPUTATION",
    "LIFECYCLECOST_PARAMETERS",
    "LIFECYCLECOST_RECURRINGCOSTS",
    "LIFECYCLECOST_NONRECURRINGCOST",
    "LIFECYCLECOST_USEPRICEESCALATION",
    "LIFECYCLECOST_USEADJUSTMENT",
    "PARAMETRIC_SETVALUEFORRUN",
    "PARAMETRIC_LOGIC",
    "PARAMETRIC_RUNCONTROL",
    "PARAMETRIC_FILENAMESUFFIX",
    "OUTPUT_VARIABLEDICTIONARY",
    "OUTPUT_SURFACES_LIST",
    "OUTPUT_SURFACES_DRAWING",
    "OUTPUT_SCHEDULES",
    "OUTPUT_CONSTRUCTIONS",
    "OUTPUT_ENERGYMANAGEMENTSYSTEM",
    "OUTPUTCONTROL_SURFACECOLORSCHEME",
    "OUTPUT_TABLE_SUMMARYREPORTS",
    "OUTPUT_TABLE_TIMEBINS",
    "OUTPUT_TABLE_MONTHLY",
    "OUTPUT_TABLE_ANNUAL",
    "OUTPUTCONTROL_TABLE_STYLE",
    "OUTPUTCONTROL_REPORTINGTOLERANCES",
    "OUTPUT_VARIABLE",
    "OUTPUT_METER",
    "OUTPUT_METER_METERFILEONLY",
    "OUTPUT_METER_CUMULATIVE",
    "OUTPUT_METER_CUMULATIVE_METERFILEONLY",
    "METER_CUSTOM",
    "METER_CUSTOMDECREMENT",
    "OUTPUTCONTROL_FILES",
    "OUTPUTCONTROL_TIMESTAMP",
    "OUTPUT_JSON",
    "OUTPUT_SQLITE",
    "OUTPUT_ENVIRONMENTALIMPACTFACTORS",
    "ENVIRONMENTALIMPACTFACTORS",
    "FUELFACTORS",
    "OUTPUT_DIAGNOSTICS",
    "OUTPUT_DEBUGGINGDATA",
    "OUTPUT_PREPROCESSORMESSAGE",
    "PYTHONPLUGIN_SEARCHPATHS",
    "PYTHONPLUGIN_INSTANCE",
    "PYTHONPLUGIN_VARIABLES",
    "PYTHONPLUGIN_TRENDVARIABLE",
    "PYTHONPLUGIN_OUTPUTVARIABLE",
    "OS_VERSION",
    "OS_COMPONENTDATA",
    "OS_CONVERGENCELIMITS",
    "OS_PROGRAMCONTROL",
    "OS_HEATBALANCEALGORITHM",
    "OS_RUNPERIOD",
    "OS_RUNPERIODCONTROL_DAYLIGHTSAVINGTIME",
    "OS_RUNPERIODCONTROL_SPECIALDAYS",
    "OS_SHADOWCALCULATION",
    "OS_SIMULATIONCONTROL",
    "OS_PERFORMANCEPRECISIONTRADEOFFS",
    "OS_SIZING_PARAMETERS",
    "OS_SURFACECONVECTIONALGORITHM_INSIDE",
    "OS_SURFACECONVECTIONALGORITHM_OUTSIDE",
    "OS_TIMESTEP",
    "OS_YEARDESCRIPTION",
    "OS_ZONEAIRCONTAMINANTBALANCE",
    "OS_ZONEAIRHEATBALANCEALGORITHM",
    "OS_ZONEAIRMASSFLOWCONSERVATION",
    "OS_ZONECAPACITANCEMULTIPLIER_RESEARCHSPECIAL",
    "OS_RADIANCEPARAMETERS",
    "OS_CLIMATEZONES",
    "OS_SITE",
    "OS_SITE_GROUNDTEMPERATURE_BUILDINGSURFACE",
    "OS_SITE_GROUNDTEMPERATURE_FCFACTORMETHOD",
    "OS_SITE_GROUNDTEMPERATURE_SHALLOW",
    "OS_SITE_GROUNDTEMPERATURE_DEEP",
    "OS_SITE_GROUNDTEMPERATURE_UNDISTURBED_KUSUDAACHENBACH",
    "OS_SITE_GROUNDTEMPERATURE_UNDISTURBED_XING",
    "OS_SITE_GROUNDREFLECTANCE",
    "OS_SITE_WATERMAINSTEMPERATURE",
    "OS_SIZINGPERIOD_DESIGNDAY",
    "OS_SIZINGPERIOD_WEATHERFILECONDITIONTYPE",
    "OS_SIZINGPERIOD_WEATHERFILEDAYS",
    "OS_WEATHERFILE",
    "OS_WEATHERPROPERTY_SKYTEMPERATURE",
    "OS_ADDITIONALPROPERTIES",
    "OS_BUILDINGSTORY",
    "OS_BUILDINGUNIT",
    "OS_DEFAULTCONSTRUCTIONSET",
    "OS_DEFAULTSCHEDULESET",
    "OS_DEFAULTSUBSURFACECONSTRUCTIONS",
    "OS_DEFAULTSURFACECONSTRUCTIONS",
    "OS_RENDERING_COLOR",
    "OS_SPACETYPE",
    "OS_MODELOBJECTLIST",
    "OS_MATERIAL",
    "OS_MATERIAL_AIRGAP",
    "OS_MATERIAL_INFRAREDTRANSPARENT",
    "OS_MATERIAL_NOMASS",
    "OS_MATERIAL_ROOFVEGETATION",
    "OS_WINDOWMATERIAL_BLIND",
    "OS_WINDOWMATERIAL_DAYLIGHTREDIRECTIONDEVICE",
    "OS_WINDOWMATERIAL_GAS",
    "OS_WINDOWMATERIAL_GASMIXTURE",
    "OS_WINDOWMATERIAL_GLAZING",
    "OS_WINDOWMATERIAL_GLAZINGGROUP_THERMOCHROMIC",
    "OS_WINDOWMATERIAL_GLAZING_REFRACTIONEXTINCTIONMETHOD",
    "OS_WINDOWMATERIAL_SCREEN",
    "OS_WINDOWMATERIAL_SHADE",
    "OS_WINDOWMATERIAL_SIMPLEGLAZINGSYSTEM",
    "OS_STANDARDSINFORMATION_MATERIAL",
    "OS_MATERIALPROPERTY_GLAZINGSPECTRALDATA",
    "OS_MATERIALPROPERTY_MOISTUREPENETRATIONDEPTH_SETTINGS",
    "OS_MATERIALPROPERTY_PHASECHANGE",
    "OS_MATERIALPROPERTY_PHASECHANGEHYSTERESIS",
    "OS_CONSTRUCTION",
    "OS_CONSTRUCTION_CFACTORUNDERGROUNDWALL",
    "OS_CONSTRUCTION_FFACTORGROUNDFLOOR",
    "OS_CONSTRUCTION_INTERNALSOURCE",
    "OS_CONSTRUCTION_AIRBOUNDARY",
    "OS_CONSTRUCTION_WINDOWDATAFILE",
    "OS_STANDARDSINFORMATION_CONSTRUCTION",
    "OS_INTERNALMASS_DEFINITION",
    "OS_PEOPLE_DEFINITION",
    "OS_LIGHTS_DEFINITION",
    "OS_LUMINAIRE_DEFINITION",
    "OS_ELECTRICEQUIPMENT_DEFINITION",
    "OS_ELECTRICEQUIPMENT_ITE_AIRCOOLED_DEFINITION",
    "OS_GASEQUIPMENT_DEFINITION",
    "OS_HOTWATEREQUIPMENT_DEFINITION",
    "OS_STEAMEQUIPMENT_DEFINITION",
    "OS_OTHEREQUIPMENT_DEFINITION",
    "OS_EXTERIOR_LIGHTS_DEFINITION",
    "OS_EXTERIOR_FUELEQUIPMENT_DEFINITION",
    "OS_EXTERIOR_WATEREQUIPMENT_DEFINITION",
    "OS_SCHEDULE_COMPACT",
    "OS_SCHEDULE_DAY",
    "OS_SCHEDULE_WEEK",
    "OS_SCHEDULE_YEAR",
    "OS_SCHEDULE_RULE",
    "OS_SCHEDULE_RULESET",
    "OS_SCHEDULE_CONSTANT",
    "OS_SCHEDULE_FIXEDINTERVAL",
    "OS_SCHEDULE_VARIABLEINTERVAL",
    "OS_SCHEDULETYPELIMITS",
    "OS_EXTERNAL_FILE",
    "OS_SCHEDULE_FILE",
    "OS_BUILDING",
    "OS_DAYLIGHTINGDEVICE_TUBULAR",
    "OS_DAYLIGHTINGDEVICE_SHELF",
    "OS_DAYLIGHTINGDEVICE_LIGHTWELL",
    "OS_FACILITY",
    "OS_INTERIORPARTITIONSURFACE",
    "OS_INTERIORPARTITIONSURFACEGROUP",
    "OS_SHADINGSURFACE",
    "OS_SHADINGSURFACEGROUP",
    "OS_SPACE",
    "OS_SURFACE",
    "OS_SUBSURFACE",
    "OS_SHADINGCONTROL",
    "OS_WINDOWPROPERTY_FRAMEANDDIVIDER",
    "OS_SURFACECONTROL_MOVABLEINSULATION",
    "OS_SURFACEPROPERTY_CONVECTIONCOEFFICIENTS",
    "OS_SURFACEPROPERTY_CONVECTIONCOEFFICIENTS_MULTIPLESURFACE",
    "OS_SURFACEPROPERTY_EXPOSEDFOUNDATIONPERIMETER",
    "OS_SURFACEPROPERTY_OTHERSIDECOEFFICIENTS",
    "OS_SURFACEPROPERTY_OTHERSIDECONDITIONSMODEL",
    "OS_FOUNDATION_KIVA",
    "OS_FOUNDATION_KIVA_SETTINGS",
    "OS_ZONEPROPERTY_USERVIEWFACTORS_BYSURFACENAME",
    "OS_SURFACEPROPERTY_SURROUNDINGSURFACES",
    "OS_SURFACEPROPERTY_LOCALENVIRONMENT",
    "OS_SURFACEPROPERTY_GROUNDSURFACES",
    "OS_SURFACEPROPERTY_INCIDENTSOLARMULTIPLIER",
    "OS_INTERNALMASS",
    "OS_PEOPLE",
    "OS_LIGHTS",
    "OS_LUMINAIRE",
    "OS_ELECTRICEQUIPMENT",
    "OS_ELECTRICEQUIPMENT_ITE_AIRCOOLED",
    "OS_GASEQUIPMENT",
    "OS_HOTWATEREQUIPMENT",
    "OS_STEAMEQUIPMENT",
    "OS_OTHEREQUIPMENT",
    "OS_SPACEINFILTRATION_DESIGNFLOWRATE",
    "OS_SPACEINFILTRATION_EFFECTIVELEAKAGEAREA",
    "OS_SPACEINFILTRATION_FLOWCOEFFICIENT",
    "OS_EXTERIOR_LIGHTS",
    "OS_EXTERIOR_FUELEQUIPMENT",
    "OS_EXTERIOR_WATEREQUIPMENT",
    "OS_DAYLIGHTING_CONTROL",
    "OS_GLARE_SENSOR",
    "OS_ILLUMINANCEMAP",
    "OS_LIGHTINGDESIGNDAY",
    "OS_LIGHTINGSIMULATIONCONTROL",
    "OS_LIGHTINGSIMULATIONZONE",
    "OS_REFRIGERATION_CASE",
    "OS_REFRIGERATION_CONDENSER_AIRCOOLED",
    "OS_REFRIGERATION_CONDENSER_EVAPORATIVECOOLED",
    "OS_REFRIGERATION_CONDENSER_WATERCOOLED",
    "OS_REFRIGERATION_CONDENSER_CASCADE",
    "OS_REFRIGERATION_COMPRESSOR",
    "OS_REFRIGERATION_COMPRESSORRACK",
    "OS_REFRIGERATION_DEFROSTCYCLEPARAMETERS",
    "OS_REFRIGERATION_GASCOOLER_AIRCOOLED",
    "OS_REFRIGERATION_SUBCOOLER_MECHANICAL",
    "OS_REFRIGERATION_SUBCOOLER_LIQUIDSUCTION",
    "OS_REFRIGERATION_SYSTEM",
    "OS_REFRIGERATION_TRANSCRITICALSYSTEM",
    "OS_REFRIGERATION_SECONDARYSYSTEM",
    "OS_REFRIGERATION_WALKIN",
    "OS_REFRIGERATION_WALKIN_ZONEBOUNDARY",
    "OS_REFRIGERATION_AIRCHILLER",
    "OS_AIRCONDITIONER_VARIABLEREFRIGERANTFLOW",
    "OS_AIRCONDITIONER_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL",
    "OS_AIRCONDITIONER_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL_LOADINGINDEX",
    "OS_AIRCONDITIONER_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL_HR",
    "OS_AIRLOOPHVAC",
    "OS_AIRLOOPHVAC_OUTDOORAIRSYSTEM",
    "OS_AIRLOOPHVAC_DEDICATEDOUTDOORAIRSYSTEM",
    "OS_AIRLOOPHVAC_RETURNPLENUM",
    "OS_AIRLOOPHVAC_SUPPLYPLENUM",
    "OS_AIRLOOPHVAC_UNITARYHEATPUMP_AIRTOAIR_MULTISPEED",
    "OS_AIRLOOPHVAC_UNITARYCOOLONLY",
    "OS_AIRLOOPHVAC_UNITARYHEATPUMP_AIRTOAIR",
    "OS_AIRLOOPHVAC_UNITARYHEATCOOL_VAVCHANGEOVERBYPASS",
    "OS_AIRLOOPHVAC_UNITARYSYSTEM",
    "OS_UNITARYSYSTEMPERFORMANCE_MULTISPEED",
    "OS_AIRLOOPHVAC_ZONEMIXER",
    "OS_AIRLOOPHVAC_ZONESPLITTER",
    "OS_AIRTERMINAL_SINGLEDUCT_VAV_HEATANDCOOL_NOREHEAT",
    "OS_AIRTERMINAL_SINGLEDUCT_VAV_HEATANDCOOL_REHEAT",
    "OS_AIRTERMINAL_SINGLEDUCT_INLETSIDEMIXER",
    "OS_AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_REHEAT",
    "OS_AIRTERMINAL_SINGLEDUCT_SERIESPIU_REHEAT",
    "OS_AIRTERMINAL_SINGLEDUCT_PARALLELPIU_REHEAT",
    "OS_AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_FOURPIPEINDUCTION",
    "OS_AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_FOURPIPEBEAM",
    "OS_AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_COOLEDBEAM",
    "OS_AIRTERMINAL_SINGLEDUCT_CONSTANTVOLUME_NOREHEAT",
    "OS_AIRTERMINAL_SINGLEDUCT_VAV_NOREHEAT",
    "OS_AIRTERMINAL_SINGLEDUCT_VAV_REHEAT",
    "OS_AIRTERMINAL_DUALDUCT_VAV_OUTDOORAIR",
    "OS_AIRTERMINAL_DUALDUCT_CONSTANTVOLUME",
    "OS_AIRTERMINAL_DUALDUCT_VAV",
    "OS_AVAILABILITYMANAGERASSIGNMENTLIST",
    "OS_AVAILABILITYMANAGER_SCHEDULED",
    "OS_AVAILABILITYMANAGER_SCHEDULEDON",
    "OS_AVAILABILITYMANAGER_SCHEDULEDOFF",
    "OS_AVAILABILITYMANAGER_NIGHTCYCLE",
    "OS_AVAILABILITYMANAGER_OPTIMUMSTART",
    "OS_AVAILABILITYMANAGER_DIFFERENTIALTHERMOSTAT",
    "OS_AVAILABILITYMANAGER_HIGHTEMPERATURETURNOFF",
    "OS_AVAILABILITYMANAGER_HIGHTEMPERATURETURNON",
    "OS_AVAILABILITYMANAGER_LOWTEMPERATURETURNOFF",
    "OS_AVAILABILITYMANAGER_LOWTEMPERATURETURNON",
    "OS_AVAILABILITYMANAGER_NIGHTVENTILATION",
    "OS_AVAILABILITYMANAGER_HYBRIDVENTILATION",
    "OS_BOILER_HOTWATER",
    "OS_BOILER_STEAM",
    "OS_HEATPUMP_WATERTOWATER_EQUATIONFIT_HEATING",
    "OS_HEATPUMP_WATERTOWATER_EQUATIONFIT_COOLING",
    "OS_HEATPUMP_PLANTLOOP_EIR_HEATING",
    "OS_HEATPUMP_PLANTLOOP_EIR_COOLING",
    "OS_HEATPUMP_AIRTOWATER_FUELFIRED_HEATING",
    "OS_HEATPUMP_AIRTOWATER_FUELFIRED_COOLING",
    "OS_CHILLER_ELECTRIC_EIR",
    "OS_CHILLER_ELECTRIC_REFORMULATEDEIR",
    "OS_CHILLER_ELECTRIC_ASHRAE205",
    "OS_CENTRALHEATPUMPSYSTEM",
    "OS_CENTRALHEATPUMPSYSTEM_MODULE",
    "OS_CHILLERHEATERPERFORMANCE_ELECTRIC_EIR",
    "OS_CHILLER_ABSORPTION_INDIRECT",
    "OS_CHILLER_ABSORPTION",
    "OS_COOLINGTOWER_SINGLESPEED",
    "OS_COOLINGTOWER_TWOSPEED",
    "OS_COOLINGTOWER_VARIABLESPEED",
    "OS_COOLINGTOWERPERFORMANCE_COOLTOOLS",
    "OS_COOLINGTOWERPERFORMANCE_YORKCALC",
    "OS_DISTRICTCOOLING",
    "OS_DISTRICTHEATING",
    "OS_GROUNDHEATEXCHANGER_VERTICAL",
    "OS_GROUNDHEATEXCHANGER_HORIZONTALTRENCH",
    "OS_COIL_COOLING_COOLEDBEAM",
    "OS_COIL_COOLING_FOURPIPEBEAM",
    "OS_COIL_HEATING_FOURPIPEBEAM",
    "OS_COIL_COOLING_DX",
    "OS_COIL_COOLING_DX_CURVEFIT_PERFORMANCE",
    "OS_COIL_COOLING_DX_CURVEFIT_OPERATINGMODE",
    "OS_COIL_COOLING_DX_CURVEFIT_SPEED",
    "OS_COIL_COOLING_DX_SINGLESPEED",
    "OS_COIL_COOLING_DX_TWOSTAGEWITHHUMIDITYCONTROLMODE",
    "OS_COILPERFORMANCE_DX_COOLING",
    "OS_COIL_COOLING_DX_MULTISPEED",
    "OS_COIL_COOLING_DX_MULTISPEED_STAGEDATA",
    "OS_COIL_COOLING_DX_TWOSPEED",
    "OS_COILSYSTEM_COOLING_WATER_HEATEXCHANGERASSISTED",
    "OS_COILSYSTEM_COOLING_DX_HEATEXCHANGERASSISTED",
    "OS_COILSYSTEM_INTEGRATEDHEATPUMP_AIRSOURCE",
    "OS_COIL_COOLING_DX_VARIABLEREFRIGERANTFLOW",
    "OS_COIL_COOLING_DX_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL",
    "OS_COIL_COOLING_LOWTEMPERATURERADIANT_CONSTANTFLOW",
    "OS_COIL_COOLING_LOWTEMPERATURERADIANT_VARIABLEFLOW",
    "OS_COIL_COOLING_WATER",
    "OS_COIL_COOLING_WATERTOAIRHEATPUMP_EQUATIONFIT",
    "OS_COIL_HEATING_DX_SINGLESPEED",
    "OS_COIL_HEATING_DX_VARIABLEREFRIGERANTFLOW",
    "OS_COIL_HEATING_DX_VARIABLEREFRIGERANTFLOW_FLUIDTEMPERATURECONTROL",
    "OS_COIL_HEATING_ELECTRIC",
    "OS_COIL_HEATING_ELECTRIC_MULTISTAGE",
    "OS_COIL_HEATING_ELECTRIC_MULTISTAGE_STAGEDATA",
    "OS_COIL_HEATING_GAS",
    "OS_COIL_HEATING_GAS_MULTISTAGE",
    "OS_COIL_HEATING_GAS_MULTISTAGE_STAGEDATA",
    "OS_COIL_HEATING_DESUPERHEATER",
    "OS_COIL_WATERHEATING_DESUPERHEATER",
    "OS_COIL_HEATING_LOWTEMPERATURERADIANT_CONSTANTFLOW",
    "OS_COIL_HEATING_LOWTEMPERATURERADIANT_VARIABLEFLOW",
    "OS_COIL_HEATING_WATER",
    "OS_COIL_HEATING_WATER_BASEBOARD",
    "OS_COIL_HEATING_WATER_BASEBOARD_RADIANT",
    "OS_COIL_HEATING_WATERTOAIRHEATPUMP_EQUATIONFIT",
    "OS_COIL_HEATING_DX_MULTISPEED",
    "OS_COIL_HEATING_DX_MULTISPEED_STAGEDATA",
    "OS_COIL_HEATING_DX_VARIABLESPEED",
    "OS_COIL_HEATING_DX_VARIABLESPEED_SPEEDDATA",
    "OS_COIL_COOLING_DX_VARIABLESPEED",
    "OS_COIL_COOLING_DX_VARIABLESPEED_SPEEDDATA",
    "OS_COIL_HEATING_WATERTOAIRHEATPUMP_VARIABLESPEEDEQUATIONFIT",
    "OS_COIL_HEATING_WATERTOAIRHEATPUMP_VARIABLESPEEDEQUATIONFIT_SPEEDDATA",
    "OS_COIL_COOLING_WATERTOAIRHEATPUMP_VARIABLESPEEDEQUATIONFIT",
    "OS_COIL_COOLING_WATERTOAIRHEATPUMP_VARIABLESPEEDEQUATIONFIT_SPEEDDATA",
    "OS_COIL_COOLING_WATER_PANEL_RADIANT",
    "OS_CONNECTION",
    "OS_CONNECTOR_MIXER",
    "OS_CONNECTOR_SPLITTER",
    "OS_CONTROLLER_MECHANICALVENTILATION",
    "OS_ZONECONTROL_CONTAMINANTCONTROLLER",
    "OS_CONTROLLER_OUTDOORAIR",
    "OS_CONTROLLER_WATERCOIL",
    "OS_CURVE_LINEAR",
    "OS_CURVE_QUADLINEAR",
    "OS_CURVE_QUINTLINEAR",
    "OS_CURVE_QUADRATIC",
    "OS_CURVE_CUBIC",
    "OS_CURVE_QUARTIC",
    "OS_CURVE_EXPONENT",
    "OS_CURVE_BICUBIC",
    "OS_CURVE_BIQUADRATIC",
    "OS_CURVE_QUADRATICLINEAR",
    "OS_CURVE_TRIQUADRATIC",
    "OS_CURVE_FUNCTIONAL_PRESSUREDROP",
    "OS_CURVE_FANPRESSURERISE",
    "OS_CURVE_EXPONENTIALSKEWNORMAL",
    "OS_CURVE_SIGMOID",
    "OS_CURVE_RECTANGULARHYPERBOLA1",
    "OS_CURVE_RECTANGULARHYPERBOLA2",
    "OS_CURVE_EXPONENTIALDECAY",
    "OS_CURVE_DOUBLEEXPONENTIALDECAY",
    "OS_TABLE_MULTIVARIABLELOOKUP",
    "OS_TABLE_INDEPENDENTVARIABLE",
    "OS_TABLE_LOOKUP",
    "OS_DESIGNSPECIFICATION_OUTDOORAIR",
    "OS_DESIGNSPECIFICATION_ZONEAIRDISTRIBUTION",
    "OS_HUMIDIFIER_STEAM_ELECTRIC",
    "OS_HUMIDIFIER_STEAM_GAS",
    "OS_EVAPORATIVECOOLER_INDIRECT_RESEARCHSPECIAL",
    "OS_EVAPORATIVECOOLER_DIRECT_RESEARCHSPECIAL",
    "OS_EVAPORATIVEFLUIDCOOLER_SINGLESPEED",
    "OS_EVAPORATIVEFLUIDCOOLER_TWOSPEED",
    "OS_FAN_SYSTEMMODEL",
    "OS_FAN_CONSTANTVOLUME",
    "OS_FAN_VARIABLEVOLUME",
    "OS_FAN_ONOFF",
    "OS_FAN_COMPONENTMODEL",
    "OS_FAN_ZONEEXHAUST",
    "OS_HEATEXCHANGER_AIRTOAIR_SENSIBLEANDLATENT",
    "OS_HEATEXCHANGER_DESICCANT_BALANCEDFLOW",
    "OS_HEATEXCHANGER_DESICCANT_BALANCEDFLOW_PERFORMANCEDATATYPE1",
    "OS_HEATEXCHANGER_FLUIDTOFLUID",
    "OS_HVACCOMPONENTLIST",
    "OS_NODE",
    "OS_PORTLIST",
    "OS_PIPE_ADIABATIC",
    "OS_PLANTLOOP",
    "OS_PLANTEQUIPMENTOPERATION_COOLINGLOAD",
    "OS_PLANTEQUIPMENTOPERATION_HEATINGLOAD",
    "OS_PLANTEQUIPMENTOPERATION_OUTDOORDRYBULB",
    "OS_PLANTEQUIPMENTOPERATION_OUTDOORWETBULB",
    "OS_PLANTEQUIPMENTOPERATION_OUTDOORRELATIVEHUMIDITY",
    "OS_PLANTEQUIPMENTOPERATION_OUTDOORDEWPOINT",
    "OS_PLANTEQUIPMENTOPERATION_OUTDOORDRYBULBDIFFERENCE",
    "OS_PLANTEQUIPMENTOPERATION_OUTDOORDEWPOINTDIFFERENCE",
    "OS_PLANTEQUIPMENTOPERATION_OUTDOORWETBULBDIFFERENCE",
    "OS_PLANTCOMPONENT_TEMPERATURESOURCE",
    "OS_LOADPROFILE_PLANT",
    "OS_PUMP_CONSTANTSPEED",
    "OS_PUMP_VARIABLESPEED",
    "OS_HEADEREDPUMPS_CONSTANTSPEED",
    "OS_HEADEREDPUMPS_VARIABLESPEED",
    "OS_TEMPERINGVALVE",
    "OS_SETPOINTMANAGER_FOLLOWOUTDOORAIRTEMPERATURE",
    "OS_SETPOINTMANAGER_MIXEDAIR",
    "OS_SETPOINTMANAGER_OUTDOORAIRRESET",
    "OS_SETPOINTMANAGER_OUTDOORAIRPRETREAT",
    "OS_SETPOINTMANAGER_SINGLEZONE_REHEAT",
    "OS_SETPOINTMANAGER_SINGLEZONE_HUMIDITY_MINIMUM",
    "OS_SETPOINTMANAGER_MULTIZONE_MINIMUMHUMIDITY_AVERAGE",
    "OS_SETPOINTMANAGER_MULTIZONE_HUMIDITY_MINIMUM",
    "OS_SETPOINTMANAGER_SCHEDULED",
    "OS_SETPOINTMANAGER_SCHEDULED_DUALSETPOINT",
    "OS_SETPOINTMANAGER_WARMEST",
    "OS_SETPOINTMANAGER_WARMESTTEMPERATUREFLOW",
    "OS_SETPOINTMANAGER_COLDEST",
    "OS_SETPOINTMANAGER_FOLLOWGROUNDTEMPERATURE",
    "OS_SETPOINTMANAGER_FOLLOWSYSTEMNODETEMPERATURE",
    "OS_SETPOINTMANAGER_MULTIZONE_COOLING_AVERAGE",
    "OS_SETPOINTMANAGER_MULTIZONE_HEATING_AVERAGE",
    "OS_SETPOINTMANAGER_MULTIZONE_HUMIDITY_MAXIMUM",
    "OS_SETPOINTMANAGER_MULTIZONE_MAXIMUMHUMIDITY_AVERAGE",
    "OS_SETPOINTMANAGER_SINGLEZONE_ONESTAGECOOLING",
    "OS_SETPOINTMANAGER_SINGLEZONE_ONESTAGEHEATING",
    "OS_SETPOINTMANAGER_SINGLEZONE_HUMIDITY_MAXIMUM",
    "OS_SETPOINTMANAGER_SINGLEZONE_COOLING",
    "OS_SETPOINTMANAGER_SINGLEZONE_HEATING",
    "OS_SETPOINTMANAGER_SYSTEMNODERESET_TEMPERATURE",
    "OS_SETPOINTMANAGER_SYSTEMNODERESET_HUMIDITY",
    "OS_SIZING_PLANT",
    "OS_SIZING_SYSTEM",
    "OS_SIZING_ZONE",
    "OS_SPLITTER",
    "OS_THERMALZONE",
    "OS_ZONECONTROL_HUMIDISTAT",
    "OS_THERMOSTATSETPOINT_DUALSETPOINT",
    "OS_ZONECONTROL_THERMOSTAT_STAGEDDUALSETPOINT",
    "OS_ZONEHVAC_EQUIPMENTLIST",
    "OS_ZONEVENTILATION_DESIGNFLOWRATE",
    "OS_ZONEVENTILATION_WINDANDSTACKOPENAREA",
    "OS_ZONEHVAC_BASEBOARD_CONVECTIVE_ELECTRIC",
    "OS_ZONEHVAC_BASEBOARD_CONVECTIVE_WATER",
    "OS_ZONEHVAC_HIGHTEMPERATURERADIANT",
    "OS_ZONEHVAC_LOWTEMPERATURERADIANT_ELECTRIC",
    "OS_ZONEHVAC_IDEALLOADSAIRSYSTEM",
    "OS_ZONEHVAC_FOURPIPEFANCOIL",
    "OS_ZONEHVAC_LOWTEMPERATURERADIANT_CONSTANTFLOW",
    "OS_ZONEHVAC_LOWTEMPERATURERADIANT_VARIABLEFLOW",
    "OS_ZONEHVAC_PACKAGEDTERMINALHEATPUMP",
    "OS_ZONEHVAC_PACKAGEDTERMINALAIRCONDITIONER",
    "OS_ZONEHVAC_TERMINALUNIT_VARIABLEREFRIGERANTFLOW",
    "OS_ZONEHVAC_WATERTOAIRHEATPUMP",
    "OS_ZONEHVAC_UNITHEATER",
    "OS_ZONEHVAC_DEHUMIDIFIER_DX",
    "OS_ZONEHVAC_ENERGYRECOVERYVENTILATOR",
    "OS_ZONEHVAC_ENERGYRECOVERYVENTILATOR_CONTROLLER",
    "OS_ZONEHVAC_UNITVENTILATOR",
    "OS_ZONEHVAC_BASEBOARD_RADIANTCONVECTIVE_WATER",
    "OS_ZONEHVAC_BASEBOARD_RADIANTCONVECTIVE_ELECTRIC",
    "OS_ZONEHVAC_COOLINGPANEL_RADIANTCONVECTIVE_WATER",
    "OS_ZONEMIXING",
    "OS_WATERHEATER_MIXED",
    "OS_WATERHEATER_HEATPUMP",
    "OS_COIL_WATERHEATING_AIRTOWATERHEATPUMP",
    "OS_WATERHEATER_HEATPUMP_WRAPPEDCONDENSER",
    "OS_COIL_WATERHEATING_AIRTOWATERHEATPUMP_VARIABLESPEED",
    "OS_COIL_WATERHEATING_AIRTOWATERHEATPUMP_VARIABLESPEED_SPEEDDATA",
    "OS_COIL_WATERHEATING_AIRTOWATERHEATPUMP_WRAPPED",
    "OS_WATERHEATER_STRATIFIED",
    "OS_WATERHEATER_SIZING",
    "OS_WATERUSE_EQUIPMENT",
    "OS_WATERUSE_EQUIPMENT_DEFINITION",
    "OS_WATERUSE_CONNECTIONS",
    "OS_THERMALSTORAGE_ICE_DETAILED",
    "OS_THERMALSTORAGE_CHILLEDWATER_STRATIFIED",
    "OS_DUCT",
    "OS_FLUIDCOOLER_SINGLESPEED",
    "OS_FLUIDCOOLER_TWOSPEED",
    "OS_PIPE_INDOOR",
    "OS_PIPE_OUTDOOR",
    "OS_SWIMMINGPOOL_INDOOR",
    "OS_SOLARCOLLECTORPERFORMANCE_FLATPLATE",
    "OS_SOLARCOLLECTOR_FLATPLATE_WATER",
    "OS_SOLARCOLLECTORPERFORMANCE_INTEGRALCOLLECTORSTORAGE",
    "OS_SOLARCOLLECTOR_INTEGRALCOLLECTORSTORAGE",
    "OS_SOLARCOLLECTORPERFORMANCE_PHOTOVOLTAICTHERMAL_SIMPLE",
    "OS_SOLARCOLLECTORPERFORMANCE_PHOTOVOLTAICTHERMAL_BIPVT",
    "OS_SOLARCOLLECTOR_FLATPLATE_PHOTOVOLTAICTHERMAL",
    "OS_GENERATOR_MICROTURBINE",
    "OS_GENERATOR_MICROTURBINE_HEATRECOVERY",
    "OS_GENERATOR_PHOTOVOLTAIC",
    "OS_PHOTOVOLTAICPERFORMANCE_SIMPLE",
    "OS_PHOTOVOLTAICPERFORMANCE_EQUIVALENTONEDIODE",
    "OS_PHOTOVOLTAICPERFORMANCE_SANDIA",
    "OS_GENERATOR_FUELCELL",
    "OS_GENERATOR_FUELCELL_POWERMODULE",
    "OS_GENERATOR_FUELCELL_AIRSUPPLY",
    "OS_GENERATOR_FUELCELL_WATERSUPPLY",
    "OS_GENERATOR_FUELCELL_AUXILIARYHEATER",
    "OS_GENERATOR_FUELCELL_EXHAUSTGASTOWATERHEATEXCHANGER",
    "OS_GENERATOR_FUELCELL_ELECTRICALSTORAGE",
    "OS_GENERATOR_FUELCELL_INVERTER",
    "OS_GENERATOR_FUELCELL_STACKCOOLER",
    "OS_GENERATOR_FUELSUPPLY",
    "OS_GENERATOR_WINDTURBINE",
    "OS_GENERATOR_PVWATTS",
    "OS_ELECTRICLOADCENTER_INVERTER_PVWATTS",
    "OS_ELECTRICLOADCENTER_INVERTER_SIMPLE",
    "OS_ELECTRICLOADCENTER_INVERTER_LOOKUPTABLE",
    "OS_ELECTRICLOADCENTER_STORAGE_SIMPLE",
    "OS_ELECTRICLOADCENTER_STORAGE_LIIONNMCBATTERY",
    "OS_ELECTRICLOADCENTER_TRANSFORMER",
    "OS_ELECTRICLOADCENTER_DISTRIBUTION",
    "OS_ELECTRICLOADCENTER_STORAGE_CONVERTER",
    "OS_COMPONENTCOST_ADJUSTMENTS",
    "OS_LIFECYCLECOST",
    "OS_CURRENCYTYPE",
    "OS_LIFECYCLECOST_PARAMETERS",
    "OS_LIFECYCLECOST_USEPRICEESCALATION",
    "OS_UTILITYBILL",
    "OS_UTILITYCOST_TARIFF",
    "OS_UTILITYCOST_QUALIFY",
    "OS_UTILITYCOST_CHARGE_SIMPLE",
    "OS_UTILITYCOST_CHARGE_BLOCK",
    "OS_UTILITYCOST_RATCHET",
    "OS_UTILITYCOST_VARIABLE",
    "OS_UTILITYCOST_COMPUTATION",
    "OS_OUTPUT_SCHEDULES",
    "OS_OUTPUT_CONSTRUCTIONS",
    "OS_OUTPUTCONTROL_FILES",
    "OS_OUTPUTCONTROL_TIMESTAMP",
    "OS_OUTPUTCONTROL_REPORTINGTOLERANCES",
    "OS_OUTPUTCONTROL_TABLE_STYLE",
    "OS_OUTPUT_METER",
    "OS_OUTPUT_VARIABLE",
    "OS_METER_CUSTOM",
    "OS_METER_CUSTOMDECREMENT",
    "OS_OUTPUT_ENVIRONMENTALIMPACTFACTORS",
    "OS_ENVIRONMENTALIMPACTFACTORS",
    "OS_FUELFACTORS",
    "OS_OUTPUT_DIAGNOSTICS",
    "OS_OUTPUT_DEBUGGINGDATA",
    "OS_OUTPUT_JSON",
    "OS_OUTPUT_SQLITE",
    "OS_OUTPUT_ENERGYMANAGEMENTSYSTEM",
    "OS_OUTPUT_TABLE_SUMMARYREPORTS",
    "OS_ENERGYMANAGEMENTSYSTEM_SENSOR",
    "OS_ENERGYMANAGEMENTSYSTEM_ACTUATOR",
    "OS_ENERGYMANAGEMENTSYSTEM_PROGRAMCALLINGMANAGER",
    "OS_ENERGYMANAGEMENTSYSTEM_PROGRAM",
    "OS_ENERGYMANAGEMENTSYSTEM_SUBROUTINE",
    "OS_ENERGYMANAGEMENTSYSTEM_GLOBALVARIABLE",
    "OS_ENERGYMANAGEMENTSYSTEM_OUTPUTVARIABLE",
    "OS_ENERGYMANAGEMENTSYSTEM_METEREDOUTPUTVARIABLE",
    "OS_ENERGYMANAGEMENTSYSTEM_TRENDVARIABLE",
    "OS_ENERGYMANAGEMENTSYSTEM_INTERNALVARIABLE",
    "OS_ENERGYMANAGEMENTSYSTEM_CURVEORTABLEINDEXVARIABLE",
    "OS_ENERGYMANAGEMENTSYSTEM_CONSTRUCTIONINDEXVARIABLE",
    "OS_PYTHONPLUGIN_INSTANCE",
    "OS_PYTHONPLUGIN_VARIABLE",
    "OS_PYTHONPLUGIN_TRENDVARIABLE",
    "OS_PYTHONPLUGIN_OUTPUTVARIABLE",
    "OS_AIRFLOWNETWORKSIMULATIONCONTROL",
    "OS_AIRFLOWNETWORKZONE",
    "OS_AIRFLOWNETWORKSURFACE",
    "OS_AIRFLOWNETWORKREFERENCECRACKCONDITIONS",
    "OS_AIRFLOWNETWORKCRACK",
    "OS_AIRFLOWNETWORKEFFECTIVELEAKAGEAREA",
    "OS_AIRFLOWNETWORKSPECIFIEDFLOWRATE",
    "OS_AIRFLOWNETWORKDETAILEDOPENING",
    "OS_AIRFLOWNETWORKSIMPLEOPENING",
    "OS_AIRFLOWNETWORKHORIZONTALOPENING",
    "OS_AIRFLOWNETWORKOUTDOORAIRFLOW",
    "OS_AIRFLOWNETWORKZONEEXHAUSTFAN",
    "OS_AIRFLOWNETWORKEXTERNALNODE",
    "OS_AIRFLOWNETWORKPRESSURECONTROLLER",
    "OS_AIRFLOWNETWORKDISTRIBUTIONNODE",
    "OS_AIRFLOWNETWORKLEAK",
    "OS_AIRFLOWNETWORKLEAKAGERATIO",
    "OS_AIRFLOWNETWORKDUCT",
    "OS_AIRFLOWNETWORKFAN",
    "OS_AIRFLOWNETWORKEQUIVALENTDUCT",
    "OS_AIRFLOWNETWORKCONSTANTPRESSUREDROP",
    "OS_AIRFLOWNETWORKRELIEFAIRFLOW",
    "OS_AIRFLOWNETWORKDISTRIBUTIONLINKAGE",
    "OS_AIRFLOWNETWORKDUCTVIEWFACTORS",
    "OS_AIRFLOWNETWORKOCCUPANTVENTILATIONCONTROL",
    "OS_EXTERNALINTERFACE",
    "OS_EXTERNALINTERFACE_SCHEDULE",
    "OS_EXTERNALINTERFACE_VARIABLE",
    "OS_EXTERNALINTERFACE_ACTUATOR",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_FROM_VARIABLE",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_TO_SCHEDULE",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_TO_ACTUATOR",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITIMPORT_TO_VARIABLE",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_FROM_VARIABLE",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_TO_SCHEDULE",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_TO_ACTUATOR",
    "OS_EXTERNALINTERFACE_FUNCTIONALMOCKUPUNITEXPORT_TO_VARIABLE",
    "OS_COIL_USERDEFINED",
    "OS_PLANTCOMPONENT_USERDEFINED",
    "COMMENTONLY",
  };

  static constexpr std::array<std::string_view, Size> descriptions{
    "",
    "",
    "Lead Input",
    "End Lead Input",
    "Simulation Data",
    "End Simulation Data",
    "Version",
    "SimulationControl",
    "PerformancePrecisionTradeoffs",
    "Building",
    "ShadowCalculation",
    "SurfaceConvectionAlgorithm:Inside",
    "SurfaceConvectionAlgorithm:Outside",
    "HeatBalanceAlgorithm",
    "HeatBalanceSettings:ConductionFiniteDifference",
    "ZoneAirHeatBalanceAlgorithm",
    "ZoneAirContaminantBalance",
    "ZoneAirMassFlowConservation",
    "ZoneCapacitanceMultiplier:ResearchSpecial",
    "Timestep",
    "ConvergenceLimits",
    "HVACSystemRootFindingAlgorithm",
    "Compliance:Building",
    "Site:Location",
    "Site:VariableLocation",
    "SizingPeriod:DesignDay",
    "SizingPeriod:WeatherFileDays",
    "SizingPeriod:WeatherFileConditionType",
    "RunPeriod",
    "RunPeriodControl:SpecialDays",
    "RunPeriodControl:DaylightSavingTime",
    "WeatherProperty:SkyTemperature",
    "Site:WeatherStation",
    "Site:HeightVariation",
    "Site:GroundTemperature:BuildingSurface",
    "Site:GroundTemperature:FCfactorMethod",
    "Site:GroundTemperature:Shallow",
    "Site:GroundTemperature:Deep",
    "Site:GroundTemperature:Undisturbed:FiniteDifference",
    "Site:GroundTemperature:Undisturbed:KusudaAchenbach",
    "Site:GroundTemperature:Undisturbed:Xing",
    "Site:GroundDomain:Slab",
    "Site:GroundDomain:Basement",
    "Site:GroundReflectance",
    "Site:GroundReflectance:SnowModifier",
    "Site:WaterMainsTemperature",
    "Site:Precipitation",
    "RoofIrrigation",
    "Site:SolarAndVisibleSpectrum",
    "Site:SpectrumData",
    "ScheduleTypeLimits",
    "Schedule:Day:Hourly",
    "Schedule:Day:Interval",
    "Schedule:Day:List",
    "Schedule:Week:Daily",
    "Schedule:Week:Compact",
    "Schedule:Year",
    "Schedule:Compact",
    "Schedule:Constant",
    "Schedule:File:Shading",
    "Schedule:File",
    "Material",
    "Material:NoMass",
    "Material:InfraredTransparent",
    "Material:AirGap",
    "Material:RoofVegetation",
    "WindowMaterial:SimpleGlazingSystem",
    "WindowMaterial:Glazing",
    "WindowMaterial:GlazingGroup:Thermochromic",
    "WindowMaterial:Glazing:RefractionExtinctionMethod",
    "WindowMaterial:Gas",
    "WindowGap:SupportPillar",
    "WindowGap:DeflectionState",
    "WindowMaterial:GasMixture",
    "WindowMaterial:Gap",
    "WindowMaterial:Shade",
    "WindowMaterial:ComplexShade",
    "WindowMaterial:Blind",
    "WindowMaterial:Screen",
    "WindowMaterial:Shade:EquivalentLayer",
    "WindowMaterial:Drape:EquivalentLayer",
    "WindowMaterial:Blind:EquivalentLayer",
    "WindowMaterial:Screen:EquivalentLayer",
    "WindowMaterial:Glazing:EquivalentLayer",
    "WindowMaterial:Gap:EquivalentLayer",
    "MaterialProperty:MoisturePenetrationDepth:Settings",
    "MaterialProperty:PhaseChange",
    "MaterialProperty:PhaseChangeHysteresis",
    "MaterialProperty:VariableThermalConductivity",
    "MaterialProperty:VariableAbsorptance",
    "MaterialProperty:HeatAndMoistureTransfer:Settings",
    "MaterialProperty:HeatAndMoistureTransfer:SorptionIsotherm",
    "MaterialProperty:HeatAndMoistureTransfer:Suction",
    "MaterialProperty:HeatAndMoistureTransfer:Redistribution",
    "MaterialProperty:HeatAndMoistureTransfer:Diffusion",
    "MaterialProperty:HeatAndMoistureTransfer:ThermalConductivity",
    "MaterialProperty:GlazingSpectralData",
    "Construction",
    "Construction:CfactorUndergroundWall",
    "Construction:FfactorGroundFloor",
    "ConstructionProperty:InternalHeatSource",
    "Construction:AirBoundary",
    "WindowThermalModel:Params",
    "WindowsCalculationEngine",
    "Construction:ComplexFenestrationState",
    "Construction:WindowEquivalentLayer",
    "Construction:WindowDataFile",
    "GlobalGeometryRules",
    "GeometryTransform",
    "Space",
    "SpaceList",
    "Zone",
    "ZoneList",
    "ZoneGroup",
    "BuildingSurface:Detailed",
    "Wall:Detailed",
    "RoofCeiling:Detailed",
    "Floor:Detailed",
    "Wall:Exterior",
    "Wall:Adiabatic",
    "Wall:Underground",
    "Wall:Interzone",
    "Roof",
    "Ceiling:Adiabatic",
    "Ceiling:Interzone",
    "Floor:GroundContact",
    "Floor:Adiabatic",
    "Floor:Interzone",
    "FenestrationSurface:Detailed",
    "Window",
    "Door",
    "GlazedDoor",
    "Window:Interzone",
    "Door:Interzone",
    "GlazedDoor:Interzone",
    "WindowShadingControl",
    "WindowProperty:FrameAndDivider",
    "WindowProperty:AirflowControl",
    "WindowProperty:StormWindow",
    "InternalMass",
    "Shading:Site",
    "Shading:Building",
    "Shading:Site:Detailed",
    "Shading:Building:Detailed",
    "Shading:Overhang",
    "Shading:Overhang:Projection",
    "Shading:Fin",
    "Shading:Fin:Projection",
    "Shading:Zone:Detailed",
    "ShadingProperty:Reflectance",
    "SurfaceProperty:HeatTransferAlgorithm",
    "SurfaceProperty:HeatTransferAlgorithm:MultipleSurface",
    "SurfaceProperty:HeatTransferAlgorithm:SurfaceList",
    "SurfaceProperty:HeatTransferAlgorithm:Construction",
    "SurfaceProperty:HeatBalanceSourceTerm",
    "SurfaceControl:MovableInsulation",
    "SurfaceProperty:OtherSideCoefficients",
    "SurfaceProperty:OtherSideConditionsModel",
    "SurfaceProperty:Underwater",
    "Foundation:Kiva",
    "Foundation:Kiva:Settings",
    "SurfaceProperty:ExposedFoundationPerimeter",
    "SurfaceConvectionAlgorithm:Inside:AdaptiveModelSelections",
    "SurfaceConvectionAlgorithm:Outside:AdaptiveModelSelections",
    "SurfaceConvectionAlgorithm:Inside:UserCurve",
    "SurfaceConvectionAlgorithm:Outside:UserCurve",
    "SurfaceProperty:ConvectionCoefficients",
    "SurfaceProperty:ConvectionCoefficients:MultipleSurface",
    "SurfaceProperties:VaporCoefficients",
    "SurfaceProperty:ExteriorNaturalVentedCavity",
    "SurfaceProperty:SolarIncidentInside",
    "SurfaceProperty:IncidentSolarMultiplier",
    "SurfaceProperty:LocalEnvironment",
    "ZoneProperty:LocalEnvironment",
    "SurfaceProperty:SurroundingSurfaces",
    "SurfaceProperty:GroundSurfaces",
    "ComplexFenestrationProperty:SolarAbsorbedLayers",
    "ZoneProperty:UserViewFactors:BySurfaceName",
    "GroundHeatTransfer:Control",
    "GroundHeatTransfer:Slab:Materials",
    "GroundHeatTransfer:Slab:MatlProps",
    "GroundHeatTransfer:Slab:BoundConds",
    "GroundHeatTransfer:Slab:BldgProps",
    "GroundHeatTransfer:Slab:Insulation",
    "GroundHeatTransfer:Slab:EquivalentSlab",
    "GroundHeatTransfer:Slab:AutoGrid",
    "GroundHeatTransfer:Slab:ManualGrid",
    "GroundHeatTransfer:Slab:XFACE",
    "GroundHeatTransfer:Slab:YFACE",
    "GroundHeatTransfer:Slab:ZFACE",
    "GroundHeatTransfer:Basement:SimParameters",
    "GroundHeatTransfer:Basement:MatlProps",
    "GroundHeatTransfer:Basement:Insulation",
    "GroundHeatTransfer:Basement:SurfaceProps",
    "GroundHeatTransfer:Basement:BldgData",
    "GroundHeatTransfer:Basement:Interior",
    "GroundHeatTransfer:Basement:ComBldg",
    "GroundHeatTransfer:Basement:EquivSlab",
    "GroundHeatTransfer:Basement:EquivAutoGrid",
    "GroundHeatTransfer:Basement:AutoGrid",
    "GroundHeatTransfer:Basement:ManualGrid",
    "GroundHeatTransfer:Basement:XFACE",
    "GroundHeatTransfer:Basement:YFACE",
    "GroundHeatTransfer:Basement:ZFACE",
    "RoomAirModelType",
    "RoomAir:TemperaturePattern:UserDefined",
    "RoomAir:TemperaturePattern:ConstantGradient",
    "RoomAir:TemperaturePattern:TwoGradient",
    "RoomAir:TemperaturePattern:NondimensionalHeight",
    "RoomAir:TemperaturePattern:SurfaceMapping",
    "RoomAir:Node",
    "RoomAirSettings:OneNodeDisplacementVentilation",
    "RoomAirSettings:ThreeNodeDisplacementVentilation",
    "RoomAirSettings:CrossVentilation",
    "RoomAirSettings:UnderFloorAirDistributionInterior",
    "RoomAirSettings:UnderFloorAirDistributionExterior",
    "RoomAir:Node:AirflowNetwork",
    "RoomAir:Node:AirflowNetwork:AdjacentSurfaceList",
    "RoomAir:Node:AirflowNetwork:InternalGains",
    "RoomAir:Node:AirflowNetwork:HVACEquipment",
    "RoomAirSettings:AirflowNetwork",
    "People",
    "ComfortViewFactorAngles",
    "Lights",
    "ElectricEquipment",
    "GasEquipment",
    "HotWaterEquipment",
    "SteamEquipment",
    "OtherEquipment",
    "ElectricEquipment:ITE:AirCooled",
    "ZoneBaseboard:OutdoorTemperatureControlled",
    "SwimmingPool:Indoor",
    "ZoneContaminantSourceAndSink:CarbonDioxide",
    "ZoneContaminantSourceAndSink:Generic:Constant",
    "SurfaceContaminantSourceAndSink:Generic:PressureDriven",
    "ZoneContaminantSourceAndSink:Generic:CutoffModel",
    "ZoneContaminantSourceAndSink:Generic:DecaySource",
    "SurfaceContaminantSourceAndSink:Generic:BoundaryLayerDiffusion",
    "SurfaceContaminantSourceAndSink:Generic:DepositionVelocitySink",
    "ZoneContaminantSourceAndSink:Generic:DepositionRateSink",
    "Daylighting:Controls",
    "Daylighting:ReferencePoint",
    "Daylighting:DELight:ComplexFenestration",
    "DaylightingDevice:Tubular",
    "DaylightingDevice:Shelf",
    "DaylightingDevice:LightWell",
    "Output:DaylightFactors",
    "Output:IlluminanceMap",
    "OutputControl:IlluminanceMap:Style",
    "ZoneInfiltration:DesignFlowRate",
    "ZoneInfiltration:EffectiveLeakageArea",
    "ZoneInfiltration:FlowCoefficient",
    "ZoneVentilation:DesignFlowRate",
    "ZoneVentilation:WindandStackOpenArea",
    "ZoneAirBalance:OutdoorAir",
    "ZoneMixing",
    "ZoneCrossMixing",
    "ZoneRefrigerationDoorMixing",
    "ZoneEarthtube",
    "ZoneCoolTower:Shower",
    "ZoneThermalChimney",
    "AirflowNetwork:SimulationControl",
    "AirflowNetwork:MultiZone:Zone",
    "AirflowNetwork:MultiZone:Surface",
    "AirflowNetwork:MultiZone:ReferenceCrackConditions",
    "AirflowNetwork:MultiZone:Surface:Crack",
    "AirflowNetwork:MultiZone:Surface:EffectiveLeakageArea",
    "AirflowNetwork:MultiZone:SpecifiedFlowRate",
    "AirflowNetwork:MultiZone:Component:DetailedOpening",
    "AirflowNetwork:MultiZone:Component:SimpleOpening",
    "AirflowNetwork:MultiZone:Component:HorizontalOpening",
    "AirflowNetwork:MultiZone:Component:ZoneExhaustFan",
    "AirflowNetwork:MultiZone:ExternalNode",
    "AirflowNetwork:MultiZone:WindPressureCoefficientArray",
    "AirflowNetwork:MultiZone:WindPressureCoefficientValues",
    "AirflowNetwork:ZoneControl:PressureController",
    "AirflowNetwork:Distribution:Node",
    "AirflowNetwork:Distribution:Component:Leak",
    "AirflowNetwork:Distribution:Component:LeakageRatio",
    "AirflowNetwork:Distribution:Component:Duct",
    "AirflowNetwork:Distribution:Component:Fan",
    "AirflowNetwork:Distribution:Component:Coil",
    "AirflowNetwork:Distribution:Component:HeatExchanger",
    "AirflowNetwork:Distribution:Component:TerminalUnit",
    "AirflowNetwork:Distribution:Component:ConstantPressureDrop",
    "AirflowNetwork:Distribution:Component:OutdoorAirFlow",
    "AirflowNetwork:Distribution:Component:ReliefAirFlow",
    "AirflowNetwork:Distribution:Linkage",
    "AirflowNetwork:Distribution:DuctViewFactors",
    "AirflowNetwork:OccupantVentilationControl",
    "AirflowNetwork:IntraZone:Node",
    "AirflowNetwork:IntraZone:Linkage",
    "Exterior:Lights",
    "Exterior:FuelEquipment",
    "Exterior:WaterEquipment",
    "HVACTemplate:Thermostat",
    "HVACTemplate:Zone:IdealLoadsAirSystem",
    "HVACTemplate:Zone:BaseboardHeat",
    "HVACTemplate:Zone:FanCoil",
    "HVACTemplate:Zone:PTAC",
    "HVACTemplate:Zone:PTHP",
    "HVACTemplate:Zone:WaterToAirHeatPump",
    "HVACTemplate:Zone:VRF",
    "HVACTemplate:Zone:Unitary",
    "HVACTemplate:Zone:VAV",
    "HVACTemplate:Zone:VAV:FanPowered",
    "HVACTemplate:Zone:VAV:HeatAndCool",
    "HVACTemplate:Zone:ConstantVolume",
    "HVACTemplate:Zone:DualDuct",
    "HVACTemplate:System:VRF",
    "HVACTemplate:System:Unitary",
    "HVACTemplate:System:UnitaryHeatPump:AirToAir",
    "HVACTemplate:System:UnitarySystem",
    "HVACTemplate:System:VAV",
    "HVACTemplate:System:PackagedVAV",
    "HVACTemplate:System:ConstantVolume",
    "HVACTemplate:System:DualDuct",
    "HVACTemplate:System:DedicatedOutdoorAir",
    "HVACTemplate:Plant:ChilledWaterLoop",
    "HVACTemplate:Plant:Chiller",
    "HVACTemplate:Plant:Chiller:ObjectReference",
    "HVACTemplate:Plant:Tower",
    "HVACTemplate:Plant:Tower:ObjectReference",
    "HVACTemplate:Plant:HotWaterLoop",
    "HVACTemplate:Plant:Boiler",
    "HVACTemplate:Plant:Boiler:ObjectReference",
    "HVACTemplate:Plant:MixedWaterLoop",
    "DesignSpecification:OutdoorAir",
    "DesignSpecification:OutdoorAir:SpaceList",
    "DesignSpecification:ZoneAirDistribution",
    "Sizing:Parameters",
    "Sizing:Zone",
    "DesignSpecification:ZoneHVAC:Sizing",
    "DesignSpecification:AirTerminal:Sizing",
    "Sizing:System",
    "Sizing:Plant",
    "OutputControl:Sizing:Style",
    "ZoneControl:Humidistat",
    "ZoneControl:Thermostat",
    "ZoneControl:Thermostat:OperativeTemperature",
    "ZoneControl:Thermostat:ThermalComfort",
    "ZoneControl:Thermostat:TemperatureAndHumidity",
    "ThermostatSetpoint:SingleHeating",
    "ThermostatSetpoint:SingleCooling",
    "ThermostatSetpoint:SingleHeatingOrCooling",
    "ThermostatSetpoint:DualSetpoint",
    "ThermostatSetpoint:ThermalComfort:Fanger:SingleHeating",
    "ThermostatSetpoint:ThermalComfort:Fanger:SingleCooling",
    "ThermostatSetpoint:ThermalComfort:Fanger:SingleHeatingOrCooling",
    "ThermostatSetpoint:ThermalComfort:Fanger:DualSetpoint",
    "ZoneControl:Thermostat:StagedDualSetpoint",
    "ZoneControl:ContaminantController",
    "ZoneHVAC:IdealLoadsAirSystem",
    "ZoneHVAC:FourPipeFanCoil",
    "ZoneHVAC:WindowAirConditioner",
    "ZoneHVAC:PackagedTerminalAirConditioner",
    "ZoneHVAC:PackagedTerminalHeatPump",
    "ZoneHVAC:WaterToAirHeatPump",
    "ZoneHVAC:Dehumidifier:DX",
    "ZoneHVAC:EnergyRecoveryVentilator",
    "ZoneHVAC:EnergyRecoveryVentilator:Controller",
    "ZoneHVAC:UnitVentilator",
    "ZoneHVAC:UnitHeater",
    "ZoneHVAC:EvaporativeCoolerUnit",
    "ZoneHVAC:HybridUnitaryHVAC",
    "ZoneHVAC:OutdoorAirUnit",
    "ZoneHVAC:OutdoorAirUnit:EquipmentList",
    "ZoneHVAC:TerminalUnit:VariableRefrigerantFlow",
    "ZoneHVAC:Baseboard:RadiantConvective:Water:Design",
    "ZoneHVAC:Baseboard:RadiantConvective:Water",
    "ZoneHVAC:Baseboard:RadiantConvective:Steam:Design",
    "ZoneHVAC:Baseboard:RadiantConvective:Steam",
    "ZoneHVAC:Baseboard:RadiantConvective:Electric",
    "ZoneHVAC:CoolingPanel:RadiantConvective:Water",
    "ZoneHVAC:Baseboard:Convective:Water",
    "ZoneHVAC:Baseboard:Convective:Electric",
    "ZoneHVAC:LowTemperatureRadiant:VariableFlow",
    "ZoneHVAC:LowTemperatureRadiant:VariableFlow:Design",
    "ZoneHVAC:LowTemperatureRadiant:ConstantFlow",
    "ZoneHVAC:LowTemperatureRadiant:ConstantFlow:Design",
    "ZoneHVAC:LowTemperatureRadiant:Electric",
    "ZoneHVAC:LowTemperatureRadiant:SurfaceGroup",
    "ZoneHVAC:HighTemperatureRadiant",
    "ZoneHVAC:VentilatedSlab",
    "ZoneHVAC:VentilatedSlab:SlabGroup",
    "AirTerminal:SingleDuct:ConstantVolume:Reheat",
    "AirTerminal:SingleDuct:ConstantVolume:NoReheat",
    "AirTerminal:SingleDuct:VAV:NoReheat",
    "AirTerminal:SingleDuct:VAV:Reheat",
    "AirTerminal:SingleDuct:VAV:Reheat:VariableSpeedFan",
    "AirTerminal:SingleDuct:VAV:HeatAndCool:NoReheat",
    "AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat",
    "AirTerminal:SingleDuct:SeriesPIU:Reheat",
    "AirTerminal:SingleDuct:ParallelPIU:Reheat",
    "AirTerminal:SingleDuct:ConstantVolume:FourPipeInduction",
    "AirTerminal:SingleDuct:ConstantVolume:FourPipeBeam",
    "AirTerminal:SingleDuct:ConstantVolume:CooledBeam",
    "AirTerminal:SingleDuct:Mixer",
    "AirTerminal:DualDuct:ConstantVolume",
    "AirTerminal:DualDuct:VAV",
    "AirTerminal:DualDuct:VAV:OutdoorAir",
    "ZoneHVAC:AirDistributionUnit",
    "ZoneHVAC:ExhaustControl",
    "ZoneHVAC:EquipmentList",
    "ZoneHVAC:EquipmentConnections",
    "Fan:SystemModel",
    "Fan:ConstantVolume",
    "Fan:VariableVolume",
    "Fan:OnOff",
    "Fan:ZoneExhaust",
    "FanPerformance:NightVentilation",
    "Fan:ComponentModel",
    "Coil:Cooling:Water",
    "Coil:Cooling:Water:DetailedGeometry",
    "CoilSystem:Cooling:Water",
    "Coil:Cooling:DX",
    "Coil:Cooling:DX:CurveFit:Performance",
    "Coil:Cooling:DX:CurveFit:OperatingMode",
    "Coil:Cooling:DX:CurveFit:Speed",
    "Coil:Cooling:DX:SingleSpeed",
    "Coil:Cooling:DX:TwoSpeed",
    "Coil:Cooling:DX:MultiSpeed",
    "Coil:Cooling:DX:VariableSpeed",
    "Coil:Cooling:DX:TwoStageWithHumidityControlMode",
    "CoilPerformance:DX:Cooling",
    "Coil:Cooling:DX:VariableRefrigerantFlow",
    "Coil:Heating:DX:VariableRefrigerantFlow",
    "Coil:Cooling:DX:VariableRefrigerantFlow:FluidTemperatureControl",
    "Coil:Heating:DX:VariableRefrigerantFlow:FluidTemperatureControl",
    "Coil:Heating:Water",
    "Coil:Heating:Steam",
    "Coil:Heating:Electric",
    "Coil:Heating:Electric:MultiStage",
    "Coil:Heating:Fuel",
    "Coil:Heating:Gas:MultiStage",
    "Coil:Heating:Desuperheater",
    "Coil:Heating:DX:SingleSpeed",
    "Coil:Heating:DX:MultiSpeed",
    "Coil:Heating:DX:VariableSpeed",
    "Coil:Cooling:WaterToAirHeatPump:ParameterEstimation",
    "Coil:Heating:WaterToAirHeatPump:ParameterEstimation",
    "Coil:Cooling:WaterToAirHeatPump:EquationFit",
    "Coil:Cooling:WaterToAirHeatPump:VariableSpeedEquationFit",
    "Coil:Heating:WaterToAirHeatPump:EquationFit",
    "Coil:Heating:WaterToAirHeatPump:VariableSpeedEquationFit",
    "Coil:WaterHeating:AirToWaterHeatPump:Pumped",
    "Coil:WaterHeating:AirToWaterHeatPump:Wrapped",
    "Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed",
    "Coil:WaterHeating:Desuperheater",
    "CoilSystem:Cooling:DX",
    "CoilSystem:Heating:DX",
    "CoilSystem:Cooling:Water:HeatExchangerAssisted",
    "CoilSystem:Cooling:DX:HeatExchangerAssisted",
    "CoilSystem:IntegratedHeatPump:AirSource",
    "Coil:Cooling:DX:SingleSpeed:ThermalStorage",
    "EvaporativeCooler:Direct:CelDekPad",
    "EvaporativeCooler:Indirect:CelDekPad",
    "EvaporativeCooler:Indirect:WetCoil",
    "EvaporativeCooler:Indirect:ResearchSpecial",
    "EvaporativeCooler:Direct:ResearchSpecial",
    "Humidifier:Steam:Electric",
    "Humidifier:Steam:Gas",
    "Dehumidifier:Desiccant:NoFans",
    "Dehumidifier:Desiccant:System",
    "HeatExchanger:AirToAir:FlatPlate",
    "HeatExchanger:AirToAir:SensibleAndLatent",
    "HeatExchanger:Desiccant:BalancedFlow",
    "HeatExchanger:Desiccant:BalancedFlow:PerformanceDataType1",
    "AirLoopHVAC:UnitarySystem",
    "UnitarySystemPerformance:Multispeed",
    "AirLoopHVAC:Unitary:Furnace:HeatOnly",
    "AirLoopHVAC:Unitary:Furnace:HeatCool",
    "AirLoopHVAC:UnitaryHeatOnly",
    "AirLoopHVAC:UnitaryHeatCool",
    "AirLoopHVAC:UnitaryHeatPump:AirToAir",
    "AirLoopHVAC:UnitaryHeatPump:WaterToAir",
    "AirLoopHVAC:UnitaryHeatCool:VAVChangeoverBypass",
    "AirLoopHVAC:UnitaryHeatPump:AirToAir:MultiSpeed",
    "AirConditioner:VariableRefrigerantFlow",
    "AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl",
    "AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl:HR",
    "ZoneTerminalUnitList",
    "Controller:WaterCoil",
    "Controller:OutdoorAir",
    "Controller:MechanicalVentilation",
    "AirLoopHVAC:ControllerList",
    "AirLoopHVAC",
    "AirLoopHVAC:OutdoorAirSystem:EquipmentList",
    "AirLoopHVAC:OutdoorAirSystem",
    "OutdoorAir:Mixer",
    "AirLoopHVAC:ZoneSplitter",
    "AirLoopHVAC:SupplyPlenum",
    "AirLoopHVAC:SupplyPath",
    "AirLoopHVAC:ZoneMixer",
    "AirLoopHVAC:ReturnPlenum",
    "AirLoopHVAC:ReturnPath",
    "AirLoopHVAC:ExhaustSystem",
    "AirLoopHVAC:DedicatedOutdoorAirSystem",
    "AirLoopHVAC:Mixer",
    "AirLoopHVAC:Splitter",
    "Branch",
    "BranchList",
    "Connector:Splitter",
    "Connector:Mixer",
    "ConnectorList",
    "NodeList",
    "OutdoorAir:Node",
    "OutdoorAir:NodeList",
    "Pipe:Adiabatic",
    "Pipe:Adiabatic:Steam",
    "Pipe:Indoor",
    "Pipe:Outdoor",
    "Pipe:Underground",
    "PipingSystem:Underground:Domain",
    "PipingSystem:Underground:PipeCircuit",
    "PipingSystem:Underground:PipeSegment",
    "Duct",
    "Pump:VariableSpeed",
    "Pump:ConstantSpeed",
    "Pump:VariableSpeed:Condensate",
    "HeaderedPumps:ConstantSpeed",
    "HeaderedPumps:VariableSpeed",
    "TemperingValve",
    "LoadProfile:Plant",
    "SolarCollectorPerformance:FlatPlate",
    "SolarCollector:FlatPlate:Water",
    "SolarCollector:FlatPlate:PhotovoltaicThermal",
    "SolarCollectorPerformance:PhotovoltaicThermal:Simple",
    "SolarCollectorPerformance:PhotovoltaicThermal:BIPVT",
    "SolarCollector:IntegralCollectorStorage",
    "SolarCollectorPerformance:IntegralCollectorStorage",
    "SolarCollector:UnglazedTranspired",
    "SolarCollector:UnglazedTranspired:Multisystem",
    "Boiler:HotWater",
    "Boiler:Steam",
    "Chiller:Electric:ASHRAE205",
    "Chiller:Electric:EIR",
    "Chiller:Electric:ReformulatedEIR",
    "Chiller:Electric",
    "Chiller:Absorption:Indirect",
    "Chiller:Absorption",
    "Chiller:ConstantCOP",
    "Chiller:EngineDriven",
    "Chiller:CombustionTurbine",
    "ChillerHeater:Absorption:DirectFired",
    "ChillerHeater:Absorption:DoubleEffect",
    "HeatPump:PlantLoop:EIR:Cooling",
    "HeatPump:PlantLoop:EIR:Heating",
    "HeatPump:AirToWater:FuelFired:Heating",
    "HeatPump:AirToWater:FuelFired:Cooling",
    "HeatPump:WaterToWater:EquationFit:Heating",
    "HeatPump:WaterToWater:EquationFit:Cooling",
    "HeatPump:WaterToWater:ParameterEstimation:Cooling",
    "HeatPump:WaterToWater:ParameterEstimation:Heating",
    "DistrictCooling",
    "DistrictHeating",
    "PlantComponent:TemperatureSource",
    "CentralHeatPumpSystem",
    "ChillerHeaterPerformance:Electric:EIR",
    "CoolingTower:SingleSpeed",
    "CoolingTower:TwoSpeed",
    "CoolingTower:VariableSpeed:Merkel",
    "CoolingTower:VariableSpeed",
    "CoolingTowerPerformance:CoolTools",
    "CoolingTowerPerformance:YorkCalc",
    "EvaporativeFluidCooler:SingleSpeed",
    "EvaporativeFluidCooler:TwoSpeed",
    "FluidCooler:SingleSpeed",
    "FluidCooler:TwoSpeed",
    "GroundHeatExchanger:System",
    "GroundHeatExchanger:Vertical:Properties",
    "GroundHeatExchanger:Vertical:Array",
    "GroundHeatExchanger:Vertical:Single",
    "GroundHeatExchanger:ResponseFactors",
    "GroundHeatExchanger:Pond",
    "GroundHeatExchanger:Surface",
    "GroundHeatExchanger:HorizontalTrench",
    "GroundHeatExchanger:Slinky",
    "HeatExchanger:FluidToFluid",
    "WaterHeater:Mixed",
    "WaterHeater:Stratified",
    "WaterHeater:Sizing",
    "WaterHeater:HeatPump:PumpedCondenser",
    "WaterHeater:HeatPump:WrappedCondenser",
    "ThermalStorage:Ice:Simple",
    "ThermalStorage:Ice:Detailed",
    "ThermalStorage:ChilledWater:Mixed",
    "ThermalStorage:ChilledWater:Stratified",
    "PlantLoop",
    "CondenserLoop",
    "PlantEquipmentList",
    "CondenserEquipmentList",
    "PlantEquipmentOperation:Uncontrolled",
    "PlantEquipmentOperation:CoolingLoad",
    "PlantEquipmentOperation:HeatingLoad",
    "PlantEquipmentOperation:OutdoorDryBulb",
    "PlantEquipmentOperation:OutdoorWetBulb",
    "PlantEquipmentOperation:OutdoorRelativeHumidity",
    "PlantEquipmentOperation:OutdoorDewpoint",
    "PlantEquipmentOperation:ComponentSetpoint",
    "PlantEquipmentOperation:ThermalEnergyStorage",
    "PlantEquipmentOperation:OutdoorDryBulbDifference",
    "PlantEquipmentOperation:OutdoorWetBulbDifference",
    "PlantEquipmentOperation:OutdoorDewpointDifference",
    "PlantEquipmentOperationSchemes",
    "CondenserEquipmentOperationSchemes",
    "EnergyManagementSystem:Sensor",
    "EnergyManagementSystem:Actuator",
    "EnergyManagementSystem:ProgramCallingManager",
    "EnergyManagementSystem:Program",
    "EnergyManagementSystem:Subroutine",
    "EnergyManagementSystem:GlobalVariable",
    "EnergyManagementSystem:OutputVariable",
    "EnergyManagementSystem:MeteredOutputVariable",
    "EnergyManagementSystem:TrendVariable",
    "EnergyManagementSystem:InternalVariable",
    "EnergyManagementSystem:CurveOrTableIndexVariable",
    "EnergyManagementSystem:ConstructionIndexVariable",
    "ExternalInterface",
    "ExternalInterface:Schedule",
    "ExternalInterface:Variable",
    "ExternalInterface:Actuator",
    "ExternalInterface:FunctionalMockupUnitImport",
    "ExternalInterface:FunctionalMockupUnitImport:From:Variable",
    "ExternalInterface:FunctionalMockupUnitImport:To:Schedule",
    "ExternalInterface:FunctionalMockupUnitImport:To:Actuator",
    "ExternalInterface:FunctionalMockupUnitImport:To:Variable",
    "ExternalInterface:FunctionalMockupUnitExport:From:Variable",
    "ExternalInterface:FunctionalMockupUnitExport:To:Schedule",
    "ExternalInterface:FunctionalMockupUnitExport:To:Actuator",
    "ExternalInterface:FunctionalMockupUnitExport:To:Variable",
    "ZoneHVAC:ForcedAir:UserDefined",
    "AirTerminal:SingleDuct:UserDefined",
    "Coil:UserDefined",
    "PlantComponent:UserDefined",
    "PlantEquipmentOperation:UserDefined",
    "AvailabilityManager:Scheduled",
    "AvailabilityManager:ScheduledOn",
    "AvailabilityManager:ScheduledOff",
    "AvailabilityManager:OptimumStart",
    "AvailabilityManager:NightCycle",
    "AvailabilityManager:DifferentialThermostat",
    "AvailabilityManager:HighTemperatureTurnOff",
    "AvailabilityManager:HighTemperatureTurnOn",
    "AvailabilityManager:LowTemperatureTurnOff",
    "AvailabilityManager:LowTemperatureTurnOn",
    "AvailabilityManager:NightVentilation",
    "AvailabilityManager:HybridVentilation",
    "AvailabilityManagerAssignmentList",
    "SetpointManager:Scheduled",
    "SetpointManager:Scheduled:DualSetpoint",
    "SetpointManager:OutdoorAirReset",
    "SetpointManager:SingleZone:Reheat",
    "SetpointManager:SingleZone:Heating",
    "SetpointManager:SingleZone:Cooling",
    "SetpointManager:SingleZone:Humidity:Minimum",
    "SetpointManager:SingleZone:Humidity:Maximum",
    "SetpointManager:MixedAir",
    "SetpointManager:OutdoorAirPretreat",
    "SetpointManager:Warmest",
    "SetpointManager:Coldest",
    "SetpointManager:ReturnAirBypassFlow",
    "SetpointManager:WarmestTemperatureFlow",
    "SetpointManager:MultiZone:Heating:Average",
    "SetpointManager:MultiZone:Cooling:Average",
    "SetpointManager:MultiZone:MinimumHumidity:Average",
    "SetpointManager:MultiZone:MaximumHumidity:Average",
    "SetpointManager:MultiZone:Humidity:Minimum",
    "SetpointManager:MultiZone:Humidity:Maximum",
    "SetpointManager:FollowOutdoorAirTemperature",
    "SetpointManager:FollowSystemNodeTemperature",
    "SetpointManager:FollowGroundTemperature",
    "SetpointManager:CondenserEnteringReset",
    "SetpointManager:CondenserEnteringReset:Ideal",
    "SetpointManager:SingleZone:OneStageCooling",
    "SetpointManager:SingleZone:OneStageHeating",
    "SetpointManager:ReturnTemperature:ChilledWater",
    "SetpointManager:ReturnTemperature:HotWater",
    "SetpointManager:SystemNodeReset:Temperature",
    "SetpointManager:SystemNodeReset:Humidity",
    "Refrigeration:Case",
    "Refrigeration:CompressorRack",
    "Refrigeration:CaseAndWalkInList",
    "Refrigeration:Condenser:AirCooled",
    "Refrigeration:Condenser:EvaporativeCooled",
    "Refrigeration:Condenser:WaterCooled",
    "Refrigeration:Condenser:Cascade",
    "Refrigeration:GasCooler:AirCooled",
    "Refrigeration:TransferLoadList",
    "Refrigeration:Subcooler",
    "Refrigeration:Compressor",
    "Refrigeration:CompressorList",
    "Refrigeration:System",
    "Refrigeration:TranscriticalSystem",
    "Refrigeration:SecondarySystem",
    "Refrigeration:WalkIn",
    "Refrigeration:AirChiller",
    "ZoneHVAC:RefrigerationChillerSet",
    "DemandManagerAssignmentList",
    "DemandManager:ExteriorLights",
    "DemandManager:Lights",
    "DemandManager:ElectricEquipment",
    "DemandManager:Thermostats",
    "DemandManager:Ventilation",
    "Generator:InternalCombustionEngine",
    "Generator:CombustionTurbine",
    "Generator:MicroTurbine",
    "Generator:Photovoltaic",
    "PhotovoltaicPerformance:Simple",
    "PhotovoltaicPerformance:EquivalentOne-Diode",
    "PhotovoltaicPerformance:Sandia",
    "Generator:PVWatts",
    "ElectricLoadCenter:Inverter:PVWatts",
    "Generator:FuelCell",
    "Generator:FuelCell:PowerModule",
    "Generator:FuelCell:AirSupply",
    "Generator:FuelCell:WaterSupply",
    "Generator:FuelCell:AuxiliaryHeater",
    "Generator:FuelCell:ExhaustGasToWaterHeatExchanger",
    "Generator:FuelCell:ElectricalStorage",
    "Generator:FuelCell:Inverter",
    "Generator:FuelCell:StackCooler",
    "Generator:MicroCHP",
    "Generator:MicroCHP:NonNormalizedParameters",
    "Generator:FuelSupply",
    "Generator:WindTurbine",
    "ElectricLoadCenter:Generators",
    "ElectricLoadCenter:Inverter:Simple",
    "ElectricLoadCenter:Inverter:FunctionOfPower",
    "ElectricLoadCenter:Inverter:LookUpTable",
    "ElectricLoadCenter:Storage:Simple",
    "ElectricLoadCenter:Storage:Battery",
    "ElectricLoadCenter:Storage:LiIonNMCBattery",
    "ElectricLoadCenter:Transformer",
    "ElectricLoadCenter:Distribution",
    "ElectricLoadCenter:Storage:Converter",
    "WaterUse:Equipment",
    "WaterUse:Connections",
    "WaterUse:Storage",
    "WaterUse:Well",
    "WaterUse:RainCollector",
    "FaultModel:TemperatureSensorOffset:OutdoorAir",
    "FaultModel:HumiditySensorOffset:OutdoorAir",
    "FaultModel:EnthalpySensorOffset:OutdoorAir",
    "FaultModel:TemperatureSensorOffset:ReturnAir",
    "FaultModel:EnthalpySensorOffset:ReturnAir",
    "FaultModel:TemperatureSensorOffset:ChillerSupplyWater",
    "FaultModel:TemperatureSensorOffset:CoilSupplyAir",
    "FaultModel:TemperatureSensorOffset:CondenserSupplyWater",
    "FaultModel:ThermostatOffset",
    "FaultModel:HumidistatOffset",
    "FaultModel:Fouling:AirFilter",
    "FaultModel:Fouling:Boiler",
    "FaultModel:Fouling:EvaporativeCooler",
    "FaultModel:Fouling:Chiller",
    "FaultModel:Fouling:CoolingTower",
    "FaultModel:Fouling:Coil",
    "Matrix:TwoDimension",
    "HybridModel:Zone",
    "Curve:Linear",
    "Curve:QuadLinear",
    "Curve:QuintLinear",
    "Curve:Quadratic",
    "Curve:Cubic",
    "Curve:Quartic",
    "Curve:Exponent",
    "Curve:Bicubic",
    "Curve:Biquadratic",
    "Curve:QuadraticLinear",
    "Curve:CubicLinear",
    "Curve:Triquadratic",
    "Curve:Functional:PressureDrop",
    "Curve:FanPressureRise",
    "Curve:ExponentialSkewNormal",
    "Curve:Sigmoid",
    "Curve:RectangularHyperbola1",
    "Curve:RectangularHyperbola2",
    "Curve:ExponentialDecay",
    "Curve:DoubleExponentialDecay",
    "Curve:ChillerPartLoadWithLift",
    "Table:IndependentVariable",
    "Table:IndependentVariableList",
    "Table:Lookup",
    "FluidProperties:Name",
    "FluidProperties:GlycolConcentration",
    "FluidProperties:Temperatures",
    "FluidProperties:Saturated",
    "FluidProperties:Superheated",
    "FluidProperties:Concentration",
    "CurrencyType",
    "ComponentCost:Adjustments",
    "ComponentCost:Reference",
    "ComponentCost:LineItem",
    "UtilityCost:Tariff",
    "UtilityCost:Qualify",
    "UtilityCost:Charge:Simple",
    "UtilityCost:Charge:Block",
    "UtilityCost:Ratchet",
    "UtilityCost:Variable",
    "UtilityCost:Computation",
    "LifeCycleCost:Parameters",
    "LifeCycleCost:RecurringCosts",
    "LifeCycleCost:NonrecurringCost",
    "LifeCycleCost:UsePriceEscalation",
    "LifeCycleCost:UseAdjustment",
    "Parametric:SetValueForRun",
    "Parametric:Logic",
    "Parametric:RunControl",
    "Parametric:FileNameSuffix",
    "Output:VariableDictionary",
    "Output:Surfaces:List",
    "Output:Surfaces:Drawing",
    "Output:Schedules",
    "Output:Constructions",
    "Output:EnergyManagementSystem",
    "OutputControl:SurfaceColorScheme",
    "Output:Table:SummaryReports",
    "Output:Table:TimeBins",
    "Output:Table:Monthly",
    "Output:Table:Annual",
    "OutputControl:Table:Style",
    "OutputControl:ReportingTolerances",
    "Output:Variable",
    "Output:Meter",
    "Output:Meter:MeterFileOnly",
    "Output:Meter:Cumulative",
    "Output:Meter:Cumulative:MeterFileOnly",
    "Meter:Custom",
    "Meter:CustomDecrement",
    "OutputControl:Files",
    "OutputControl:Timestamp",
    "Output:JSON",
    "Output:SQLite",
    "Output:EnvironmentalImpactFactors",
    "EnvironmentalImpactFactors",
    "FuelFactors",
    "Output:Diagnostics",
    "Output:DebuggingData",
    "Output:PreprocessorMessage",
    "PythonPlugin:SearchPaths",
    "PythonPlugin:Instance",
    "PythonPlugin:Variables",
    "PythonPlugin:TrendVariable",
    "PythonPlugin:OutputVariable",
    "OS:Version",
    "OS:ComponentData",
    "OS:ConvergenceLimits",
    "OS:ProgramControl",
    "OS:HeatBalanceAlgorithm",
    "OS:RunPeriod",
    "OS:RunPeriodControl:DaylightSavingTime",
    "OS:RunPeriodControl:SpecialDays",
    "OS:ShadowCalculation",
    "OS:SimulationControl",
    "OS:PerformancePrecisionTradeoffs",
    "OS:Sizing:Parameters",
    "OS:SurfaceConvectionAlgorithm:Inside",
    "OS:SurfaceConvectionAlgorithm:Outside",
    "OS:Timestep",
    "OS:YearDescription",
    "OS:ZoneAirContaminantBalance",
    "OS:ZoneAirHeatBalanceAlgorithm",
    "OS:ZoneAirMassFlowConservation",
    "OS:ZoneCapacitanceMultiplier:ResearchSpecial",
    "OS:RadianceParameters",
    "OS:ClimateZones",
    "OS:Site",
    "OS:Site:GroundTemperature:BuildingSurface",
    "OS:Site:GroundTemperature:FCfactorMethod",
    "OS:Site:GroundTemperature:Shallow",
    "OS:Site:GroundTemperature:Deep",
    "OS:Site:GroundTemperature:Undisturbed:KusudaAchenbach",
    "OS:Site:GroundTemperature:Undisturbed:Xing",
    "OS:Site:GroundReflectance",
    "OS:Site:WaterMainsTemperature",
    "OS:SizingPeriod:DesignDay",
    "OS:SizingPeriod:WeatherFileConditionType",
    "OS:SizingPeriod:WeatherFileDays",
    "OS:WeatherFile",
    "OS:WeatherProperty:SkyTemperature",
    "OS:AdditionalProperties",
    "OS:BuildingStory",
    "OS:BuildingUnit",
    "OS:DefaultConstructionSet",
    "OS:DefaultScheduleSet",
    "OS:DefaultSubSurfaceConstructions",
    "OS:DefaultSurfaceConstructions",
    "OS:Rendering:Color",
    "OS:SpaceType",
    "OS:ModelObjectList",
    "OS:Material",
    "OS:Material:AirGap",
    "OS:Material:InfraredTransparent",
    "OS:Material:NoMass",
    "OS:Material:RoofVegetation",
    "OS:WindowMaterial:Blind",
    "OS:WindowMaterial:DaylightRedirectionDevice",
    "OS:WindowMaterial:Gas",
    "OS:WindowMaterial:GasMixture",
    "OS:WindowMaterial:Glazing",
    "OS:WindowMaterial:GlazingGroup:Thermochromic",
    "OS:WindowMaterial:Glazing:RefractionExtinctionMethod",
    "OS:WindowMaterial:Screen",
    "OS:WindowMaterial:Shade",
    "OS:WindowMaterial:SimpleGlazingSystem",
    "OS:StandardsInformation:Material",
    "OS:MaterialProperty:GlazingSpectralData",
    "OS:MaterialProperty:MoisturePenetrationDepth:Settings",
    "OS:MaterialProperty:PhaseChange",
    "OS:MaterialProperty:PhaseChangeHysteresis",
    "OS:Construction",
    "OS:Construction:CfactorUndergroundWall",
    "OS:Construction:FfactorGroundFloor",
    "OS:Construction:InternalSource",
    "OS:Construction:AirBoundary",
    "OS:Construction:WindowDataFile",
    "OS:StandardsInformation:Construction",
    "OS:InternalMass:Definition",
    "OS:People:Definition",
    "OS:Lights:Definition",
    "OS:Luminaire:Definition",
    "OS:ElectricEquipment:Definition",
    "OS:ElectricEquipment:ITE:AirCooled:Definition",
    "OS:GasEquipment:Definition",
    "OS:HotWaterEquipment:Definition",
    "OS:SteamEquipment:Definition",
    "OS:OtherEquipment:Definition",
    "OS:Exterior:Lights:Definition",
    "OS:Exterior:FuelEquipment:Definition",
    "OS:Exterior:WaterEquipment:Definition",
    "OS:Schedule:Compact",
    "OS:Schedule:Day",
    "OS:Schedule:Week",
    "OS:Schedule:Year",
    "OS:Schedule:Rule",
    "OS:Schedule:Ruleset",
    "OS:Schedule:Constant",
    "OS:Schedule:FixedInterval",
    "OS:Schedule:VariableInterval",
    "OS:ScheduleTypeLimits",
    "OS:External:File",
    "OS:Schedule:File",
    "OS:Building",
    "OS:DaylightingDevice:Tubular",
    "OS:DaylightingDevice:Shelf",
    "OS:DaylightingDevice:LightWell",
    "OS:Facility",
    "OS:InteriorPartitionSurface",
    "OS:InteriorPartitionSurfaceGroup",
    "OS:ShadingSurface",
    "OS:ShadingSurfaceGroup",
    "OS:Space",
    "OS:Surface",
    "OS:SubSurface",
    "OS:ShadingControl",
    "OS:WindowProperty:FrameAndDivider",
    "OS:SurfaceControl:MovableInsulation",
    "OS:SurfaceProperty:ConvectionCoefficients",
    "OS:SurfaceProperty:ConvectionCoefficients:MultipleSurface",
    "OS:SurfaceProperty:ExposedFoundationPerimeter",
    "OS:SurfaceProperty:OtherSideCoefficients",
    "OS:SurfaceProperty:OtherSideConditionsModel",
    "OS:Foundation:Kiva",
    "OS:Foundation:Kiva:Settings",
    "OS:ZoneProperty:UserViewFactors:BySurfaceName",
    "OS:SurfaceProperty:SurroundingSurfaces",
    "OS:SurfaceProperty:LocalEnvironment",
    "OS:SurfaceProperty:GroundSurfaces",
    "OS:SurfaceProperty:IncidentSolarMultiplier",
    "OS:InternalMass",
    "OS:People",
    "OS:Lights",
    "OS:Luminaire",
    "OS:ElectricEquipment",
    "OS:ElectricEquipment:ITE:AirCooled",
    "OS:GasEquipment",
    "OS:HotWaterEquipment",
    "OS:SteamEquipment",
    "OS:OtherEquipment",
    "OS:SpaceInfiltration:DesignFlowRate",
    "OS:SpaceInfiltration:EffectiveLeakageArea",
    "OS:SpaceInfiltration:FlowCoefficient",
    "OS:Exterior:Lights",
    "OS:Exterior:FuelEquipment",
    "OS:Exterior:WaterEquipment",
    "OS:Daylighting:Control",
    "OS:Glare:Sensor",
    "OS:IlluminanceMap",
    "OS:LightingDesignDay",
    "OS:LightingSimulationControl",
    "OS:LightingSimulationZone",
    "OS:Refrigeration:Case",
    "OS:Refrigeration:Condenser:AirCooled",
    "OS:Refrigeration:Condenser:EvaporativeCooled",
    "OS:Refrigeration:Condenser:WaterCooled",
    "OS:Refrigeration:Condenser:Cascade",
    "OS:Refrigeration:Compressor",
    "OS:Refrigeration:CompressorRack",
    "OS:Refrigeration:DefrostCycleParameters",
    "OS:Refrigeration:GasCooler:AirCooled",
    "OS:Refrigeration:Subcooler:Mechanical",
    "OS:Refrigeration:Subcooler:LiquidSuction",
    "OS:Refrigeration:System",
    "OS:Refrigeration:TranscriticalSystem",
    "OS:Refrigeration:SecondarySystem",
    "OS:Refrigeration:WalkIn",
    "OS:Refrigeration:WalkIn:ZoneBoundary",
    "OS:Refrigeration:AirChiller",
    "OS:AirConditioner:VariableRefrigerantFlow",
    "OS:AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl",
    "OS:AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl:LoadingIndex",
    "OS:AirConditioner:VariableRefrigerantFlow:FluidTemperatureControl:HR",
    "OS:AirLoopHVAC",
    "OS:AirLoopHVAC:OutdoorAirSystem",
    "OS:AirLoopHVAC:DedicatedOutdoorAirSystem",
    "OS:AirLoopHVAC:ReturnPlenum",
    "OS:AirLoopHVAC:SupplyPlenum",
    "OS:AirLoopHVAC:UnitaryHeatPump:AirToAir:MultiSpeed",
    "OS:AirLoopHVAC:UnitaryCoolOnly",
    "OS:AirLoopHVAC:UnitaryHeatPump:AirToAir",
    "OS:AirLoopHVAC:UnitaryHeatCool:VAVChangeoverBypass",
    "OS:AirLoopHVAC:UnitarySystem",
    "OS:UnitarySystemPerformance:Multispeed",
    "OS:AirLoopHVAC:ZoneMixer",
    "OS:AirLoopHVAC:ZoneSplitter",
    "OS:AirTerminal:SingleDuct:VAV:HeatAndCool:NoReheat",
    "OS:AirTerminal:SingleDuct:VAV:HeatAndCool:Reheat",
    "OS:AirTerminal:SingleDuct:InletSideMixer",
    "OS:AirTerminal:SingleDuct:ConstantVolume:Reheat",
    "OS:AirTerminal:SingleDuct:SeriesPIU:Reheat",
    "OS:AirTerminal:SingleDuct:ParallelPIU:Reheat",
    "OS:AirTerminal:SingleDuct:ConstantVolume:FourPipeInduction",
    "OS:AirTerminal:SingleDuct:ConstantVolume:FourPipeBeam",
    "OS:AirTerminal:SingleDuct:ConstantVolume:CooledBeam",
    "OS:AirTerminal:SingleDuct:ConstantVolume:NoReheat",
    "OS:AirTerminal:SingleDuct:VAV:NoReheat",
    "OS:AirTerminal:SingleDuct:VAV:Reheat",
    "OS:AirTerminal:DualDuct:VAV:OutdoorAir",
    "OS:AirTerminal:DualDuct:ConstantVolume",
    "OS:AirTerminal:DualDuct:VAV",
    "OS:AvailabilityManagerAssignmentList",
    "OS:AvailabilityManager:Scheduled",
    "OS:AvailabilityManager:ScheduledOn",
    "OS:AvailabilityManager:ScheduledOff",
    "OS:AvailabilityManager:NightCycle",
    "OS:AvailabilityManager:OptimumStart",
    "OS:AvailabilityManager:DifferentialThermostat",
    "OS:AvailabilityManager:HighTemperatureTurnOff",
    "OS:AvailabilityManager:HighTemperatureTurnOn",
    "OS:AvailabilityManager:LowTemperatureTurnOff",
    "OS:AvailabilityManager:LowTemperatureTurnOn",
    "OS:AvailabilityManager:NightVentilation",
    "OS:AvailabilityManager:HybridVentilation",
    "OS:Boiler:HotWater",
    "OS:Boiler:Steam",
    "OS:HeatPump:WaterToWater:EquationFit:Heating",
    "OS:HeatPump:WaterToWater:EquationFit:Cooling",
    "OS:HeatPump:PlantLoop:EIR:Heating",
    "OS:HeatPump:PlantLoop:EIR:Cooling",
    "OS:HeatPump:AirToWater:FuelFired:Heating",
    "OS:HeatPump:AirToWater:FuelFired:Cooling",
    "OS:Chiller:Electric:EIR",
    "OS:Chiller:Electric:ReformulatedEIR",
    "OS:Chiller:Electric:ASHRAE205",
    "OS:CentralHeatPumpSystem",
    "OS:CentralHeatPumpSystem:Module",
    "OS:ChillerHeaterPerformance:Electric:EIR",
    "OS:Chiller:Absorption:Indirect",
    "OS:Chiller:Absorption",
    "OS:CoolingTower:SingleSpeed",
    "OS:CoolingTower:TwoSpeed",
    "OS:CoolingTower:VariableSpeed",
    "OS:CoolingTowerPerformance:CoolTools",
    "OS:CoolingTowerPerformance:YorkCalc",
    "OS:DistrictCooling",
    "OS:DistrictHeating",
    "OS:GroundHeatExchanger:Vertical",
    "OS:GroundHeatExchanger:HorizontalTrench",
    "OS:Coil:Cooling:CooledBeam",
    "OS:Coil:Cooling:FourPipeBeam",
    "OS:Coil:Heating:FourPipeBeam",
    "OS:Coil:Cooling:DX",
    "OS:Coil:Cooling:DX:CurveFit:Performance",
    "OS:Coil:Cooling:DX:CurveFit:OperatingMode",
    "OS:Coil:Cooling:DX:CurveFit:Speed",
    "OS:Coil:Cooling:DX:SingleSpeed",
    "OS:Coil:Cooling:DX:TwoStageWithHumidityControlMode",
    "OS:CoilPerformance:DX:Cooling",
    "OS:Coil:Cooling:DX:MultiSpeed",
    "OS:Coil:Cooling:DX:MultiSpeed:StageData",
    "OS:Coil:Cooling:DX:TwoSpeed",
    "OS:CoilSystem:Cooling:Water:HeatExchangerAssisted",
    "OS:CoilSystem:Cooling:DX:HeatExchangerAssisted",
    "OS:CoilSystem:IntegratedHeatPump:AirSource",
    "OS:Coil:Cooling:DX:VariableRefrigerantFlow",
    "OS:Coil:Cooling:DX:VariableRefrigerantFlow:FluidTemperatureControl",
    "OS:Coil:Cooling:LowTemperatureRadiant:ConstantFlow",
    "OS:Coil:Cooling:LowTemperatureRadiant:VariableFlow",
    "OS:Coil:Cooling:Water",
    "OS:Coil:Cooling:WaterToAirHeatPump:EquationFit",
    "OS:Coil:Heating:DX:SingleSpeed",
    "OS:Coil:Heating:DX:VariableRefrigerantFlow",
    "OS:Coil:Heating:DX:VariableRefrigerantFlow:FluidTemperatureControl",
    "OS:Coil:Heating:Electric",
    "OS:Coil:Heating:Electric:MultiStage",
    "OS:Coil:Heating:Electric:MultiStage:StageData",
    "OS:Coil:Heating:Gas",
    "OS:Coil:Heating:Gas:MultiStage",
    "OS:Coil:Heating:Gas:MultiStage:StageData",
    "OS:Coil:Heating:Desuperheater",
    "OS:Coil:WaterHeating:Desuperheater",
    "OS:Coil:Heating:LowTemperatureRadiant:ConstantFlow",
    "OS:Coil:Heating:LowTemperatureRadiant:VariableFlow",
    "OS:Coil:Heating:Water",
    "OS:Coil:Heating:Water:Baseboard",
    "OS:Coil:Heating:Water:Baseboard:Radiant",
    "OS:Coil:Heating:WaterToAirHeatPump:EquationFit",
    "OS:Coil:Heating:DX:MultiSpeed",
    "OS:Coil:Heating:DX:MultiSpeed:StageData",
    "OS:Coil:Heating:DX:VariableSpeed",
    "OS:Coil:Heating:DX:VariableSpeed:SpeedData",
    "OS:Coil:Cooling:DX:VariableSpeed",
    "OS:Coil:Cooling:DX:VariableSpeed:SpeedData",
    "OS:Coil:Heating:WaterToAirHeatPump:VariableSpeedEquationFit",
    "OS:Coil:Heating:WaterToAirHeatPump:VariableSpeedEquationFit:SpeedData",
    "OS:Coil:Cooling:WaterToAirHeatPump:VariableSpeedEquationFit",
    "OS:Coil:Cooling:WaterToAirHeatPump:VariableSpeedEquationFit:SpeedData",
    "OS:Coil:Cooling:Water:Panel:Radiant",
    "OS:Connection",
    "OS:Connector:Mixer",
    "OS:Connector:Splitter",
    "OS:Controller:MechanicalVentilation",
    "OS:ZoneControl:ContaminantController",
    "OS:Controller:OutdoorAir",
    "OS:Controller:WaterCoil",
    "OS:Curve:Linear",
    "OS:Curve:QuadLinear",
    "OS:Curve:QuintLinear",
    "OS:Curve:Quadratic",
    "OS:Curve:Cubic",
    "OS:Curve:Quartic",
    "OS:Curve:Exponent",
    "OS:Curve:Bicubic",
    "OS:Curve:Biquadratic",
    "OS:Curve:QuadraticLinear",
    "OS:Curve:Triquadratic",
    "OS:Curve:Functional:PressureDrop",
    "OS:Curve:FanPressureRise",
    "OS:Curve:ExponentialSkewNormal",
    "OS:Curve:Sigmoid",
    "OS:Curve:RectangularHyperbola1",
    "OS:Curve:RectangularHyperbola2",
    "OS:Curve:ExponentialDecay",
    "OS:Curve:DoubleExponentialDecay",
    "OS:Table:MultiVariableLookup",
    "OS:Table:IndependentVariable",
    "OS:Table:Lookup",
    "OS:DesignSpecification:OutdoorAir",
    "OS:DesignSpecification:ZoneAirDistribution",
    "OS:Humidifier:Steam:Electric",
    "OS:Humidifier:Steam:Gas",
    "OS:EvaporativeCooler:Indirect:ResearchSpecial",
    "OS:EvaporativeCooler:Direct:ResearchSpecial",
    "OS:EvaporativeFluidCooler:SingleSpeed",
    "OS:EvaporativeFluidCooler:TwoSpeed",
    "OS:Fan:SystemModel",
    "OS:Fan:ConstantVolume",
    "OS:Fan:VariableVolume",
    "OS:Fan:OnOff",
    "OS:Fan:ComponentModel",
    "OS:Fan:ZoneExhaust",
    "OS:HeatExchanger:AirToAir:SensibleAndLatent",
    "OS:HeatExchanger:Desiccant:BalancedFlow",
    "OS:HeatExchanger:Desiccant:BalancedFlow:PerformanceDataType1",
    "OS:HeatExchanger:FluidToFluid",
    "OS:HVACComponentList",
    "OS:Node",
    "OS:PortList",
    "OS:Pipe:Adiabatic",
    "OS:PlantLoop",
    "OS:PlantEquipmentOperation:CoolingLoad",
    "OS:PlantEquipmentOperation:HeatingLoad",
    "OS:PlantEquipmentOperation:OutdoorDryBulb",
    "OS:PlantEquipmentOperation:OutdoorWetBulb",
    "OS:PlantEquipmentOperation:OutdoorRelativeHumidity",
    "OS:PlantEquipmentOperation:OutdoorDewpoint",
    "OS:PlantEquipmentOperation:OutdoorDryBulbDifference",
    "OS:PlantEquipmentOperation:OutdoorDewpointDifference",
    "OS:PlantEquipmentOperation:OutdoorWetBulbDifference",
    "OS:PlantComponent:TemperatureSource",
    "OS:LoadProfile:Plant",
    "OS:Pump:ConstantSpeed",
    "OS:Pump:VariableSpeed",
    "OS:HeaderedPumps:ConstantSpeed",
    "OS:HeaderedPumps:VariableSpeed",
    "OS:TemperingValve",
    "OS:SetpointManager:FollowOutdoorAirTemperature",
    "OS:SetpointManager:MixedAir",
    "OS:SetpointManager:OutdoorAirReset",
    "OS:SetpointManager:OutdoorAirPretreat",
    "OS:SetpointManager:SingleZone:Reheat",
    "OS:SetpointManager:SingleZone:Humidity:Minimum",
    "OS:SetpointManager:MultiZone:MinimumHumidity:Average",
    "OS:SetpointManager:MultiZone:Humidity:Minimum",
    "OS:SetpointManager:Scheduled",
    "OS:SetpointManager:Scheduled:DualSetpoint",
    "OS:SetpointManager:Warmest",
    "OS:SetpointManager:WarmestTemperatureFlow",
    "OS:SetpointManager:Coldest",
    "OS:SetpointManager:FollowGroundTemperature",
    "OS:SetpointManager:FollowSystemNodeTemperature",
    "OS:SetpointManager:MultiZone:Cooling:Average",
    "OS:SetpointManager:MultiZone:Heating:Average",
    "OS:SetpointManager:MultiZone:Humidity:Maximum",
    "OS:SetpointManager:MultiZone:MaximumHumidity:Average",
    "OS:SetpointManager:SingleZone:OneStageCooling",
    "OS:SetpointManager:SingleZone:OneStageHeating",
    "OS:SetpointManager:SingleZone:Humidity:Maximum",
    "OS:SetpointManager:SingleZone:Cooling",
    "OS:SetpointManager:SingleZone:Heating",
    "OS:SetpointManager:SystemNodeReset:Temperature",
    "OS:SetpointManager:SystemNodeReset:Humidity",
    "OS:Sizing:Plant",
    "OS:Sizing:System",
    "OS:Sizing:Zone",
    "OS:Splitter",
    "OS:ThermalZone",
    "OS:ZoneControl:Humidistat",
    "OS:ThermostatSetpoint:DualSetpoint",
    "OS:ZoneControl:Thermostat:StagedDualSetpoint",
    "OS:ZoneHVAC:EquipmentList",
    "OS:ZoneVentilation:DesignFlowRate",
    "OS:ZoneVentilation:WindandStackOpenArea",
    "OS:ZoneHVAC:Baseboard:Convective:Electric",
    "OS:ZoneHVAC:Baseboard:Convective:Water",
    "OS:ZoneHVAC:HighTemperatureRadiant",
    "OS:ZoneHVAC:LowTemperatureRadiant:Electric",
    "OS:ZoneHVAC:IdealLoadsAirSystem",
    "OS:ZoneHVAC:FourPipeFanCoil",
    "OS:ZoneHVAC:LowTemperatureRadiant:ConstantFlow",
    "OS:ZoneHVAC:LowTemperatureRadiant:VariableFlow",
    "OS:ZoneHVAC:PackagedTerminalHeatPump",
    "OS:ZoneHVAC:PackagedTerminalAirConditioner",
    "OS:ZoneHVAC:TerminalUnit:VariableRefrigerantFlow",
    "OS:ZoneHVAC:WaterToAirHeatPump",
    "OS:ZoneHVAC:UnitHeater",
    "OS:ZoneHVAC:Dehumidifier:DX",
    "OS:ZoneHVAC:EnergyRecoveryVentilator",
    "OS:ZoneHVAC:EnergyRecoveryVentilator:Controller",
    "OS:ZoneHVAC:UnitVentilator",
    "OS:ZoneHVAC:Baseboard:RadiantConvective:Water",
    "OS:ZoneHVAC:Baseboard:RadiantConvective:Electric",
    "OS:ZoneHVAC:CoolingPanel:RadiantConvective:Water",
    "OS:ZoneMixing",
    "OS:WaterHeater:Mixed",
    "OS:WaterHeater:HeatPump",
    "OS:Coil:WaterHeating:AirToWaterHeatPump",
    "OS:WaterHeater:HeatPump:WrappedCondenser",
    "OS:Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed",
    "OS:Coil:WaterHeating:AirToWaterHeatPump:VariableSpeed:SpeedData",
    "OS:Coil:WaterHeating:AirToWaterHeatPump:Wrapped",
    "OS:WaterHeater:Stratified",
    "OS:WaterHeater:Sizing",
    "OS:WaterUse:Equipment",
    "OS:WaterUse:Equipment:Definition",
    "OS:WaterUse:Connections",
    "OS:ThermalStorage:Ice:Detailed",
    "OS:ThermalStorage:ChilledWater:Stratified",
    "OS:Duct",
    "OS:FluidCooler:SingleSpeed",
    "OS:FluidCooler:TwoSpeed",
    "OS:Pipe:Indoor",
    "OS:Pipe:Outdoor",
    "OS:SwimmingPool:Indoor",
    "OS:SolarCollectorPerformance:FlatPlate",
    "OS:SolarCollector:FlatPlate:Water",
    "OS:SolarCollectorPerformance:IntegralCollectorStorage",
    "OS:SolarCollector:IntegralCollectorStorage",
    "OS:SolarCollectorPerformance:PhotovoltaicThermal:Simple",
    "OS:SolarCollectorPerformance:PhotovoltaicThermal:BIPVT",
    "OS:SolarCollector:FlatPlate:PhotovoltaicThermal",
    "OS:Generator:MicroTurbine",
    "OS:Generator:MicroTurbine:HeatRecovery",
    "OS:Generator:Photovoltaic",
    "OS:PhotovoltaicPerformance:Simple",
    "OS:PhotovoltaicPerformance:EquivalentOne-Diode",
    "OS:PhotovoltaicPerformance:Sandia",
    "OS:Generator:FuelCell",
    "OS:Generator:FuelCell:PowerModule",
    "OS:Generator:FuelCell:AirSupply",
    "OS:Generator:FuelCell:WaterSupply",
    "OS:Generator:FuelCell:AuxiliaryHeater",
    "OS:Generator:FuelCell:ExhaustGasToWaterHeatExchanger",
    "OS:Generator:FuelCell:ElectricalStorage",
    "OS:Generator:FuelCell:Inverter",
    "OS:Generator:FuelCell:StackCooler",
    "OS:Generator:FuelSupply",
    "OS:Generator:WindTurbine",
    "OS:Generator:PVWatts",
    "OS:ElectricLoadCenter:Inverter:PVWatts",
    "OS:ElectricLoadCenter:Inverter:Simple",
    "OS:ElectricLoadCenter:Inverter:LookUpTable",
    "OS:ElectricLoadCenter:Storage:Simple",
    "OS:ElectricLoadCenter:Storage:LiIonNMCBattery",
    "OS:ElectricLoadCenter:Transformer",
    "OS:ElectricLoadCenter:Distribution",
    "OS:ElectricLoadCenter:Storage:Converter",
    "OS:ComponentCost:Adjustments",
    "OS:LifeCycleCost",
    "OS:CurrencyType",
    "OS:LifeCycleCost:Parameters",
    "OS:LifeCycleCost:UsePriceEscalation",
    "OS:UtilityBill",
    "OS:UtilityCost:Tariff",
    "OS:UtilityCost:Qualify",
    "OS:UtilityCost:Charge:Simple",
    "OS:UtilityCost:Charge:Block",
    "OS:UtilityCost:Ratchet",
    "OS:UtilityCost:Variable",
    "OS:UtilityCost:Computation",
    "OS:Output:Schedules",
    "OS:Output:Constructions",
    "OS:OutputControl:Files",
    "OS:OutputControl:Timestamp",
    "OS:OutputControl:ReportingTolerances",
    "OS:OutputControl:Table:Style",
    "OS:Output:Meter",
    "OS:Output:Variable",
    "OS:Meter:Custom",
    "OS:Meter:CustomDecrement",
    "OS:Output:EnvironmentalImpactFactors",
    "OS:EnvironmentalImpactFactors",
    "OS:FuelFactors",
    "OS:Output:Diagnostics",
    "OS:Output:DebuggingData",
    "OS:Output:JSON",
    "OS:Output:SQLite",
    "OS:Output:EnergyManagementSystem",
    "OS:Output:Table:SummaryReports",
    "OS:EnergyManagementSystem:Sensor",
    "OS:EnergyManagementSystem:Actuator",
    "OS:EnergyManagementSystem:ProgramCallingManager",
    "OS:EnergyManagementSystem:Program",
    "OS:EnergyManagementSystem:Subroutine",
    "OS:EnergyManagementSystem:GlobalVariable",
    "OS:EnergyManagementSystem:OutputVariable",
    "OS:EnergyManagementSystem:MeteredOutputVariable",
    "OS:EnergyManagementSystem:TrendVariable",
    "OS:EnergyManagementSystem:InternalVariable",
    "OS:EnergyManagementSystem:CurveOrTableIndexVariable",
    "OS:EnergyManagementSystem:ConstructionIndexVariable",
    "OS:PythonPlugin:Instance",
    "OS:PythonPlugin:Variable",
    "OS:PythonPlugin:TrendVariable",
    "OS:PythonPlugin:OutputVariable",
    "OS:AirflowNetworkSimulationControl",
    "OS:AirflowNetworkZone",
    "OS:AirflowNetworkSurface",
    "OS:AirflowNetworkReferenceCrackConditions",
    "OS:AirflowNetworkCrack",
    "OS:AirflowNetworkEffectiveLeakageArea",
    "OS:AirflowNetworkSpecifiedFlowRate",
    "OS:AirflowNetworkDetailedOpening",
    "OS:AirflowNetworkSimpleOpening",
    "OS:AirflowNetworkHorizontalOpening",
    "OS:AirflowNetworkOutdoorAirflow",
    "OS:AirflowNetworkZoneExhaustFan",
    "OS:AirflowNetworkExternalNode",
    "OS:AirflowNetworkPressureController",
    "OS:AirflowNetworkDistributionNode",
    "OS:AirflowNetworkLeak",
    "OS:AirflowNetworkLeakageRatio",
    "OS:AirflowNetworkDuct",
    "OS:AirflowNetworkFan",
    "OS:AirflowNetworkEquivalentDuct",
    "OS:AirflowNetworkConstantPressureDrop",
    "OS:AirflowNetworkReliefAirFlow",
    "OS:AirflowNetworkDistributionLinkage",
    "OS:AirflowNetworkDuctViewFactors",
    "OS:AirflowNetworkOccupantVentilationControl",
    "OS:ExternalInterface",
    "OS:ExternalInterface:Schedule",
    "OS:ExternalInterface:Variable",
    "OS:ExternalInterface:Actuator",
    "OS:ExternalInterface:FunctionalMockupUnitImport",
    "OS:ExternalInterface:FunctionalMockupUnitImport:From:Variable",
    "OS:ExternalInterface:FunctionalMockupUnitImport:To:Schedule",
    "OS:ExternalInterface:FunctionalMockupUnitImport:To:Actuator",
    "OS:ExternalInterface:FunctionalMockupUnitImport:To:Variable",
    "OS:ExternalInterface:FunctionalMockupUnitExport:From:Variable",
    "OS:ExternalInterface:FunctionalMockupUnitExport:To:Schedule",
    "OS:ExternalInterface:FunctionalMockupUnitExport:To:Actuator",
    "OS:ExternalInterface:FunctionalMockupUnitExport:To:Variable",
    "OS:Coil:UserDefined",
    "OS:PlantComponent:UserDefined",
    "",
  };

  static constexpr std::array<std::string_view, Size> descriptionsUC{
    "",
    "",
    "LEAD INPUT",
    "END LEAD INPUT",
    "SIMULATION DATA",
    "END SIMULATION DATA",
    "VERSION",
    "SIMULATIONCONTROL",
    "PERFORMANCEPRECISIONTRADEOFFS",
    "BUILDING",
    "SHADOWCALCULATION",
    "SURFACECONVECTIONALGORITHM:INSIDE",
    "SURFACECONVECTIONALGORITHM:OUTSIDE",
    "HEATBALANCEALGORITHM",
    "HEATBALANCESETTINGS:CONDUCTIONFINITEDIFFERENCE",
    "ZONEAIRHEATBALANCEALGORITHM",
    "ZONEAIRCONTAMINANTBALANCE",
    "ZONEAIRMASSFLOWCONSERVATION",
    "ZONECAPACITANCEMULTIPLIER:RESEARCHSPECIAL",
    "TIMESTEP",
    "CONVERGENCELIMITS",
    "HVACSYSTEMROOTFINDINGALGORITHM",
    "COMPLIANCE:BUILDING",
    "SITE:LOCATION",
    "SITE:VARIABLELOCATION",
    "SIZINGPERIOD:DESIGNDAY",
    "SIZINGPERIOD:WEATHERFILEDAYS",
    "SIZINGPERIOD:WEATHERFILECONDITIONTYPE",
    "RUNPERIOD",
    "RUNPERIODCONTROL:SPECIALDAYS",
    "RUNPERIODCONTROL:DAYLIGHTSAVINGTIME",
    "WEATHERPROPERTY:SKYTEMPERATURE",
    "SITE:WEATHERSTATION",
    "SITE:HEIGHTVARIATION",
    "SITE:GROUNDTEMPERATURE:BUILDINGSURFACE",
    "SITE:GROUNDTEMPERATURE:FCFACTORMETHOD",
    "SITE:GROUNDTEMPERATURE:SHALLOW",
    "SITE:GROUNDTEMPERATURE:DEEP",
    "SITE:GROUNDTEMPERATURE:UNDISTURBED:FINITEDIFFERENCE",
    "SITE:GROUNDTEMPERATURE:UNDISTURBED:KUSUDAACHENBACH",
    "SITE:GROUNDTEMPERATURE:UNDISTURBED:XING",
    "SITE:GROUNDDOMAIN:SLAB",
    "SITE:GROUNDDOMAIN:BASEMENT",
    "SITE:GROUNDREFLECTANCE",
    "SITE:GROUNDREFLECTANCE:SNOWMODIFIER",
    "SITE:WATERMAINSTEMPERATURE",
    "SITE:PRECIPITATION",
    "ROOFIRRIGATION",
    "SITE:SOLARANDVISIBLESPECTRUM",
    "SITE:SPECTRUMDATA",
    "SCHEDULETYPELIMITS",
    "SCHEDULE:DAY:HOURLY",
    "SCHEDULE:DAY:INTERVAL",
    "SCHEDULE:DAY:LIST",
    "SCHEDULE:WEEK:DAILY",
    "SCHEDULE:WEEK:COMPACT",
    "SCHEDULE:YEAR",
    "SCHEDULE:COMPACT",
    "SCHEDULE:CONSTANT",
    "SCHEDULE:FILE:SHADING",
    "SCHEDULE:FILE",
    "MATERIAL",
    "MATERIAL:NOMASS",
    "MATERIAL:INFRAREDTRANSPARENT",
    "MATERIAL:AIRGAP",
    "MATERIAL:ROOFVEGETATION",
    "WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
    "WINDOWMATERIAL:GLAZING",
    "WINDOWMATERIAL:GLAZINGGROUP:THERMOCHROMIC",
    "WINDOWMATERIAL:GLAZING:REFRACTIONEXTINCTIONMETHOD",
    "WINDOWMATERIAL:GAS",
    "WINDOWGAP:SUPPORTPILLAR",
    "WINDOWGAP:DEFLECTIONSTATE",
    "WINDOWMATERIAL:GASMIXTURE",
    "WINDOWMATERIAL:GAP",
    "WINDOWMATERIAL:SHADE",
    "WINDOWMATERIAL:COMPLEXSHADE",
    "WINDOWMATERIAL:BLIND",
    "WINDOWMATERIAL:SCREEN",
    "WINDOWMATERIAL:SHADE:EQUIVALENTLAYER",
    "WINDOWMATERIAL:DRAPE:EQUIVALENTLAYER",
    "WINDOWMATERIAL:BLIND:EQUIVALENTLAYER",
    "WINDOWMATERIAL:SCREEN:EQUIVALENTLAYER",
    "WINDOWMATERIAL:GLAZING:EQUIVALENTLAYER",
    "WINDOWMATERIAL:GAP:EQUIVALENTLAYER",
    "MATERIALPROPERTY:MOISTUREPENETRATIONDEPTH:SETTINGS",
    "MATERIALPROPERTY:PHASECHANGE",
    "MATERIALPROPERTY:PHASECHANGEHYSTERESIS",
    "MATERIALPROPERTY:VARIABLETHERMALCONDUCTIVITY",
    "MATERIALPROPERTY:VARIABLEABSORPTANCE",
    "MATERIALPROPERTY:HEATANDMOISTURETRANSFER:SETTINGS",
    "MATERIALPROPERTY:HEATANDMOISTURETRANSFER:SORPTIONISOTHERM",
    "MATERIALPROPERTY:HEATANDMOISTURETRANSFER:SUCTION",
    "MATERIALPROPERTY:HEATANDMOISTURETRANSFER:REDISTRIBUTION",
    "MATERIALPROPERTY:HEATANDMOISTURETRANSFER:DIFFUSION",
    "MATERIALPROPERTY:HEATANDMOISTURETRANSFER:THERMALCONDUCTIVITY",
    "MATERIALPROPERTY:GLAZINGSPECTRALDATA",
    "CONSTRUCTION",
    "CONSTRUCTION:CFACTORUNDERGROUNDWALL",
    "CONSTRUCTION:FFACTORGROUNDFLOOR",
    "CONSTRUCTIONPROPERTY:INTERNALHEATSOURCE",
    "CONSTRUCTION:AIRBOUNDARY",
    "WINDOWTHERMALMODEL:PARAMS",
    "WINDOWSCALCULATIONENGINE",
    "CONSTRUCTION:COMPLEXFENESTRATIONSTATE",
    "CONSTRUCTION:WINDOWEQUIVALENTLAYER",
    "CONSTRUCTION:WINDOWDATAFILE",
    "GLOBALGEOMETRYRULES",
    "GEOMETRYTRANSFORM",
    "SPACE",
    "SPACELIST",
    "ZONE",
    "ZONELIST",
    "ZONEGROUP",
    "BUILDINGSURFACE:DETAILED",
    "WALL:DETAILED",
    "ROOFCEILING:DETAILED",
    "FLOOR:DETAILED",
    "WALL:EXTERIOR",
    "WALL:ADIABATIC",
    "WALL:UNDERGROUND",
    "WALL:INTERZONE",
    "ROOF",
    "CEILING:ADIABATIC",
    "CEILING:INTERZONE",
    "FLOOR:GROUNDCONTACT",
    "FLOOR:ADIABATIC",
    "FLOOR:INTERZONE",
    "FENESTRATIONSURFACE:DETAILED",
    "WINDOW",
    "DOOR",
    "GLAZEDDOOR",
    "WINDOW:INTERZONE",
    "DOOR:INTERZONE",
    "GLAZEDDOOR:INTERZONE",
    "WINDOWSHADINGCONTROL",
    "WINDOWPROPERTY:FRAMEANDDIVIDER",
    "WINDOWPROPERTY:AIRFLOWCONTROL",
    "WINDOWPROPERTY:STORMWINDOW",
    "INTERNALMASS",
    "SHADING:SITE",
    "SHADING:BUILDING",
    "SHADING:SITE:DETAILED",
    "SHADING:BUILDING:DETAILED",
    "SHADING:OVERHANG",
    "SHADING:OVERHANG:PROJECTION",
    "SHADING:FIN",
    "SHADING:FIN:PROJECTION",
    "SHADING:ZONE:DETAILED",
    "SHADINGPROPERTY:REFLECTANCE",
    "SURFACEPROPERTY:HEATTRANSFERALGORITHM",
    "SURFACEPROPERTY:HEATTRANSFERALGORITHM:MULTIPLESURFACE",
    "SURFACEPROPERTY:HEATTRANSFERALGORITHM:SURFACELIST",
    "SURFACEPROPERTY:HEATTRANSFERALGORITHM:CONSTRUCTION",
    "SURFACEPROPERTY:HEATBALANCESOURCETERM",
    "SURFACECONTROL:MOVABLEINSULATION",
    "SURFACEPROPERTY:OTHERSIDECOEFFICIENTS",
    "SURFACEPROPERTY:OTHERSIDECONDITIONSMODEL",
    "SURFACEPROPERTY:UNDERWATER",
    "FOUNDATION:KIVA",
    "FOUNDATION:KIVA:SETTINGS",
    "SURFACEPROPERTY:EXPOSEDFOUNDATIONPERIMETER",
    "SURFACECONVECTIONALGORITHM:INSIDE:ADAPTIVEMODELSELECTIONS",
    "SURFACECONVECTIONALGORITHM:OUTSIDE:ADAPTIVEMODELSELECTIONS",
    "SURFACECONVECTIONALGORITHM:INSIDE:USERCURVE",
    "SURFACECONVECTIONALGORITHM:OUTSIDE:USERCURVE",
    "SURFACEPROPERTY:CONVECTIONCOEFFICIENTS",
    "SURFACEPROPERTY:CONVECTIONCOEFFICIENTS:MULTIPLESURFACE",
    "SURFACEPROPERTIES:VAPORCOEFFICIENTS",
    "SURFACEPROPERTY:EXTERIORNATURALVENTEDCAVITY",
    "SURFACEPROPERTY:SOLARINCIDENTINSIDE",
    "SURFACEPROPERTY:INCIDENTSOLARMULTIPLIER",
    "SURFACEPROPERTY:LOCALENVIRONMENT",
    "ZONEPROPERTY:LOCALENVIRONMENT",
    "SURFACEPROPERTY:SURROUNDINGSURFACES",
    "SURFACEPROPERTY:GROUNDSURFACES",
    "COMPLEXFENESTRATIONPROPERTY:SOLARABSORBEDLAYERS",
    "ZONEPROPERTY:USERVIEWFACTORS:BYSURFACENAME",
    "GROUNDHEATTRANSFER:CONTROL",
    "GROUNDHEATTRANSFER:SLAB:MATERIALS",
    "GROUNDHEATTRANSFER:SLAB:MATLPROPS",
    "GROUNDHEATTRANSFER:SLAB:BOUNDCONDS",
    "GROUNDHEATTRANSFER:SLAB:BLDGPROPS",
    "GROUNDHEATTRANSFER:SLAB:INSULATION",
    "GROUNDHEATTRANSFER:SLAB:EQUIVALENTSLAB",
    "GROUNDHEATTRANSFER:SLAB:AUTOGRID",
    "GROUNDHEATTRANSFER:SLAB:MANUALGRID",
    "GROUNDHEATTRANSFER:SLAB:XFACE",
    "GROUNDHEATTRANSFER:SLAB:YFACE",
    "GROUNDHEATTRANSFER:SLAB:ZFACE",
    "GROUNDHEATTRANSFER:BASEMENT:SIMPARAMETERS",
    "GROUNDHEATTRANSFER:BASEMENT:MATLPROPS",
    "GROUNDHEATTRANSFER:BASEMENT:INSULATION",
    "GROUNDHEATTRANSFER:BASEMENT:SURFACEPROPS",
    "GROUNDHEATTRANSFER:BASEMENT:BLDGDATA",
    "GROUNDHEATTRANSFER:BASEMENT:INTERIOR",
    "GROUNDHEATTRANSFER:BASEMENT:COMBLDG",
    "GROUNDHEATTRANSFER:BASEMENT:EQUIVSLAB",
    "GROUNDHEATTRANSFER:BASEMENT:EQUIVAUTOGRID",
    "GROUNDHEATTRANSFER:BASEMENT:AUTOGRID",
    "GROUNDHEATTRANSFER:BASEMENT:MANUALGRID",
    "GROUNDHEATTRANSFER:BASEMENT:XFACE",
    "GROUNDHEATTRANSFER:BASEMENT:YFACE",
    "GROUNDHEATTRANSFER:BASEMENT:ZFACE",
    "ROOMAIRMODELTYPE",
    "ROOMAIR:TEMPERATUREPATTERN:USERDEFINED",
    "ROOMAIR:TEMPERATUREPATTERN:CONSTANTGRADIENT",
    "ROOMAIR:TEMPERATUREPATTERN:TWOGRADIENT",
    "ROOMAIR:TEMPERATUREPATTERN:NONDIMENSIONALHEIGHT",
    "ROOMAIR:TEMPERATUREPATTERN:SURFACEMAPPING",
    "ROOMAIR:NODE",
    "ROOMAIRSETTINGS:ONENODEDISPLACEMENTVENTILATION",
    "ROOMAIRSETTINGS:THREENODEDISPLACEMENTVENTILATION",
    "ROOMAIRSETTINGS:CROSSVENTILATION",
    "ROOMAIRSETTINGS:UNDERFLOORAIRDISTRIBUTIONINTERIOR",
    "ROOMAIRSETTINGS:UNDERFLOORAIRDISTRIBUTIONEXTERIOR",
    "ROOMAIR:NODE:AIRFLOWNETWORK",
    "ROOMAIR:NODE:AIRFLOWNETWORK:ADJACENTSURFACELIST",
    "ROOMAIR:NODE:AIRFLOWNETWORK:INTERNALGAINS",
    "ROOMAIR:NODE:AIRFLOWNETWORK:HVACEQUIPMENT",
    "ROOMAIRSETTINGS:AIRFLOWNETWORK",
    "PEOPLE",
    "COMFORTVIEWFACTORANGLES",
    "LIGHTS",
    "ELECTRICEQUIPMENT",
    "GASEQUIPMENT",
    "HOTWATEREQUIPMENT",
    "STEAMEQUIPMENT",
    "OTHEREQUIPMENT",
    "ELECTRICEQUIPMENT:ITE:AIRCOOLED",
    "ZONEBASEBOARD:OUTDOORTEMPERATURECONTROLLED",
    "SWIMMINGPOOL:INDOOR",
    "ZONECONTAMINANTSOURCEANDSINK:CARBONDIOXIDE",
    "ZONECONTAMINANTSOURCEANDSINK:GENERIC:CONSTANT",
    "SURFACECONTAMINANTSOURCEANDSINK:GENERIC:PRESSUREDRIVEN",
    "ZONECONTAMINANTSOURCEANDSINK:GENERIC:CUTOFFMODEL",
    "ZONECONTAMINANTSOURCEANDSINK:GENERIC:DECAYSOURCE",
    "SURFACECONTAMINANTSOURCEANDSINK:GENERIC:BOUNDARYLAYERDIFFUSION",
    "SURFACECONTAMINANTSOURCEANDSINK:GENERIC:DEPOSITIONVELOCITYSINK",
    "ZONECONTAMINANTSOURCEANDSINK:GENERIC:DEPOSITIONRATESINK",
    "DAYLIGHTING:CONTROLS",
    "DAYLIGHTING:REFERENCEPOINT",
    "DAYLIGHTING:DELIGHT:COMPLEXFENESTRATION",
    "DAYLIGHTINGDEVICE:TUBULAR",
    "DAYLIGHTINGDEVICE:SHELF",
    "DAYLIGHTINGDEVICE:LIGHTWELL",
    "OUTPUT:DAYLIGHTFACTORS",
    "OUTPUT:ILLUMINANCEMAP",
    "OUTPUTCONTROL:ILLUMINANCEMAP:STYLE",
    "ZONEINFILTRATION:DESIGNFLOWRATE",
    "ZONEINFILTRATION:EFFECTIVELEAKAGEAREA",
    "ZONEINFILTRATION:FLOWCOEFFICIENT",
    "ZONEVENTILATION:DESIGNFLOWRATE",
    "ZONEVENTILATION:WINDANDSTACKOPENAREA",
    "ZONEAIRBALANCE:OUTDOORAIR",
    "ZONEMIXING",
    "ZONECROSSMIXING",
    "ZONEREFRIGERATIONDOORMIXING",
    "ZONEEARTHTUBE",
    "ZONECOOLTOWER:SHOWER",
    "ZONETHERMALCHIMNEY",
    "AIRFLOWNETWORK:SIMULATIONCONTROL",
    "AIRFLOWNETWORK:MULTIZONE:ZONE",
    "AIRFLOWNETWORK:MULTIZONE:SURFACE",
    "AIRFLOWNETWORK:MULTIZONE:REFERENCECRACKCONDITIONS",
    "AIRFLOWNETWORK:MULTIZONE:SURFACE:CRACK",
    "AIRFLOWNETWORK:MULTIZONE:SURFACE:EFFECTIVELEAKAGEAREA",
    "AIRFLOWNETWORK:MULTIZONE:SPECIFIEDFLOWRATE",
    "AIRFLOWNETWORK:MULTIZONE:COMPONENT:DETAILEDOPENING",
    "AIRFLOWNETWORK:MULTIZONE:COMPONENT:SIMPLEOPENING",
    "AIRFLOWNETWORK:MULTIZONE:COMPONENT:HORIZONTALOPENING",
    "AIRFLOWNETWORK:MULTIZONE:COMPONENT:ZONEEXHAUSTFAN",
    "AIRFLOWNETWORK:MULTIZONE:EXTERNALNODE",
    "AIRFLOWNETWORK:MULTIZONE:WINDPRESSURECOEFFICIENTARRAY",
    "AIRFLOWNETWORK:MULTIZONE:WINDPRESSURECOEFFICIENTVALUES",
    "AIRFLOWNETWORK:ZONECONTROL:PRESSURECONTROLLER",
    "AIRFLOWNETWORK:DISTRIBUTION:NODE",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:LEAK",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:LEAKAGERATIO",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:DUCT",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:FAN",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:COIL",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:HEATEXCHANGER",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:TERMINALUNIT",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:CONSTANTPRESSUREDROP",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:OUTDOORAIRFLOW",
    "AIRFLOWNETWORK:DISTRIBUTION:COMPONENT:RELIEFAIRFLOW",
    "AIRFLOWNETWORK:DISTRIBUTION:LINKAGE",
    "AIRFLOWNETWORK:DISTRIBUTION:DUCTVIEWFACTORS",
    "AIRFLOWNETWORK:OCCUPANTVENTILATIONCONTROL",
    "AIRFLOWNETWORK:INTRAZONE:NODE",
    "AIRFLOWNETWORK:INTRAZONE:LINKAGE",
    "EXTERIOR:LIGHTS",
    "EXTERIOR:FUELEQUIPMENT",
    "EXTERIOR:WATEREQUIPMENT",
    "HVACTEMPLATE:THERMOSTAT",
    "HVACTEMPLATE:ZONE:IDEALLOADSAIRSYSTEM",
    "HVACTEMPLATE:ZONE:BASEBOARDHEAT",
    "HVACTEMPLATE:ZONE:FANCOIL",
    "HVACTEMPLATE:ZONE:PTAC",
    "HVACTEMPLATE:ZONE:PTHP",
    "HVACTEMPLATE:ZONE:WATERTOAIRHEATPUMP",
    "HVACTEMPLATE:ZONE:VRF",
    "HVACTEMPLATE:ZONE:UNITARY",
    "HVACTEMPLATE:ZONE:VAV",
    "HVACTEMPLATE:ZONE:VAV:FANPOWERED",
    "HVACTEMPLATE:ZONE:VAV:HEATANDCOOL",
    "HVACTEMPLATE:ZONE:CONSTANTVOLUME",
    "HVACTEMPLATE:ZONE:DUALDUCT",
    "HVACTEMPLATE:SYSTEM:VRF",
    "HVACTEMPLATE:SYSTEM:UNITARY",
    "HVACTEMPLATE:SYSTEM:UNITARYHEATPUMP:AIRTOAIR",
    "HVACTEMPLATE:SYSTEM:UNITARYSYSTEM",
    "HVACTEMPLATE:SYSTEM:VAV",
    "HVACTEMPLATE:SYSTEM:PACKAGEDVAV",
    "HVACTEMPLATE:SYSTEM:CONSTANTVOLUME",
    "HVACTEMPLATE:SYSTEM:DUALDUCT",
    "HVACTEMPLATE:SYSTEM:DEDICATEDOUTDOORAIR",
    "HVACTEMPLATE:PLANT:CHILLEDWATERLOOP",
    "HVACTEMPLATE:PLANT:CHILLER",
    "HVACTEMPLATE:PLANT:CHILLER:OBJECTREFERENCE",
    "HVACTEMPLATE:PLANT:TOWER",
    "HVACTEMPLATE:PLANT:TOWER:OBJECTREFERENCE",
    "HVACTEMPLATE:PLANT:HOTWATERLOOP",
    "HVACTEMPLATE:PLANT:BOILER",
    "HVACTEMPLATE:PLANT:BOILER:OBJECTREFERENCE",
    "HVACTEMPLATE:PLANT:MIXEDWATERLOOP",
    "DESIGNSPECIFICATION:OUTDOORAIR",
    "DESIGNSPECIFICATION:OUTDOORAIR:SPACELIST",
    "DESIGNSPECIFICATION:ZONEAIRDISTRIBUTION",
    "SIZING:PARAMETERS",
    "SIZING:ZONE",
    "DESIGNSPECIFICATION:ZONEHVAC:SIZING",
    "DESIGNSPECIFICATION:AIRTERMINAL:SIZING",
    "SIZING:SYSTEM",
    "SIZING:PLANT",
    "OUTPUTCONTROL:SIZING:STYLE",
    "ZONECONTROL:HUMIDISTAT",
    "ZONECONTROL:THERMOSTAT",
    "ZONECONTROL:THERMOSTAT:OPERATIVETEMPERATURE",
    "ZONECONTROL:THERMOSTAT:THERMALCOMFORT",
    "ZONECONTROL:THERMOSTAT:TEMPERATUREANDHUMIDITY",
    "THERMOSTATSETPOINT:SINGLEHEATING",
    "THERMOSTATSETPOINT:SINGLECOOLING",
    "THERMOSTATSETPOINT:SINGLEHEATINGORCOOLING",
    "THERMOSTATSETPOINT:DUALSETPOINT",
    "THERMOSTATSETPOINT:THERMALCOMFORT:FANGER:SINGLEHEATING",
    "THERMOSTATSETPOINT:THERMALCOMFORT:FANGER:SINGLECOOLING",
    "THERMOSTATSETPOINT:THERMALCOMFORT:FANGER:SINGLEHEATINGORCOOLING",
    "THERMOSTATSETPOINT:THERMALCOMFORT:FANGER:DUALSETPOINT",
    "ZONECONTROL:THERMOSTAT:STAGEDDUALSETPOINT",
    "ZONECONTROL:CONTAMINANTCONTROLLER",
    "ZONEHVAC:IDEALLOADSAIRSYSTEM",
    "ZONEHVAC:FOURPIPEFANCOIL",
    "ZONEHVAC:WINDOWAIRCONDITIONER",
    "ZONEHVAC:PACKAGEDTERMINALAIRCONDITIONER",
    "ZONEHVAC:PACKAGEDTERMINALHEATPUMP",
    "ZONEHVAC:WATERTOAIRHEATPUMP",
    "ZONEHVAC:DEHUMIDIFIER:DX",
    "ZONEHVAC:ENERGYRECOVERYVENTILATOR",
    "ZONEHVAC:ENERGYRECOVERYVENTILATOR:CONTROLLER",
    "ZONEHVAC:UNITVENTILATOR",
    "ZONEHVAC:UNITHEATER",
    "ZONEHVAC:EVAPORATIVECOOLERUNIT",
    "ZONEHVAC:HYBRIDUNITARYHVAC",
    "ZONEHVAC:OUTDOORAIRUNIT",
    "ZONEHVAC:OUTDOORAIRUNIT:EQUIPMENTLIST",
    "ZONEHVAC:TERMINALUNIT:VARIABLEREFRIGERANTFLOW",
    "ZONEHVAC:BASEBOARD:RADIANTCONVECTIVE:WATER:DESIGN",
    "ZONEHVAC:BASEBOARD:RADIANTCONVECTIVE:WATER",
    "ZONEHVAC:BASEBOARD:RADIANTCONVECTIVE:STEAM:DESIGN",
    "ZONEHVAC:BASEBOARD:RADIANTCONVECTIVE:STEAM",
    "ZONEHVAC:BASEBOARD:RADIANTCONVECTIVE:ELECTRIC",
    "ZONEHVAC:COOLINGPANEL:RADIANTCONVECTIVE:WATER",
    "ZONEHVAC:BASEBOARD:CONVECTIVE:WATER",
    "ZONEHVAC:BASEBOARD:CONVECTIVE:ELECTRIC",
    "ZONEHVAC:LOWTEMPERATURERADIANT:VARIABLEFLOW",
    "ZONEHVAC:LOWTEMPERATURERADIANT:VARIABLEFLOW:DESIGN",
    "ZONEHVAC:LOWTEMPERATURERADIANT:CONSTANTFLOW",
    "ZONEHVAC:LOWTEMPERATURERADIANT:CONSTANTFLOW:DESIGN",
    "ZONEHVAC:LOWTEMPERATURERADIANT:ELECTRIC",
    "ZONEHVAC:LOWTEMPERATURERADIANT:SURFACEGROUP",
    "ZONEHVAC:HIGHTEMPERATURERADIANT",
    "ZONEHVAC:VENTILATEDSLAB",
    "ZONEHVAC:VENTILATEDSLAB:SLABGROUP",
    "AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:REHEAT",
    "AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:NOREHEAT",
    "AIRTERMINAL:SINGLEDUCT:VAV:NOREHEAT",
    "AIRTERMINAL:SINGLEDUCT:VAV:REHEAT",
    "AIRTERMINAL:SINGLEDUCT:VAV:REHEAT:VARIABLESPEEDFAN",
    "AIRTERMINAL:SINGLEDUCT:VAV:HEATANDCOOL:NOREHEAT",
    "AIRTERMINAL:SINGLEDUCT:VAV:HEATANDCOOL:REHEAT",
    "AIRTERMINAL:SINGLEDUCT:SERIESPIU:REHEAT",
    "AIRTERMINAL:SINGLEDUCT:PARALLELPIU:REHEAT",
    "AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:FOURPIPEINDUCTION",
    "AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:FOURPIPEBEAM",
    "AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:COOLEDBEAM",
    "AIRTERMINAL:SINGLEDUCT:MIXER",
    "AIRTERMINAL:DUALDUCT:CONSTANTVOLUME",
    "AIRTERMINAL:DUALDUCT:VAV",
    "AIRTERMINAL:DUALDUCT:VAV:OUTDOORAIR",
    "ZONEHVAC:AIRDISTRIBUTIONUNIT",
    "ZONEHVAC:EXHAUSTCONTROL",
    "ZONEHVAC:EQUIPMENTLIST",
    "ZONEHVAC:EQUIPMENTCONNECTIONS",
    "FAN:SYSTEMMODEL",
    "FAN:CONSTANTVOLUME",
    "FAN:VARIABLEVOLUME",
    "FAN:ONOFF",
    "FAN:ZONEEXHAUST",
    "FANPERFORMANCE:NIGHTVENTILATION",
    "FAN:COMPONENTMODEL",
    "COIL:COOLING:WATER",
    "COIL:COOLING:WATER:DETAILEDGEOMETRY",
    "COILSYSTEM:COOLING:WATER",
    "COIL:COOLING:DX",
    "COIL:COOLING:DX:CURVEFIT:PERFORMANCE",
    "COIL:COOLING:DX:CURVEFIT:OPERATINGMODE",
    "COIL:COOLING:DX:CURVEFIT:SPEED",
    "COIL:COOLING:DX:SINGLESPEED",
    "COIL:COOLING:DX:TWOSPEED",
    "COIL:COOLING:DX:MULTISPEED",
    "COIL:COOLING:DX:VARIABLESPEED",
    "COIL:COOLING:DX:TWOSTAGEWITHHUMIDITYCONTROLMODE",
    "COILPERFORMANCE:DX:COOLING",
    "COIL:COOLING:DX:VARIABLEREFRIGERANTFLOW",
    "COIL:HEATING:DX:VARIABLEREFRIGERANTFLOW",
    "COIL:COOLING:DX:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL",
    "COIL:HEATING:DX:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL",
    "COIL:HEATING:WATER",
    "COIL:HEATING:STEAM",
    "COIL:HEATING:ELECTRIC",
    "COIL:HEATING:ELECTRIC:MULTISTAGE",
    "COIL:HEATING:FUEL",
    "COIL:HEATING:GAS:MULTISTAGE",
    "COIL:HEATING:DESUPERHEATER",
    "COIL:HEATING:DX:SINGLESPEED",
    "COIL:HEATING:DX:MULTISPEED",
    "COIL:HEATING:DX:VARIABLESPEED",
    "COIL:COOLING:WATERTOAIRHEATPUMP:PARAMETERESTIMATION",
    "COIL:HEATING:WATERTOAIRHEATPUMP:PARAMETERESTIMATION",
    "COIL:COOLING:WATERTOAIRHEATPUMP:EQUATIONFIT",
    "COIL:COOLING:WATERTOAIRHEATPUMP:VARIABLESPEEDEQUATIONFIT",
    "COIL:HEATING:WATERTOAIRHEATPUMP:EQUATIONFIT",
    "COIL:HEATING:WATERTOAIRHEATPUMP:VARIABLESPEEDEQUATIONFIT",
    "COIL:WATERHEATING:AIRTOWATERHEATPUMP:PUMPED",
    "COIL:WATERHEATING:AIRTOWATERHEATPUMP:WRAPPED",
    "COIL:WATERHEATING:AIRTOWATERHEATPUMP:VARIABLESPEED",
    "COIL:WATERHEATING:DESUPERHEATER",
    "COILSYSTEM:COOLING:DX",
    "COILSYSTEM:HEATING:DX",
    "COILSYSTEM:COOLING:WATER:HEATEXCHANGERASSISTED",
    "COILSYSTEM:COOLING:DX:HEATEXCHANGERASSISTED",
    "COILSYSTEM:INTEGRATEDHEATPUMP:AIRSOURCE",
    "COIL:COOLING:DX:SINGLESPEED:THERMALSTORAGE",
    "EVAPORATIVECOOLER:DIRECT:CELDEKPAD",
    "EVAPORATIVECOOLER:INDIRECT:CELDEKPAD",
    "EVAPORATIVECOOLER:INDIRECT:WETCOIL",
    "EVAPORATIVECOOLER:INDIRECT:RESEARCHSPECIAL",
    "EVAPORATIVECOOLER:DIRECT:RESEARCHSPECIAL",
    "HUMIDIFIER:STEAM:ELECTRIC",
    "HUMIDIFIER:STEAM:GAS",
    "DEHUMIDIFIER:DESICCANT:NOFANS",
    "DEHUMIDIFIER:DESICCANT:SYSTEM",
    "HEATEXCHANGER:AIRTOAIR:FLATPLATE",
    "HEATEXCHANGER:AIRTOAIR:SENSIBLEANDLATENT",
    "HEATEXCHANGER:DESICCANT:BALANCEDFLOW",
    "HEATEXCHANGER:DESICCANT:BALANCEDFLOW:PERFORMANCEDATATYPE1",
    "AIRLOOPHVAC:UNITARYSYSTEM",
    "UNITARYSYSTEMPERFORMANCE:MULTISPEED",
    "AIRLOOPHVAC:UNITARY:FURNACE:HEATONLY",
    "AIRLOOPHVAC:UNITARY:FURNACE:HEATCOOL",
    "AIRLOOPHVAC:UNITARYHEATONLY",
    "AIRLOOPHVAC:UNITARYHEATCOOL",
    "AIRLOOPHVAC:UNITARYHEATPUMP:AIRTOAIR",
    "AIRLOOPHVAC:UNITARYHEATPUMP:WATERTOAIR",
    "AIRLOOPHVAC:UNITARYHEATCOOL:VAVCHANGEOVERBYPASS",
    "AIRLOOPHVAC:UNITARYHEATPUMP:AIRTOAIR:MULTISPEED",
    "AIRCONDITIONER:VARIABLEREFRIGERANTFLOW",
    "AIRCONDITIONER:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL",
    "AIRCONDITIONER:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL:HR",
    "ZONETERMINALUNITLIST",
    "CONTROLLER:WATERCOIL",
    "CONTROLLER:OUTDOORAIR",
    "CONTROLLER:MECHANICALVENTILATION",
    "AIRLOOPHVAC:CONTROLLERLIST",
    "AIRLOOPHVAC",
    "AIRLOOPHVAC:OUTDOORAIRSYSTEM:EQUIPMENTLIST",
    "AIRLOOPHVAC:OUTDOORAIRSYSTEM",
    "OUTDOORAIR:MIXER",
    "AIRLOOPHVAC:ZONESPLITTER",
    "AIRLOOPHVAC:SUPPLYPLENUM",
    "AIRLOOPHVAC:SUPPLYPATH",
    "AIRLOOPHVAC:ZONEMIXER",
    "AIRLOOPHVAC:RETURNPLENUM",
    "AIRLOOPHVAC:RETURNPATH",
    "AIRLOOPHVAC:EXHAUSTSYSTEM",
    "AIRLOOPHVAC:DEDICATEDOUTDOORAIRSYSTEM",
    "AIRLOOPHVAC:MIXER",
    "AIRLOOPHVAC:SPLITTER",
    "BRANCH",
    "BRANCHLIST",
    "CONNECTOR:SPLITTER",
    "CONNECTOR:MIXER",
    "CONNECTORLIST",
    "NODELIST",
    "OUTDOORAIR:NODE",
    "OUTDOORAIR:NODELIST",
    "PIPE:ADIABATIC",
    "PIPE:ADIABATIC:STEAM",
    "PIPE:INDOOR",
    "PIPE:OUTDOOR",
    "PIPE:UNDERGROUND",
    "PIPINGSYSTEM:UNDERGROUND:DOMAIN",
    "PIPINGSYSTEM:UNDERGROUND:PIPECIRCUIT",
    "PIPINGSYSTEM:UNDERGROUND:PIPESEGMENT",
    "DUCT",
    "PUMP:VARIABLESPEED",
    "PUMP:CONSTANTSPEED",
    "PUMP:VARIABLESPEED:CONDENSATE",
    "HEADEREDPUMPS:CONSTANTSPEED",
    "HEADEREDPUMPS:VARIABLESPEED",
    "TEMPERINGVALVE",
    "LOADPROFILE:PLANT",
    "SOLARCOLLECTORPERFORMANCE:FLATPLATE",
    "SOLARCOLLECTOR:FLATPLATE:WATER",
    "SOLARCOLLECTOR:FLATPLATE:PHOTOVOLTAICTHERMAL",
    "SOLARCOLLECTORPERFORMANCE:PHOTOVOLTAICTHERMAL:SIMPLE",
    "SOLARCOLLECTORPERFORMANCE:PHOTOVOLTAICTHERMAL:BIPVT",
    "SOLARCOLLECTOR:INTEGRALCOLLECTORSTORAGE",
    "SOLARCOLLECTORPERFORMANCE:INTEGRALCOLLECTORSTORAGE",
    "SOLARCOLLECTOR:UNGLAZEDTRANSPIRED",
    "SOLARCOLLECTOR:UNGLAZEDTRANSPIRED:MULTISYSTEM",
    "BOILER:HOTWATER",
    "BOILER:STEAM",
    "CHILLER:ELECTRIC:ASHRAE205",
    "CHILLER:ELECTRIC:EIR",
    "CHILLER:ELECTRIC:REFORMULATEDEIR",
    "CHILLER:ELECTRIC",
    "CHILLER:ABSORPTION:INDIRECT",
    "CHILLER:ABSORPTION",
    "CHILLER:CONSTANTCOP",
    "CHILLER:ENGINEDRIVEN",
    "CHILLER:COMBUSTIONTURBINE",
    "CHILLERHEATER:ABSORPTION:DIRECTFIRED",
    "CHILLERHEATER:ABSORPTION:DOUBLEEFFECT",
    "HEATPUMP:PLANTLOOP:EIR:COOLING",
    "HEATPUMP:PLANTLOOP:EIR:HEATING",
    "HEATPUMP:AIRTOWATER:FUELFIRED:HEATING",
    "HEATPUMP:AIRTOWATER:FUELFIRED:COOLING",
    "HEATPUMP:WATERTOWATER:EQUATIONFIT:HEATING",
    "HEATPUMP:WATERTOWATER:EQUATIONFIT:COOLING",
    "HEATPUMP:WATERTOWATER:PARAMETERESTIMATION:COOLING",
    "HEATPUMP:WATERTOWATER:PARAMETERESTIMATION:HEATING",
    "DISTRICTCOOLING",
    "DISTRICTHEATING",
    "PLANTCOMPONENT:TEMPERATURESOURCE",
    "CENTRALHEATPUMPSYSTEM",
    "CHILLERHEATERPERFORMANCE:ELECTRIC:EIR",
    "COOLINGTOWER:SINGLESPEED",
    "COOLINGTOWER:TWOSPEED",
    "COOLINGTOWER:VARIABLESPEED:MERKEL",
    "COOLINGTOWER:VARIABLESPEED",
    "COOLINGTOWERPERFORMANCE:COOLTOOLS",
    "COOLINGTOWERPERFORMANCE:YORKCALC",
    "EVAPORATIVEFLUIDCOOLER:SINGLESPEED",
    "EVAPORATIVEFLUIDCOOLER:TWOSPEED",
    "FLUIDCOOLER:SINGLESPEED",
    "FLUIDCOOLER:TWOSPEED",
    "GROUNDHEATEXCHANGER:SYSTEM",
    "GROUNDHEATEXCHANGER:VERTICAL:PROPERTIES",
    "GROUNDHEATEXCHANGER:VERTICAL:ARRAY",
    "GROUNDHEATEXCHANGER:VERTICAL:SINGLE",
    "GROUNDHEATEXCHANGER:RESPONSEFACTORS",
    "GROUNDHEATEXCHANGER:POND",
    "GROUNDHEATEXCHANGER:SURFACE",
    "GROUNDHEATEXCHANGER:HORIZONTALTRENCH",
    "GROUNDHEATEXCHANGER:SLINKY",
    "HEATEXCHANGER:FLUIDTOFLUID",
    "WATERHEATER:MIXED",
    "WATERHEATER:STRATIFIED",
    "WATERHEATER:SIZING",
    "WATERHEATER:HEATPUMP:PUMPEDCONDENSER",
    "WATERHEATER:HEATPUMP:WRAPPEDCONDENSER",
    "THERMALSTORAGE:ICE:SIMPLE",
    "THERMALSTORAGE:ICE:DETAILED",
    "THERMALSTORAGE:CHILLEDWATER:MIXED",
    "THERMALSTORAGE:CHILLEDWATER:STRATIFIED",
    "PLANTLOOP",
    "CONDENSERLOOP",
    "PLANTEQUIPMENTLIST",
    "CONDENSEREQUIPMENTLIST",
    "PLANTEQUIPMENTOPERATION:UNCONTROLLED",
    "PLANTEQUIPMENTOPERATION:COOLINGLOAD",
    "PLANTEQUIPMENTOPERATION:HEATINGLOAD",
    "PLANTEQUIPMENTOPERATION:OUTDOORDRYBULB",
    "PLANTEQUIPMENTOPERATION:OUTDOORWETBULB",
    "PLANTEQUIPMENTOPERATION:OUTDOORRELATIVEHUMIDITY",
    "PLANTEQUIPMENTOPERATION:OUTDOORDEWPOINT",
    "PLANTEQUIPMENTOPERATION:COMPONENTSETPOINT",
    "PLANTEQUIPMENTOPERATION:THERMALENERGYSTORAGE",
    "PLANTEQUIPMENTOPERATION:OUTDOORDRYBULBDIFFERENCE",
    "PLANTEQUIPMENTOPERATION:OUTDOORWETBULBDIFFERENCE",
    "PLANTEQUIPMENTOPERATION:OUTDOORDEWPOINTDIFFERENCE",
    "PLANTEQUIPMENTOPERATIONSCHEMES",
    "CONDENSEREQUIPMENTOPERATIONSCHEMES",
    "ENERGYMANAGEMENTSYSTEM:SENSOR",
    "ENERGYMANAGEMENTSYSTEM:ACTUATOR",
    "ENERGYMANAGEMENTSYSTEM:PROGRAMCALLINGMANAGER",
    "ENERGYMANAGEMENTSYSTEM:PROGRAM",
    "ENERGYMANAGEMENTSYSTEM:SUBROUTINE",
    "ENERGYMANAGEMENTSYSTEM:GLOBALVARIABLE",
    "ENERGYMANAGEMENTSYSTEM:OUTPUTVARIABLE",
    "ENERGYMANAGEMENTSYSTEM:METEREDOUTPUTVARIABLE",
    "ENERGYMANAGEMENTSYSTEM:TRENDVARIABLE",
    "ENERGYMANAGEMENTSYSTEM:INTERNALVARIABLE",
    "ENERGYMANAGEMENTSYSTEM:CURVEORTABLEINDEXVARIABLE",
    "ENERGYMANAGEMENTSYSTEM:CONSTRUCTIONINDEXVARIABLE",
    "EXTERNALINTERFACE",
    "EXTERNALINTERFACE:SCHEDULE",
    "EXTERNALINTERFACE:VARIABLE",
    "EXTERNALINTERFACE:ACTUATOR",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:FROM:VARIABLE",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:TO:SCHEDULE",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:TO:ACTUATOR",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:TO:VARIABLE",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:FROM:VARIABLE",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:TO:SCHEDULE",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:TO:ACTUATOR",
    "EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:TO:VARIABLE",
    "ZONEHVAC:FORCEDAIR:USERDEFINED",
    "AIRTERMINAL:SINGLEDUCT:USERDEFINED",
    "COIL:USERDEFINED",
    "PLANTCOMPONENT:USERDEFINED",
    "PLANTEQUIPMENTOPERATION:USERDEFINED",
    "AVAILABILITYMANAGER:SCHEDULED",
    "AVAILABILITYMANAGER:SCHEDULEDON",
    "AVAILABILITYMANAGER:SCHEDULEDOFF",
    "AVAILABILITYMANAGER:OPTIMUMSTART",
    "AVAILABILITYMANAGER:NIGHTCYCLE",
    "AVAILABILITYMANAGER:DIFFERENTIALTHERMOSTAT",
    "AVAILABILITYMANAGER:HIGHTEMPERATURETURNOFF",
    "AVAILABILITYMANAGER:HIGHTEMPERATURETURNON",
    "AVAILABILITYMANAGER:LOWTEMPERATURETURNOFF",
    "AVAILABILITYMANAGER:LOWTEMPERATURETURNON",
    "AVAILABILITYMANAGER:NIGHTVENTILATION",
    "AVAILABILITYMANAGER:HYBRIDVENTILATION",
    "AVAILABILITYMANAGERASSIGNMENTLIST",
    "SETPOINTMANAGER:SCHEDULED",
    "SETPOINTMANAGER:SCHEDULED:DUALSETPOINT",
    "SETPOINTMANAGER:OUTDOORAIRRESET",
    "SETPOINTMANAGER:SINGLEZONE:REHEAT",
    "SETPOINTMANAGER:SINGLEZONE:HEATING",
    "SETPOINTMANAGER:SINGLEZONE:COOLING",
    "SETPOINTMANAGER:SINGLEZONE:HUMIDITY:MINIMUM",
    "SETPOINTMANAGER:SINGLEZONE:HUMIDITY:MAXIMUM",
    "SETPOINTMANAGER:MIXEDAIR",
    "SETPOINTMANAGER:OUTDOORAIRPRETREAT",
    "SETPOINTMANAGER:WARMEST",
    "SETPOINTMANAGER:COLDEST",
    "SETPOINTMANAGER:RETURNAIRBYPASSFLOW",
    "SETPOINTMANAGER:WARMESTTEMPERATUREFLOW",
    "SETPOINTMANAGER:MULTIZONE:HEATING:AVERAGE",
    "SETPOINTMANAGER:MULTIZONE:COOLING:AVERAGE",
    "SETPOINTMANAGER:MULTIZONE:MINIMUMHUMIDITY:AVERAGE",
    "SETPOINTMANAGER:MULTIZONE:MAXIMUMHUMIDITY:AVERAGE",
    "SETPOINTMANAGER:MULTIZONE:HUMIDITY:MINIMUM",
    "SETPOINTMANAGER:MULTIZONE:HUMIDITY:MAXIMUM",
    "SETPOINTMANAGER:FOLLOWOUTDOORAIRTEMPERATURE",
    "SETPOINTMANAGER:FOLLOWSYSTEMNODETEMPERATURE",
    "SETPOINTMANAGER:FOLLOWGROUNDTEMPERATURE",
    "SETPOINTMANAGER:CONDENSERENTERINGRESET",
    "SETPOINTMANAGER:CONDENSERENTERINGRESET:IDEAL",
    "SETPOINTMANAGER:SINGLEZONE:ONESTAGECOOLING",
    "SETPOINTMANAGER:SINGLEZONE:ONESTAGEHEATING",
    "SETPOINTMANAGER:RETURNTEMPERATURE:CHILLEDWATER",
    "SETPOINTMANAGER:RETURNTEMPERATURE:HOTWATER",
    "SETPOINTMANAGER:SYSTEMNODERESET:TEMPERATURE",
    "SETPOINTMANAGER:SYSTEMNODERESET:HUMIDITY",
    "REFRIGERATION:CASE",
    "REFRIGERATION:COMPRESSORRACK",
    "REFRIGERATION:CASEANDWALKINLIST",
    "REFRIGERATION:CONDENSER:AIRCOOLED",
    "REFRIGERATION:CONDENSER:EVAPORATIVECOOLED",
    "REFRIGERATION:CONDENSER:WATERCOOLED",
    "REFRIGERATION:CONDENSER:CASCADE",
    "REFRIGERATION:GASCOOLER:AIRCOOLED",
    "REFRIGERATION:TRANSFERLOADLIST",
    "REFRIGERATION:SUBCOOLER",
    "REFRIGERATION:COMPRESSOR",
    "REFRIGERATION:COMPRESSORLIST",
    "REFRIGERATION:SYSTEM",
    "REFRIGERATION:TRANSCRITICALSYSTEM",
    "REFRIGERATION:SECONDARYSYSTEM",
    "REFRIGERATION:WALKIN",
    "REFRIGERATION:AIRCHILLER",
    "ZONEHVAC:REFRIGERATIONCHILLERSET",
    "DEMANDMANAGERASSIGNMENTLIST",
    "DEMANDMANAGER:EXTERIORLIGHTS",
    "DEMANDMANAGER:LIGHTS",
    "DEMANDMANAGER:ELECTRICEQUIPMENT",
    "DEMANDMANAGER:THERMOSTATS",
    "DEMANDMANAGER:VENTILATION",
    "GENERATOR:INTERNALCOMBUSTIONENGINE",
    "GENERATOR:COMBUSTIONTURBINE",
    "GENERATOR:MICROTURBINE",
    "GENERATOR:PHOTOVOLTAIC",
    "PHOTOVOLTAICPERFORMANCE:SIMPLE",
    "PHOTOVOLTAICPERFORMANCE:EQUIVALENTONE-DIODE",
    "PHOTOVOLTAICPERFORMANCE:SANDIA",
    "GENERATOR:PVWATTS",
    "ELECTRICLOADCENTER:INVERTER:PVWATTS",
    "GENERATOR:FUELCELL",
    "GENERATOR:FUELCELL:POWERMODULE",
    "GENERATOR:FUELCELL:AIRSUPPLY",
    "GENERATOR:FUELCELL:WATERSUPPLY",
    "GENERATOR:FUELCELL:AUXILIARYHEATER",
    "GENERATOR:FUELCELL:EXHAUSTGASTOWATERHEATEXCHANGER",
    "GENERATOR:FUELCELL:ELECTRICALSTORAGE",
    "GENERATOR:FUELCELL:INVERTER",
    "GENERATOR:FUELCELL:STACKCOOLER",
    "GENERATOR:MICROCHP",
    "GENERATOR:MICROCHP:NONNORMALIZEDPARAMETERS",
    "GENERATOR:FUELSUPPLY",
    "GENERATOR:WINDTURBINE",
    "ELECTRICLOADCENTER:GENERATORS",
    "ELECTRICLOADCENTER:INVERTER:SIMPLE",
    "ELECTRICLOADCENTER:INVERTER:FUNCTIONOFPOWER",
    "ELECTRICLOADCENTER:INVERTER:LOOKUPTABLE",
    "ELECTRICLOADCENTER:STORAGE:SIMPLE",
    "ELECTRICLOADCENTER:STORAGE:BATTERY",
    "ELECTRICLOADCENTER:STORAGE:LIIONNMCBATTERY",
    "ELECTRICLOADCENTER:TRANSFORMER",
    "ELECTRICLOADCENTER:DISTRIBUTION",
    "ELECTRICLOADCENTER:STORAGE:CONVERTER",
    "WATERUSE:EQUIPMENT",
    "WATERUSE:CONNECTIONS",
    "WATERUSE:STORAGE",
    "WATERUSE:WELL",
    "WATERUSE:RAINCOLLECTOR",
    "FAULTMODEL:TEMPERATURESENSOROFFSET:OUTDOORAIR",
    "FAULTMODEL:HUMIDITYSENSOROFFSET:OUTDOORAIR",
    "FAULTMODEL:ENTHALPYSENSOROFFSET:OUTDOORAIR",
    "FAULTMODEL:TEMPERATURESENSOROFFSET:RETURNAIR",
    "FAULTMODEL:ENTHALPYSENSOROFFSET:RETURNAIR",
    "FAULTMODEL:TEMPERATURESENSOROFFSET:CHILLERSUPPLYWATER",
    "FAULTMODEL:TEMPERATURESENSOROFFSET:COILSUPPLYAIR",
    "FAULTMODEL:TEMPERATURESENSOROFFSET:CONDENSERSUPPLYWATER",
    "FAULTMODEL:THERMOSTATOFFSET",
    "FAULTMODEL:HUMIDISTATOFFSET",
    "FAULTMODEL:FOULING:AIRFILTER",
    "FAULTMODEL:FOULING:BOILER",
    "FAULTMODEL:FOULING:EVAPORATIVECOOLER",
    "FAULTMODEL:FOULING:CHILLER",
    "FAULTMODEL:FOULING:COOLINGTOWER",
    "FAULTMODEL:FOULING:COIL",
    "MATRIX:TWODIMENSION",
    "HYBRIDMODEL:ZONE",
    "CURVE:LINEAR",
    "CURVE:QUADLINEAR",
    "CURVE:QUINTLINEAR",
    "CURVE:QUADRATIC",
    "CURVE:CUBIC",
    "CURVE:QUARTIC",
    "CURVE:EXPONENT",
    "CURVE:BICUBIC",
    "CURVE:BIQUADRATIC",
    "CURVE:QUADRATICLINEAR",
    "CURVE:CUBICLINEAR",
    "CURVE:TRIQUADRATIC",
    "CURVE:FUNCTIONAL:PRESSUREDROP",
    "CURVE:FANPRESSURERISE",
    "CURVE:EXPONENTIALSKEWNORMAL",
    "CURVE:SIGMOID",
    "CURVE:RECTANGULARHYPERBOLA1",
    "CURVE:RECTANGULARHYPERBOLA2",
    "CURVE:EXPONENTIALDECAY",
    "CURVE:DOUBLEEXPONENTIALDECAY",
    "CURVE:CHILLERPARTLOADWITHLIFT",
    "TABLE:INDEPENDENTVARIABLE",
    "TABLE:INDEPENDENTVARIABLELIST",
    "TABLE:LOOKUP",
    "FLUIDPROPERTIES:NAME",
    "FLUIDPROPERTIES:GLYCOLCONCENTRATION",
    "FLUIDPROPERTIES:TEMPERATURES",
    "FLUIDPROPERTIES:SATURATED",
    "FLUIDPROPERTIES:SUPERHEATED",
    "FLUIDPROPERTIES:CONCENTRATION",
    "CURRENCYTYPE",
    "COMPONENTCOST:ADJUSTMENTS",
    "COMPONENTCOST:REFERENCE",
    "COMPONENTCOST:LINEITEM",
    "UTILITYCOST:TARIFF",
    "UTILITYCOST:QUALIFY",
    "UTILITYCOST:CHARGE:SIMPLE",
    "UTILITYCOST:CHARGE:BLOCK",
    "UTILITYCOST:RATCHET",
    "UTILITYCOST:VARIABLE",
    "UTILITYCOST:COMPUTATION",
    "LIFECYCLECOST:PARAMETERS",
    "LIFECYCLECOST:RECURRINGCOSTS",
    "LIFECYCLECOST:NONRECURRINGCOST",
    "LIFECYCLECOST:USEPRICEESCALATION",
    "LIFECYCLECOST:USEADJUSTMENT",
    "PARAMETRIC:SETVALUEFORRUN",
    "PARAMETRIC:LOGIC",
    "PARAMETRIC:RUNCONTROL",
    "PARAMETRIC:FILENAMESUFFIX",
    "OUTPUT:VARIABLEDICTIONARY",
    "OUTPUT:SURFACES:LIST",
    "OUTPUT:SURFACES:DRAWING",
    "OUTPUT:SCHEDULES",
    "OUTPUT:CONSTRUCTIONS",
    "OUTPUT:ENERGYMANAGEMENTSYSTEM",
    "OUTPUTCONTROL:SURFACECOLORSCHEME",
    "OUTPUT:TABLE:SUMMARYREPORTS",
    "OUTPUT:TABLE:TIMEBINS",
    "OUTPUT:TABLE:MONTHLY",
    "OUTPUT:TABLE:ANNUAL",
    "OUTPUTCONTROL:TABLE:STYLE",
    "OUTPUTCONTROL:REPORTINGTOLERANCES",
    "OUTPUT:VARIABLE",
    "OUTPUT:METER",
    "OUTPUT:METER:METERFILEONLY",
    "OUTPUT:METER:CUMULATIVE",
    "OUTPUT:METER:CUMULATIVE:METERFILEONLY",
    "METER:CUSTOM",
    "METER:CUSTOMDECREMENT",
    "OUTPUTCONTROL:FILES",
    "OUTPUTCONTROL:TIMESTAMP",
    "OUTPUT:JSON",
    "OUTPUT:SQLITE",
    "OUTPUT:ENVIRONMENTALIMPACTFACTORS",
    "ENVIRONMENTALIMPACTFACTORS",
    "FUELFACTORS",
    "OUTPUT:DIAGNOSTICS",
    "OUTPUT:DEBUGGINGDATA",
    "OUTPUT:PREPROCESSORMESSAGE",
    "PYTHONPLUGIN:SEARCHPATHS",
    "PYTHONPLUGIN:INSTANCE",
    "PYTHONPLUGIN:VARIABLES",
    "PYTHONPLUGIN:TRENDVARIABLE",
    "PYTHONPLUGIN:OUTPUTVARIABLE",
    "OS:VERSION",
    "OS:COMPONENTDATA",
    "OS:CONVERGENCELIMITS",
    "OS:PROGRAMCONTROL",
    "OS:HEATBALANCEALGORITHM",
    "OS:RUNPERIOD",
    "OS:RUNPERIODCONTROL:DAYLIGHTSAVINGTIME",
    "OS:RUNPERIODCONTROL:SPECIALDAYS",
    "OS:SHADOWCALCULATION",
    "OS:SIMULATIONCONTROL",
    "OS:PERFORMANCEPRECISIONTRADEOFFS",
    "OS:SIZING:PARAMETERS",
    "OS:SURFACECONVECTIONALGORITHM:INSIDE",
    "OS:SURFACECONVECTIONALGORITHM:OUTSIDE",
    "OS:TIMESTEP",
    "OS:YEARDESCRIPTION",
    "OS:ZONEAIRCONTAMINANTBALANCE",
    "OS:ZONEAIRHEATBALANCEALGORITHM",
    "OS:ZONEAIRMASSFLOWCONSERVATION",
    "OS:ZONECAPACITANCEMULTIPLIER:RESEARCHSPECIAL",
    "OS:RADIANCEPARAMETERS",
    "OS:CLIMATEZONES",
    "OS:SITE",
    "OS:SITE:GROUNDTEMPERATURE:BUILDINGSURFACE",
    "OS:SITE:GROUNDTEMPERATURE:FCFACTORMETHOD",
    "OS:SITE:GROUNDTEMPERATURE:SHALLOW",
    "OS:SITE:GROUNDTEMPERATURE:DEEP",
    "OS:SITE:GROUNDTEMPERATURE:UNDISTURBED:KUSUDAACHENBACH",
    "OS:SITE:GROUNDTEMPERATURE:UNDISTURBED:XING",
    "OS:SITE:GROUNDREFLECTANCE",
    "OS:SITE:WATERMAINSTEMPERATURE",
    "OS:SIZINGPERIOD:DESIGNDAY",
    "OS:SIZINGPERIOD:WEATHERFILECONDITIONTYPE",
    "OS:SIZINGPERIOD:WEATHERFILEDAYS",
    "OS:WEATHERFILE",
    "OS:WEATHERPROPERTY:SKYTEMPERATURE",
    "OS:ADDITIONALPROPERTIES",
    "OS:BUILDINGSTORY",
    "OS:BUILDINGUNIT",
    "OS:DEFAULTCONSTRUCTIONSET",
    "OS:DEFAULTSCHEDULESET",
    "OS:DEFAULTSUBSURFACECONSTRUCTIONS",
    "OS:DEFAULTSURFACECONSTRUCTIONS",
    "OS:RENDERING:COLOR",
    "OS:SPACETYPE",
    "OS:MODELOBJECTLIST",
    "OS:MATERIAL",
    "OS:MATERIAL:AIRGAP",
    "OS:MATERIAL:INFRAREDTRANSPARENT",
    "OS:MATERIAL:NOMASS",
    "OS:MATERIAL:ROOFVEGETATION",
    "OS:WINDOWMATERIAL:BLIND",
    "OS:WINDOWMATERIAL:DAYLIGHTREDIRECTIONDEVICE",
    "OS:WINDOWMATERIAL:GAS",
    "OS:WINDOWMATERIAL:GASMIXTURE",
    "OS:WINDOWMATERIAL:GLAZING",
    "OS:WINDOWMATERIAL:GLAZINGGROUP:THERMOCHROMIC",
    "OS:WINDOWMATERIAL:GLAZING:REFRACTIONEXTINCTIONMETHOD",
    "OS:WINDOWMATERIAL:SCREEN",
    "OS:WINDOWMATERIAL:SHADE",
    "OS:WINDOWMATERIAL:SIMPLEGLAZINGSYSTEM",
    "OS:STANDARDSINFORMATION:MATERIAL",
    "OS:MATERIALPROPERTY:GLAZINGSPECTRALDATA",
    "OS:MATERIALPROPERTY:MOISTUREPENETRATIONDEPTH:SETTINGS",
    "OS:MATERIALPROPERTY:PHASECHANGE",
    "OS:MATERIALPROPERTY:PHASECHANGEHYSTERESIS",
    "OS:CONSTRUCTION",
    "OS:CONSTRUCTION:CFACTORUNDERGROUNDWALL",
    "OS:CONSTRUCTION:FFACTORGROUNDFLOOR",
    "OS:CONSTRUCTION:INTERNALSOURCE",
    "OS:CONSTRUCTION:AIRBOUNDARY",
    "OS:CONSTRUCTION:WINDOWDATAFILE",
    "OS:STANDARDSINFORMATION:CONSTRUCTION",
    "OS:INTERNALMASS:DEFINITION",
    "OS:PEOPLE:DEFINITION",
    "OS:LIGHTS:DEFINITION",
    "OS:LUMINAIRE:DEFINITION",
    "OS:ELECTRICEQUIPMENT:DEFINITION",
    "OS:ELECTRICEQUIPMENT:ITE:AIRCOOLED:DEFINITION",
    "OS:GASEQUIPMENT:DEFINITION",
    "OS:HOTWATEREQUIPMENT:DEFINITION",
    "OS:STEAMEQUIPMENT:DEFINITION",
    "OS:OTHEREQUIPMENT:DEFINITION",
    "OS:EXTERIOR:LIGHTS:DEFINITION",
    "OS:EXTERIOR:FUELEQUIPMENT:DEFINITION",
    "OS:EXTERIOR:WATEREQUIPMENT:DEFINITION",
    "OS:SCHEDULE:COMPACT",
    "OS:SCHEDULE:DAY",
    "OS:SCHEDULE:WEEK",
    "OS:SCHEDULE:YEAR",
    "OS:SCHEDULE:RULE",
    "OS:SCHEDULE:RULESET",
    "OS:SCHEDULE:CONSTANT",
    "OS:SCHEDULE:FIXEDINTERVAL",
    "OS:SCHEDULE:VARIABLEINTERVAL",
    "OS:SCHEDULETYPELIMITS",
    "OS:EXTERNAL:FILE",
    "OS:SCHEDULE:FILE",
    "OS:BUILDING",
    "OS:DAYLIGHTINGDEVICE:TUBULAR",
    "OS:DAYLIGHTINGDEVICE:SHELF",
    "OS:DAYLIGHTINGDEVICE:LIGHTWELL",
    "OS:FACILITY",
    "OS:INTERIORPARTITIONSURFACE",
    "OS:INTERIORPARTITIONSURFACEGROUP",
    "OS:SHADINGSURFACE",
    "OS:SHADINGSURFACEGROUP",
    "OS:SPACE",
    "OS:SURFACE",
    "OS:SUBSURFACE",
    "OS:SHADINGCONTROL",
    "OS:WINDOWPROPERTY:FRAMEANDDIVIDER",
    "OS:SURFACECONTROL:MOVABLEINSULATION",
    "OS:SURFACEPROPERTY:CONVECTIONCOEFFICIENTS",
    "OS:SURFACEPROPERTY:CONVECTIONCOEFFICIENTS:MULTIPLESURFACE",
    "OS:SURFACEPROPERTY:EXPOSEDFOUNDATIONPERIMETER",
    "OS:SURFACEPROPERTY:OTHERSIDECOEFFICIENTS",
    "OS:SURFACEPROPERTY:OTHERSIDECONDITIONSMODEL",
    "OS:FOUNDATION:KIVA",
    "OS:FOUNDATION:KIVA:SETTINGS",
    "OS:ZONEPROPERTY:USERVIEWFACTORS:BYSURFACENAME",
    "OS:SURFACEPROPERTY:SURROUNDINGSURFACES",
    "OS:SURFACEPROPERTY:LOCALENVIRONMENT",
    "OS:SURFACEPROPERTY:GROUNDSURFACES",
    "OS:SURFACEPROPERTY:INCIDENTSOLARMULTIPLIER",
    "OS:INTERNALMASS",
    "OS:PEOPLE",
    "OS:LIGHTS",
    "OS:LUMINAIRE",
    "OS:ELECTRICEQUIPMENT",
    "OS:ELECTRICEQUIPMENT:ITE:AIRCOOLED",
    "OS:GASEQUIPMENT",
    "OS:HOTWATEREQUIPMENT",
    "OS:STEAMEQUIPMENT",
    "OS:OTHEREQUIPMENT",
    "OS:SPACEINFILTRATION:DESIGNFLOWRATE",
    "OS:SPACEINFILTRATION:EFFECTIVELEAKAGEAREA",
    "OS:SPACEINFILTRATION:FLOWCOEFFICIENT",
    "OS:EXTERIOR:LIGHTS",
    "OS:EXTERIOR:FUELEQUIPMENT",
    "OS:EXTERIOR:WATEREQUIPMENT",
    "OS:DAYLIGHTING:CONTROL",
    "OS:GLARE:SENSOR",
    "OS:ILLUMINANCEMAP",
    "OS:LIGHTINGDESIGNDAY",
    "OS:LIGHTINGSIMULATIONCONTROL",
    "OS:LIGHTINGSIMULATIONZONE",
    "OS:REFRIGERATION:CASE",
    "OS:REFRIGERATION:CONDENSER:AIRCOOLED",
    "OS:REFRIGERATION:CONDENSER:EVAPORATIVECOOLED",
    "OS:REFRIGERATION:CONDENSER:WATERCOOLED",
    "OS:REFRIGERATION:CONDENSER:CASCADE",
    "OS:REFRIGERATION:COMPRESSOR",
    "OS:REFRIGERATION:COMPRESSORRACK",
    "OS:REFRIGERATION:DEFROSTCYCLEPARAMETERS",
    "OS:REFRIGERATION:GASCOOLER:AIRCOOLED",
    "OS:REFRIGERATION:SUBCOOLER:MECHANICAL",
    "OS:REFRIGERATION:SUBCOOLER:LIQUIDSUCTION",
    "OS:REFRIGERATION:SYSTEM",
    "OS:REFRIGERATION:TRANSCRITICALSYSTEM",
    "OS:REFRIGERATION:SECONDARYSYSTEM",
    "OS:REFRIGERATION:WALKIN",
    "OS:REFRIGERATION:WALKIN:ZONEBOUNDARY",
    "OS:REFRIGERATION:AIRCHILLER",
    "OS:AIRCONDITIONER:VARIABLEREFRIGERANTFLOW",
    "OS:AIRCONDITIONER:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL",
    "OS:AIRCONDITIONER:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL:LOADINGINDEX",
    "OS:AIRCONDITIONER:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL:HR",
    "OS:AIRLOOPHVAC",
    "OS:AIRLOOPHVAC:OUTDOORAIRSYSTEM",
    "OS:AIRLOOPHVAC:DEDICATEDOUTDOORAIRSYSTEM",
    "OS:AIRLOOPHVAC:RETURNPLENUM",
    "OS:AIRLOOPHVAC:SUPPLYPLENUM",
    "OS:AIRLOOPHVAC:UNITARYHEATPUMP:AIRTOAIR:MULTISPEED",
    "OS:AIRLOOPHVAC:UNITARYCOOLONLY",
    "OS:AIRLOOPHVAC:UNITARYHEATPUMP:AIRTOAIR",
    "OS:AIRLOOPHVAC:UNITARYHEATCOOL:VAVCHANGEOVERBYPASS",
    "OS:AIRLOOPHVAC:UNITARYSYSTEM",
    "OS:UNITARYSYSTEMPERFORMANCE:MULTISPEED",
    "OS:AIRLOOPHVAC:ZONEMIXER",
    "OS:AIRLOOPHVAC:ZONESPLITTER",
    "OS:AIRTERMINAL:SINGLEDUCT:VAV:HEATANDCOOL:NOREHEAT",
    "OS:AIRTERMINAL:SINGLEDUCT:VAV:HEATANDCOOL:REHEAT",
    "OS:AIRTERMINAL:SINGLEDUCT:INLETSIDEMIXER",
    "OS:AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:REHEAT",
    "OS:AIRTERMINAL:SINGLEDUCT:SERIESPIU:REHEAT",
    "OS:AIRTERMINAL:SINGLEDUCT:PARALLELPIU:REHEAT",
    "OS:AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:FOURPIPEINDUCTION",
    "OS:AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:FOURPIPEBEAM",
    "OS:AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:COOLEDBEAM",
    "OS:AIRTERMINAL:SINGLEDUCT:CONSTANTVOLUME:NOREHEAT",
    "OS:AIRTERMINAL:SINGLEDUCT:VAV:NOREHEAT",
    "OS:AIRTERMINAL:SINGLEDUCT:VAV:REHEAT",
    "OS:AIRTERMINAL:DUALDUCT:VAV:OUTDOORAIR",
    "OS:AIRTERMINAL:DUALDUCT:CONSTANTVOLUME",
    "OS:AIRTERMINAL:DUALDUCT:VAV",
    "OS:AVAILABILITYMANAGERASSIGNMENTLIST",
    "OS:AVAILABILITYMANAGER:SCHEDULED",
    "OS:AVAILABILITYMANAGER:SCHEDULEDON",
    "OS:AVAILABILITYMANAGER:SCHEDULEDOFF",
    "OS:AVAILABILITYMANAGER:NIGHTCYCLE",
    "OS:AVAILABILITYMANAGER:OPTIMUMSTART",
    "OS:AVAILABILITYMANAGER:DIFFERENTIALTHERMOSTAT",
    "OS:AVAILABILITYMANAGER:HIGHTEMPERATURETURNOFF",
    "OS:AVAILABILITYMANAGER:HIGHTEMPERATURETURNON",
    "OS:AVAILABILITYMANAGER:LOWTEMPERATURETURNOFF",
    "OS:AVAILABILITYMANAGER:LOWTEMPERATURETURNON",
    "OS:AVAILABILITYMANAGER:NIGHTVENTILATION",
    "OS:AVAILABILITYMANAGER:HYBRIDVENTILATION",
    "OS:BOILER:HOTWATER",
    "OS:BOILER:STEAM",
    "OS:HEATPUMP:WATERTOWATER:EQUATIONFIT:HEATING",
    "OS:HEATPUMP:WATERTOWATER:EQUATIONFIT:COOLING",
    "OS:HEATPUMP:PLANTLOOP:EIR:HEATING",
    "OS:HEATPUMP:PLANTLOOP:EIR:COOLING",
    "OS:HEATPUMP:AIRTOWATER:FUELFIRED:HEATING",
    "OS:HEATPUMP:AIRTOWATER:FUELFIRED:COOLING",
    "OS:CHILLER:ELECTRIC:EIR",
    "OS:CHILLER:ELECTRIC:REFORMULATEDEIR",
    "OS:CHILLER:ELECTRIC:ASHRAE205",
    "OS:CENTRALHEATPUMPSYSTEM",
    "OS:CENTRALHEATPUMPSYSTEM:MODULE",
    "OS:CHILLERHEATERPERFORMANCE:ELECTRIC:EIR",
    "OS:CHILLER:ABSORPTION:INDIRECT",
    "OS:CHILLER:ABSORPTION",
    "OS:COOLINGTOWER:SINGLESPEED",
    "OS:COOLINGTOWER:TWOSPEED",
    "OS:COOLINGTOWER:VARIABLESPEED",
    "OS:COOLINGTOWERPERFORMANCE:COOLTOOLS",
    "OS:COOLINGTOWERPERFORMANCE:YORKCALC",
    "OS:DISTRICTCOOLING",
    "OS:DISTRICTHEATING",
    "OS:GROUNDHEATEXCHANGER:VERTICAL",
    "OS:GROUNDHEATEXCHANGER:HORIZONTALTRENCH",
    "OS:COIL:COOLING:COOLEDBEAM",
    "OS:COIL:COOLING:FOURPIPEBEAM",
    "OS:COIL:HEATING:FOURPIPEBEAM",
    "OS:COIL:COOLING:DX",
    "OS:COIL:COOLING:DX:CURVEFIT:PERFORMANCE",
    "OS:COIL:COOLING:DX:CURVEFIT:OPERATINGMODE",
    "OS:COIL:COOLING:DX:CURVEFIT:SPEED",
    "OS:COIL:COOLING:DX:SINGLESPEED",
    "OS:COIL:COOLING:DX:TWOSTAGEWITHHUMIDITYCONTROLMODE",
    "OS:COILPERFORMANCE:DX:COOLING",
    "OS:COIL:COOLING:DX:MULTISPEED",
    "OS:COIL:COOLING:DX:MULTISPEED:STAGEDATA",
    "OS:COIL:COOLING:DX:TWOSPEED",
    "OS:COILSYSTEM:COOLING:WATER:HEATEXCHANGERASSISTED",
    "OS:COILSYSTEM:COOLING:DX:HEATEXCHANGERASSISTED",
    "OS:COILSYSTEM:INTEGRATEDHEATPUMP:AIRSOURCE",
    "OS:COIL:COOLING:DX:VARIABLEREFRIGERANTFLOW",
    "OS:COIL:COOLING:DX:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL",
    "OS:COIL:COOLING:LOWTEMPERATURERADIANT:CONSTANTFLOW",
    "OS:COIL:COOLING:LOWTEMPERATURERADIANT:VARIABLEFLOW",
    "OS:COIL:COOLING:WATER",
    "OS:COIL:COOLING:WATERTOAIRHEATPUMP:EQUATIONFIT",
    "OS:COIL:HEATING:DX:SINGLESPEED",
    "OS:COIL:HEATING:DX:VARIABLEREFRIGERANTFLOW",
    "OS:COIL:HEATING:DX:VARIABLEREFRIGERANTFLOW:FLUIDTEMPERATURECONTROL",
    "OS:COIL:HEATING:ELECTRIC",
    "OS:COIL:HEATING:ELECTRIC:MULTISTAGE",
    "OS:COIL:HEATING:ELECTRIC:MULTISTAGE:STAGEDATA",
    "OS:COIL:HEATING:GAS",
    "OS:COIL:HEATING:GAS:MULTISTAGE",
    "OS:COIL:HEATING:GAS:MULTISTAGE:STAGEDATA",
    "OS:COIL:HEATING:DESUPERHEATER",
    "OS:COIL:WATERHEATING:DESUPERHEATER",
    "OS:COIL:HEATING:LOWTEMPERATURERADIANT:CONSTANTFLOW",
    "OS:COIL:HEATING:LOWTEMPERATURERADIANT:VARIABLEFLOW",
    "OS:COIL:HEATING:WATER",
    "OS:COIL:HEATING:WATER:BASEBOARD",
    "OS:COIL:HEATING:WATER:BASEBOARD:RADIANT",
    "OS:COIL:HEATING:WATERTOAIRHEATPUMP:EQUATIONFIT",
    "OS:COIL:HEATING:DX:MULTISPEED",
    "OS:COIL:HEATING:DX:MULTISPEED:STAGEDATA",
    "OS:COIL:HEATING:DX:VARIABLESPEED",
    "OS:COIL:HEATING:DX:VARIABLESPEED:SPEEDDATA",
    "OS:COIL:COOLING:DX:VARIABLESPEED",
    "OS:COIL:COOLING:DX:VARIABLESPEED:SPEEDDATA",
    "OS:COIL:HEATING:WATERTOAIRHEATPUMP:VARIABLESPEEDEQUATIONFIT",
    "OS:COIL:HEATING:WATERTOAIRHEATPUMP:VARIABLESPEEDEQUATIONFIT:SPEEDDATA",
    "OS:COIL:COOLING:WATERTOAIRHEATPUMP:VARIABLESPEEDEQUATIONFIT",
    "OS:COIL:COOLING:WATERTOAIRHEATPUMP:VARIABLESPEEDEQUATIONFIT:SPEEDDATA",
    "OS:COIL:COOLING:WATER:PANEL:RADIANT",
    "OS:CONNECTION",
    "OS:CONNECTOR:MIXER",
    "OS:CONNECTOR:SPLITTER",
    "OS:CONTROLLER:MECHANICALVENTILATION",
    "OS:ZONECONTROL:CONTAMINANTCONTROLLER",
    "OS:CONTROLLER:OUTDOORAIR",
    "OS:CONTROLLER:WATERCOIL",
    "OS:CURVE:LINEAR",
    "OS:CURVE:QUADLINEAR",
    "OS:CURVE:QUINTLINEAR",
    "OS:CURVE:QUADRATIC",
    "OS:CURVE:CUBIC",
    "OS:CURVE:QUARTIC",
    "OS:CURVE:EXPONENT",
    "OS:CURVE:BICUBIC",
    "OS:CURVE:BIQUADRATIC",
    "OS:CURVE:QUADRATICLINEAR",
    "OS:CURVE:TRIQUADRATIC",
    "OS:CURVE:FUNCTIONAL:PRESSUREDROP",
    "OS:CURVE:FANPRESSURERISE",
    "OS:CURVE:EXPONENTIALSKEWNORMAL",
    "OS:CURVE:SIGMOID",
    "OS:CURVE:RECTANGULARHYPERBOLA1",
    "OS:CURVE:RECTANGULARHYPERBOLA2",
    "OS:CURVE:EXPONENTIALDECAY",
    "OS:CURVE:DOUBLEEXPONENTIALDECAY",
    "OS:TABLE:MULTIVARIABLELOOKUP",
    "OS:TABLE:INDEPENDENTVARIABLE",
    "OS:TABLE:LOOKUP",
    "OS:DESIGNSPECIFICATION:OUTDOORAIR",
    "OS:DESIGNSPECIFICATION:ZONEAIRDISTRIBUTION",
    "OS:HUMIDIFIER:STEAM:ELECTRIC",
    "OS:HUMIDIFIER:STEAM:GAS",
    "OS:EVAPORATIVECOOLER:INDIRECT:RESEARCHSPECIAL",
    "OS:EVAPORATIVECOOLER:DIRECT:RESEARCHSPECIAL",
    "OS:EVAPORATIVEFLUIDCOOLER:SINGLESPEED",
    "OS:EVAPORATIVEFLUIDCOOLER:TWOSPEED",
    "OS:FAN:SYSTEMMODEL",
    "OS:FAN:CONSTANTVOLUME",
    "OS:FAN:VARIABLEVOLUME",
    "OS:FAN:ONOFF",
    "OS:FAN:COMPONENTMODEL",
    "OS:FAN:ZONEEXHAUST",
    "OS:HEATEXCHANGER:AIRTOAIR:SENSIBLEANDLATENT",
    "OS:HEATEXCHANGER:DESICCANT:BALANCEDFLOW",
    "OS:HEATEXCHANGER:DESICCANT:BALANCEDFLOW:PERFORMANCEDATATYPE1",
    "OS:HEATEXCHANGER:FLUIDTOFLUID",
    "OS:HVACCOMPONENTLIST",
    "OS:NODE",
    "OS:PORTLIST",
    "OS:PIPE:ADIABATIC",
    "OS:PLANTLOOP",
    "OS:PLANTEQUIPMENTOPERATION:COOLINGLOAD",
    "OS:PLANTEQUIPMENTOPERATION:HEATINGLOAD",
    "OS:PLANTEQUIPMENTOPERATION:OUTDOORDRYBULB",
    "OS:PLANTEQUIPMENTOPERATION:OUTDOORWETBULB",
    "OS:PLANTEQUIPMENTOPERATION:OUTDOORRELATIVEHUMIDITY",
    "OS:PLANTEQUIPMENTOPERATION:OUTDOORDEWPOINT",
    "OS:PLANTEQUIPMENTOPERATION:OUTDOORDRYBULBDIFFERENCE",
    "OS:PLANTEQUIPMENTOPERATION:OUTDOORDEWPOINTDIFFERENCE",
    "OS:PLANTEQUIPMENTOPERATION:OUTDOORWETBULBDIFFERENCE",
    "OS:PLANTCOMPONENT:TEMPERATURESOURCE",
    "OS:LOADPROFILE:PLANT",
    "OS:PUMP:CONSTANTSPEED",
    "OS:PUMP:VARIABLESPEED",
    "OS:HEADEREDPUMPS:CONSTANTSPEED",
    "OS:HEADEREDPUMPS:VARIABLESPEED",
    "OS:TEMPERINGVALVE",
    "OS:SETPOINTMANAGER:FOLLOWOUTDOORAIRTEMPERATURE",
    "OS:SETPOINTMANAGER:MIXEDAIR",
    "OS:SETPOINTMANAGER:OUTDOORAIRRESET",
    "OS:SETPOINTMANAGER:OUTDOORAIRPRETREAT",
    "OS:SETPOINTMANAGER:SINGLEZONE:REHEAT",
    "OS:SETPOINTMANAGER:SINGLEZONE:HUMIDITY:MINIMUM",
    "OS:SETPOINTMANAGER:MULTIZONE:MINIMUMHUMIDITY:AVERAGE",
    "OS:SETPOINTMANAGER:MULTIZONE:HUMIDITY:MINIMUM",
    "OS:SETPOINTMANAGER:SCHEDULED",
    "OS:SETPOINTMANAGER:SCHEDULED:DUALSETPOINT",
    "OS:SETPOINTMANAGER:WARMEST",
    "OS:SETPOINTMANAGER:WARMESTTEMPERATUREFLOW",
    "OS:SETPOINTMANAGER:COLDEST",
    "OS:SETPOINTMANAGER:FOLLOWGROUNDTEMPERATURE",
    "OS:SETPOINTMANAGER:FOLLOWSYSTEMNODETEMPERATURE",
    "OS:SETPOINTMANAGER:MULTIZONE:COOLING:AVERAGE",
    "OS:SETPOINTMANAGER:MULTIZONE:HEATING:AVERAGE",
    "OS:SETPOINTMANAGER:MULTIZONE:HUMIDITY:MAXIMUM",
    "OS:SETPOINTMANAGER:MULTIZONE:MAXIMUMHUMIDITY:AVERAGE",
    "OS:SETPOINTMANAGER:SINGLEZONE:ONESTAGECOOLING",
    "OS:SETPOINTMANAGER:SINGLEZONE:ONESTAGEHEATING",
    "OS:SETPOINTMANAGER:SINGLEZONE:HUMIDITY:MAXIMUM",
    "OS:SETPOINTMANAGER:SINGLEZONE:COOLING",
    "OS:SETPOINTMANAGER:SINGLEZONE:HEATING",
    "OS:SETPOINTMANAGER:SYSTEMNODERESET:TEMPERATURE",
    "OS:SETPOINTMANAGER:SYSTEMNODERESET:HUMIDITY",
    "OS:SIZING:PLANT",
    "OS:SIZING:SYSTEM",
    "OS:SIZING:ZONE",
    "OS:SPLITTER",
    "OS:THERMALZONE",
    "OS:ZONECONTROL:HUMIDISTAT",
    "OS:THERMOSTATSETPOINT:DUALSETPOINT",
    "OS:ZONECONTROL:THERMOSTAT:STAGEDDUALSETPOINT",
    "OS:ZONEHVAC:EQUIPMENTLIST",
    "OS:ZONEVENTILATION:DESIGNFLOWRATE",
    "OS:ZONEVENTILATION:WINDANDSTACKOPENAREA",
    "OS:ZONEHVAC:BASEBOARD:CONVECTIVE:ELECTRIC",
    "OS:ZONEHVAC:BASEBOARD:CONVECTIVE:WATER",
    "OS:ZONEHVAC:HIGHTEMPERATURERADIANT",
    "OS:ZONEHVAC:LOWTEMPERATURERADIANT:ELECTRIC",
    "OS:ZONEHVAC:IDEALLOADSAIRSYSTEM",
    "OS:ZONEHVAC:FOURPIPEFANCOIL",
    "OS:ZONEHVAC:LOWTEMPERATURERADIANT:CONSTANTFLOW",
    "OS:ZONEHVAC:LOWTEMPERATURERADIANT:VARIABLEFLOW",
    "OS:ZONEHVAC:PACKAGEDTERMINALHEATPUMP",
    "OS:ZONEHVAC:PACKAGEDTERMINALAIRCONDITIONER",
    "OS:ZONEHVAC:TERMINALUNIT:VARIABLEREFRIGERANTFLOW",
    "OS:ZONEHVAC:WATERTOAIRHEATPUMP",
    "OS:ZONEHVAC:UNITHEATER",
    "OS:ZONEHVAC:DEHUMIDIFIER:DX",
    "OS:ZONEHVAC:ENERGYRECOVERYVENTILATOR",
    "OS:ZONEHVAC:ENERGYRECOVERYVENTILATOR:CONTROLLER",
    "OS:ZONEHVAC:UNITVENTILATOR",
    "OS:ZONEHVAC:BASEBOARD:RADIANTCONVECTIVE:WATER",
    "OS:ZONEHVAC:BASEBOARD:RADIANTCONVECTIVE:ELECTRIC",
    "OS:ZONEHVAC:COOLINGPANEL:RADIANTCONVECTIVE:WATER",
    "OS:ZONEMIXING",
    "OS:WATERHEATER:MIXED",
    "OS:WATERHEATER:HEATPUMP",
    "OS:COIL:WATERHEATING:AIRTOWATERHEATPUMP",
    "OS:WATERHEATER:HEATPUMP:WRAPPEDCONDENSER",
    "OS:COIL:WATERHEATING:AIRTOWATERHEATPUMP:VARIABLESPEED",
    "OS:COIL:WATERHEATING:AIRTOWATERHEATPUMP:VARIABLESPEED:SPEEDDATA",
    "OS:COIL:WATERHEATING:AIRTOWATERHEATPUMP:WRAPPED",
    "OS:WATERHEATER:STRATIFIED",
    "OS:WATERHEATER:SIZING",
    "OS:WATERUSE:EQUIPMENT",
    "OS:WATERUSE:EQUIPMENT:DEFINITION",
    "OS:WATERUSE:CONNECTIONS",
    "OS:THERMALSTORAGE:ICE:DETAILED",
    "OS:THERMALSTORAGE:CHILLEDWATER:STRATIFIED",
    "OS:DUCT",
    "OS:FLUIDCOOLER:SINGLESPEED",
    "OS:FLUIDCOOLER:TWOSPEED",
    "OS:PIPE:INDOOR",
    "OS:PIPE:OUTDOOR",
    "OS:SWIMMINGPOOL:INDOOR",
    "OS:SOLARCOLLECTORPERFORMANCE:FLATPLATE",
    "OS:SOLARCOLLECTOR:FLATPLATE:WATER",
    "OS:SOLARCOLLECTORPERFORMANCE:INTEGRALCOLLECTORSTORAGE",
    "OS:SOLARCOLLECTOR:INTEGRALCOLLECTORSTORAGE",
    "OS:SOLARCOLLECTORPERFORMANCE:PHOTOVOLTAICTHERMAL:SIMPLE",
    "OS:SOLARCOLLECTORPERFORMANCE:PHOTOVOLTAICTHERMAL:BIPVT",
    "OS:SOLARCOLLECTOR:FLATPLATE:PHOTOVOLTAICTHERMAL",
    "OS:GENERATOR:MICROTURBINE",
    "OS:GENERATOR:MICROTURBINE:HEATRECOVERY",
    "OS:GENERATOR:PHOTOVOLTAIC",
    "OS:PHOTOVOLTAICPERFORMANCE:SIMPLE",
    "OS:PHOTOVOLTAICPERFORMANCE:EQUIVALENTONE-DIODE",
    "OS:PHOTOVOLTAICPERFORMANCE:SANDIA",
    "OS:GENERATOR:FUELCELL",
    "OS:GENERATOR:FUELCELL:POWERMODULE",
    "OS:GENERATOR:FUELCELL:AIRSUPPLY",
    "OS:GENERATOR:FUELCELL:WATERSUPPLY",
    "OS:GENERATOR:FUELCELL:AUXILIARYHEATER",
    "OS:GENERATOR:FUELCELL:EXHAUSTGASTOWATERHEATEXCHANGER",
    "OS:GENERATOR:FUELCELL:ELECTRICALSTORAGE",
    "OS:GENERATOR:FUELCELL:INVERTER",
    "OS:GENERATOR:FUELCELL:STACKCOOLER",
    "OS:GENERATOR:FUELSUPPLY",
    "OS:GENERATOR:WINDTURBINE",
    "OS:GENERATOR:PVWATTS",
    "OS:ELECTRICLOADCENTER:INVERTER:PVWATTS",
    "OS:ELECTRICLOADCENTER:INVERTER:SIMPLE",
    "OS:ELECTRICLOADCENTER:INVERTER:LOOKUPTABLE",
    "OS:ELECTRICLOADCENTER:STORAGE:SIMPLE",
    "OS:ELECTRICLOADCENTER:STORAGE:LIIONNMCBATTERY",
    "OS:ELECTRICLOADCENTER:TRANSFORMER",
    "OS:ELECTRICLOADCENTER:DISTRIBUTION",
    "OS:ELECTRICLOADCENTER:STORAGE:CONVERTER",
    "OS:COMPONENTCOST:ADJUSTMENTS",
    "OS:LIFECYCLECOST",
    "OS:CURRENCYTYPE",
    "OS:LIFECYCLECOST:PARAMETERS",
    "OS:LIFECYCLECOST:USEPRICEESCALATION",
    "OS:UTILITYBILL",
    "OS:UTILITYCOST:TARIFF",
    "OS:UTILITYCOST:QUALIFY",
    "OS:UTILITYCOST:CHARGE:SIMPLE",
    "OS:UTILITYCOST:CHARGE:BLOCK",
    "OS:UTILITYCOST:RATCHET",
    "OS:UTILITYCOST:VARIABLE",
    "OS:UTILITYCOST:COMPUTATION",
    "OS:OUTPUT:SCHEDULES",
    "OS:OUTPUT:CONSTRUCTIONS",
    "OS:OUTPUTCONTROL:FILES",
    "OS:OUTPUTCONTROL:TIMESTAMP",
    "OS:OUTPUTCONTROL:REPORTINGTOLERANCES",
    "OS:OUTPUTCONTROL:TABLE:STYLE",
    "OS:OUTPUT:METER",
    "OS:OUTPUT:VARIABLE",
    "OS:METER:CUSTOM",
    "OS:METER:CUSTOMDECREMENT",
    "OS:OUTPUT:ENVIRONMENTALIMPACTFACTORS",
    "OS:ENVIRONMENTALIMPACTFACTORS",
    "OS:FUELFACTORS",
    "OS:OUTPUT:DIAGNOSTICS",
    "OS:OUTPUT:DEBUGGINGDATA",
    "OS:OUTPUT:JSON",
    "OS:OUTPUT:SQLITE",
    "OS:OUTPUT:ENERGYMANAGEMENTSYSTEM",
    "OS:OUTPUT:TABLE:SUMMARYREPORTS",
    "OS:ENERGYMANAGEMENTSYSTEM:SENSOR",
    "OS:ENERGYMANAGEMENTSYSTEM:ACTUATOR",
    "OS:ENERGYMANAGEMENTSYSTEM:PROGRAMCALLINGMANAGER",
    "OS:ENERGYMANAGEMENTSYSTEM:PROGRAM",
    "OS:ENERGYMANAGEMENTSYSTEM:SUBROUTINE",
    "OS:ENERGYMANAGEMENTSYSTEM:GLOBALVARIABLE",
    "OS:ENERGYMANAGEMENTSYSTEM:OUTPUTVARIABLE",
    "OS:ENERGYMANAGEMENTSYSTEM:METEREDOUTPUTVARIABLE",
    "OS:ENERGYMANAGEMENTSYSTEM:TRENDVARIABLE",
    "OS:ENERGYMANAGEMENTSYSTEM:INTERNALVARIABLE",
    "OS:ENERGYMANAGEMENTSYSTEM:CURVEORTABLEINDEXVARIABLE",
    "OS:ENERGYMANAGEMENTSYSTEM:CONSTRUCTIONINDEXVARIABLE",
    "OS:PYTHONPLUGIN:INSTANCE",
    "OS:PYTHONPLUGIN:VARIABLE",
    "OS:PYTHONPLUGIN:TRENDVARIABLE",
    "OS:PYTHONPLUGIN:OUTPUTVARIABLE",
    "OS:AIRFLOWNETWORKSIMULATIONCONTROL",
    "OS:AIRFLOWNETWORKZONE",
    "OS:AIRFLOWNETWORKSURFACE",
    "OS:AIRFLOWNETWORKREFERENCECRACKCONDITIONS",
    "OS:AIRFLOWNETWORKCRACK",
    "OS:AIRFLOWNETWORKEFFECTIVELEAKAGEAREA",
    "OS:AIRFLOWNETWORKSPECIFIEDFLOWRATE",
    "OS:AIRFLOWNETWORKDETAILEDOPENING",
    "OS:AIRFLOWNETWORKSIMPLEOPENING",
    "OS:AIRFLOWNETWORKHORIZONTALOPENING",
    "OS:AIRFLOWNETWORKOUTDOORAIRFLOW",
    "OS:AIRFLOWNETWORKZONEEXHAUSTFAN",
    "OS:AIRFLOWNETWORKEXTERNALNODE",
    "OS:AIRFLOWNETWORKPRESSURECONTROLLER",
    "OS:AIRFLOWNETWORKDISTRIBUTIONNODE",
    "OS:AIRFLOWNETWORKLEAK",
    "OS:AIRFLOWNETWORKLEAKAGERATIO",
    "OS:AIRFLOWNETWORKDUCT",
    "OS:AIRFLOWNETWORKFAN",
    "OS:AIRFLOWNETWORKEQUIVALENTDUCT",
    "OS:AIRFLOWNETWORKCONSTANTPRESSUREDROP",
    "OS:AIRFLOWNETWORKRELIEFAIRFLOW",
    "OS:AIRFLOWNETWORKDISTRIBUTIONLINKAGE",
    "OS:AIRFLOWNETWORKDUCTVIEWFACTORS",
    "OS:AIRFLOWNETWORKOCCUPANTVENTILATIONCONTROL",
    "OS:EXTERNALINTERFACE",
    "OS:EXTERNALINTERFACE:SCHEDULE",
    "OS:EXTERNALINTERFACE:VARIABLE",
    "OS:EXTERNALINTERFACE:ACTUATOR",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:FROM:VARIABLE",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:TO:SCHEDULE",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:TO:ACTUATOR",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITIMPORT:TO:VARIABLE",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:FROM:VARIABLE",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:TO:SCHEDULE",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:TO:ACTUATOR",
    "OS:EXTERNALINTERFACE:FUNCTIONALMOCKUPUNITEXPORT:TO:VARIABLE",
    "OS:COIL:USERDEFINED",
    "OS:PLANTCOMPONENT:USERDEFINED",
    "",
  };

  static constexpr size_t max_element_size = []() -> std::size_t {
    auto size_predicate = [](const auto& lhs, const auto& rhs) { return lhs.size() < rhs.size(); };
    auto it = std::max_element(namesUC.cbegin(), namesUC.cend(), size_predicate);
    auto it2 = std::max_element(descriptionsUC.cbegin(), descriptionsUC.cend(), size_predicate);
    return std::max(it->size(), it2->size());
  }() + 1;
};

inline std::ostream& operator<<(std::ostream& os, const IddObjectType& e) {
  return os << e.valueName() << "(" << e.value() << ")";
}

/** \class IddObjectType */
enum IddObjectType::domain : IddObjectType::SizeType
{
  Catchall,
  UserCustom,
  LeadInput,
  EndLeadInput,
  SimulationData,
  EndSimulationData,
  Version,
  SimulationControl,
  PerformancePrecisionTradeoffs,
  Building,
  ShadowCalculation,
  SurfaceConvectionAlgorithm_Inside,
  SurfaceConvectionAlgorithm_Outside,
  HeatBalanceAlgorithm,
  HeatBalanceSettings_ConductionFiniteDifference,
  ZoneAirHeatBalanceAlgorithm,
  ZoneAirContaminantBalance,
  ZoneAirMassFlowConservation,
  ZoneCapacitanceMultiplier_ResearchSpecial,
  Timestep,
  ConvergenceLimits,
  HVACSystemRootFindingAlgorithm,
  Compliance_Building,
  Site_Location,
  Site_VariableLocation,
  SizingPeriod_DesignDay,
  SizingPeriod_WeatherFileDays,
  SizingPeriod_WeatherFileConditionType,
  RunPeriod,
  RunPeriodControl_SpecialDays,
  RunPeriodControl_DaylightSavingTime,
  WeatherProperty_SkyTemperature,
  Site_WeatherStation,
  Site_HeightVariation,
  Site_GroundTemperature_BuildingSurface,
  Site_GroundTemperature_FCfactorMethod,
  Site_GroundTemperature_Shallow,
  Site_GroundTemperature_Deep,
  Site_GroundTemperature_Undisturbed_FiniteDifference,
  Site_GroundTemperature_Undisturbed_KusudaAchenbach,
  Site_GroundTemperature_Undisturbed_Xing,
  Site_GroundDomain_Slab,
  Site_GroundDomain_Basement,
  Site_GroundReflectance,
  Site_GroundReflectance_SnowModifier,
  Site_WaterMainsTemperature,
  Site_Precipitation,
  RoofIrrigation,
  Site_SolarAndVisibleSpectrum,
  Site_SpectrumData,
  ScheduleTypeLimits,
  Schedule_Day_Hourly,
  Schedule_Day_Interval,
  Schedule_Day_List,
  Schedule_Week_Daily,
  Schedule_Week_Compact,
  Schedule_Year,
  Schedule_Compact,
  Schedule_Constant,
  Schedule_File_Shading,
  Schedule_File,
  Material,
  Material_NoMass,
  Material_InfraredTransparent,
  Material_AirGap,
  Material_RoofVegetation,
  WindowMaterial_SimpleGlazingSystem,
  WindowMaterial_Glazing,
  WindowMaterial_GlazingGroup_Thermochromic,
  WindowMaterial_Glazing_RefractionExtinctionMethod,
  WindowMaterial_Gas,
  WindowGap_SupportPillar,
  WindowGap_DeflectionState,
  WindowMaterial_GasMixture,
  WindowMaterial_Gap,
  WindowMaterial_Shade,
  WindowMaterial_ComplexShade,
  WindowMaterial_Blind,
  WindowMaterial_Screen,
  WindowMaterial_Shade_EquivalentLayer,
  WindowMaterial_Drape_EquivalentLayer,
  WindowMaterial_Blind_EquivalentLayer,
  WindowMaterial_Screen_EquivalentLayer,
  WindowMaterial_Glazing_EquivalentLayer,
  WindowMaterial_Gap_EquivalentLayer,
  MaterialProperty_MoisturePenetrationDepth_Settings,
  MaterialProperty_PhaseChange,
  MaterialProperty_PhaseChangeHysteresis,
  MaterialProperty_VariableThermalConductivity,
  MaterialProperty_VariableAbsorptance,
  MaterialProperty_HeatAndMoistureTransfer_Settings,
  MaterialProperty_HeatAndMoistureTransfer_SorptionIsotherm,
  MaterialProperty_HeatAndMoistureTransfer_Suction,
  MaterialProperty_HeatAndMoistureTransfer_Redistribution,
  MaterialProperty_HeatAndMoistureTransfer_Diffusion,
  MaterialProperty_HeatAndMoistureTransfer_ThermalConductivity,
  MaterialProperty_GlazingSpectralData,
  Construction,
  Construction_CfactorUndergroundWall,
  Construction_FfactorGroundFloor,
  ConstructionProperty_InternalHeatSource,
  Construction_AirBoundary,
  WindowThermalModel_Params,
  WindowsCalculationEngine,
  Construction_ComplexFenestrationState,
  Construction_WindowEquivalentLayer,
  Construction_WindowDataFile,
  GlobalGeometryRules,
  GeometryTransform,
  Space,
  SpaceList,
  Zone,
  ZoneList,
  ZoneGroup,
  BuildingSurface_Detailed,
  Wall_Detailed,
  RoofCeiling_Detailed,
  Floor_Detailed,
  Wall_Exterior,
  Wall_Adiabatic,
  Wall_Underground,
  Wall_Interzone,
  Roof,
  Ceiling_Adiabatic,
  Ceiling_Interzone,
  Floor_GroundContact,
  Floor_Adiabatic,
  Floor_Interzone,
  FenestrationSurface_Detailed,
  Window,
  Door,
  GlazedDoor,
  Window_Interzone,
  Door_Interzone,
  GlazedDoor_Interzone,
  WindowShadingControl,
  WindowProperty_FrameAndDivider,
  WindowProperty_AirflowControl,
  WindowProperty_StormWindow,
  InternalMass,
  Shading_Site,
  Shading_Building,
  Shading_Site_Detailed,
  Shading_Building_Detailed,
  Shading_Overhang,
  Shading_Overhang_Projection,
  Shading_Fin,
  Shading_Fin_Projection,
  Shading_Zone_Detailed,
  ShadingProperty_Reflectance,
  SurfaceProperty_HeatTransferAlgorithm,
  SurfaceProperty_HeatTransferAlgorithm_MultipleSurface,
  SurfaceProperty_HeatTransferAlgorithm_SurfaceList,
  SurfaceProperty_HeatTransferAlgorithm_Construction,
  SurfaceProperty_HeatBalanceSourceTerm,
  SurfaceControl_MovableInsulation,
  SurfaceProperty_OtherSideCoefficients,
  SurfaceProperty_OtherSideConditionsModel,
  SurfaceProperty_Underwater,
  Foundation_Kiva,
  Foundation_Kiva_Settings,
  SurfaceProperty_ExposedFoundationPerimeter,
  SurfaceConvectionAlgorithm_Inside_AdaptiveModelSelections,
  SurfaceConvectionAlgorithm_Outside_AdaptiveModelSelections,
  SurfaceConvectionAlgorithm_Inside_UserCurve,
  SurfaceConvectionAlgorithm_Outside_UserCurve,
  SurfaceProperty_ConvectionCoefficients,
  SurfaceProperty_ConvectionCoefficients_MultipleSurface,
  SurfaceProperties_VaporCoefficients,
  SurfaceProperty_ExteriorNaturalVentedCavity,
  SurfaceProperty_SolarIncidentInside,
  SurfaceProperty_IncidentSolarMultiplier,
  SurfaceProperty_LocalEnvironment,
  ZoneProperty_LocalEnvironment,
  SurfaceProperty_SurroundingSurfaces,
  SurfaceProperty_GroundSurfaces,
  ComplexFenestrationProperty_SolarAbsorbedLayers,
  ZoneProperty_UserViewFactors_BySurfaceName,
  GroundHeatTransfer_Control,
  GroundHeatTransfer_Slab_Materials,
  GroundHeatTransfer_Slab_MatlProps,
  GroundHeatTransfer_Slab_BoundConds,
  GroundHeatTransfer_Slab_BldgProps,
  GroundHeatTransfer_Slab_Insulation,
  GroundHeatTransfer_Slab_EquivalentSlab,
  GroundHeatTransfer_Slab_AutoGrid,
  GroundHeatTransfer_Slab_ManualGrid,
  GroundHeatTransfer_Slab_XFACE,
  GroundHeatTransfer_Slab_YFACE,
  GroundHeatTransfer_Slab_ZFACE,
  GroundHeatTransfer_Basement_SimParameters,
  GroundHeatTransfer_Basement_MatlProps,
  GroundHeatTransfer_Basement_Insulation,
  GroundHeatTransfer_Basement_SurfaceProps,
  GroundHeatTransfer_Basement_BldgData,
  GroundHeatTransfer_Basement_Interior,
  GroundHeatTransfer_Basement_ComBldg,
  GroundHeatTransfer_Basement_EquivSlab,
  GroundHeatTransfer_Basement_EquivAutoGrid,
  GroundHeatTransfer_Basement_AutoGrid,
  GroundHeatTransfer_Basement_ManualGrid,
  GroundHeatTransfer_Basement_XFACE,
  GroundHeatTransfer_Basement_YFACE,
  GroundHeatTransfer_Basement_ZFACE,
  RoomAirModelType,
  RoomAir_TemperaturePattern_UserDefined,
  RoomAir_TemperaturePattern_ConstantGradient,
  RoomAir_TemperaturePattern_TwoGradient,
  RoomAir_TemperaturePattern_NondimensionalHeight,
  RoomAir_TemperaturePattern_SurfaceMapping,
  RoomAir_Node,
  RoomAirSettings_OneNodeDisplacementVentilation,
  RoomAirSettings_ThreeNodeDisplacementVentilation,
  RoomAirSettings_CrossVentilation,
  RoomAirSettings_UnderFloorAirDistributionInterior,
  RoomAirSettings_UnderFloorAirDistributionExterior,
  RoomAir_Node_AirflowNetwork,
  RoomAir_Node_AirflowNetwork_AdjacentSurfaceList,
  RoomAir_Node_AirflowNetwork_InternalGains,
  RoomAir_Node_AirflowNetwork_HVACEquipment,
  RoomAirSettings_AirflowNetwork,
  People,
  ComfortViewFactorAngles,
  Lights,
  ElectricEquipment,
  GasEquipment,
  HotWaterEquipment,
  SteamEquipment,
  OtherEquipment,
  ElectricEquipment_ITE_AirCooled,
  ZoneBaseboard_OutdoorTemperatureControlled,
  SwimmingPool_Indoor,
  ZoneContaminantSourceAndSink_CarbonDioxide,
  ZoneContaminantSourceAndSink_Generic_Constant,
  SurfaceContaminantSourceAndSink_Generic_PressureDriven,
  ZoneContaminantSourceAndSink_Generic_CutoffModel,
  ZoneContaminantSourceAndSink_Generic_DecaySource,
  SurfaceContaminantSourceAndSink_Generic_BoundaryLayerDiffusion,
  SurfaceContaminantSourceAndSink_Generic_DepositionVelocitySink,
  ZoneContaminantSourceAndSink_Generic_DepositionRateSink,
  Daylighting_Controls,
  Daylighting_ReferencePoint,
  Daylighting_DELight_ComplexFenestration,
  DaylightingDevice_Tubular,
  DaylightingDevice_Shelf,
  DaylightingDevice_LightWell,
  Output_DaylightFactors,
  Output_IlluminanceMap,
  OutputControl_IlluminanceMap_Style,
  ZoneInfiltration_DesignFlowRate,
  ZoneInfiltration_EffectiveLeakageArea,
  ZoneInfiltration_FlowCoefficient,
  ZoneVentilation_DesignFlowRate,
  ZoneVentilation_WindandStackOpenArea,
  ZoneAirBalance_OutdoorAir,
  ZoneMixing,
  ZoneCrossMixing,
  ZoneRefrigerationDoorMixing,
  ZoneEarthtube,
  ZoneCoolTower_Shower,
  ZoneThermalChimney,
  AirflowNetwork_SimulationControl,
  AirflowNetwork_MultiZone_Zone,
  AirflowNetwork_MultiZone_Surface,
  AirflowNetwork_MultiZone_ReferenceCrackConditions,
  AirflowNetwork_MultiZone_Surface_Crack,
  AirflowNetwork_MultiZone_Surface_EffectiveLeakageArea,
  AirflowNetwork_MultiZone_SpecifiedFlowRate,
  AirflowNetwork_MultiZone_Component_DetailedOpening,
  AirflowNetwork_MultiZone_Component_SimpleOpening,
  AirflowNetwork_MultiZone_Component_HorizontalOpening,
  AirflowNetwork_MultiZone_Component_ZoneExhaustFan,
  AirflowNetwork_MultiZone_ExternalNode,
  AirflowNetwork_MultiZone_WindPressureCoefficientArray,
  AirflowNetwork_MultiZone_WindPressureCoefficientValues,
  AirflowNetwork_ZoneControl_PressureController,
  AirflowNetwork_Distribution_Node,
  AirflowNetwork_Distribution_Component_Leak,
  AirflowNetwork_Distribution_Component_LeakageRatio,
  AirflowNetwork_Distribution_Component_Duct,
  AirflowNetwork_Distribution_Component_Fan,
  AirflowNetwork_Distribution_Component_Coil,
  AirflowNetwork_Distribution_Component_HeatExchanger,
  AirflowNetwork_Distribution_Component_TerminalUnit,
  AirflowNetwork_Distribution_Component_ConstantPressureDrop,
  AirflowNetwork_Distribution_Component_OutdoorAirFlow,
  AirflowNetwork_Distribution_Component_ReliefAirFlow,
  AirflowNetwork_Distribution_Linkage,
  AirflowNetwork_Distribution_DuctViewFactors,
  AirflowNetwork_OccupantVentilationControl,
  AirflowNetwork_IntraZone_Node,
  AirflowNetwork_IntraZone_Linkage,
  Exterior_Lights,
  Exterior_FuelEquipment,
  Exterior_WaterEquipment,
  HVACTemplate_Thermostat,
  HVACTemplate_Zone_IdealLoadsAirSystem,
  HVACTemplate_Zone_BaseboardHeat,
  HVACTemplate_Zone_FanCoil,
  HVACTemplate_Zone_PTAC,
  HVACTemplate_Zone_PTHP,
  HVACTemplate_Zone_WaterToAirHeatPump,
  HVACTemplate_Zone_VRF,
  HVACTemplate_Zone_Unitary,
  HVACTemplate_Zone_VAV,
  HVACTemplate_Zone_VAV_FanPowered,
  HVACTemplate_Zone_VAV_HeatAndCool,
  HVACTemplate_Zone_ConstantVolume,
  HVACTemplate_Zone_DualDuct,
  HVACTemplate_System_VRF,
  HVACTemplate_System_Unitary,
  HVACTemplate_System_UnitaryHeatPump_AirToAir,
  HVACTemplate_System_UnitarySystem,
  HVACTemplate_System_VAV,
  HVACTemplate_System_PackagedVAV,
  HVACTemplate_System_ConstantVolume,
  HVACTemplate_System_DualDuct,
  HVACTemplate_System_DedicatedOutdoorAir,
  HVACTemplate_Plant_ChilledWaterLoop,
  HVACTemplate_Plant_Chiller,
  HVACTemplate_Plant_Chiller_ObjectReference,
  HVACTemplate_Plant_Tower,
  HVACTemplate_Plant_Tower_ObjectReference,
  HVACTemplate_Plant_HotWaterLoop,
  HVACTemplate_Plant_Boiler,
  HVACTemplate_Plant_Boiler_ObjectReference,
  HVACTemplate_Plant_MixedWaterLoop,
  DesignSpecification_OutdoorAir,
  DesignSpecification_OutdoorAir_SpaceList,
  DesignSpecification_ZoneAirDistribution,
  Sizing_Parameters,
  Sizing_Zone,
  DesignSpecification_ZoneHVAC_Sizing,
  DesignSpecification_AirTerminal_Sizing,
  Sizing_System,
  Sizing_Plant,
  OutputControl_Sizing_Style,
  ZoneControl_Humidistat,
  ZoneControl_Thermostat,
  ZoneControl_Thermostat_OperativeTemperature,
  ZoneControl_Thermostat_ThermalComfort,
  ZoneControl_Thermostat_TemperatureAndHumidity,
  ThermostatSetpoint_SingleHeating,
  ThermostatSetpoint_SingleCooling,
  ThermostatSetpoint_SingleHeatingOrCooling,
  ThermostatSetpoint_DualSetpoint,
  ThermostatSetpoint_ThermalComfort_Fanger_SingleHeating,
  ThermostatSetpoint_ThermalComfort_Fanger_SingleCooling,
  ThermostatSetpoint_ThermalComfort_Fanger_SingleHeatingOrCooling,
  ThermostatSetpoint_ThermalComfort_Fanger_DualSetpoint,
  ZoneControl_Thermostat_StagedDualSetpoint,
  ZoneControl_ContaminantController,
  ZoneHVAC_IdealLoadsAirSystem,
  ZoneHVAC_FourPipeFanCoil,
  ZoneHVAC_WindowAirConditioner,
  ZoneHVAC_PackagedTerminalAirConditioner,
  ZoneHVAC_PackagedTerminalHeatPump,
  ZoneHVAC_WaterToAirHeatPump,
  ZoneHVAC_Dehumidifier_DX,
  ZoneHVAC_EnergyRecoveryVentilator,
  ZoneHVAC_EnergyRecoveryVentilator_Controller,
  ZoneHVAC_UnitVentilator,
  ZoneHVAC_UnitHeater,
  ZoneHVAC_EvaporativeCoolerUnit,
  ZoneHVAC_HybridUnitaryHVAC,
  ZoneHVAC_OutdoorAirUnit,
  ZoneHVAC_OutdoorAirUnit_EquipmentList,
  ZoneHVAC_TerminalUnit_VariableRefrigerantFlow,
  ZoneHVAC_Baseboard_RadiantConvective_Water_Design,
  ZoneHVAC_Baseboard_RadiantConvective_Water,
  ZoneHVAC_Baseboard_RadiantConvective_Steam_Design,
  ZoneHVAC_Baseboard_RadiantConvective_Steam,
  ZoneHVAC_Baseboard_RadiantConvective_Electric,
  ZoneHVAC_CoolingPanel_RadiantConvective_Water,
  ZoneHVAC_Baseboard_Convective_Water,
  ZoneHVAC_Baseboard_Convective_Electric,
  ZoneHVAC_LowTemperatureRadiant_VariableFlow,
  ZoneHVAC_LowTemperatureRadiant_VariableFlow_Design,
  ZoneHVAC_LowTemperatureRadiant_ConstantFlow,
  ZoneHVAC_LowTemperatureRadiant_ConstantFlow_Design,
  ZoneHVAC_LowTemperatureRadiant_Electric,
  ZoneHVAC_LowTemperatureRadiant_SurfaceGroup,
  ZoneHVAC_HighTemperatureRadiant,
  ZoneHVAC_VentilatedSlab,
  ZoneHVAC_VentilatedSlab_SlabGroup,
  AirTerminal_SingleDuct_ConstantVolume_Reheat,
  AirTerminal_SingleDuct_ConstantVolume_NoReheat,
  AirTerminal_SingleDuct_VAV_NoReheat,
  AirTerminal_SingleDuct_VAV_Reheat,
  AirTerminal_SingleDuct_VAV_Reheat_VariableSpeedFan,
  AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat,
  AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat,
  AirTerminal_SingleDuct_SeriesPIU_Reheat,
  AirTerminal_SingleDuct_ParallelPIU_Reheat,
  AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction,
  AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam,
  AirTerminal_SingleDuct_ConstantVolume_CooledBeam,
  AirTerminal_SingleDuct_Mixer,
  AirTerminal_DualDuct_ConstantVolume,
  AirTerminal_DualDuct_VAV,
  AirTerminal_DualDuct_VAV_OutdoorAir,
  ZoneHVAC_AirDistributionUnit,
  ZoneHVAC_ExhaustControl,
  ZoneHVAC_EquipmentList,
  ZoneHVAC_EquipmentConnections,
  Fan_SystemModel,
  Fan_ConstantVolume,
  Fan_VariableVolume,
  Fan_OnOff,
  Fan_ZoneExhaust,
  FanPerformance_NightVentilation,
  Fan_ComponentModel,
  Coil_Cooling_Water,
  Coil_Cooling_Water_DetailedGeometry,
  CoilSystem_Cooling_Water,
  Coil_Cooling_DX,
  Coil_Cooling_DX_CurveFit_Performance,
  Coil_Cooling_DX_CurveFit_OperatingMode,
  Coil_Cooling_DX_CurveFit_Speed,
  Coil_Cooling_DX_SingleSpeed,
  Coil_Cooling_DX_TwoSpeed,
  Coil_Cooling_DX_MultiSpeed,
  Coil_Cooling_DX_VariableSpeed,
  Coil_Cooling_DX_TwoStageWithHumidityControlMode,
  CoilPerformance_DX_Cooling,
  Coil_Cooling_DX_VariableRefrigerantFlow,
  Coil_Heating_DX_VariableRefrigerantFlow,
  Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  Coil_Heating_Water,
  Coil_Heating_Steam,
  Coil_Heating_Electric,
  Coil_Heating_Electric_MultiStage,
  Coil_Heating_Fuel,
  Coil_Heating_Gas_MultiStage,
  Coil_Heating_Desuperheater,
  Coil_Heating_DX_SingleSpeed,
  Coil_Heating_DX_MultiSpeed,
  Coil_Heating_DX_VariableSpeed,
  Coil_Cooling_WaterToAirHeatPump_ParameterEstimation,
  Coil_Heating_WaterToAirHeatPump_ParameterEstimation,
  Coil_Cooling_WaterToAirHeatPump_EquationFit,
  Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit,
  Coil_Heating_WaterToAirHeatPump_EquationFit,
  Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit,
  Coil_WaterHeating_AirToWaterHeatPump_Pumped,
  Coil_WaterHeating_AirToWaterHeatPump_Wrapped,
  Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed,
  Coil_WaterHeating_Desuperheater,
  CoilSystem_Cooling_DX,
  CoilSystem_Heating_DX,
  CoilSystem_Cooling_Water_HeatExchangerAssisted,
  CoilSystem_Cooling_DX_HeatExchangerAssisted,
  CoilSystem_IntegratedHeatPump_AirSource,
  Coil_Cooling_DX_SingleSpeed_ThermalStorage,
  EvaporativeCooler_Direct_CelDekPad,
  EvaporativeCooler_Indirect_CelDekPad,
  EvaporativeCooler_Indirect_WetCoil,
  EvaporativeCooler_Indirect_ResearchSpecial,
  EvaporativeCooler_Direct_ResearchSpecial,
  Humidifier_Steam_Electric,
  Humidifier_Steam_Gas,
  Dehumidifier_Desiccant_NoFans,
  Dehumidifier_Desiccant_System,
  HeatExchanger_AirToAir_FlatPlate,
  HeatExchanger_AirToAir_SensibleAndLatent,
  HeatExchanger_Desiccant_BalancedFlow,
  HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1,
  AirLoopHVAC_UnitarySystem,
  UnitarySystemPerformance_Multispeed,
  AirLoopHVAC_Unitary_Furnace_HeatOnly,
  AirLoopHVAC_Unitary_Furnace_HeatCool,
  AirLoopHVAC_UnitaryHeatOnly,
  AirLoopHVAC_UnitaryHeatCool,
  AirLoopHVAC_UnitaryHeatPump_AirToAir,
  AirLoopHVAC_UnitaryHeatPump_WaterToAir,
  AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass,
  AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed,
  AirConditioner_VariableRefrigerantFlow,
  AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl,
  AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR,
  ZoneTerminalUnitList,
  Controller_WaterCoil,
  Controller_OutdoorAir,
  Controller_MechanicalVentilation,
  AirLoopHVAC_ControllerList,
  AirLoopHVAC,
  AirLoopHVAC_OutdoorAirSystem_EquipmentList,
  AirLoopHVAC_OutdoorAirSystem,
  OutdoorAir_Mixer,
  AirLoopHVAC_ZoneSplitter,
  AirLoopHVAC_SupplyPlenum,
  AirLoopHVAC_SupplyPath,
  AirLoopHVAC_ZoneMixer,
  AirLoopHVAC_ReturnPlenum,
  AirLoopHVAC_ReturnPath,
  AirLoopHVAC_ExhaustSystem,
  AirLoopHVAC_DedicatedOutdoorAirSystem,
  AirLoopHVAC_Mixer,
  AirLoopHVAC_Splitter,
  Branch,
  BranchList,
  Connector_Splitter,
  Connector_Mixer,
  ConnectorList,
  NodeList,
  OutdoorAir_Node,
  OutdoorAir_NodeList,
  Pipe_Adiabatic,
  Pipe_Adiabatic_Steam,
  Pipe_Indoor,
  Pipe_Outdoor,
  Pipe_Underground,
  PipingSystem_Underground_Domain,
  PipingSystem_Underground_PipeCircuit,
  PipingSystem_Underground_PipeSegment,
  Duct,
  Pump_VariableSpeed,
  Pump_ConstantSpeed,
  Pump_VariableSpeed_Condensate,
  HeaderedPumps_ConstantSpeed,
  HeaderedPumps_VariableSpeed,
  TemperingValve,
  LoadProfile_Plant,
  SolarCollectorPerformance_FlatPlate,
  SolarCollector_FlatPlate_Water,
  SolarCollector_FlatPlate_PhotovoltaicThermal,
  SolarCollectorPerformance_PhotovoltaicThermal_Simple,
  SolarCollectorPerformance_PhotovoltaicThermal_BIPVT,
  SolarCollector_IntegralCollectorStorage,
  SolarCollectorPerformance_IntegralCollectorStorage,
  SolarCollector_UnglazedTranspired,
  SolarCollector_UnglazedTranspired_Multisystem,
  Boiler_HotWater,
  Boiler_Steam,
  Chiller_Electric_ASHRAE205,
  Chiller_Electric_EIR,
  Chiller_Electric_ReformulatedEIR,
  Chiller_Electric,
  Chiller_Absorption_Indirect,
  Chiller_Absorption,
  Chiller_ConstantCOP,
  Chiller_EngineDriven,
  Chiller_CombustionTurbine,
  ChillerHeater_Absorption_DirectFired,
  ChillerHeater_Absorption_DoubleEffect,
  HeatPump_PlantLoop_EIR_Cooling,
  HeatPump_PlantLoop_EIR_Heating,
  HeatPump_AirToWater_FuelFired_Heating,
  HeatPump_AirToWater_FuelFired_Cooling,
  HeatPump_WaterToWater_EquationFit_Heating,
  HeatPump_WaterToWater_EquationFit_Cooling,
  HeatPump_WaterToWater_ParameterEstimation_Cooling,
  HeatPump_WaterToWater_ParameterEstimation_Heating,
  DistrictCooling,
  DistrictHeating,
  PlantComponent_TemperatureSource,
  CentralHeatPumpSystem,
  ChillerHeaterPerformance_Electric_EIR,
  CoolingTower_SingleSpeed,
  CoolingTower_TwoSpeed,
  CoolingTower_VariableSpeed_Merkel,
  CoolingTower_VariableSpeed,
  CoolingTowerPerformance_CoolTools,
  CoolingTowerPerformance_YorkCalc,
  EvaporativeFluidCooler_SingleSpeed,
  EvaporativeFluidCooler_TwoSpeed,
  FluidCooler_SingleSpeed,
  FluidCooler_TwoSpeed,
  GroundHeatExchanger_System,
  GroundHeatExchanger_Vertical_Properties,
  GroundHeatExchanger_Vertical_Array,
  GroundHeatExchanger_Vertical_Single,
  GroundHeatExchanger_ResponseFactors,
  GroundHeatExchanger_Pond,
  GroundHeatExchanger_Surface,
  GroundHeatExchanger_HorizontalTrench,
  GroundHeatExchanger_Slinky,
  HeatExchanger_FluidToFluid,
  WaterHeater_Mixed,
  WaterHeater_Stratified,
  WaterHeater_Sizing,
  WaterHeater_HeatPump_PumpedCondenser,
  WaterHeater_HeatPump_WrappedCondenser,
  ThermalStorage_Ice_Simple,
  ThermalStorage_Ice_Detailed,
  ThermalStorage_ChilledWater_Mixed,
  ThermalStorage_ChilledWater_Stratified,
  PlantLoop,
  CondenserLoop,
  PlantEquipmentList,
  CondenserEquipmentList,
  PlantEquipmentOperation_Uncontrolled,
  PlantEquipmentOperation_CoolingLoad,
  PlantEquipmentOperation_HeatingLoad,
  PlantEquipmentOperation_OutdoorDryBulb,
  PlantEquipmentOperation_OutdoorWetBulb,
  PlantEquipmentOperation_OutdoorRelativeHumidity,
  PlantEquipmentOperation_OutdoorDewpoint,
  PlantEquipmentOperation_ComponentSetpoint,
  PlantEquipmentOperation_ThermalEnergyStorage,
  PlantEquipmentOperation_OutdoorDryBulbDifference,
  PlantEquipmentOperation_OutdoorWetBulbDifference,
  PlantEquipmentOperation_OutdoorDewpointDifference,
  PlantEquipmentOperationSchemes,
  CondenserEquipmentOperationSchemes,
  EnergyManagementSystem_Sensor,
  EnergyManagementSystem_Actuator,
  EnergyManagementSystem_ProgramCallingManager,
  EnergyManagementSystem_Program,
  EnergyManagementSystem_Subroutine,
  EnergyManagementSystem_GlobalVariable,
  EnergyManagementSystem_OutputVariable,
  EnergyManagementSystem_MeteredOutputVariable,
  EnergyManagementSystem_TrendVariable,
  EnergyManagementSystem_InternalVariable,
  EnergyManagementSystem_CurveOrTableIndexVariable,
  EnergyManagementSystem_ConstructionIndexVariable,
  ExternalInterface,
  ExternalInterface_Schedule,
  ExternalInterface_Variable,
  ExternalInterface_Actuator,
  ExternalInterface_FunctionalMockupUnitImport,
  ExternalInterface_FunctionalMockupUnitImport_From_Variable,
  ExternalInterface_FunctionalMockupUnitImport_To_Schedule,
  ExternalInterface_FunctionalMockupUnitImport_To_Actuator,
  ExternalInterface_FunctionalMockupUnitImport_To_Variable,
  ExternalInterface_FunctionalMockupUnitExport_From_Variable,
  ExternalInterface_FunctionalMockupUnitExport_To_Schedule,
  ExternalInterface_FunctionalMockupUnitExport_To_Actuator,
  ExternalInterface_FunctionalMockupUnitExport_To_Variable,
  ZoneHVAC_ForcedAir_UserDefined,
  AirTerminal_SingleDuct_UserDefined,
  Coil_UserDefined,
  PlantComponent_UserDefined,
  PlantEquipmentOperation_UserDefined,
  AvailabilityManager_Scheduled,
  AvailabilityManager_ScheduledOn,
  AvailabilityManager_ScheduledOff,
  AvailabilityManager_OptimumStart,
  AvailabilityManager_NightCycle,
  AvailabilityManager_DifferentialThermostat,
  AvailabilityManager_HighTemperatureTurnOff,
  AvailabilityManager_HighTemperatureTurnOn,
  AvailabilityManager_LowTemperatureTurnOff,
  AvailabilityManager_LowTemperatureTurnOn,
  AvailabilityManager_NightVentilation,
  AvailabilityManager_HybridVentilation,
  AvailabilityManagerAssignmentList,
  SetpointManager_Scheduled,
  SetpointManager_Scheduled_DualSetpoint,
  SetpointManager_OutdoorAirReset,
  SetpointManager_SingleZone_Reheat,
  SetpointManager_SingleZone_Heating,
  SetpointManager_SingleZone_Cooling,
  SetpointManager_SingleZone_Humidity_Minimum,
  SetpointManager_SingleZone_Humidity_Maximum,
  SetpointManager_MixedAir,
  SetpointManager_OutdoorAirPretreat,
  SetpointManager_Warmest,
  SetpointManager_Coldest,
  SetpointManager_ReturnAirBypassFlow,
  SetpointManager_WarmestTemperatureFlow,
  SetpointManager_MultiZone_Heating_Average,
  SetpointManager_MultiZone_Cooling_Average,
  SetpointManager_MultiZone_MinimumHumidity_Average,
  SetpointManager_MultiZone_MaximumHumidity_Average,
  SetpointManager_MultiZone_Humidity_Minimum,
  SetpointManager_MultiZone_Humidity_Maximum,
  SetpointManager_FollowOutdoorAirTemperature,
  SetpointManager_FollowSystemNodeTemperature,
  SetpointManager_FollowGroundTemperature,
  SetpointManager_CondenserEnteringReset,
  SetpointManager_CondenserEnteringReset_Ideal,
  SetpointManager_SingleZone_OneStageCooling,
  SetpointManager_SingleZone_OneStageHeating,
  SetpointManager_ReturnTemperature_ChilledWater,
  SetpointManager_ReturnTemperature_HotWater,
  SetpointManager_SystemNodeReset_Temperature,
  SetpointManager_SystemNodeReset_Humidity,
  Refrigeration_Case,
  Refrigeration_CompressorRack,
  Refrigeration_CaseAndWalkInList,
  Refrigeration_Condenser_AirCooled,
  Refrigeration_Condenser_EvaporativeCooled,
  Refrigeration_Condenser_WaterCooled,
  Refrigeration_Condenser_Cascade,
  Refrigeration_GasCooler_AirCooled,
  Refrigeration_TransferLoadList,
  Refrigeration_Subcooler,
  Refrigeration_Compressor,
  Refrigeration_CompressorList,
  Refrigeration_System,
  Refrigeration_TranscriticalSystem,
  Refrigeration_SecondarySystem,
  Refrigeration_WalkIn,
  Refrigeration_AirChiller,
  ZoneHVAC_RefrigerationChillerSet,
  DemandManagerAssignmentList,
  DemandManager_ExteriorLights,
  DemandManager_Lights,
  DemandManager_ElectricEquipment,
  DemandManager_Thermostats,
  DemandManager_Ventilation,
  Generator_InternalCombustionEngine,
  Generator_CombustionTurbine,
  Generator_MicroTurbine,
  Generator_Photovoltaic,
  PhotovoltaicPerformance_Simple,
  PhotovoltaicPerformance_EquivalentOneDiode,
  PhotovoltaicPerformance_Sandia,
  Generator_PVWatts,
  ElectricLoadCenter_Inverter_PVWatts,
  Generator_FuelCell,
  Generator_FuelCell_PowerModule,
  Generator_FuelCell_AirSupply,
  Generator_FuelCell_WaterSupply,
  Generator_FuelCell_AuxiliaryHeater,
  Generator_FuelCell_ExhaustGasToWaterHeatExchanger,
  Generator_FuelCell_ElectricalStorage,
  Generator_FuelCell_Inverter,
  Generator_FuelCell_StackCooler,
  Generator_MicroCHP,
  Generator_MicroCHP_NonNormalizedParameters,
  Generator_FuelSupply,
  Generator_WindTurbine,
  ElectricLoadCenter_Generators,
  ElectricLoadCenter_Inverter_Simple,
  ElectricLoadCenter_Inverter_FunctionOfPower,
  ElectricLoadCenter_Inverter_LookUpTable,
  ElectricLoadCenter_Storage_Simple,
  ElectricLoadCenter_Storage_Battery,
  ElectricLoadCenter_Storage_LiIonNMCBattery,
  ElectricLoadCenter_Transformer,
  ElectricLoadCenter_Distribution,
  ElectricLoadCenter_Storage_Converter,
  WaterUse_Equipment,
  WaterUse_Connections,
  WaterUse_Storage,
  WaterUse_Well,
  WaterUse_RainCollector,
  FaultModel_TemperatureSensorOffset_OutdoorAir,
  FaultModel_HumiditySensorOffset_OutdoorAir,
  FaultModel_EnthalpySensorOffset_OutdoorAir,
  FaultModel_TemperatureSensorOffset_ReturnAir,
  FaultModel_EnthalpySensorOffset_ReturnAir,
  FaultModel_TemperatureSensorOffset_ChillerSupplyWater,
  FaultModel_TemperatureSensorOffset_CoilSupplyAir,
  FaultModel_TemperatureSensorOffset_CondenserSupplyWater,
  FaultModel_ThermostatOffset,
  FaultModel_HumidistatOffset,
  FaultModel_Fouling_AirFilter,
  FaultModel_Fouling_Boiler,
  FaultModel_Fouling_EvaporativeCooler,
  FaultModel_Fouling_Chiller,
  FaultModel_Fouling_CoolingTower,
  FaultModel_Fouling_Coil,
  Matrix_TwoDimension,
  HybridModel_Zone,
  Curve_Linear,
  Curve_QuadLinear,
  Curve_QuintLinear,
  Curve_Quadratic,
  Curve_Cubic,
  Curve_Quartic,
  Curve_Exponent,
  Curve_Bicubic,
  Curve_Biquadratic,
  Curve_QuadraticLinear,
  Curve_CubicLinear,
  Curve_Triquadratic,
  Curve_Functional_PressureDrop,
  Curve_FanPressureRise,
  Curve_ExponentialSkewNormal,
  Curve_Sigmoid,
  Curve_RectangularHyperbola1,
  Curve_RectangularHyperbola2,
  Curve_ExponentialDecay,
  Curve_DoubleExponentialDecay,
  Curve_ChillerPartLoadWithLift,
  Table_IndependentVariable,
  Table_IndependentVariableList,
  Table_Lookup,
  FluidProperties_Name,
  FluidProperties_GlycolConcentration,
  FluidProperties_Temperatures,
  FluidProperties_Saturated,
  FluidProperties_Superheated,
  FluidProperties_Concentration,
  CurrencyType,
  ComponentCost_Adjustments,
  ComponentCost_Reference,
  ComponentCost_LineItem,
  UtilityCost_Tariff,
  UtilityCost_Qualify,
  UtilityCost_Charge_Simple,
  UtilityCost_Charge_Block,
  UtilityCost_Ratchet,
  UtilityCost_Variable,
  UtilityCost_Computation,
  LifeCycleCost_Parameters,
  LifeCycleCost_RecurringCosts,
  LifeCycleCost_NonrecurringCost,
  LifeCycleCost_UsePriceEscalation,
  LifeCycleCost_UseAdjustment,
  Parametric_SetValueForRun,
  Parametric_Logic,
  Parametric_RunControl,
  Parametric_FileNameSuffix,
  Output_VariableDictionary,
  Output_Surfaces_List,
  Output_Surfaces_Drawing,
  Output_Schedules,
  Output_Constructions,
  Output_EnergyManagementSystem,
  OutputControl_SurfaceColorScheme,
  Output_Table_SummaryReports,
  Output_Table_TimeBins,
  Output_Table_Monthly,
  Output_Table_Annual,
  OutputControl_Table_Style,
  OutputControl_ReportingTolerances,
  Output_Variable,
  Output_Meter,
  Output_Meter_MeterFileOnly,
  Output_Meter_Cumulative,
  Output_Meter_Cumulative_MeterFileOnly,
  Meter_Custom,
  Meter_CustomDecrement,
  OutputControl_Files,
  OutputControl_Timestamp,
  Output_JSON,
  Output_SQLite,
  Output_EnvironmentalImpactFactors,
  EnvironmentalImpactFactors,
  FuelFactors,
  Output_Diagnostics,
  Output_DebuggingData,
  Output_PreprocessorMessage,
  PythonPlugin_SearchPaths,
  PythonPlugin_Instance,
  PythonPlugin_Variables,
  PythonPlugin_TrendVariable,
  PythonPlugin_OutputVariable,
  OS_Version,
  OS_ComponentData,
  OS_ConvergenceLimits,
  OS_ProgramControl,
  OS_HeatBalanceAlgorithm,
  OS_RunPeriod,
  OS_RunPeriodControl_DaylightSavingTime,
  OS_RunPeriodControl_SpecialDays,
  OS_ShadowCalculation,
  OS_SimulationControl,
  OS_PerformancePrecisionTradeoffs,
  OS_Sizing_Parameters,
  OS_SurfaceConvectionAlgorithm_Inside,
  OS_SurfaceConvectionAlgorithm_Outside,
  OS_Timestep,
  OS_YearDescription,
  OS_ZoneAirContaminantBalance,
  OS_ZoneAirHeatBalanceAlgorithm,
  OS_ZoneAirMassFlowConservation,
  OS_ZoneCapacitanceMultiplier_ResearchSpecial,
  OS_RadianceParameters,
  OS_ClimateZones,
  OS_Site,
  OS_Site_GroundTemperature_BuildingSurface,
  OS_Site_GroundTemperature_FCfactorMethod,
  OS_Site_GroundTemperature_Shallow,
  OS_Site_GroundTemperature_Deep,
  OS_Site_GroundTemperature_Undisturbed_KusudaAchenbach,
  OS_Site_GroundTemperature_Undisturbed_Xing,
  OS_Site_GroundReflectance,
  OS_Site_WaterMainsTemperature,
  OS_SizingPeriod_DesignDay,
  OS_SizingPeriod_WeatherFileConditionType,
  OS_SizingPeriod_WeatherFileDays,
  OS_WeatherFile,
  OS_WeatherProperty_SkyTemperature,
  OS_AdditionalProperties,
  OS_BuildingStory,
  OS_BuildingUnit,
  OS_DefaultConstructionSet,
  OS_DefaultScheduleSet,
  OS_DefaultSubSurfaceConstructions,
  OS_DefaultSurfaceConstructions,
  OS_Rendering_Color,
  OS_SpaceType,
  OS_ModelObjectList,
  OS_Material,
  OS_Material_AirGap,
  OS_Material_InfraredTransparent,
  OS_Material_NoMass,
  OS_Material_RoofVegetation,
  OS_WindowMaterial_Blind,
  OS_WindowMaterial_DaylightRedirectionDevice,
  OS_WindowMaterial_Gas,
  OS_WindowMaterial_GasMixture,
  OS_WindowMaterial_Glazing,
  OS_WindowMaterial_GlazingGroup_Thermochromic,
  OS_WindowMaterial_Glazing_RefractionExtinctionMethod,
  OS_WindowMaterial_Screen,
  OS_WindowMaterial_Shade,
  OS_WindowMaterial_SimpleGlazingSystem,
  OS_StandardsInformation_Material,
  OS_MaterialProperty_GlazingSpectralData,
  OS_MaterialProperty_MoisturePenetrationDepth_Settings,
  OS_MaterialProperty_PhaseChange,
  OS_MaterialProperty_PhaseChangeHysteresis,
  OS_Construction,
  OS_Construction_CfactorUndergroundWall,
  OS_Construction_FfactorGroundFloor,
  OS_Construction_InternalSource,
  OS_Construction_AirBoundary,
  OS_Construction_WindowDataFile,
  OS_StandardsInformation_Construction,
  OS_InternalMass_Definition,
  OS_People_Definition,
  OS_Lights_Definition,
  OS_Luminaire_Definition,
  OS_ElectricEquipment_Definition,
  OS_ElectricEquipment_ITE_AirCooled_Definition,
  OS_GasEquipment_Definition,
  OS_HotWaterEquipment_Definition,
  OS_SteamEquipment_Definition,
  OS_OtherEquipment_Definition,
  OS_Exterior_Lights_Definition,
  OS_Exterior_FuelEquipment_Definition,
  OS_Exterior_WaterEquipment_Definition,
  OS_Schedule_Compact,
  OS_Schedule_Day,
  OS_Schedule_Week,
  OS_Schedule_Year,
  OS_Schedule_Rule,
  OS_Schedule_Ruleset,
  OS_Schedule_Constant,
  OS_Schedule_FixedInterval,
  OS_Schedule_VariableInterval,
  OS_ScheduleTypeLimits,
  OS_External_File,
  OS_Schedule_File,
  OS_Building,
  OS_DaylightingDevice_Tubular,
  OS_DaylightingDevice_Shelf,
  OS_DaylightingDevice_LightWell,
  OS_Facility,
  OS_InteriorPartitionSurface,
  OS_InteriorPartitionSurfaceGroup,
  OS_ShadingSurface,
  OS_ShadingSurfaceGroup,
  OS_Space,
  OS_Surface,
  OS_SubSurface,
  OS_ShadingControl,
  OS_WindowProperty_FrameAndDivider,
  OS_SurfaceControl_MovableInsulation,
  OS_SurfaceProperty_ConvectionCoefficients,
  OS_SurfaceProperty_ConvectionCoefficients_MultipleSurface,
  OS_SurfaceProperty_ExposedFoundationPerimeter,
  OS_SurfaceProperty_OtherSideCoefficients,
  OS_SurfaceProperty_OtherSideConditionsModel,
  OS_Foundation_Kiva,
  OS_Foundation_Kiva_Settings,
  OS_ZoneProperty_UserViewFactors_BySurfaceName,
  OS_SurfaceProperty_SurroundingSurfaces,
  OS_SurfaceProperty_LocalEnvironment,
  OS_SurfaceProperty_GroundSurfaces,
  OS_SurfaceProperty_IncidentSolarMultiplier,
  OS_InternalMass,
  OS_People,
  OS_Lights,
  OS_Luminaire,
  OS_ElectricEquipment,
  OS_ElectricEquipment_ITE_AirCooled,
  OS_GasEquipment,
  OS_HotWaterEquipment,
  OS_SteamEquipment,
  OS_OtherEquipment,
  OS_SpaceInfiltration_DesignFlowRate,
  OS_SpaceInfiltration_EffectiveLeakageArea,
  OS_SpaceInfiltration_FlowCoefficient,
  OS_Exterior_Lights,
  OS_Exterior_FuelEquipment,
  OS_Exterior_WaterEquipment,
  OS_Daylighting_Control,
  OS_Glare_Sensor,
  OS_IlluminanceMap,
  OS_LightingDesignDay,
  OS_LightingSimulationControl,
  OS_LightingSimulationZone,
  OS_Refrigeration_Case,
  OS_Refrigeration_Condenser_AirCooled,
  OS_Refrigeration_Condenser_EvaporativeCooled,
  OS_Refrigeration_Condenser_WaterCooled,
  OS_Refrigeration_Condenser_Cascade,
  OS_Refrigeration_Compressor,
  OS_Refrigeration_CompressorRack,
  OS_Refrigeration_DefrostCycleParameters,
  OS_Refrigeration_GasCooler_AirCooled,
  OS_Refrigeration_Subcooler_Mechanical,
  OS_Refrigeration_Subcooler_LiquidSuction,
  OS_Refrigeration_System,
  OS_Refrigeration_TranscriticalSystem,
  OS_Refrigeration_SecondarySystem,
  OS_Refrigeration_WalkIn,
  OS_Refrigeration_WalkIn_ZoneBoundary,
  OS_Refrigeration_AirChiller,
  OS_AirConditioner_VariableRefrigerantFlow,
  OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl,
  OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_LoadingIndex,
  OS_AirConditioner_VariableRefrigerantFlow_FluidTemperatureControl_HR,
  OS_AirLoopHVAC,
  OS_AirLoopHVAC_OutdoorAirSystem,
  OS_AirLoopHVAC_DedicatedOutdoorAirSystem,
  OS_AirLoopHVAC_ReturnPlenum,
  OS_AirLoopHVAC_SupplyPlenum,
  OS_AirLoopHVAC_UnitaryHeatPump_AirToAir_MultiSpeed,
  OS_AirLoopHVAC_UnitaryCoolOnly,
  OS_AirLoopHVAC_UnitaryHeatPump_AirToAir,
  OS_AirLoopHVAC_UnitaryHeatCool_VAVChangeoverBypass,
  OS_AirLoopHVAC_UnitarySystem,
  OS_UnitarySystemPerformance_Multispeed,
  OS_AirLoopHVAC_ZoneMixer,
  OS_AirLoopHVAC_ZoneSplitter,
  OS_AirTerminal_SingleDuct_VAV_HeatAndCool_NoReheat,
  OS_AirTerminal_SingleDuct_VAV_HeatAndCool_Reheat,
  OS_AirTerminal_SingleDuct_InletSideMixer,
  OS_AirTerminal_SingleDuct_ConstantVolume_Reheat,
  OS_AirTerminal_SingleDuct_SeriesPIU_Reheat,
  OS_AirTerminal_SingleDuct_ParallelPIU_Reheat,
  OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeInduction,
  OS_AirTerminal_SingleDuct_ConstantVolume_FourPipeBeam,
  OS_AirTerminal_SingleDuct_ConstantVolume_CooledBeam,
  OS_AirTerminal_SingleDuct_ConstantVolume_NoReheat,
  OS_AirTerminal_SingleDuct_VAV_NoReheat,
  OS_AirTerminal_SingleDuct_VAV_Reheat,
  OS_AirTerminal_DualDuct_VAV_OutdoorAir,
  OS_AirTerminal_DualDuct_ConstantVolume,
  OS_AirTerminal_DualDuct_VAV,
  OS_AvailabilityManagerAssignmentList,
  OS_AvailabilityManager_Scheduled,
  OS_AvailabilityManager_ScheduledOn,
  OS_AvailabilityManager_ScheduledOff,
  OS_AvailabilityManager_NightCycle,
  OS_AvailabilityManager_OptimumStart,
  OS_AvailabilityManager_DifferentialThermostat,
  OS_AvailabilityManager_HighTemperatureTurnOff,
  OS_AvailabilityManager_HighTemperatureTurnOn,
  OS_AvailabilityManager_LowTemperatureTurnOff,
  OS_AvailabilityManager_LowTemperatureTurnOn,
  OS_AvailabilityManager_NightVentilation,
  OS_AvailabilityManager_HybridVentilation,
  OS_Boiler_HotWater,
  OS_Boiler_Steam,
  OS_HeatPump_WaterToWater_EquationFit_Heating,
  OS_HeatPump_WaterToWater_EquationFit_Cooling,
  OS_HeatPump_PlantLoop_EIR_Heating,
  OS_HeatPump_PlantLoop_EIR_Cooling,
  OS_HeatPump_AirToWater_FuelFired_Heating,
  OS_HeatPump_AirToWater_FuelFired_Cooling,
  OS_Chiller_Electric_EIR,
  OS_Chiller_Electric_ReformulatedEIR,
  OS_Chiller_Electric_ASHRAE205,
  OS_CentralHeatPumpSystem,
  OS_CentralHeatPumpSystem_Module,
  OS_ChillerHeaterPerformance_Electric_EIR,
  OS_Chiller_Absorption_Indirect,
  OS_Chiller_Absorption,
  OS_CoolingTower_SingleSpeed,
  OS_CoolingTower_TwoSpeed,
  OS_CoolingTower_VariableSpeed,
  OS_CoolingTowerPerformance_CoolTools,
  OS_CoolingTowerPerformance_YorkCalc,
  OS_DistrictCooling,
  OS_DistrictHeating,
  OS_GroundHeatExchanger_Vertical,
  OS_GroundHeatExchanger_HorizontalTrench,
  OS_Coil_Cooling_CooledBeam,
  OS_Coil_Cooling_FourPipeBeam,
  OS_Coil_Heating_FourPipeBeam,
  OS_Coil_Cooling_DX,
  OS_Coil_Cooling_DX_CurveFit_Performance,
  OS_Coil_Cooling_DX_CurveFit_OperatingMode,
  OS_Coil_Cooling_DX_CurveFit_Speed,
  OS_Coil_Cooling_DX_SingleSpeed,
  OS_Coil_Cooling_DX_TwoStageWithHumidityControlMode,
  OS_CoilPerformance_DX_Cooling,
  OS_Coil_Cooling_DX_MultiSpeed,
  OS_Coil_Cooling_DX_MultiSpeed_StageData,
  OS_Coil_Cooling_DX_TwoSpeed,
  OS_CoilSystem_Cooling_Water_HeatExchangerAssisted,
  OS_CoilSystem_Cooling_DX_HeatExchangerAssisted,
  OS_CoilSystem_IntegratedHeatPump_AirSource,
  OS_Coil_Cooling_DX_VariableRefrigerantFlow,
  OS_Coil_Cooling_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  OS_Coil_Cooling_LowTemperatureRadiant_ConstantFlow,
  OS_Coil_Cooling_LowTemperatureRadiant_VariableFlow,
  OS_Coil_Cooling_Water,
  OS_Coil_Cooling_WaterToAirHeatPump_EquationFit,
  OS_Coil_Heating_DX_SingleSpeed,
  OS_Coil_Heating_DX_VariableRefrigerantFlow,
  OS_Coil_Heating_DX_VariableRefrigerantFlow_FluidTemperatureControl,
  OS_Coil_Heating_Electric,
  OS_Coil_Heating_Electric_MultiStage,
  OS_Coil_Heating_Electric_MultiStage_StageData,
  OS_Coil_Heating_Gas,
  OS_Coil_Heating_Gas_MultiStage,
  OS_Coil_Heating_Gas_MultiStage_StageData,
  OS_Coil_Heating_Desuperheater,
  OS_Coil_WaterHeating_Desuperheater,
  OS_Coil_Heating_LowTemperatureRadiant_ConstantFlow,
  OS_Coil_Heating_LowTemperatureRadiant_VariableFlow,
  OS_Coil_Heating_Water,
  OS_Coil_Heating_Water_Baseboard,
  OS_Coil_Heating_Water_Baseboard_Radiant,
  OS_Coil_Heating_WaterToAirHeatPump_EquationFit,
  OS_Coil_Heating_DX_MultiSpeed,
  OS_Coil_Heating_DX_MultiSpeed_StageData,
  OS_Coil_Heating_DX_VariableSpeed,
  OS_Coil_Heating_DX_VariableSpeed_SpeedData,
  OS_Coil_Cooling_DX_VariableSpeed,
  OS_Coil_Cooling_DX_VariableSpeed_SpeedData,
  OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit,
  OS_Coil_Heating_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData,
  OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit,
  OS_Coil_Cooling_WaterToAirHeatPump_VariableSpeedEquationFit_SpeedData,
  OS_Coil_Cooling_Water_Panel_Radiant,
  OS_Connection,
  OS_Connector_Mixer,
  OS_Connector_Splitter,
  OS_Controller_MechanicalVentilation,
  OS_ZoneControl_ContaminantController,
  OS_Controller_OutdoorAir,
  OS_Controller_WaterCoil,
  OS_Curve_Linear,
  OS_Curve_QuadLinear,
  OS_Curve_QuintLinear,
  OS_Curve_Quadratic,
  OS_Curve_Cubic,
  OS_Curve_Quartic,
  OS_Curve_Exponent,
  OS_Curve_Bicubic,
  OS_Curve_Biquadratic,
  OS_Curve_QuadraticLinear,
  OS_Curve_Triquadratic,
  OS_Curve_Functional_PressureDrop,
  OS_Curve_FanPressureRise,
  OS_Curve_ExponentialSkewNormal,
  OS_Curve_Sigmoid,
  OS_Curve_RectangularHyperbola1,
  OS_Curve_RectangularHyperbola2,
  OS_Curve_ExponentialDecay,
  OS_Curve_DoubleExponentialDecay,
  OS_Table_MultiVariableLookup,
  OS_Table_IndependentVariable,
  OS_Table_Lookup,
  OS_DesignSpecification_OutdoorAir,
  OS_DesignSpecification_ZoneAirDistribution,
  OS_Humidifier_Steam_Electric,
  OS_Humidifier_Steam_Gas,
  OS_EvaporativeCooler_Indirect_ResearchSpecial,
  OS_EvaporativeCooler_Direct_ResearchSpecial,
  OS_EvaporativeFluidCooler_SingleSpeed,
  OS_EvaporativeFluidCooler_TwoSpeed,
  OS_Fan_SystemModel,
  OS_Fan_ConstantVolume,
  OS_Fan_VariableVolume,
  OS_Fan_OnOff,
  OS_Fan_ComponentModel,
  OS_Fan_ZoneExhaust,
  OS_HeatExchanger_AirToAir_SensibleAndLatent,
  OS_HeatExchanger_Desiccant_BalancedFlow,
  OS_HeatExchanger_Desiccant_BalancedFlow_PerformanceDataType1,
  OS_HeatExchanger_FluidToFluid,
  OS_HVACComponentList,
  OS_Node,
  OS_PortList,
  OS_Pipe_Adiabatic,
  OS_PlantLoop,
  OS_PlantEquipmentOperation_CoolingLoad,
  OS_PlantEquipmentOperation_HeatingLoad,
  OS_PlantEquipmentOperation_OutdoorDryBulb,
  OS_PlantEquipmentOperation_OutdoorWetBulb,
  OS_PlantEquipmentOperation_OutdoorRelativeHumidity,
  OS_PlantEquipmentOperation_OutdoorDewpoint,
  OS_PlantEquipmentOperation_OutdoorDryBulbDifference,
  OS_PlantEquipmentOperation_OutdoorDewpointDifference,
  OS_PlantEquipmentOperation_OutdoorWetBulbDifference,
  OS_PlantComponent_TemperatureSource,
  OS_LoadProfile_Plant,
  OS_Pump_ConstantSpeed,
  OS_Pump_VariableSpeed,
  OS_HeaderedPumps_ConstantSpeed,
  OS_HeaderedPumps_VariableSpeed,
  OS_TemperingValve,
  OS_SetpointManager_FollowOutdoorAirTemperature,
  OS_SetpointManager_MixedAir,
  OS_SetpointManager_OutdoorAirReset,
  OS_SetpointManager_OutdoorAirPretreat,
  OS_SetpointManager_SingleZone_Reheat,
  OS_SetpointManager_SingleZone_Humidity_Minimum,
  OS_SetpointManager_MultiZone_MinimumHumidity_Average,
  OS_SetpointManager_MultiZone_Humidity_Minimum,
  OS_SetpointManager_Scheduled,
  OS_SetpointManager_Scheduled_DualSetpoint,
  OS_SetpointManager_Warmest,
  OS_SetpointManager_WarmestTemperatureFlow,
  OS_SetpointManager_Coldest,
  OS_SetpointManager_FollowGroundTemperature,
  OS_SetpointManager_FollowSystemNodeTemperature,
  OS_SetpointManager_MultiZone_Cooling_Average,
  OS_SetpointManager_MultiZone_Heating_Average,
  OS_SetpointManager_MultiZone_Humidity_Maximum,
  OS_SetpointManager_MultiZone_MaximumHumidity_Average,
  OS_SetpointManager_SingleZone_OneStageCooling,
  OS_SetpointManager_SingleZone_OneStageHeating,
  OS_SetpointManager_SingleZone_Humidity_Maximum,
  OS_SetpointManager_SingleZone_Cooling,
  OS_SetpointManager_SingleZone_Heating,
  OS_SetpointManager_SystemNodeReset_Temperature,
  OS_SetpointManager_SystemNodeReset_Humidity,
  OS_Sizing_Plant,
  OS_Sizing_System,
  OS_Sizing_Zone,
  OS_Splitter,
  OS_ThermalZone,
  OS_ZoneControl_Humidistat,
  OS_ThermostatSetpoint_DualSetpoint,
  OS_ZoneControl_Thermostat_StagedDualSetpoint,
  OS_ZoneHVAC_EquipmentList,
  OS_ZoneVentilation_DesignFlowRate,
  OS_ZoneVentilation_WindandStackOpenArea,
  OS_ZoneHVAC_Baseboard_Convective_Electric,
  OS_ZoneHVAC_Baseboard_Convective_Water,
  OS_ZoneHVAC_HighTemperatureRadiant,
  OS_ZoneHVAC_LowTemperatureRadiant_Electric,
  OS_ZoneHVAC_IdealLoadsAirSystem,
  OS_ZoneHVAC_FourPipeFanCoil,
  OS_ZoneHVAC_LowTemperatureRadiant_ConstantFlow,
  OS_ZoneHVAC_LowTemperatureRadiant_VariableFlow,
  OS_ZoneHVAC_PackagedTerminalHeatPump,
  OS_ZoneHVAC_PackagedTerminalAirConditioner,
  OS_ZoneHVAC_TerminalUnit_VariableRefrigerantFlow,
  OS_ZoneHVAC_WaterToAirHeatPump,
  OS_ZoneHVAC_UnitHeater,
  OS_ZoneHVAC_Dehumidifier_DX,
  OS_ZoneHVAC_EnergyRecoveryVentilator,
  OS_ZoneHVAC_EnergyRecoveryVentilator_Controller,
  OS_ZoneHVAC_UnitVentilator,
  OS_ZoneHVAC_Baseboard_RadiantConvective_Water,
  OS_ZoneHVAC_Baseboard_RadiantConvective_Electric,
  OS_ZoneHVAC_CoolingPanel_RadiantConvective_Water,
  OS_ZoneMixing,
  OS_WaterHeater_Mixed,
  OS_WaterHeater_HeatPump,
  OS_Coil_WaterHeating_AirToWaterHeatPump,
  OS_WaterHeater_HeatPump_WrappedCondenser,
  OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed,
  OS_Coil_WaterHeating_AirToWaterHeatPump_VariableSpeed_SpeedData,
  OS_Coil_WaterHeating_AirToWaterHeatPump_Wrapped,
  OS_WaterHeater_Stratified,
  OS_WaterHeater_Sizing,
  OS_WaterUse_Equipment,
  OS_WaterUse_Equipment_Definition,
  OS_WaterUse_Connections,
  OS_ThermalStorage_Ice_Detailed,
  OS_ThermalStorage_ChilledWater_Stratified,
  OS_Duct,
  OS_FluidCooler_SingleSpeed,
  OS_FluidCooler_TwoSpeed,
  OS_Pipe_Indoor,
  OS_Pipe_Outdoor,
  OS_SwimmingPool_Indoor,
  OS_SolarCollectorPerformance_FlatPlate,
  OS_SolarCollector_FlatPlate_Water,
  OS_SolarCollectorPerformance_IntegralCollectorStorage,
  OS_SolarCollector_IntegralCollectorStorage,
  OS_SolarCollectorPerformance_PhotovoltaicThermal_Simple,
  OS_SolarCollectorPerformance_PhotovoltaicThermal_BIPVT,
  OS_SolarCollector_FlatPlate_PhotovoltaicThermal,
  OS_Generator_MicroTurbine,
  OS_Generator_MicroTurbine_HeatRecovery,
  OS_Generator_Photovoltaic,
  OS_PhotovoltaicPerformance_Simple,
  OS_PhotovoltaicPerformance_EquivalentOneDiode,
  OS_PhotovoltaicPerformance_Sandia,
  OS_Generator_FuelCell,
  OS_Generator_FuelCell_PowerModule,
  OS_Generator_FuelCell_AirSupply,
  OS_Generator_FuelCell_WaterSupply,
  OS_Generator_FuelCell_AuxiliaryHeater,
  OS_Generator_FuelCell_ExhaustGasToWaterHeatExchanger,
  OS_Generator_FuelCell_ElectricalStorage,
  OS_Generator_FuelCell_Inverter,
  OS_Generator_FuelCell_StackCooler,
  OS_Generator_FuelSupply,
  OS_Generator_WindTurbine,
  OS_Generator_PVWatts,
  OS_ElectricLoadCenter_Inverter_PVWatts,
  OS_ElectricLoadCenter_Inverter_Simple,
  OS_ElectricLoadCenter_Inverter_LookUpTable,
  OS_ElectricLoadCenter_Storage_Simple,
  OS_ElectricLoadCenter_Storage_LiIonNMCBattery,
  OS_ElectricLoadCenter_Transformer,
  OS_ElectricLoadCenter_Distribution,
  OS_ElectricLoadCenter_Storage_Converter,
  OS_ComponentCost_Adjustments,
  OS_LifeCycleCost,
  OS_CurrencyType,
  OS_LifeCycleCost_Parameters,
  OS_LifeCycleCost_UsePriceEscalation,
  OS_UtilityBill,
  OS_UtilityCost_Tariff,
  OS_UtilityCost_Qualify,
  OS_UtilityCost_Charge_Simple,
  OS_UtilityCost_Charge_Block,
  OS_UtilityCost_Ratchet,
  OS_UtilityCost_Variable,
  OS_UtilityCost_Computation,
  OS_Output_Schedules,
  OS_Output_Constructions,
  OS_OutputControl_Files,
  OS_OutputControl_Timestamp,
  OS_OutputControl_ReportingTolerances,
  OS_OutputControl_Table_Style,
  OS_Output_Meter,
  OS_Output_Variable,
  OS_Meter_Custom,
  OS_Meter_CustomDecrement,
  OS_Output_EnvironmentalImpactFactors,
  OS_EnvironmentalImpactFactors,
  OS_FuelFactors,
  OS_Output_Diagnostics,
  OS_Output_DebuggingData,
  OS_Output_JSON,
  OS_Output_SQLite,
  OS_Output_EnergyManagementSystem,
  OS_Output_Table_SummaryReports,
  OS_EnergyManagementSystem_Sensor,
  OS_EnergyManagementSystem_Actuator,
  OS_EnergyManagementSystem_ProgramCallingManager,
  OS_EnergyManagementSystem_Program,
  OS_EnergyManagementSystem_Subroutine,
  OS_EnergyManagementSystem_GlobalVariable,
  OS_EnergyManagementSystem_OutputVariable,
  OS_EnergyManagementSystem_MeteredOutputVariable,
  OS_EnergyManagementSystem_TrendVariable,
  OS_EnergyManagementSystem_InternalVariable,
  OS_EnergyManagementSystem_CurveOrTableIndexVariable,
  OS_EnergyManagementSystem_ConstructionIndexVariable,
  OS_PythonPlugin_Instance,
  OS_PythonPlugin_Variable,
  OS_PythonPlugin_TrendVariable,
  OS_PythonPlugin_OutputVariable,
  OS_AirflowNetworkSimulationControl,
  OS_AirflowNetworkZone,
  OS_AirflowNetworkSurface,
  OS_AirflowNetworkReferenceCrackConditions,
  OS_AirflowNetworkCrack,
  OS_AirflowNetworkEffectiveLeakageArea,
  OS_AirflowNetworkSpecifiedFlowRate,
  OS_AirflowNetworkDetailedOpening,
  OS_AirflowNetworkSimpleOpening,
  OS_AirflowNetworkHorizontalOpening,
  OS_AirflowNetworkOutdoorAirflow,
  OS_AirflowNetworkZoneExhaustFan,
  OS_AirflowNetworkExternalNode,
  OS_AirflowNetworkPressureController,
  OS_AirflowNetworkDistributionNode,
  OS_AirflowNetworkLeak,
  OS_AirflowNetworkLeakageRatio,
  OS_AirflowNetworkDuct,
  OS_AirflowNetworkFan,
  OS_AirflowNetworkEquivalentDuct,
  OS_AirflowNetworkConstantPressureDrop,
  OS_AirflowNetworkReliefAirFlow,
  OS_AirflowNetworkDistributionLinkage,
  OS_AirflowNetworkDuctViewFactors,
  OS_AirflowNetworkOccupantVentilationControl,
  OS_ExternalInterface,
  OS_ExternalInterface_Schedule,
  OS_ExternalInterface_Variable,
  OS_ExternalInterface_Actuator,
  OS_ExternalInterface_FunctionalMockupUnitImport,
  OS_ExternalInterface_FunctionalMockupUnitImport_From_Variable,
  OS_ExternalInterface_FunctionalMockupUnitImport_To_Schedule,
  OS_ExternalInterface_FunctionalMockupUnitImport_To_Actuator,
  OS_ExternalInterface_FunctionalMockupUnitImport_To_Variable,
  OS_ExternalInterface_FunctionalMockupUnitExport_From_Variable,
  OS_ExternalInterface_FunctionalMockupUnitExport_To_Schedule,
  OS_ExternalInterface_FunctionalMockupUnitExport_To_Actuator,
  OS_ExternalInterface_FunctionalMockupUnitExport_To_Variable,
  OS_Coil_UserDefined,
  OS_PlantComponent_UserDefined,
  CommentOnly,
};

}  // namespace constexpr_enum

#endif  // UTILITIES_IDD_IDDENUMS_CONSTEXPR_HXX
